=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2011 Fredo6 - Designed and written Jul 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   FredoTools.rb
# Original Date	:   06 Jul 2011
# Description	:   Loader for the Fredo Tools suite
#
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

PlugInfo = Struct.new :hmod, :name, :loaded, :files, :hsh, :plug_name, :version

#---------------------------------------------------------------------------------------------------------------------------
#Constants for texts (do not translate here, use Translation Dialog Box instead)	
#---------------------------------------------------------------------------------------------------------------------------

T6[:MNU_MainList] = "FredoTools Launcher..."
T6[:TIP_MainList] = "Launch FredoTools plugins from a list"

T6[:MSG_Processing] = "PROCESSING"
T6[:MSG_Finished] = "DONE"
T6[:MSG_NowPartFredoTools] = "Now part of FredoTools"
T6[:MSG_NoFaceInSelection] = "NO face in the selection"
T6[:MSG_NoVertices] = "NO vertex in the selection"

#---------------------------------------------------------------------------------------
#Class to provide name space for the strings of each Plug
#---------------------------------------------------------------------------------------

class T7Mod

def initialize(name) ; @name = name ; end
def []=(symb, sval) ; T6[("#{@name}__#{symb}").intern] = sval ; end
def [](symb, *args)
	ss = ("#{@name}__#{symb}").intern
	tx = T6[ss, *args]
	(tx == ss.to_s) ? T6[symb, *args] : tx
end

end	#class T7Mod

#--------------------------------------------------------------------------------------------------------------------------------------
# Plugs Registration
#--------------------------------------------------------------------------------------------------------------------------------------

#Register the module and attached files
def F6_FredoTools.register_plug(hmod, lst_files)
	name = (hmod.name =~ /::(.+)\Z/) ? $1 : hmod.name
	hmod.module_eval "T7 = F6_FredoTools::T7Mod.new '#{name}'"
	info = PlugInfo.new
	info.hmod = hmod
	info.name = name
	folder = MYPLUGIN.folder
	info.hsh = {}
	info.files = (lst_files.class == Array) ? lst_files : [lst_files]
	info.files = info.files.collect { |f| File.basename(f) }
	@hsh_plugs = {} unless @hsh_plugs
	@hsh_plugs[hmod.to_s] = info
end
	
#Register Additional info about the plug module
def F6_FredoTools.register_info(hmod, hsh_info)	
	info = @hsh_plugs[hmod.to_s]
	return unless info
	
	#Check SU version
	unless F6_FredoTools.check_su_version(hsh_info)
		LibFredo6.log "!FredoTools: Sub-plugin #{info.name} is not compatible with Sketchup Version #{Sketchup.version} - Required #{hsh_info[:su]}"
		@hsh_plugs.delete hmod.to_s
		return
	end
	
	#Transfering the other parameters
	hsh = info.hsh
	plug_name = hsh[:name] = hmod::T7[:PlugName]
	plug_name = info.name unless plug_name
	hsh[:description] = hmod::T7[:PlugDesc]
	hsh[:author] = "Fredo6 Tools"
	hsh = info.hsh.update hsh_info
	info.plug_name = plug_name
	info.version = hsh[:version]
	
	MYPLUGIN.add_credits hsh[:credits]
	MYPLUGIN.add_videos hsh[:videos]
	patdoc = hsh[:doc_pattern]
	patdoc = info.name unless patdoc
	MYPLUGIN.register_documentation_pattern plug_name, patdoc
	Traductor::AllPlugins.register hmod.name.sub("F6_", ''), hsh
end

#Check if the SU version is compatible
def F6_FredoTools.check_su_version(hsh_info)
	v = hsh_info[:su]
	return true unless v
	v = v.to_f if v.class == String
	suv = Sketchup.version.to_f
	(v <= suv)
end

#Declare a main command - Called by the sub modules	
def F6_FredoTools.declare_command(hmod, symb=nil, txt_mnu=nil, txt_tip=nil, icon=nil)
	info = @hsh_plugs[hmod.to_s]
	return unless info
	symb = info.name.intern unless symb
	txt_mnu = hmod::T7[:PlugName] unless txt_mnu
	txt_tip = hmod::T7[:PlugDesc] unless txt_tip
	txt_tip = info.name + ": " + txt_tip
	icon = info.name unless icon
	@lst_commands = [] unless @lst_commands
	tpc = MYPLUGIN.declare_command_long(symb, txt_mnu, txt_tip, icon) { F6_FredoTools.execute_command(hmod, symb) }
	@lst_commands.push [hmod, symb, txt_mnu, txt_tip, tpc]
end
	
#Register a context handlers for the Plugins
def self.register_context_handlers
	submenu = "FredoTools"
	@lst_plugs.each do |info|
		hmod = info.hmod
		next unless defined?(hmod.contextual_menu_info)
		ls = hmod.contextual_menu_info
		next unless ls && ls.length > 0
		ls = [ls] unless ls[0].class == Array
		ls.each do |a|
			cond, name, tip, symb = a
			next unless cond
			tc = Traductor::TestCondition.new() { cond.call }
			name = hmod::T7[:PlugName] unless name
			tip = hmod::T7[:PlugDesc] unless tip
			symb = info.name unless symb
			MYPLUGIN.declare_context_handler_long(symb, name, tip, tc, nil, submenu) { F6_FredoTools.execute_command(hmod, symb) }
		end	
	end	
end	
	
#Load Submodule	
def F6_FredoTools.load_submodule(hmod)
	#Load plug files if not done yet
	info = @hsh_plugs[hmod.to_s]
	folder = MYPLUGIN.folder
	return unless info
	unless info.loaded || defined?(hmod._execution)
		t0 = Time.now
		begin
			text = "#{info.name}  #{info.version}: Loading implementation files"
			Sketchup.set_status_text text
			info.files.each { |f| require File.join folder, "body_#{f}" }
			delta = ((Time.now - t0) * 1000).to_i
			LibFredo6.log text + " [#{delta} ms]"
			Sketchup.set_status_text ""
		rescue Exception => e
			err = LibFredo6.mask_dir e.message
			text = "#{info.name}  #{info.version}: ERROR in Loading ruby implementation file"
			LibFredo6.log "?#{text}", err
			LibFredo6.log_messagebox text, err
		end
		MYPLUGIN.add_load_time_second(Time.now - t0)
		info.loaded = true
	end
end

#Execute a command for the plug	
def F6_FredoTools.execute_command(hmod, symb)
	#Load plug files if not done yet
	F6_FredoTools.load_submodule(hmod)
	
	#Execute the command
	hmod._execution(symb) if defined?(hmod._execution)
end
	
#Invoke the dialog box to pick up the command	
def F6_FredoTools.call_list_dialog
	FredoToolsTopDialog.invoke
end
	
#Get the list of commands
def F6_FredoTools.list_commands
	@lst_commands
end
	
#Pin dialog box
def self.get_pin
	@pin_dialog
end
	
def self.toggle_pin
	@pin_dialog = !@pin_dialog
end	

#--------------------------------------------------------
# Startup routine (called from LibFredo6)
#--------------------------------------------------------
	
#Create all commands with menus and toolbar icons
def F6_FredoTools.startup	
	#Top menu
	MYPLUGIN.declare_topmenu "Fredo Tools", ["Tools", "Plugins", :T_MenuTools_Fredo6Collection, :T_MenuPlugins_Fredo6Collection]

	#Build the list of Plug modules
	@lst_plugs = @hsh_plugs.values.sort { |a, b| a.name <=> b.name }
	
	#Declaring the main command (menu and icon)
	MYPLUGIN.declare_command_long(:FredoTools_List, ">> " + T6[:MNU_MainList], T6[:TIP_MainList], "FredoTools_Main") { F6_FredoTools.call_list_dialog }
	#MYPLUGIN.declare_command_long(:FredoTools_List, ">> " + T6[:MNU_MainList], T6[:TIP_MainList], "FredoTools_Main", :quick_launcher) { F6_FredoTools.call_list_dialog }
	@lst_plugs.each { |info| info.hmod.declare_commands if defined?(info.hmod.declare_commands) }
	MYPLUGIN.default_icons_visible [:FredoTools_List]
	
	#Contextual menu handler
	register_context_handlers
	
	#Declaring the default parameters for the Tools (if any)
	@lst_plugs.each { |info| info.hmod.default_param if defined?(info.hmod.default_param)	}
	
	#Startup of the Plugin
	MYPLUGIN.go		
end
	
#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
# Top Dialog box for Report Area
#--------------------------------------------------------------------------------------------------------------			 
#--------------------------------------------------------------------------------------------------------------			 

class FredoToolsTopDialog

HTML = Traductor::HTML

@@top_dialog = nil
@@unique_key = "FredoToolsTopDialog_DLG"

#Invoke the Report Dialog
def FredoToolsTopDialog.invoke
	@@top_dialog = FredoToolsTopDialog.new unless Traductor::Wdlg.check_instance_displayed(@@unique_key)
end	

#initialization of the dialog box 
def initialize
	@wdlg = create_dialog_top
end
	
#--------------------------------------------------------------------------------------------------------------
# Dialog box configuration
#--------------------------------------------------------------------------------------------------------------			 

#Create the dialog box
def create_dialog_top
	@lst_commands = F6_FredoTools.list_commands
	@icmd = nil
	init_dialog_top
	wdlg_key = @@unique_key
	@title = T6[:MNU_MainList]
	@wdlg = Traductor::Wdlg.new @title, wdlg_key, false
	@wdlg.set_unique_key @@unique_key
	@wdlg.set_size @wid_total, @hgt_total
	@wdlg.set_background_color 'whitesmoke'
	@wdlg.set_callback self.method('topdialog_callback') 
	@wdlg.set_on_close { on_close_top() }
	@wdlg.set_html format_html_top
	@wdlg.initial_focus "ListBox", true
	@wdlg.show
	@wdlg
end

#Initialize parameters of the dialog box
def init_dialog_top
	#Table Heights
	@hgt_row = 25
	@hgt_table = 300
	h = @lst_commands.length * @hgt_row
	@hgt_table = h if h < @hgt_table
	
	#Table width
	nbchar = @lst_commands.collect { |a| a[2].length }
	max_char = nbchar.max
	@wid_extra = (RUN_ON_MAC) ? 20 : 10
	@wid_col_name = max_char * 8 + 50
	@wid_total = @wid_col_name + @wid_extra
	@hgt_total = @hgt_table + 0	
end

#Notification of window closure
def on_close_top
	@@top_dialog = nil
	launch_command
end

#Close the dialog box
def close_dialog_top
	@wdlg.close
end

#Call back for Dialog
def topdialog_callback(event, type, id, svalue)
	case event
		
	#Command buttons
	when /onclick/i
		case id
		when /IMG_PIN/
			update_pin
		when /IMG_WRENCH/
			MYPLUGIN.invoke_default_parameter_dialog
			
		when /ListBox/i
			@icmd = svalue.to_i
			close_or_not
		end	
	
	when /onchange/i
		case id
		when /ListBox/i
			@icmd = svalue.to_i
		end	

	when /onKeyUp/i	#Escape
		@icmd = nil
		close_or_not if svalue =~ /\A27\*/
	when /onKeyDown/i	#Return key
		@icmd = 0 unless @icmd
		close_or_not if svalue =~ /\A13\*/
		
	end
	svalue
end

def launch_command
	return unless @icmd
	#hmod, symb = @lst_commands[@icmd]
	hmod, symb, txt_mnu, txt_tip, tpc = @lst_commands[@icmd]
	#F6_FredoTools.execute_command hmod, symb
	MYPLUGIN.command_invoke tpc
end

#Close or not the dialog box when executing the action
def close_or_not
	(F6_FredoTools.get_pin) ? launch_command : @wdlg.close
end

#Build the HTML for Dialog
def format_html_top
	#Creating the HTML stream	
	html = Traductor::HTML.new
		
	#Styles used in the dialog box
	html.create_style 'Element', nil, 'K: navy', 'F-SZ: 10', 'align: center'
	
	#Inserting the main table and unit choosers
	@img_pin_on = HTML.image_file Traductor::MYPLUGIN.picture_get("Img_Pin_On")
	@img_pin_off = HTML.image_file Traductor::MYPLUGIN.picture_get("Img_Pin_Off")
	@tip_pin_on = T6[:T_TIP_Pin_On]
	@tip_pin_off = T6[:T_TIP_Pin_Off]
	px = py = 36
	if F6_FredoTools.get_pin
		img_pin = @img_pin_on
		tip = @tip_pin_on
	else	
		img_pin = @img_pin_off
		tip = @tip_pin_off
	end	
	himg_pin = HTML.format_imagelink img_pin, px, py, "IMG_PIN", nil, nil, tip

	px = py = 24
	img_wrench = HTML.image_file Traductor::MYPLUGIN.picture_get("Button_Wrench")
	himg_wrench = HTML.format_imagelink img_wrench, px, py, "IMG_WRENCH", nil, nil, T6[:T_STR_DefaultParamDialog]
	
	klist = []
	hextras = []
	for i in 0..@lst_commands.length-1
		klist.push [i, @lst_commands[i][2]]
		hextras[i] = { "tip" => @lst_commands[i][3] }
	end
	nb_item = @lst_commands.length
	icmd = @icmd
	icmd = 0 unless icmd
	hlb = HTML.format_listbox(icmd, klist, nb_item, "ListBox", "Element") { |i| hextras[i] }
	html.body_add "<table>"
	html.body_add "<tr><td valign='top'>#{himg_pin}<br><br>#{himg_wrench}</td><td valign='top'>#{hlb}</td></tr>"
	html.body_add "</table>"
	
	#Returning the HTML object
	html	
end
	
#Tiggle the pin dialog options and update the dialog box	
def update_pin
	if F6_FredoTools.toggle_pin
		img_pin = @img_pin_on
		tip = @tip_pin_on
	else	
		img_pin = @img_pin_off
		tip = @tip_pin_off
	end	
	cmd = "document.getElementById('IMG_PIN').src = '#{img_pin}' ;"
	cmd += "document.getElementById('IMG_PIN').title = '#{tip}' ;"
	@wdlg.execute_script cmd
end
	
end	#class FredoToolsTopDialog
	
	
end	#End Module F6_FredoTools
