=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2013 Fredo6 - Designed and written Oct 2013 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   FredoTools__EdgeInspector.rb
# Original Date	:   15 Oct 2013 
# Description	:   Detect and correct defects in Edges - Works with a Glass magnifier
# IMPORTANT		:	DO NOT TRANSLATE STRINGS in the source code
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

#====================================================================================================
#----------------------------------------------------------------------------------------------------
# Plugin Registration and startup
#----------------------------------------------------------------------------------------------------
#====================================================================================================

module EdgeInspector

F6_FredoTools.register_plug self, __FILE__

T7[:PlugName] = "Edge Inspector"
T7[:PlugDesc] = "Inspect and Fix defects about Edges"

F6_FredoTools.register_info self, 	
	{	
		:version => "1.2a",
		:date => "12 Oct 14",	
		:credits => ["EdgeInspector: Daniel S. for icons and cursors",
		             "EdgeInspector: Roberto Padulazzi and blippy for providing models with (lots of) edge defects",
		             "EdgeInspector: Daniel S., driven, Pilou and Jeff Hammond for beta testing and useful feedback"],
		:videos => "Edge Inspector: Overview ; http://www.youtube.com/watch?v=1gLGaqO712Q",
		:link_info => "http://sketchucation.com/forums/viewtopic.php?f=323&t=56004#p508307"
	}
 
#Declare the main menu
def self.declare_commands
	F6_FredoTools.declare_command self
end

#Condition for contextual menu as ["menu", test, proc]
def self.contextual_menu_info
	[proc { contextual_on_selection? } ]
end

#Check if selection contains a vertex
def self.contextual_on_selection?
	Sketchup.active_model.selection.find do |e| 
		e.instance_of?(Sketchup::Edge) || e.instance_of?(Sketchup::Face) ||
		e.instance_of?(Sketchup::Group) || e.instance_of?(Sketchup::ComponentInstance)
	end
end

#Default Parameters
def self.default_param
	MYDEFPARAM.separator :DEFAULT_EdgeInspector_Section, nil, T7[:PlugName]
	MYDEFPARAM.declare :DEFAULT_EdgeInspector_NativeDistance, true, 'B', nil, T7[:DEF_NativeDistance]
end

#Default Parameters
T7[:DEF_NativeDistance] = "Use native distance (versus real distance) for tolerance comparison"

end	#End Module EdgeInspector

end	#End Module F6_FredoTools
