=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2011 Fredo6 - Designed and written Jul 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   FredoTools__RemoveLonelyVertices.rb
# Original Date	:   15 Jul 2011 (based on work published in Nov 2009)
# Description	:   Mark all vertices of the selection with a guide point
#                   This is useful to visualize the vertices
# IMPORTANT		:	DO NOT TRANSLATE STRINGS in the source code
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

#==================================================
#--------------------------------------------------
# Plugin Registration and startup
#--------------------------------------------------
#==================================================

module RemoveLonelyVertices

F6_FredoTools.register_plug self, __FILE__

T7[:PlugName] = "Remove Lonely Vertices"
T7[:PlugDesc] = "Remove lonely vertices in the selection (vertex joining collinear edges)"

F6_FredoTools.register_info self, 	
	{	
		:version => "1.2a",
		:date => "20 Jul 11",	
		:videos => "Remove Lonely Vertices: Overview ; http://www.youtube.com/watch?v=WbxKv4WJfbc",
		:link_info => "http://forums.sketchucation.com/viewtopic.php?f=323&t=37619#p332047"
	}
 
#Declare the main menu
def self.declare_commands
	F6_FredoTools.declare_command self
	F6_FredoTools::MYPLUGIN.register_obsolete_files "RemoveLonelyVertices.rb"
	Traductor::AllPlugins.declare_obsolete_plugin "RemoveLonelyVertices", T6[:MSG_NowPartFredoTools]
end

#Condition for contextual menu as ["menu", test, proc]
def self.contextual_menu_info
	[proc { contextual_on_selection? } ]
end

#Check if selection contains a vertex
def self.contextual_on_selection?
	Sketchup.active_model.selection.find do |e| 
		e.instance_of?(Sketchup::Edge) || e.instance_of?(Sketchup::Face) ||
		e.instance_of?(Sketchup::Group) || e.instance_of?(Sketchup::ComponentInstance)
	end
end

end	#End Module RemoveLonelyVertices

end	#End Module F6_FredoTools
