=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2011 Fredo6 - Designed and written Aug 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   FredoTools__ReportLabelArea.rb
# Original Date	:   26 Aug 2011 (based on work published in June 2011)
# Description	:   Built reports on Areas for selection or whole model
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

#==================================================
#--------------------------------------------------
# Plugin Registration and startup
#--------------------------------------------------
#==================================================

module ReportLabelArea

F6_FredoTools.register_plug self, __FILE__

T7[:PlugName] = "Report Label Area"
T7[:PlugDesc] = "Report on Areas and Label surfaces with text tags"

F6_FredoTools.register_info self, 	
	{	
		:version => "2.0a",
		:date => "19 Dec 11",	
		:credits => "ReportLabelArea: TIG who pioneered his Area Formatter",
		:videos => ["ReportLabelArea: Reporting ; http://www.youtube.com/watch?v=bjbJO717Cyo",
					"ReportLabelArea: Labelling ; http://www.youtube.com/watch?v=Bu31TpMq3pY"],
		:link_info => "http://forums.sketchucation.com/viewtopic.php?f=323&t=40025#p354050"
	}
 
#Declare the main menu
def self.declare_commands
	F6_FredoTools.declare_command self, :ReportLabelArea_report, T7[:MNU_Report], T7[:TIP_Report], "ReportLabelArea_Report"
	F6_FredoTools.declare_command self, :ReportLabelArea_label, T7[:MNU_Label], T7[:TIP_Label], "ReportLabelArea_Label"
	Traductor::AllPlugins.declare_obsolete_plugin "ReportArea", T7[:MSG_NowPartFredoTools]
	F6_FredoTools::MYPLUGIN.register_obsolete_files "ReportArea.rb"
end

#Condition for contextual menu as ["menu", test, proc]
def self.contextual_menu_info
	cond_proc = proc { contextual_on_selection? }
	[[cond_proc, T7[:MNU_Report], T7[:TIP_Report], :ReportLabelArea_report], [cond_proc, T7[:MNU_Label], T7[:TIP_Label], :ReportLabelArea_label]]
end

#Check if entity is an edge, face, component or group
def self.contextual_on_selection?
	Sketchup.active_model.selection.each do |e|
		return true if e.instance_of?(Sketchup::Face) || e.instance_of?(Sketchup::Group) || 
		               e.instance_of?(Sketchup::ComponentInstance)
	end
	false	
end

#default Parameters
def self.default_param
	MYDEFPARAM.separator :DEFAULT_ReportArea_Section, nil, T7[:MNU_Label]
	MYDEFPARAM.declare :DEFAULT_ReportArea_Recurse, true, 'B', nil, T7[:DEF_Recurse]
	MYDEFPARAM.declare :DEFAULT_ReportArea_SingleFace, false, 'B', nil, T7[:DEF_SingleFace]
	MYDEFPARAM.declare :DEFAULT_ReportArea_ShowBoxlines, false, 'B', nil, T7[:DEF_ShowBoxlines]
	MYDEFPARAM.declare :DEFAULT_ReportArea_MultiArrow, true, 'B', nil, T7[:DEF_MultiArrow]
end

#Texts for the module
T7[:MNU_Report] = "Report on Areas"
T7[:TIP_Report]= "Calculate and report on areas for the selection or the whole model"
T7[:MNU_Label] = "Label with Areas"
T7[:TIP_Label] = "Label surfaces and elements with Areas text tags"

T7[:DEF_Recurse] = "Calculate areas recursively down through components and groups"
T7[:DEF_SingleFace] = "Selection of Single face (vs. surfaces)"
T7[:DEF_ShowBoxlines] = "Show box lines for containers"
T7[:DEF_MultiArrow] = "Multi-Arrow label when more than 2 objects are selected"

end	#End Module ReportLabelArea

end	#End Module F6_FredoTools
