=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2011 Fredo6 - Designed and written Oct 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   FredoTools__ThruPaint.rb
# Original Date	:   20 Oct 2011
# Description	:   Extend the SU Paint tools by painting faces within components and groups
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

#==================================================
#--------------------------------------------------
# Plugin Registration and startup
#--------------------------------------------------
#==================================================

module ThruPaint

F6_FredoTools.register_plug self, __FILE__

T7[:PlugName] = "ThruPaint"
T7[:PlugDesc] = "Extension of Sketchup Paint Tool"

F6_FredoTools.register_info self, 	
	{
		:version => "1.2a",
		:date => "08 May 12",	
		:credits => ["ThruPaint: Thomthom who came up with a similar concept in Feb 2010",
		             "ThruPaint: Thomthom for convention on diagonals of Quad meshes and other tips",
		             "ThruPaint: Jeff Hammond for troubleshooting on Mac",
		             "ThruPaint: Daniel S for helping me in troubleshooting some bugs"],
		:videos => "ThruPaint: Overview ; http://www.youtube.com/watch?v=l2QjLlUQVrY",
		:link_info => "http://forums.sketchucation.com/viewtopic.php?f=323&t=44552#p397448"
	}
 
#Declare the main menu
def self.declare_commands
	F6_FredoTools.declare_command self
	Traductor::MaterialManager.manager
	@tracker = ThruPaintObserver.new
	load_tracking_auto
end

#Getting  / Setting the Tracking modes
def self.set_tracking_auto(onoff) ; @tracker.tracking_auto = onoff ; end
def self.get_tracking_auto ; @tracker.tracking_auto ; end
def self.toggle_tracking_auto ; @tracker.tracking_auto = !@tracker.tracking_auto ; end
def self.set_tracking_tempo(onoff) ; @tracker.tracking_tempo = onoff ; end
def self.get_tracking_tempo ; @tracker.tracking_tempo ; end
def self.get_tool_id ; @tool_id ; end
def self.set_tool_id(id) ; @tool_id = id ; end
def self.supaint_id ; @tracker.supaint_id ; end
def self.alien_tool_before ; @tracker.alien_tool_id ; end

def self.load_tracking_auto
	s = Sketchup.read_default "FredoTools_ThruPaint", "Tracking_auto"
	tracking_auto = (s == 'true') ? true : false
	set_tracking_auto tracking_auto
end

def self.save_tracking_auto(tracking_auto)
	val = (tracking_auto) ? "true" : ""
	Sketchup.write_default "FredoTools_ThruPaint", "Tracking_auto", val
end

#Default Parameters
def self.default_param
	MYDEFPARAM.separator :DEFAULT_ThruPaint_Section, nil, T7[:PlugName]
	MYDEFPARAM.declare :DEFAULT_ThruPaint_TrackingAuto, false, 'B', nil, T7[:DEF_TrackingAuto]
	MYDEFPARAM.declare :DEFAULT_ThruPaint_NbHotFaces, 1000, 'I:>=50<5000', nil, T7[:DEF_NbHotFaces]
end

#Texts for the module
T7[:DEF_TrackingAuto] = "Automatically activate ThruPaint when invoking the SU native Paint Tool"
T7[:DEF_NbHotFaces] = "Maximum number of faces interactively refreshed"

#-----------------------------------------------------------------------------------------------------------------------
# Observer
#-----------------------------------------------------------------------------------------------------------------------

class ThruPaintObserver 

attr_accessor :tracking_auto, :tracking_tempo, :supaint_id, :alien_tool_id

#ThruPaintObserver: Add the Application Observer to track change of models
def initialize
	@alien_tool_id = true
	Sketchup.add_observer self
	model_changing Sketchup.active_model, 'Boot'
end

#ThruPaintObserver: Observers events
def onNewModel(model) ; model_changing(model, 'new') ; end
def onOpenModel(model) ; model_changing(model, 'open') ; end

#ThruPaintObserver: New model or Opening the model
def model_changing(model, mode)
	#LibFredo6.debug "begin model changing #{mode} #{model}"
	Sketchup.active_model.tools.add_observer self	
	#LibFredo6.debug "after model changing #{mode} #{model}"
end
	
#ThruPaintObserver: Observer callback when the tool is changed	
def onActiveToolChanged(tools, tool_name, tool_id) ; tool_changed(tool_name, tool_id) unless tool_id == 0 ; end
 
#ThruPaintObserver: Check the current tool and intercept if it is the SU Paint Tool 
def tool_changed(tool_name, tool_id)
	#LibFredo6.debug "begin tool changed #{tool_name} #{tool_id}"
	@alien_tool_id = false if tool_id == ThruPaint.get_tool_id
	return if @observer_active || tool_id == ThruPaint.get_tool_id
	@observer_active = true
	@tracking_auto = ThruPaint.get_tracking_auto
	@tracking_auto = MYDEFPARAM[:DEFAULT_ThruPaint_TrackingAuto] if @tracking_auto == nil
	ok = false
	orbit = false
	
	#Current tool is SU Paint tool
	if tool_name =~ /PaintTool\Z/i
		ok = true
		@supaint_id = tool_id
	elsif tool_name =~ /PaintMatchTool\Z/i
		#ok = true
		@paint_match_id = tool_id
	elsif tool_name =~ /OrbitTool\Z/i
		orbit = true
		@orbit_id = tool_id
	else
		@alien_tool_id = true
	end
	unless @tracking_auto || @tracking_tempo
		@observer_active = false
		return
	end	
	
	#Intercepting the SU Paint tool
	if !ok && RUN_ON_MAC	#bug on Mac
		ok = true if @supaint_id && tool_id == @supaint_id
		ok = true if @supaint_match_id && tool_id == @supaint_match_id
	end
	
	#invoking ThurPaint instead of SU Paint
	if ok
		unless @loaded
			F6_FredoTools.load_submodule F6_FredoTools::ThruPaint
			@loaded = true
		end	
		@observer_active = false
		ThruPaint.execute_direct
	elsif tool_id != ThruPaint.get_tool_id && !orbit
		ThruPaint.set_tracking_tempo false
	end	
	@observer_active = false
	#LibFredo6.debug "End tool changed"
end
	
end	#class ThruPaintObserver

#==================================================
#--------------------------------------------------
# Plugin Class Methods
#--------------------------------------------------
#==================================================

#----------------------------------------------------------------------------------------------------
# Plugin Top Level
#----------------------------------------------------------------------------------------------------

#Direct Launch of the tool
def ThruPaint.execute_direct(selection=nil)
	#LibFredo6.debug "Begin Execute direct"
	Traductor.use_libraries
	@thru_tool = ThruPaintTool.new #####unless @thru_tool
	Sketchup.active_model.select_tool @thru_tool
	#Regression in SU2015Reg "End Execute direct"
end

end	#End Module ThruPaint

end	#End Module F6_FredoTools
