=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2011 Fredo6 - Designed and written Jul 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_FredoTools__ConstructFaceNormal.rb
# Original Date	:   15 Jul 2011 (based on work published in May 2011)
# Description	:   Create a construction line along the noormal and barycenter of each face selected
# IMPORTANT		:	DO NOT TRANSLATE STRINGS in the source code
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

#====================================================================================================
#----------------------------------------------------------------------------------------------------
# Plugin Implementation
#----------------------------------------------------------------------------------------------------
#====================================================================================================

module ConstructFaceNormal

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

def self._execution(symb)
	execute
end

#============================================================
# Top Processing method
#============================================================

#Initialize and execute the computation of normals
def self.execute(selection=nil)
	model = Sketchup.active_model
	selection = model.selection unless selection
	@menutitle = T7[:PlugName]
	
	#Create the list of faces
	lst_faces = selection.find_all { |e| e.instance_of?(Sketchup::Face) }
		
	#No faces
	if lst_faces.length == 0
		Sketchup.set_status_text T6[:MSG_NoFaceInSelection]
		return
	end
	
	#processing the components
	Sketchup.set_status_text @msg_processing, SB_VCB_LABEL
	entities = model.active_entities
	model.start_operation @menutitle

	lst_faces.each do |face|
		pts = face.outer_loop.vertices.collect { |v| v.position }
		bary = face_centroid face
		normal = face.normal
		len = 0
		pts.each { |pt| len += bary.distance(pt) }
		len /= pts.length
		pt1 = bary.offset normal, len
		pt2 = bary.offset normal, -len
		entities.add_cline pt1, pt2, "-"
		entities.add_cpoint bary
	end
	
	model.commit_operation
	Sketchup.set_status_text @msg_done, SB_VCB_LABEL
	Sketchup.set_status_text "#{lst_faces.length}", SB_VCB_VALUE
end

#Compute the centroid of a face
def self.face_centroid(face)
	lpt = face.outer_loop.vertices.collect { |v| v.position }
	axes = face.normal.axes
	tr_axe = Geom::Transformation.axes lpt[0], axes[0], axes[1], axes[2]
	tr_axe_inv = tr_axe.inverse
	lpt = lpt.collect { |pt| tr_axe_inv * pt }
	lpt.push lpt[0]
	area = 0.0
	cx = cy = 0
	for i in 0..lpt.length-2
		pt1 = lpt[i]
		pt2 = lpt[i+1]
		a = pt1.x * pt2.y - pt2.x * pt1.y
		area += a
		cx += (pt1.x + pt2.x) * a
		cy += (pt1.y + pt2.y) * a
	end
	area *= 3
	tr_axe * Geom::Point3d.new(cx / area, cy / area, 0)
end

end	#End Module ConstructFaceNormal

end	#End Module F6_FredoTools
