=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2011 Fredo6 - Designed and written Aug 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_FredoTools__CountFacesBySides.rb
# Original Date	:   23 Aug 2011 (based on work published in Sept. 2010)
# Description	:   Count faces by number of sides - Calculate Area
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

module CountFacesBySides

#Texts for the module
T7[:MSG_CFBS_Faces] = "Faces by number of sides: all (unique)"
T7[:MSG_CFBS_Sides] = "Sides"
T7[:MSG_CFBS_Poly] = "Tri/quad"
T7[:MSG_CFBS_Total] = "Total"
T7[:MSG_CFBS_Lone] = "Lonely edges"
T7[:MSG_CFBS_Area] = "Total Area"
T7[:MSG_CFBS_Vertices] = "Vertices"

#====================================================================================================
#----------------------------------------------------------------------------------------------------
# Plugin Implementation
#----------------------------------------------------------------------------------------------------
#====================================================================================================

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

def self._execution(symb)
	execute
end

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

#Initialize and execute the marking of vertices
def self.execute(selection=nil)
	hfaces = []
	[0, 3, 4, 5, 6].each { |n| hfaces[n] = 0 }
	
	model = Sketchup.active_model
	selection = model.selection

	entities = (selection == nil || selection.empty?) ? model.active_entities : selection
	
	Sketchup.set_status_text T6[:MSG_Processing], SB_VCB_LABEL
	
	#Exploring the selection
	@hsh_info = Hash.new 0
	@hsh_cdef = {}
	process_selection entities, nil, nil, Geom::Transformation.new
	
	#Assembling the results
	t3 = @hsh_info[3]
	t3_t = t3 + @hsh_info[-3]
	t4 = @hsh_info[4]
	t4_t = t4 + @hsh_info[-4]
	t5 = @hsh_info[5]
	t5_t = t5 + @hsh_info[-5]
	poly = t3 + t4
	poly_t = t3_t + t4_t
	total = poly + t5
	total_t = poly_t + t5_t
	lone = @hsh_info[1]
	lone_t = lone + @hsh_info[-1]
	area = @hsh_info[:area]		
	tx_area = Sketchup.format_area area
	nvx = @hsh_info[2]
	nvx_t = nvx + @hsh_info[-2]
	
	#Displaying the results
	Sketchup.set_status_text T6[:MSG_Finished], SB_VCB_LABEL
			
	msg_sides = T7[:MSG_CFBS_Sides]	
	msg_poly = T7[:MSG_CFBS_Poly]
	msg_total = T7[:MSG_CFBS_Total]
	msg_lone = T7[:MSG_CFBS_Lone]
	msg_area = T7[:MSG_CFBS_Area]
	msg_vertices = T7[:MSG_CFBS_Vertices]
	
	textd = T7[:MSG_CFBS_Faces]
	textd += "\n3 " + msg_sides + "  \t ->  #{t3_t}\t (#{t3})"
	textd += "\n4 " + msg_sides + "  \t ->  #{t4_t}\t (#{t4})"
	textd += "\n5+ " + msg_sides + " \t ->  #{t5_t}\t (#{t5})"
	textd += "\n" + msg_poly + "\t ->  #{poly_t}\t (#{poly})"
	textd += "\n" + msg_total + "\t ->  #{total_t}\t (#{total})"
	textd += "\n\n" + msg_lone + " ->  #{lone_t} (#{lone})" if lone > 0
	textd += "\n\n" + msg_vertices + " ->  #{nvx_t} (#{nvx})"
	textd += "\n\n" + msg_area + " ->  #{tx_area}" if area > 0
	
	text = "3 #{msg_sides}: #{t3_t} (#{t3})  " 
	text += "4 #{msg_sides}: #{t4_t} (#{t4})  " 
	text += "5+ #{msg_sides}: #{t5_t} (#{t5})  " 
	text += "#{msg_poly}: #{poly_t} (#{poly})  " 
	text += "#{msg_total}: #{total_t} (#{total})  " 
	text += "#{msg_lone}: #{lone_t} (#{lone})" if lone > 0
	text += "#{msg_area}: #{tx_area}" if area > 0
	Sketchup.set_status_text text
		
	UI.messagebox textd
	Sketchup.set_status_text '', SB_VCB_LABEL
	
end
		
#Recursive counting	
def self.process_selection(entities, comp, cdef, t)
	hvertices = {}
	fac = (cdef && @hsh_cdef[cdef.entityID]) ? -1 : 1
	
	entities.each do |e|
		if e.instance_of?(Sketchup::Face)
			n = e.outer_loop.edges.length
			n = 5 if n >= 5
			@hsh_info[fac*n] += 1
			@hsh_info[fac*2] += register_vertex(e.vertices, hvertices)
			@hsh_info[:area] += G6.face_area(e, t)
		elsif e.instance_of?(Sketchup::Edge)
			@hsh_info[fac] += 1 if e.faces.length == 0
			@hsh_info[fac*2] += register_vertex([e.start, e.end], hvertices)
		elsif e.instance_of?(Sketchup::Group)
			process_selection e.entities, e, nil, t * e.transformation
		elsif e.instance_of?(Sketchup::ComponentInstance)
			cdef = e.definition
			process_selection cdef.entities, e, cdef, t * e.transformation
			@hsh_cdef[cdef.entityID] = true 
		end			
	end
end
	
#Counting vertex	
def self.register_vertex(lvx, hvertices)
	n = 0
	lvx.each do |vx|
		next if hvertices[vx.entityID]
		hvertices[vx.entityID] = true
		n += 1
	end
	n	
end
	
end	#End Module CountFacesBySides

end	#End Module F6_FredoTools
