=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2014 Fredo6 - Designed and written Mar 2014 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_FredoTools__DrawAlong.rb
# Original Date	:   11 Mar 2014 
# Description	:   Analog to SU Draw tool, but can force the direction along a defined plane
# IMPORTANT		:	DO NOT TRANSLATE STRINGS in the source code
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

module DrawAlong
				   
#Texts for the module
T7[:TIP_ToggleProtractorMode] = "Toggle Protractor for drawing by an angle"

T7[:TIP_PencilRedo] = "Redo Draw Line"
T7[:TIP_GuideToggle] = "Drawing with Guide lines and Guide points"
T7[:TIP_GenFaceToggle] = "Toggle automatic Generation of Faces"
T7[:TIP_GuideMode] = "Toggle Guide versus Edge drawing"
T7[:TIP_NativeDraw] = "Call the native Sketchup Draw tool"

#====================================================================================================
#----------------------------------------------------------------------------------------------------
# Plugin Implementation
#----------------------------------------------------------------------------------------------------
#====================================================================================================

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

def self._execution(symb)
	Sketchup.active_model.select_tool DrawAlongTool.new
end

#=============================================================================================
#=============================================================================================
# Class DrawAlongTool: main class for the Interactive tool
#=============================================================================================
#=============================================================================================

class DrawAlongTool < Traductor::PaletteSuperTool

#----------------------------------------------------------------------------------------------------
# INIT: Initialization
#----------------------------------------------------------------------------------------------------

#INIT: Class instance initialization
def initialize(*args)
	#Basic Initialization
	@tr_id = Geom::Transformation.new
	@model = Sketchup.active_model
	@entities = @model.active_entities
	@view = @model.active_view
	@rendering_options = Sketchup.active_model.rendering_options
	@dico_name = "Fredo6_FredoTools_DrawAlong"
	@dico_attr = "Parameters"
	init_flags
	MYDEFPARAM.add_notify_proc(self.method("notify_from_defparam"), true)	
	@duration_long_click = 0.8
	
	#Creating the protractor shape
	hsh = { :default_color => 'red' }
	@protractor = G6::ShapeProtractor.new hsh
	
	#Parsing the arguments
	args.each { |arg| arg.each { |key, value|  parse_args(key, value) } if arg.class == Hash }

	#Loading parameters
	parameter_load
	
	#Static initialization
	init_cursors
	init_colors
	init_messages

	#Initializing the Zoom Manager
	@zoom_void = G6::ZoomVoid.new
	
	#Initializing the Key manager
	@keyman = Traductor::KeyManager.new() { |event, key, view| key_manage(event, key, view) }

	#Create the Origin-Target Picker
	hsh = { :notify_proc => self.method("notify_from_otpicker"), :line_mode => true, :ignore_selection => true,
	        :color_box_distance => @color_box_distance }
	@otpicker = Traductor::OriginTargetPicker.new @hsh_parameters, hsh
	guide_edge_after

	notify_from_defparam
	
	#Creating the palette manager and texts
	init_palette
end

def init_flags
	@flag_anchor_text = false
end
	
#INIT: Parse the arguments of the initialize method
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /from/i
		@invoked_from = value
	end	
end

#INIT: Initialize texts and messages
def init_messages
	@menutitle = T7[:PlugName]
	
	@mnu_exit = T6[:T_BUTTON_Exit]	
	@mnu_abort = T6[:T_STR_AbortTool]	
	
	@msg_origin = [T6[:T_MSG_OriginPick], T6[:T_MSG_PickFromModel], T6[:T_MSG_ArrowModeAxis], T6[:T_MSG_DoubleClickRepeat]].join " - "
	@msg_target = [T6[:T_MSG_TargetPick], T6[:T_MSG_ShiftLockMode], T6[:T_MSG_ArrowModeAxis]].join " - "
	
	@tip_protractor_mode = T7[:TIP_ToggleProtractorMode] + " [#{T6[:T_KEY_CTRL]}]"
	
	@tip_pencil_redo = T7[:TIP_PencilRedo] + " [#{T6[:T_KEY_DoubleClick]}]"
			
	@txt_offset = T6[:T_TXT_Offset]
	@txt_offset_remote = T6[:T_TXT_Distance]
	@txt_angle = T6[:T_TXT_Angle]
	
	@msg_doubleclick_exit = T6[:T_INFO_DoubleClickExit]
end

#INIT: Initialize texts and messages
def init_colors
	@color_box_distance = ['lightgreen', 'green']
	@color_box_angle = ['pink', 'red']
	
	@color_but_title = 'lightblue'
	@color_but_hi = 'lightgreen'	
end

#INIT: Initialize the cursors
def init_cursors
	@id_cursor_edge = MYPLUGIN.create_cursor "DrawAlong_cursor", 0, 24	
	@id_cursor_edge_noinf = MYPLUGIN.create_cursor "DrawAlong_cursor_noinf", 0, 24	
	@id_cursor_cline = MYPLUGIN.create_cursor "DrawAlong_cursor_guide_line", 0, 24	
	@id_cursor_cline_noinf = MYPLUGIN.create_cursor "DrawAlong_cursor_guide_line_noinf", 0, 24	
	@id_cursor_cpoint = MYPLUGIN.create_cursor "DrawAlong_cursor_guide_point", 0, 24	
	@id_cursor_cpoint_noinf = MYPLUGIN.create_cursor "DrawAlong_cursor_guide_point_noinf", 0, 24	
end

#--------------------------------------------------------------
# PARAMETER: Tolerance and other settings and parameters
#--------------------------------------------------------------

#INIT: Notification from default parameters
def notify_from_defparam(key=nil, val=nil)

end

#Persistent parameters and defaults
@@hsh_parameters_persist = {} unless defined?(@@hsh_parameters_persist) && @@hsh_parameters_persist.length > 0

#PARAMETER: Load the parameters from the model attribute and the defaults
def parameter_load
	@hsh_parameters = {}
	
	#Defaults
	@hsh_param_default = {
		:edge_prop => 'P',
		:guide_mode => false,
		:guide_mode_line => true,
		:guide_mode_point => true,
		:protractor_mode => false
	}
		
	#Calculating the appropriate parameters
	@hsh_parameters = {}
	@hsh_parameters.update @hsh_param_default
	sparam = Traductor::Default.read @dico_name, @dico_attr
	hsh = eval(sparam) rescue {}
	hsh = {} unless hsh
	@@hsh_parameters_persist.update hsh
	@hsh_parameters.update @@hsh_parameters_persist 
	
	@edge_prop = @hsh_parameters[:edge_prop]
	@protractor_mode = false
	@option_gen_faces = @hsh_parameters.fetch :option_gen_faces, true
	@guide_mode = @hsh_parameters.fetch :guide_mode, false
	@guide_mode_line = @hsh_parameters.fetch :guide_mode_line, true
	@guide_mode_point = @hsh_parameters.fetch :guide_mode_point, true
end

#PARAMETER: Save the parameters as a model attribute
def parameter_save
	@otpicker.option_update_hash(@hsh_parameters)
	@hsh_parameters[:edge_prop] = @edge_prop 
	@hsh_parameters[:option_gen_faces] = @option_gen_faces 
	@hsh_parameters[:guide_mode] = @guide_mode 
	@hsh_parameters[:guide_mode_line] = @guide_mode_line
	@hsh_parameters[:guide_mode_point] = @guide_mode_point

	@@hsh_parameters_persist.update @hsh_parameters
	sparam = @@hsh_parameters_persist.inspect.gsub('"', "'")
	Traductor::Default.store @dico_name, @dico_attr, sparam
end

#----------------------------------------------------------------------------------------------------
# EDGE_PROP: Manage Edge Properties
#----------------------------------------------------------------------------------------------------

#EDGE_PROP: Insert a flag in the Edge_prop property
def edge_prop_insert(c)
	if c == 'P'
		@edge_prop = (@edge_prop == 'P') ? 'SM' : 'P'
	elsif c == 'D'
		@edge_prop = (@edge_prop == 'D') ? 'P' : 'D'
	elsif @edge_prop.include?(c)
		@edge_prop = @edge_prop.sub(c, '')
	else
		@edge_prop = @edge_prop.sub('P', '')
		@edge_prop = @edge_prop.sub('D', '')
		@edge_prop += c
	end
	@edge_prop = 'P' if @edge_prop == ''
	edge_prop_set @edge_prop
end

#EDGE_PROP: Set the Edge properties
def edge_prop_set(edge_prop)
	@guide_mode = false
	@edge_prop = edge_prop
	guide_edge_after
	onMouseMove_zero
end

#EDGE_PROP: Toggle across edge properties
def edge_prop_toggle
	return guide_toggle if @guide_mode
		
	if @edge_prop == 'P'
		edge_prop_set 'SM'
	elsif @edge_prop.include?('S') && @edge_prop.include?('M')
		edge_prop_set 'S'
	elsif @edge_prop == 'S'
		edge_prop_set 'D'
	else
		edge_prop_set 'P'
	end
end

#----------------------------------------------------------------------------------------------------
# GUIDE: Manage the Guide flags
#----------------------------------------------------------------------------------------------------

#GUIDE: Toggle Guide mode versus Edge mode
def guide_toggle
	@guide_mode = !@guide_mode
	guide_edge_after
end

#GUIDE: Toggle Guide Line mode
def guide_toggle_line
	@guide_mode = true
	@guide_mode_line = !@guide_mode_line
	@guide_mode_point = true unless @guide_mode_line
	guide_edge_after
end

#GUIDE: Force Guide Line mode alone
def guide_set_line
	@guide_mode = true
	@guide_mode_line = true
	@guide_mode_point = false
	guide_edge_after
end

#GUIDE: Toggle Guide Point mode
def guide_toggle_point
	@guide_mode = true
	@guide_mode_point = !@guide_mode_point
	@guide_mode_line = true unless @guide_mode_point
	guide_edge_after
end

#GUIDE: Toggle Guide Point mode
def guide_set_point
	@guide_mode = true
	@guide_mode_point = true
	@guide_mode_line = false
	guide_edge_after
end

#GUIDE: Setting the drawing parameters depending on Guide or Edge mode
def guide_edge_after
	if @guide_mode
		stipple = '_'
		if @guide_mode_line
			width = 2
			color = 'black'
		else
			width = 1
			color = 'gray'
		end
	else
		width = 2
		color = 'black'
		if @edge_prop == 'P'
			stipple = ''
		else
			stipple = '-'
		end
	end
	@otpicker.set_stipple_color_width stipple, color, width	
	@view.invalidate	
end

#GENERATE FACE: Toggle generation of faces
def generate_faces_toggle
	@option_gen_faces = !@option_gen_faces
end
	
#----------------------------------------------------------------------------------------------------
# NOTIFY: Call back from Otpicker
#----------------------------------------------------------------------------------------------------

def notify_from_otpicker(event)
	case event
	when :shift_down
		@keyman.shift_down?
	when :ctrl_down	
		@keyman.ctrl_down?
	when :onMouseMove_zero
		onMouseMove_zero
	else
		:no_event
	end	
end
	
#----------------------------------------------------------------------------------------------------
# PROTRACTOR: Manage the Protractor
#----------------------------------------------------------------------------------------------------

#PROTRACTOR: Toggle the Protractor mode
def protractor_toggle
	@protractor_mode = !@protractor_mode
	protractor_manage_after
end

#PROTRACTOR: Set the Protractor mode
def protractor_set(protractor_mode)
	@protractor_mode = protractor_mode
	protractor_manage_after
end

#PROTRACTOR: Tasks after changing the Protractor mode flag
def protractor_manage_after
	@otpicker.protractor_enable @protractor_mode
	onMouseMove_zero
end

#----------------------------------------------------------------------------------------------------
# ACTIVATION: Plugin Activation / Deactivation
#----------------------------------------------------------------------------------------------------

#ACTIVATION: Tool activation
def activate
	LibFredo6.register_ruby "FredoTools::DrawAlong"
	Traductor::Hilitor.stop_all_active
			
	#Initializing the Executor of Operation
	hsh = { :title => @menutitle, :palette => @palette, :no_commit => true }
	@suops = Traductor::SUOperation.new hsh

	@otpicker.protractor_enable @protractor_mode

	#Initializing the environment
	reset_env
	
	#Refreshing the view
	refresh_viewport
end

#ACTIVATION: Deactivation of the tool
def deactivate(view)
	parameter_save
	
	if @aborting && @geometry_generated
		text = T6[:T_MSG_ConfirmAbortText] + "\n\n" + T6[:T_MSG_ConfirmAbortQuestion]
		if UI.messagebox(text, MB_YESNO) != 6
			@suops.abort_operation
		end	
	elsif @geometry_generated
		commit_operation
	end	
	view.invalidate
	@view.remove_observer self
end

#ACTIVATION: Reset the picking environment
def reset_env
	@mode = @otpicker.set_mode(:origin)
	@pt_origin = @pt_target = nil
	@otpicker.reset
	@dragging = false
	@bb_extents = Geom::BoundingBox.new.add @model.bounds
end	

#ACTIVATION
def inference_reset
	@otpicker.inference_reset
	refresh_viewport
end

#ACTIVATION: Exit the tool
def exit_tool
	@aborting = false
	@model.select_tool nil
end

#ACTIVATION: Abort the operation and exit
def abort_tool
	@aborting = true
	@model.select_tool nil
end

#ACTIVATION: Activate the native SU Line tool
def native_tool
	Sketchup.send_action "selectLineTool:"
end

def toogle_draw_hidden
	@rendering_options["DrawHidden"] = !@rendering_options["DrawHidden"]
end
	
#--------------------------------------------------------------
# VIEWPORT: View Management
#--------------------------------------------------------------

#VIEWPORT: Computing Current cursor
def onSetCursor
	#Palette cursors
	ic = super
	return (ic != 0) if ic
			
	#Other cursors depending on state
	if @guide_mode
		if @guide_mode_line
			id_cursor = (@otpicker && @otpicker.no_inference?) ? @id_cursor_cline_noinf : @id_cursor_cline
		else		
			id_cursor = (@otpicker && @otpicker.no_inference?) ? @id_cursor_cpoint_noinf : @id_cursor_cpoint
		end	
	else
		id_cursor = (@otpicker && @otpicker.no_inference?) ? @id_cursor_edge_noinf : @id_cursor_edge
	end	
	UI.set_cursor id_cursor
end

#VIEWPORT: Refresh the viewport
def refresh_viewport
	onSetCursor
	show_message
	@view.invalidate
end
	
#VIEWPORT: Show message in the palette
def show_message
	#Mouse if either in the palette or out of the viewport
	if @mouseOut
		@palette.set_message nil
		return
	elsif @mouse_in_palette
		@palette.set_message @palette.get_main_tooltip, 'aliceblue'
		return
	end	
	
	show_message_pencil
end

#VIEWPORT: Show message for Move
def show_message_pencil
	#Displaying the Offset or Distance
	tx_lab, tx_len = @otpicker.vcb_label_length
	Sketchup.set_status_text tx_lab, SB_VCB_LABEL
	Sketchup.set_status_text tx_len, SB_VCB_VALUE

	#Main status bar - Help text
	text = (@mode == :origin) ? @msg_origin : @msg_target
	Sketchup.set_status_text text
	
	#Displaying tooltip in the palette
	tip_view = @otpicker.get_view_tooltip
	tip_dir = @otpicker.direction_tooltip
	tip = tip_view
	if tip_dir && tip_dir != tip_view
		if tip
			tip += " - " + tip_dir
		else
			tip = tip_dir
		end
	end
	code = 'lightgrey'
	if !tip_view && @mode == :origin
		tip = @msg_doubleclick_exit
		code = 'palegreen'
	elsif tip_dir
		code = 'lightblue'
	else
		code = 'lightyellow'
	end	
	tip = '' unless tip
	@palette.set_message tip, code
end

#---------------------------------------------------------------------------------------------
# MOVE: Mouse Move Methods
#---------------------------------------------------------------------------------------------

#MOVE: Mouse Movements
def onMouseMove_zero(flag=nil) ; onMouseMove(@flags, @xmove, @ymove, @view) if @xmove ; end
def onMouseMove(flags, x, y, view)
	#Event for the palette
	@xmove = x
	@ymove = y
	@flags = flags
	
	#Mouse in palette
	if super
		@mouse_in_palette = true
		refresh_viewport
		return	
	end
	@mouse_in_palette = false
			
	return if @moving
	
	#Tracking the origin with possible auto-selection
	if @mode == :origin
		#Pick origin with inference
		@origin_info = @otpicker.origin_pick view, x, y, false, @keyman.shift_down?, true
		@pt_origin, @type_origin = @origin_info
		@bb_extents.add @pt_origin if @pt_origin
		
	#Tracking the target	
	elsif @mode == :target
		@target_info = @otpicker.target_pick(view, x, y, false)
		@pt_target, = @target_info
		pencil_dragging if @pt_origin != @pt_target

		#Updating the Bounding box for drawing
		@bb_extents.add @pt_target if @pt_target		
	end
	
	#Refreshing the view
	@moving = true
	refresh_viewport
end	

#MOVE: Mouse leaves the viewport
def onMouseLeave(view)
	@mouseOut = true
	view.invalidate
end

#MOVE: Mouse enters the viewport
def onMouseEnter(view)
	@mouseOut = false
	view.invalidate
end
	
#MOVE: Before Zooming or changing view
def suspend(view)
	@keyman.reset
	@zoom_void.suspend(view, @xmove, @ymove)
end

#MOVE: After changing view
def resume(view)
	@zoom_void.resume(view)
	refresh_viewport
end
	
#VIEW: Notification of view changed
def onViewChanged(view)

end

#---------------------------------------------------------------------------------------------
# VCB: VCB Management
#---------------------------------------------------------------------------------------------

#VCB: Enable or disable the VCB
def enableVCB?
	true
end

#VCB: Handle VCB input
def onUserText(text, view)
	@vcb = Traductor::VCB.new
	
	remote = @otpicker.vcb_origin_remote?
	
	if @mode == :origin && remote
		@vcb.declare_input_format :distance, "l"			#offset length		
	else
		@vcb.declare_input_format :offset, "l"				#offset length
		@vcb.declare_input_format :multiple, "i__x"			#multiple copy
		@vcb.declare_input_format :multiple2, "star"		#multiple copy
		@vcb.declare_input_format :divide, "slash"			#divide copy
		if @otpicker.protractor_shown?
			@vcb.declare_input_format :angle, "a"			#angle round for round mode
		end	
	end
	
	nb_errors = @vcb.process_parsing(text)
	
	if nb_errors > 0
		lmsg = []
		@vcb.each_error { |symb, s| lmsg.push "<#{s}>" }
		msg = "#{T6[:T_ERROR_InVCB]} \"#{text}\"  --> #{lmsg.join(' - ')}"
		@palette.set_message msg, 'E'
		view.invalidate
	else
		action_from_VCB
	end	
end

#VCB: Execute actions from the VCB inputs	
def action_from_VCB
	angle = multiple = divide = distance = nil
	loffset = []
	@vcb.each_result do |symb, val, suffix|
		case symb
		when :distance
			distance = val
		when :offset
			loffset.push val
		when :multiple, :multiple2
			multiple = val
		when :divide
			multiple = val
			divide = true
		when :angle
			angle = val
		end
	end
	
	#Changing the angle if protractor is on
	if angle
		@otpicker.protractor_force_angle angle.degrees
	end
	
	#Changing the remote distance and place origin
	if distance
		pencil_change_from_remote distance
		return
	end
			
	#Changing the offset and multiple
	noffset = loffset.length
	if noffset > 0 || multiple
		if noffset <= 1
			pencil_post_execute true, loffset[0], multiple, divide
		else
			x, y, z = loffset
			x = 0 unless x
			y = 0 unless y
			z = 0 unless z
			new_pt = Geom::Point3d.new x, y, z
			if @mode == :origin
				pencil_change_origin new_pt
			else	
				pencil_change_target new_pt
			end	
		end	
	end	
end	

#---------------------------------------------------------------------------------------------
# SU TOOL: Click Management
#---------------------------------------------------------------------------------------------

#SU TOOL: Button click DOWN
def onLButtonDown(flags, x, y, view)
	#Palette management
	if super
		refresh_viewport
		return
	end
	
	#Storing the reference
	@button_down = true
	@xdown = x
	@ydown = y	
	
	#Changing mode
	if @mode == :origin
		onMouseMove(0, x, y, view) unless @pt_origin
		action_start
	elsif @mode == :target
		action_stop true
	end	
	@time_down_start = Time.now
end

#SU TOOL: Button click UP - Means that we end the selection
def onLButtonUp(flags, x, y, view)
	#Palette buttons
	if super
		@time_down_start = nil
		onMouseMove_zero
		return
	end
	return unless @button_down
	@button_down = false
	
	#Logic for dragging
	far_from_origin = (@xdown - x).abs > 3 || (@ydown - y).abs > 3
	if @time_down_start && (Time.now - @time_down_start) > @duration_long_click && !far_from_origin
		reset_env
		onMouseMove_zero
		@otpicker.direction_pick_from_model
	elsif @dragging
		if @time_down_start && (Time.now - @time_down_start) > 0.3 && ((@xdown - x).abs > 1 || (@ydown - y).abs > 1)
			action_stop
		end
	end	
end

#SU TOOL: Double Click received
def onLButtonDoubleClick(flags, x, y, view)
	return if super
	if @mode == :target && @pt_target && @pt_origin == @pt_target
		reset_env
	else
		if @type_origin == :void_origin
			exit_tool
		else
			pencil_redo
		end	
	end	
end

#---------------------------------------------------------------------------------------------
# ACTION: Control of the Movement or Rotation
#---------------------------------------------------------------------------------------------

#ACTION: Start either Move or Rotate
def action_start
	pencil_start
	@time_action_start = Time.now
end

#ACTION: Stop either Move or Rotate
def action_stop(pursue=false)
	pursue = action_pursue? if pursue
	pencil_finish
	pencil_process pursue
end

#ACTION: Check if we can chain the inputs for lines
def action_pursue?
	return false if @otpicker.direction_forced?
	
	pt, type_ori, info_comp = @origin_info
	elt_ori, comp, = info_comp
	return true if comp
	
	pt, type_targ, info_comp = @target_info
	elt_targ, comp, = info_comp
	return true if comp
	
	cond_ori = (elt_ori.instance_of?(Sketchup::Vertex) || (elt_ori.instance_of?(Sketchup::Edge) && elt_ori.faces.length > 0))
	cond_targ = (elt_targ.instance_of?(Sketchup::Vertex) || (elt_targ.instance_of?(Sketchup::Edge) && elt_targ.faces.length > 0))
	return false if cond_ori && cond_targ
	true
end

#---------------------------------------------------------------------------------------------
# PENCIL: Manage creation of Line
#---------------------------------------------------------------------------------------------

#PENCIL: Start of drawing
def pencil_start
	@multiple = @divide = nil
	
	#Initializing the move environment
	pencil_prepare
	
	refresh_viewport
end

#PENCIL: During the dragging process
def pencil_dragging
	@action_stopped = false
	@dragging = true
end

#PENCIL: Prepare the interactive creation of line
def pencil_prepare
	@mode = @otpicker.set_mode(:target)
end

#PENCIL: Cancel the current operation
def pencil_cancel
	@otpicker.vcb_restore_after_cancel
	reset_env
	@just_cancelled = true
	@action_stopped = true
	onMouseMove_zero
end

#PENCIL: Store current information
def pencil_finish
	if @pt_target && @pt_origin == @pt_target
		reset_env
		return
	end	
	@otpicker.vcb_freeze
	commit_operation
	@action_stopped = true
end

#PENCIL: Terminate the drawing and create the line
def pencil_process(pursue=false)
	#Edge has Zero length
	if !@pt_target || @pt_origin == @pt_target
		reset_env
		return
	end
	
	pursue = action_pursue? if pursue
	
	#Starting the operation
	start_operation
	
	#Creating the edges or the construction lines
	new_origin_info = (@guide_mode) ? pencil_create_guides(pursue) : pencil_create_edges(pursue)

	#Continuing or stopping the interactive input
	if new_origin_info
		vec = @pt_origin.vector_to @pt_target
		@otpicker.protractor_set_direction(vec)
		@origin_info = @otpicker.origin_set_info new_origin_info
		@pt_origin, = @origin_info
		pencil_prepare
	else	
		reset_env
	end	
	
	#Refreshing the view
	@geometry_generated = true
	refresh_viewport
end

#PENCIL: Continue from the last origin
def pencil_post_execute(pursue, offset, multiple=nil, divide=nil, new_vec=nil, new_origin=nil)
	@multiple = multiple if multiple
	@divide = divide if divide
	
	#Finishing if not already done
	pencil_finish if !@action_stopped
	
	#Getting the applicable origin and target
	@pt_origin, @pt_target = @otpicker.vcb_change_specifications(offset, new_origin, new_vec)
	
	#Not valid post-execution - Resetting the previous environment
	if !@pt_origin || !@pt_target || @pt_origin == @pt_target
		reset_env
		return
	end
		
	#Cancelling and restarting operation if vector has not changed
	if @mode == :origin && @just_cancelled
		Sketchup.undo
	else
		abort_operation
	end	
	@just_cancelled = false
	
	#Performing the transformation
	pencil_prepare
	pencil_process pursue
	
	#Storing the multiple and divide flags
	@multiple_prev = @multiple
	@divide_prev = @divide	
	show_message
end

#PENCIL: Redo the same line specs on another origin
def pencil_redo
	commit_operation
	pencil_post_execute false, nil, @multiple_prev, @divide_prev, nil, @pt_origin
end

#PENCIL: Change the origin from remote point
def pencil_change_from_remote(distance)
	new_ori, targ = @otpicker.vcb_change_specifications(distance)
	
	@origin_info = @otpicker.origin_set_info [new_ori, :remote, nil]
	@pt_origin, = new_ori
	pencil_start
	onMouseMove_zero
end

#PENCIL: Create the line geometry
def pencil_create_edges(pursue=false)
	#Creating the line and finding potential faces
	vec_draw = @pt_origin.vector_to @pt_target
	ptend = @pt_target
	ptbeg = @pt_origin
	offset = vec_draw.length
	nseg = (@multiple) ? @multiple.abs : 1
	range = 1..nseg
	offset = offset / nseg if @divide
	g = @entities.add_group
	for i in range
		pt = ptbeg.offset vec_draw, offset
		g.entities.add_line ptbeg, pt
		ptbeg = pt	
	end
	if @multiple && @multiple < 0
		ptbeg = @pt_origin
		for i in range
			pt = ptbeg.offset vec_draw, -offset
			g.entities.add_line ptbeg, pt
			ptbeg = pt	
		end
	end
	
	#Generating faces for the new created edges
	le = G6.grouponent_explode(g)
	edges = le.grep(Sketchup::Edge)
	nfaces = 0
	edges.each { |e| nfaces += e.find_faces } if @option_gen_faces
	pursue = false if nfaces > 0 || nseg > 1
	
	#Assigning Edge properties
	smooth = @edge_prop.include?('M')
	soft = @edge_prop.include?('S')
	hidden = @edge_prop.include?('H')
	edges.each do |e|
		e.soft = true if soft
		e.smooth = true if smooth
		e.hidden = true if hidden
	end	
	
	#Continuing from last target or Resetting the environment
	vxend = nil
	if pursue
		edges.each do |e| 
			vxend = [e.start, e.end].find { |vx| vx.position == ptend }
			break if vxend
		end	
	end

	(vxend) ? [@pt_target, :vertex, [vxend, nil, @tr_id]] : nil
end

#PENCIL: Create the line geometry
def pencil_create_guides(pursue=false)
	#Creating the line and finding potential faces
	vec_draw = @pt_origin.vector_to @pt_target
	ptend = @pt_target
	ptbeg = @pt_origin
	offset = vec_draw.length
	nseg = (@multiple) ? @multiple.abs : 1
	range = 1..nseg
	offset = offset / nseg if @divide
	last_cline = last_pt = nil
	g = @entities.add_group
	gentities = g.entities
	pencil_create_cpoint(gentities, ptbeg)
	for i in range
		pt = ptbeg.offset vec_draw, offset
		last_cline = pencil_create_cline(gentities, ptbeg, pt)
		last_cpoint = pencil_create_cpoint(gentities, pt)
		ptbeg = last_pt = pt	
	end
	if @multiple && @multiple < 0
		ptbeg = @pt_origin
		for i in range
			pt = ptbeg.offset vec_draw, -offset
			pencil_create_cline(gentities, ptbeg, pt)
			pencil_create_cpoint(gentities, pt)
			ptbeg = pt	
		end
	end
	
	#Generating faces for the new created edges
	G6.grouponent_explode(g)
	
	#Returning information for next guide
	return nil unless pursue
	(last_cline) ? [last_pt, :end_cline, [last_cline, nil, @tr_id]] : [last_pt, :cpoint, [last_cpoint, nil, @tr_id]]
end

#PENCIL: Create a Construction point
def pencil_create_cpoint(entities, pt)
	return unless @guide_mode_point
	@lcpoints = @entities.grep(Sketchup::ConstructionPoint) unless @lcpoints
	return if @lcpoints.find { |cpoint| cpoint.position == pt }
	cp = entities.add_cpoint pt
	@lcpoints.push cp
	cp
end

#PENCIL: Create a Construction point
def pencil_create_cline(entities, pt1, pt2)
	return unless @guide_mode_line
	entities.add_cline pt1, pt2
end

#PENCIL: Change the origin by full point specifications
def pencil_change_origin(new_ori)
	@mode = @otpicker.set_mode :origin
	@origin_info = @otpicker.origin_set_info [new_ori, :point, nil]
	@pt_origin, = new_ori
	pencil_start
	onMouseMove_zero
end

#PENCIL: Change the origin by full point specifications
def pencil_change_target(new_target)
	@mode = @otpicker.set_mode :target
	@target_info = @otpicker.target_set_info [new_target, :point, nil]
	@pt_target, = @target_info
	pencil_process
end

#---------------------------------------------------------------------------------------------
# KEY: Keyboard handling
#---------------------------------------------------------------------------------------------

#KEY: Return key pressed
def onReturn(view)
	@otpicker.direction_pick_from_model
	refresh_viewport
end

#KEY: Manage key events
def key_manage(event, key, view)
	case event
	
	#SHIFT key
	when :shift_down
		@otpicker.direction_changed
	when :shift_toggle
		if @mode == :origin
			@otpicker.direction_pick_from_model
		else	
			@otpicker.direction_toggle_lock
		end	
	when :shift_up
		@otpicker.direction_changed
		
	#CTRL key
	when :ctrl_down	
		@protractor_mode_at_down = @protractor_mode
		protractor_toggle
	when :ctrl_up
		protractor_set @protractor_mode_at_down
	when :ctrl_other_up
		protractor_set @protractor_mode_at_down
	
	#ALT key
	when :alt_down
		@otpicker.no_inference_toggle
	when :alt_up	
		@otpicker.no_inference_set false
		
	#Arrows	
	when :arrow_down	
		@otpicker.option_from_arrow key, @keyman.shift_down?, @keyman.ctrl_down?
		
	#Backspace	
	when :backspace
		@otpicker.inference_reset

	#TAB	
	when :tab
		edge_prop_toggle
	end
end

#KEY: Handle Key down
def onKeyDown(key, rpt, flags, view)
	ret_val = @keyman.onKeyDown(key, rpt, flags, view)
	onSetCursor
	@view.invalidate
	ret_val
end

#KEY: Handle Key up
def onKeyUp(key, rpt, flags, view)
	ret_val = @keyman.onKeyUp(key, rpt, flags, view)
	onSetCursor
	@view.invalidate
	ret_val
end

#---------------------------------------------------------------------------------------
# UNDO: Undo and Rollback Management
#---------------------------------------------------------------------------------------

#UNDO: Cancel and undo methods
def onCancel(flag, view)
	case flag
	when 1, 2	#Undo or reselect the tool
		handle_undo
	when 0		#user pressed Escape	
		handle_escape
	end
end

#UNDO: Handle an undo
def handle_undo
	reset_env
	UI.start_timer(0.1) { rollback }
end

#UNDO: Handle the escape key depending on current mode
def handle_escape
	if @otpicker.remote_active_dir
		@otpicker.remote_reset
	elsif @dragging
		pencil_cancel
		@otpicker.direction_cancel_all
	elsif @mode == :origin
		@otpicker.direction_cancel_all
	end	
end

#UNDO: Execute a rollback
def rollback(perform_undo=false)
	if perform_undo
		Sketchup.undo unless @geometry_generated
	end	
	@geometry_generated = false
	reset_env
	onMouseMove_zero
end
		
#---------------------------------------------------------------------------------------------
# DRAW: Drawing Methods
#---------------------------------------------------------------------------------------------

#DRAW: Get the extents for drawing in the viewport
def getExtents
	@bb_extents
end

#DRAW: Draw top method
def draw(view)	
	@otpicker.draw_all view
	
	#Flag for fluidity
	@moving = false
	
	#Drawing the palette
	super
end

#--------------------------------------------------
# MENU: Contextual menu
#--------------------------------------------------

#MENU: Contextual menu
def getMenu(menu)
	cxmenu = Traductor::ContextMenu.new

	#Privileged directions
	cxmenu.add_sepa
	@otpicker.menu_contribution cxmenu
	
	#Protractor modes
	cxmenu.add_sepa
	cxmenu.add_item(@tip_protractor_mode) { protractor_toggle }
	
	#Enter same offset
	if @otpicker.vcb_redo_possible?
		cxmenu.add_sepa
		cxmenu.add_item(@tip_pencil_redo) { pencil_start ; pencil_redo }	
	end
	
	#Exit and Abort
	cxmenu.add_sepa
	cxmenu.add_item(@mnu_abort) { exit_tool }
	cxmenu.add_item(@mnu_exit) { abort_tool }
	
	#Showing the menu
	cxmenu.show menu
	true
end

#============================================================
# OPERATION: handling SU operation
#============================================================
	
#OPERATION: Start an operation based on symbol	
def start_operation
	@suops.start_operation nil, false
end

#OPERATION: Commit and save the fixing
def commit_operation
	return unless @geometry_generated
	@suops.commit_operation
	@geometry_generated = false
end

#OPERATION: Abort the operation
def abort_operation
	@suops.abort_operation
end

#-------------------------------------------------------------- 
#--------------------------------------------------------------
# Palette Management
#--------------------------------------------------------------
#--------------------------------------------------------------

#PALETTE: Separator for the Main and floating palette
def pal_separator ;	@palette.declare_separator ; end

#PALETTE: Initialize the main palette and top level method
def init_palette
	hshpal = { :width_message => 0, :width_message_min => 150, :key_registry => :toposhaper }
	@palette = Traductor::Palette.new hshpal
	@draw_local = self.method "draw_button_opengl"
	@blason_proc = self.method "draw_button_blason"
	@symb_blason = :blason
	@bk_color_tool = "lemonchiffon"
	@info_text1 = ""
	@info_text2 = ""
	@info_message = ""
		
	#Blason Button
	tip = T7[:PlugName] + ' ' + T6[:T_STR_DefaultParamDialog]
	hsh = { :blason => true, :draw_proc => @blason_proc, :tooltip => tip }
	@palette.declare_button(@symb_blason, hsh) { MYPLUGIN.invoke_default_parameter_dialog }
	
	#Plane and Vector settings
	pal_separator
	@otpicker.palette_contribution_planar @palette
	pal_separator
	@otpicker.palette_contribution_vector @palette
	
	#Edge properties and Protractor
	palette_edge_prop
	palette_guide
	palette_protractor
	palette_options
	palette_actions
	palette_native
	
	#Abort and Exit
	pal_separator
	hsh = { :draw_proc => :std_abortexit, :main_color => 'red', :frame_color => 'green', 
	        :tooltip => T6[:T_STR_AbortTool] }
	@palette.declare_button(:pal_abort, hsh) { abort_tool }

	#Rollback
	tip = T6[:T_STR_Undo] + " [#{T6[:T_KEY_CTRL]}-Z]"
	hsh = { :tooltip => tip, :draw_proc => :rollback }
	@palette.declare_button(:pal_back, hsh) { rollback true }
	
	#Exit tool
	hsh = { :draw_proc => :std_exit, :tooltip => T6[:T_STR_ExitTool] }
	@palette.declare_button(:pal_exit, hsh) { exit_tool }
		
	#Associating the palette
	set_palette @palette
end

#PALETTE: Button set for edge property control
def palette_edge_prop
	pal_separator
	
	#Title button
	symb_master = :pal_edge_prop
	tip = T6[:T_TIP_EdgePropertieCreated]
	value_proc = proc { !@guide_mode }
	double_proc = proc { edge_prop_set 'P' }
	hsh = { :type => 'multi_free', :text => T6[:T_BOX_EdgePropertie], :tooltip => tip, :bk_color => @color_but_title,
            :value_proc => value_proc, :hi_color => @color_but_hi, :double_click_proc => double_proc }
	@palette.declare_button(symb_master, hsh) { edge_prop_toggle }
	
	hshp = { :parent => symb_master, :width => 27, :hi_color => @color_but_hi }
	list = []
	list.push ['P', T6[:T_DLG_EdgePlain], :edge_prop_plain]
	list.push [:sepa0]
	list.push ['S', T6[:T_DLG_EdgeSoft], :edge_prop_soft]
	list.push ['M', T6[:T_DLG_EdgeSmooth], :edge_prop_smooth]
	list.push ['H', T6[:T_DLG_EdgeHidden], :edge_prop_hidden]
	list.push [:sepa1]
	list.push ['D', T6[:T_DLG_EdgeDiagonal], :edge_prop_diagonal]
	list.each do |ll|
		code, tip, sdraw = ll
		symb = "#{symb_master}_#{code}".intern
		if sdraw
			hsh = { :draw_proc => sdraw, :tooltip => tip }
			@palette.declare_button symb, hshp, hsh, palette_edge_prop_button_proc(code)
		else	
			hsh = { :draw_proc => :separator_V, :passive => true, :width => 8 }
			@palette.declare_button symb, hshp, hsh
		end	
	end	
end

#PALETTE: Compute the proc for buttons related to Edge properties
def palette_edge_prop_button_proc(code)
	vproc = proc { @edge_prop.include?(code) }
	aproc = proc { edge_prop_insert(code) }
	doubleproc = proc { edge_prop_set(code) }
	{ :value_proc => vproc, :action_proc => aproc, :double_click_proc => doubleproc }
end

#PALETTE: Button for Guide line generation
def palette_guide
	pal_separator
	
	#Title button for Options
	symb_master = :pal_guide_mode
	value_proc = proc { @guide_mode }
	hsh = { :type => 'multi_free', :value_proc => value_proc, :text => T6[:T_BOX_Guide], :tooltip => T7[:TIP_GuideMode],
	        :hi_color => @color_but_hi, :bk_color => @color_but_title }
	@palette.declare_button(symb_master, hsh) { guide_toggle }
	
	hshp = { :parent => symb_master, :width => 28, :hi_color => @color_but_hi }

	#Guide Lines
	symb = "#{symb_master}_Line".intern
	value_proc = proc { @guide_mode_line }
	double_proc = proc { guide_set_line }
	hsh = { :value_proc => value_proc, :draw_proc => @draw_local, :tooltip => T6[:T_TIP_GuideLines],
            :double_click_proc => double_proc }
	@palette.declare_button(symb, hshp, hsh) { guide_toggle_line }
	
	#Guide Points
	symb = "#{symb_master}_Point".intern
	value_proc = proc { @guide_mode_point }
	double_proc = proc { guide_set_point }
	hsh = { :value_proc => value_proc, :draw_proc => @draw_local, :tooltip => T6[:T_TIP_GuidePoints],
            :double_click_proc => double_proc }
	@palette.declare_button(symb, hshp, hsh) { guide_toggle_point }
end
	
#PALETTE: Button for native SU Draw tool	
def palette_native
	pal_separator
	hsh = { :tooltip => T7[:TIP_NativeDraw], :draw_proc => :std_edition_pencil, :main_color => 'firebrick' }
	@palette.declare_button(:pal_native, hsh) { native_tool }
end	

#PALETTE: Button for Actions
def palette_options
	pal_separator
	
	#Title button for Options
	symb_master = :pal_options
	hsh = { :type => 'multi_free', :passive => true, :text => T6[:T_BOX_Options], :tooltip => T6[:T_TIP_Options],
	        :bk_color => @color_but_title }
	@palette.declare_button(symb_master, hsh)
	
	hshp = { :parent => symb_master, :width => 28, :hi_color => @color_but_hi }

	#Generate Faces
	symb = "#{symb_master}_GenFace".intern
	value_proc = proc { @option_gen_faces }
	hsh = { :value_proc => value_proc, :draw_proc => :std_generate_face, :tooltip => T7[:TIP_GenFaceToggle] }
	@palette.declare_button(symb, hshp, hsh) { generate_faces_toggle }
	
	#Remote Inferences
	@otpicker.palette_contribute_button(:inference_remote, @palette, symb_master, hshp)
	
	#Inferences on Components
	@otpicker.palette_contribute_button(:inference_on_comp, @palette, symb_master, hshp)
	
	#No Inferences
	@otpicker.palette_contribute_button(:no_inference, @palette, symb_master, hshp)	
end
	
#PALETTE: Button for Actions
def palette_actions
	pal_separator

	#Title button for Actions
	symb_master = :pal_actions
	hsh = { :type => 'multi_free', :passive => true, :text => T6[:T_BOX_Actions], :tooltip => T6[:T_TIP_Actions],
	        :bk_color => @color_but_title }
	@palette.declare_button(symb_master, hsh)
	
	hshp = { :parent => symb_master, :width => 28, :hi_color => @color_but_hi }

	#Toggle Hidden geometry
	@otpicker.palette_contribute_button(:toggle_show_hidden, @palette, symb_master, hshp)
	
	#Erase all inference marks
	@otpicker.palette_contribute_button(:reset_all_marks, @palette, symb_master, hshp)
end

#PALETTE: Button for Protractor mode
def palette_protractor
	pal_separator
	hgt = 32
	wid = hgt

	val_proc = proc { @protractor_mode }
	hsh = { :tooltip => @tip_protractor_mode, :height => hgt, :width => wid, :draw_proc => :std_protractor,
			:hi_color => @color_but_hi, :value_proc => val_proc }
	@palette.declare_button(:pal_protractor_mode, hsh) { protractor_toggle }
end
	
#PALETTE: Drawing the Blason
def draw_button_blason(symb, dx, dy)
	lst_gl = []
	dx2 = dx/2
	dy2 = dy/2
	
	lpti = [[1, dy2], [dx2, dy-1], [dx-1, dy2], [dx2, 1]]
	pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
	lst_gl.push [GL_POLYGON, pts, 'lightyellow']

	dec = 2
	lpti = [[1, dy2+dec], [dx2-dec, dy-1], [dx2+dec, dy-1], [dx-1, dy2+dec], [dx-1, dy2-dec], [dx2+dec, 1], [dx2-dec, 1], [1, dy2-dec]]
	pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
	lst_gl.push [GL_LINES, pts, 'lightblue', 1, '']	

	#Drawing the pencil
	maincolor = 'blue'
	ptxy = Geom::Point3d.new dx2, dy2
	decx = 2
	decy = 4
	y1 = decy
	lpti = [[dx2, 0], [dx2-decx, y1], [dx2+decx, y1]]
	pts_pointe = lpti.collect { |a| Geom::Point3d.new(*a) }
	y2 = y1 + 4
	lpti = [[dx2-decx, y1], [dx2+decx, y1], [dx2+decx, y2], [dx2-decx, y2]]
	pts_mine = lpti.collect { |a| Geom::Point3d.new(*a) }
	y3 = dy-decy
	lpti = [[dx2-decx, y2], [dx2+decx, y2], [dx2+decx, y3], [dx2-decx, y3]]
	pts_tube = lpti.collect { |a| Geom::Point3d.new(*a) }
	y4 = dy-1
	lpti = [[dx2-decx, y3], [dx2+decx, y3], [dx2+decx, y4], [dx2-decx, y4]]
	pts_top = lpti.collect { |a| Geom::Point3d.new(*a) }

	t = Geom::Transformation.rotation ptxy, Z_AXIS, -45.degrees
	pts_pointe = pts_pointe.collect { |pt| t * pt }
	pts_mine = pts_mine.collect { |pt| t * pt }
	pts_tube = pts_tube.collect { |pt| t * pt }
	pts_top = pts_top.collect { |pt| t * pt }

	lst_gl.push [GL_POLYGON, pts_pointe, 'black']
	lst_gl.push [GL_LINE_LOOP, pts_pointe, 'gray', 1, '']
	lst_gl.push [GL_POLYGON, pts_mine, 'beige']
	lst_gl.push [GL_LINE_LOOP, pts_mine, 'gray', 1, '']
	lst_gl.push [GL_POLYGON, pts_tube, maincolor]
	lst_gl.push [GL_LINE_LOOP, pts_tube, 'black', 1, '']
	lst_gl.push [GL_POLYGON, pts_top, 'beige']
	lst_gl.push [GL_LINE_LOOP, pts_top, 'black', 1, '']
	
	lst_gl
end

#PALETTE: Custom drawing of buttons
def draw_button_opengl(symb, dx, dy, main_color, frame_color, selected)
	code = symb.to_s
	lst_gl = []
	dx2 = dx / 2
	dy2 = dy / 2
	grayed = @palette.button_is_grayed?(symb)
	color = (grayed) ? 'gray' : frame_color
	
	case code

	when /pal_guide_mode_Line/
		lpti = [[1, 1], [dx-1, dy-1]]
		pts = lpti.collect { |a| Geom::Point3d.new(*a) }
		lst_gl.push [GL_LINE_STRIP, pts, 'black', 2, '_']		
	
	when /pal_guide_mode_Point/
		dec = 4
		lpti = [[dx2-dec, dy2], [dx2+dec, dy2], [dx2, dy2-dec], [dx2, dy2+dec]]
		pts = lpti.collect { |a| Geom::Point3d.new(*a) }
		lst_gl.push [GL_LINES, pts, 'black', 2, '']		
		
	end	#case code
	
	lst_gl
end

end	#class DrawAlongTool

end	#End Module DrawAlong

end	#End Module F6_FredoTools
