=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2013 Fredo6 - Designed and written Oct 2013 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_FredoTools__EdgeInspector.rb
# Original Date	:   15 Oct 2013 
# Description	:   Detect and correct defects in Edges - Works with a Glass magnifier
# IMPORTANT		:	DO NOT TRANSLATE STRINGS in the source code
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

module EdgeInspector
							   
#Texts for the module
T7[:MSG_NoOcurrence] = "NO Occurence of overlapping edges found"
T7[:MSG_ZeroLengthEdges] = "Very Small Edges = %1"
T7[:MSG_OccurencesNb] = "Overlapping Edges = %1"
T7[:MSG_Confirm] = "Do you wish to fix the model?"
T7[:TXT_PreparingDisplay] = "Preparing Display"
T7[:TXT_VisualElements] = "Visual Elements"

T7[:BOX_Original] = "Original"
T7[:BOX_BeforeFix] = "Current"
T7[:BOX_AfterFix] = "After Fix"

T7[:BOX_Inspect] = "INSPECT"
T7[:TIP_Inspect] = "Launch the Inspection of selection with current settings"
T7[:BOX_Fix] = "FIX"
T7[:TIP_Fix] = "Fix all defects calculated"
T7[:TIP_RollBack] = "Undo last repair"

T7[:TIT_FixingResults] = "EdgeInspector: Inspection and Fixing Status"
T7[:TIP_Reinspect] = "Inspect again after repairing"
T7[:BOX_TypeRepair] = "Type of defects"
T7[:TIP_TypeRepair] = "Type of defects and repairing"
T7[:BOX_Results] = "Results"
T7[:TIP_Results] = "Number of defects found after inspection"
T7[:TIP_PercentTolerance] = "Percentage of average length of edges"
T7[:TIP_TogglePercentLength] = "Toggle between percent of average length and absolute length value"

T7[:BOX_Repair_overlap] = "(Quasi) Overlap"
T7[:TIP_Repair_overlap] = "Overlapping and quasi-overlapping or disjoint edges"
T7[:PROMPT_Repair_overlap] = "Maximum distance between disjoint or overlapping edges"

T7[:BOX_Repair_lost_junction] = "Lost Junction"
T7[:TIP_Repair_lost_junction] = "Missing junction between extremities of standalone edges"
T7[:PROMPT_Repair_lost_junction] = "Maximum distance between extremities of edges"

T7[:BOX_Repair_zero_edge] = "Tiny Edge"
T7[:TIP_Repair_zero_edge] = "Edges with length lower than given threshold"
T7[:PROMPT_Repair_zero_edge] = "Maximum length of tiny edge"

T7[:BOX_Repair_tiny_gap] = "Tiny Gap"
T7[:TIP_Repair_tiny_gap] = "Vertices that are very close"
T7[:PROMPT_Repair_tiny_gap] = "Maximum distance between close vertices"

T7[:BOX_Repair_split_edge] = "Split Edge"
T7[:TIP_Repair_split_edge] = "Lonely vertices joining collinear edges"

T7[:BOX_Repair_coplanar_edge] = "Coplanar Edge"
T7[:TIP_Repair_coplanar_edge] = "Coplanar edges allows merging faces when removed"
T7[:BOX_Repair_Flag_coplanar_edge] = "Same Mat"
T7[:TIP_Repair_Flag_coplanar_edge] = "Coplanar Edge erased if adjacent faces have the same material"

T7[:BOX_Repair_needle_eye] = "Needle Eye - Ear"
T7[:TIP_Repair_needle_eye] = "Interlaced edges following more or less the same path"

T7[:TIP_NbEdgesTitle] = "Edges in Selection"
T7[:TIP_NbFacesTitle] = "Faces in Selection"
T7[:TIP_NbEdgesOriginal] = "Original number of Edges"
T7[:TIP_NbFacesOriginal] = "Original number of Faces"
T7[:TIP_NbEdgesCurrent] = "Number of Edges BEFORE Fixing"
T7[:TIP_NbFacesCurrent] = "Number of Faces BEFORE Fixing"
T7[:TIP_NbEdgesFix] = "Number of Edges AFTER Fixing"
T7[:TIP_NbFacesFix] = "Number of Faces AFTER Fixing"

T7[:MSG_ConfirmAbortText] = "Some fixes have been performed"
T7[:MSG_ConfirmAbortQuestion] = "Do you wish to SAVE them before exiting"

T7[:MSG_Inspecting] = "Inspecting"
T7[:MSG_Repairing] = "Repairing"

T7[:MNU_ExitGlassMode] = "Exit Zoom Mode"

T7[:OPS_other] = "Others"
T7[:OPS_global_fixing] = "Global Fixing"
T7[:OPS_single_fixing] = "Single Fixing"
T6[:OPS_UndoBeyond] = "Undo beyond changes done in EdgeInspector"
T7[:OPS_DELKey] = "Erase Selection (DEL Key)"

T7[:ERR_DuringInspection] = "Error during Inspection"
T7[:ERR_DuringFixing] = "Error during Fixing"

T7[:MSG_TooSmall_Edge1] = "Too many edges [%1] are very small [< %2]"
T7[:MSG_TooSmall_Edge2] = "You should scale the selection by a factor > %1"

#====================================================================================================
#----------------------------------------------------------------------------------------------------
# Plugin Implementation
#----------------------------------------------------------------------------------------------------
#====================================================================================================

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

def self._execution(symb)
	Sketchup.active_model.select_tool EdgeInspectorTool.new
end

#----------------------------------------------------------------------------------------------------
# Reparation structure
#----------------------------------------------------------------------------------------------------

Reparation = Struct.new :rep_id, :lent_id, :lent, :type, :tr, :parent, :d, :resolver, :ignored, 
                        :jpts, :lst_fake_ids

#=============================================================================================
#=============================================================================================
# Class EdgeInspectorTool: main class for the Interactive tool
#=============================================================================================
#=============================================================================================

class EdgeInspectorTool < Traductor::PaletteSuperTool

#----------------------------------------------------------------------------------------------------
# INIT: Initialization
#----------------------------------------------------------------------------------------------------

#INIT: Class instance initialization
def initialize(*args)
	#Basic Initialization
	@tr_id = Geom::Transformation.new
	@model = Sketchup.active_model
	@selection = @model.selection
	@rendering_options = Sketchup.active_model.rendering_options
	@entities = @model.active_entities
	@view = @model.active_view
	@view_tracker = G6::ViewTracker.new
	@vtrack_last = nil
	@dico_name = "Fredo6_FredoTools_EdgeInspector"
	@dico_attr = "Parameters"
	@ip = Sketchup::InputPoint.new
	
	#Observers
	@model.selection.add_observer self
	@view.add_observer self

	@hsh_nb_repairs_text = {}
	@hsh_nb_repairs_color = {}
	@geometry_counting = 0
	@geometry_last_ops = []
	
	#Parsing the arguments
	args.each { |arg| arg.each { |key, value|  parse_args(key, value) } if arg.class == Hash }

	#Loading parameters
	parameter_load
	
	#Static initialization
	init_cursors
	init_colors
	init_messages

	#Initialize the magnifier
	init_magnifier
	
	#Creating the palette manager and texts
	init_palette
	
	MYDEFPARAM.add_notify_proc(self.method("notify_from_defparam"), true)	
end

#INIT: Notification from default parameters
def notify_from_defparam(key, val)
	@inspection_allowed = true
	@view.invalidate
end

#INIT: Parse the arguments of the initialize method
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /from/i
		@invoked_from = value
	end	
end

#INIT: Initialize texts and messages
def init_messages
	@menutitle = T7[:PlugName]
	
	@mnu_exit = T6[:T_BUTTON_Exit]	
	@mnu_abort = T6[:T_STR_AbortTool]	
	
	@hsh_ops_titles = Hash.new T7[:OPS_other]
	@hsh_ops_titles[:global_fixing] = T7[:OPS_global_fixing]
	@hsh_ops_titles[:single_fixing] = T7[:OPS_single_fixing]
	
	@hsh_tips = {}
	@lst_params.each do |symb| 
		tip = T7["BOX_Repair_#{symb}".intern]
		case symb
		when :overlap
			bkcolor = 'pink'
			frcolor = @color_wf_overlap
		when :needle_eye
			bkcolor = 'sandybrown'
			frcolor = @color_wf_needle_eye
		when :lost_junction
			bkcolor = 'gold'
			frcolor = @color_wf_lost_junction
		when :coplanar_edge
			bkcolor = 'gray'
			frcolor = @color_wf_coplanar_edge
		when :zero_edge
			bkcolor = 'thistle'
			frcolor = @color_wf_zero_edge
		when :tiny_gap
			bkcolor = 'lightgreen'
			frcolor = @color_wf_tiny_gap
		when :split_edge
			bkcolor = 'lightpink'
			frcolor = @color_wf_split_edge
		else
			bkcolor = 'gray'
			frcolor = 'black'
		end	
		@hsh_tips[symb] = [tip, bkcolor, frcolor]
	end	
end

#INIT: Initialize texts and messages
def init_colors
	@color_wf_overlap = 'red'
	@color_wf_needle_eye = 'brown'
	@color_wf_lost_junction = 'darkorange'
	@color_wf_lost_junction_s = 'orange'
	@color_wf_coplanar_edge = 'slategray'
	@color_wf_zero_edge = 'magenta'
	@color_wf_zero_edge_line = 'palevioletred'
	@color_wf_tiny_gap = 'green'
	@color_wf_split_edge = 'red'

	@hsh_marking = {}
	@hsh_marking[:overlap] = [@color_wf_overlap]
	@hsh_marking[:needle_eye] = [@color_wf_needle_eye, 3]
	@hsh_marking[:lost_junction] = [@color_wf_lost_junction, 3, '-']
	@hsh_marking[:coplanar_edge] = [@color_wf_coplanar_edge, 5, '_']
	@hsh_marking[:zero_edge] = [@color_wf_zero_edge, 3]
	@hsh_marking[:tiny_gap] = [@color_wf_tiny_gap, 3, '_']
	@hsh_marking[:split_edge] = [@color_wf_split_edge, 3]
		
	#Reference circle
	pts = G6.pts_circle 0, 0, 9, 6
	@circle_segs = []
	for i in 0..pts.length-2
		@circle_segs.push pts[i], pts[i+1]
	end	

	#Reference square
	pts = G6.pts_circle 0, 0, 9, 4
	@rect_segs = []
	for i in 0..pts.length-2
		@rect_segs.push pts[i], pts[i+1]
	end	
	
end

#INIT: Initialize the cursors
def init_cursors
	@id_cursor_hourglass_red = Traductor.create_cursor "Cursor_hourGlass_Red", 16, 16		
	@id_cursor_magnifier = Traductor.create_cursor "Cursor_Magnifier_32", 12, 12	
	@id_cursor = @id_cursor_magnifier
end

#INIT: Initialization of the magnifier
def init_magnifier
	notify_proc = self.method "notify_from_magnifier"		
	hsh = { :notify_proc => notify_proc }
	@magnifier = Traductor::Magnifier.new hsh
end

#--------------------------------------------------------------
# PARAMETER: Tolerance and other settings and parameters
#--------------------------------------------------------------

#Persistent parameters and defaults
@@hsh_parameters_persist = {} unless defined?(@@hsh_parameters_persist) && @@hsh_parameters_persist.length > 0

#PARAMETER: Load the parameters from the model attribute and the defaults
def parameter_load
	#Default values
	@lst_params = [:overlap, :lost_junction, :needle_eye, :zero_edge, :tiny_gap, :split_edge, :coplanar_edge]
	@lst_params_tol = [:overlap, :zero_edge, :tiny_gap]
	@lst_params_tol_big = [:lost_junction]
	@lst_params_flag = [:coplanar_edge]
	@lst_params_bool = [:needle_eye, :split_edge]
	
	@hsh_param_default = {
		:reinspect => true,
		
		:tolerance_overlap => 0.1,
		:check_overlap => true,
		:percent_flag_overlap => true,
		:percent_value_overlap => 0.01,
		
		:tolerance_lost_junction => 10,
		:check_lost_junction => true,
		:percent_flag_lost_junction => true,
		:percent_value_lost_junction => 0.05,
		
		:check_needle_eye => false,
		
		:tolerance_zero_edge => 0.5,
		:check_zero_edge => true,
		:percent_flag_zero_edge => true,
		:percent_value_zero_edge => 0.01,
		
		:tolerance_tiny_gap => 0.5,
		:check_tiny_gap => true,
		:percent_flag_tiny_gap => true,
		:percent_value_tiny_gap => 0.01,
		
		:check_split_edge => false,
		
		:check_coplanar_edge => false,
		:flag_coplanar_edge => true
	}
	
	#Hard-coded parameters
	@hsh_hard_param = {}
	@tolerance_min = 0.05
	@tolerance_max = 10
	@tolerance_incr = 0.05
	@percent_min = 0.001
	@percent_max = 0.05
	@percent_incr = 0.005
	@percent_max_big = 0.2
	@percent_min_big = 0.01
	@percent_incr_big = 0.01
	
	#Calculating the appropriate parameters
	@hsh_parameters = {}
	@hsh_parameters.update @hsh_param_default
	sparam = Sketchup.read_default @dico_name, @dico_attr
	hsh = eval(sparam) rescue {}
	hsh = {} unless hsh
	@@hsh_parameters_persist.update hsh
	@hsh_parameters.update @@hsh_parameters_persist 
	
	#Checking if inspection is allowed
	parameter_checking
end

#PARAMETER: Save the parameters as a model attribute
def parameter_save
	sparam = @@hsh_parameters_persist.inspect.gsub('"', "'")
	Sketchup.write_default @dico_name, @dico_attr, sparam
end

#PARAMETER: Set a value for a key
def parameter_set(symb, val)
	@@hsh_parameters_persist[symb] = @hsh_parameters[symb] = val
	parameter_after
	val
end

#PARAMETER: Toggle a value for a key
def parameter_toggle(symb)
	val = @@hsh_parameters_persist[symb] = @hsh_parameters[symb] = !@hsh_parameters[symb]
	parameter_after
	val
end

#PARAMETER: Set the current parameter to be the only one selected
def parameter_set_only(symb)
	@hsh_parameters.keys.each do |key|
		@@hsh_parameters_persist[key] = @hsh_parameters[key] = false if key != symb && key.to_s =~ /\Acheck_/
	end	
	@@hsh_parameters_persist[symb] = @hsh_parameters[symb] = true
	parameter_after
end

#PARAMETER: Post-processing after changing parameters
def parameter_after
	parameter_checking
	repair_checking
	#refresh_viewport
	onMouseMove_zero
end

#PARAMETER: Verification after changing parameters
def parameter_checking
	@inspection_allowed = false
		
	@lst_params.each do |symb|
		if @hsh_parameters["check_#{symb}".intern]
			@inspection_allowed = true
			break
		end	
	end
	@inspection_allowed
end

#PARAMETER: Select all parameters
def parameter_select_all(val)
	@lst_params.each { |symb| parameter_set("check_#{symb}".intern, val) }
	parameter_checking
end

#-----------------------------------------------
# Observer Events for selection
#-----------------------------------------------

def onSelectionBulkChange(selection)
	n = selection.length
	@selection_number = n if n > 0
end

def onSelectionAdded(selection, elt)
	@selection_number = selection.length
end

def onSelectionCleared(selection)
	@selection_number = 0
end

#----------------------------------------------------------------------------------------------------
# ACTIVATION: Plugin Activation / Deactivation
#----------------------------------------------------------------------------------------------------

#ACTIVATION: Tool activation
def activate
	LibFredo6.register_ruby "FredoTools::EdgeInspector"
	Traductor::Hilitor.stop_all_active
		
	#Resetting the environment
	#reset_context
	
	#Initializing the Executor of Operation
	hsh = { :title => @menutitle, :palette => @palette, :no_commit => true }
	@suops = Traductor::SUOperation.new hsh

	#Handling the initial selection
	handle_initial_selection
		
	refresh_viewport
end

#ACTIVATION: Deactivation of the tool
def deactivate(view)
	@magnifier.activation false 
	parameter_save
	
	if @geometry_counting > 0 && @aborting
		text = T7[:MSG_ConfirmAbortText] + "\n\n" + T7[:MSG_ConfirmAbortQuestion]
		if UI.messagebox(text, MB_YESNO) != 6
			Sketchup.undo
		end	
	end	
	view.invalidate
	@view.remove_observer self
	@model.selection.remove_observer self
end

#ACTIVATION: Exit the tool
def exit_tool
	@aborting = false
	@model.select_tool nil
end

#ACTIVATION: Abort the operation and exit
def abort_tool
	@aborting = true
	@model.select_tool nil
end

#--------------------------------------------------------------
# MAGNIFIER: Magnifier Management
#--------------------------------------------------------------

#MAGNIFIER: Start the Magnifier
def magnifier_start
	@magnifier_on = true
	magnifier_after
end

#MAGNIFIER: Stop the Magnifier
def magnifier_stop
	@magnifier_on = false
	magnifier_after
end

#MAGNIFIER: Common actions performed after a start or stop of the Magnifier
def magnifier_after
	palette_show !@magnifier_on
	@magnifier.activation @magnifier_on
	onMouseMove_zero
end

#MAGNIFIER: Toggle magnifier on / off
def magnifier_toggle
	(@magnifier_on) ? magnifier_stop : magnifier_start
end

#MAGNIFIER: Notification and exchange with Magnifier
def notify_from_magnifier(action, *args)
	case action
	when :exit
		magnifier_stop
	when :ctrl_down
		return @ctrl_down
	when :shift_down
		return @shift_down
	when :entity_color
		return nil unless @hsh_entities_colors
		e = args[0]
		if e.class == String
			@hsh_marking[@hsh_entities_colors[e]]
		else
			return nil unless e.valid?
			@hsh_marking[@hsh_entities_colors[e.entityID]]
		end	
	when :zoom_auto
		vx, edge = args
		return nil unless @hsh_typical_distance
		[vx, edge].each do |e|
			next unless e && e.valid?
			d = @hsh_typical_distance[e.entityID]
			return d if d && d > 0
		end	
		return nil
	when :before_operation
		start_operation args[0]
	when :after_operation
		commit_operation
	when :abort_operation
		@suops.abort_operation
	when :is_repair
		return is_repair(args[0])
	when :tooltip_repair
		return tooltip_repair(args[0])
	when :is_ignored
		return is_ignored(args[0])
	when :repair_now
		execute_single_fixing args[0]
	when :ignore
		ignore_repairs(args[0])
	when :gray_check
		(@something_to_fix) ? false : true
	when :gray_undo
		(rollback_allowed?) ? false : true
	when :undo
		rollback true
	when :fake_elements
		@fake_elements
	else
		return nil
	end	
end

#MAGNIFIER: Tooltip of a repair or nil if not
def tooltip_repair(e)
	rep_id = get_repair_id(e)
	return nil unless rep_id
	rep = @hsh_all_reparations[rep_id]
	return nil unless rep
	@hsh_tips[rep.type]
end

#MAGNFIER: Compute repair id for true or fake elements
def get_repair_id(e)
	return nil unless e && @hsh_all_rep_ref
	if e.class == String
		id = e
	else	
		return nil unless e.valid?
		id = e.entityID
	end	
	rep_id = @hsh_all_rep_ref[id]
end

#MAGNIFIER: Tells if the entity is about a repair
def is_repair(e)
	rep_id = get_repair_id(e)
	return nil unless rep_id
	rep = @hsh_all_reparations[rep_id]
	return nil unless rep
	lent = rep.lent
	return nil if lent.compact.length != lent.length || lent.find { |ee| !ee.valid? }
	lent = lent + rep.lst_fake_ids if rep.lst_fake_ids
	[lent, rep_id]
end

#MAGNIFIER: Tells if the entity is about a repair ignored
def is_ignored(e)
	rep_id = get_repair_id(e)
	return false unless rep_id
	rep = @hsh_all_reparations[rep_id]
	rep.ignored
end

#MAGNIFIER: Toggle the Ignore state of a repair
def ignore_repairs(lrepair_info)
	lrepair_info.each do |repair_info|
		lent, rep_id = repair_info
		next unless rep_id
		rep = @hsh_all_reparations[rep_id]
		rep.ignored = !rep.ignored if rep
	end
end

#--------------------------------------------------------------
# VIEWPORT: View Management
#--------------------------------------------------------------

#VIEWPORT: Computing Current cursor
def onSetCursor
	#Palette cursors
	ic = super
	return (ic != 0) if ic
		
	#Geometry processing
	if @geometry_computing
		@id_cursor = @id_cursor_hourglass_red
		return UI.set_cursor(@id_cursor)	
	end

	#Other cursors depending on state
	@id_cursor = @id_cursor_magnifier
	
	UI.set_cursor @id_cursor
end

#VIEWPORT: Refresh the viewport
def refresh_viewport
	onSetCursor
	show_message
	@view.invalidate
end
	
#VIEWPORT: Show message in the palette
def show_message
	#Mouse if either in the palette or out of the viewport
	if @mouseOut
		@palette.set_message nil
		return
	elsif @mouse_in_palette
		@palette.set_message @palette.get_main_tooltip, 'aliceblue'
		return
	end			
end

#---------------------------------------------------------------------------------------------
# MOVE: Mouse Movement Methods
#---------------------------------------------------------------------------------------------

#MOVE: Mouse Movements
def onMouseMove_zero(flag=nil) ; onMouseMove(@flags, @xmove, @ymove, @view) if @xmove ; end
def onMouseMove(flags, x, y, view)
	#Event for the palette
	@xmove = x
	@ymove = y
	@flags = flags
	
	#Mouse in palette
	if super
		@mouse_in_palette = true
		refresh_viewport
		return	
	end
	@mouse_in_palette = false
			
	return if @moving
		
	#Mouse in magnifier
	@magnifier.onMouseMove(flags, x, y, view)	
	
	#Tooltip for repair
	tooltip_repair_compute(x, y)
	
	#Refreshing the view
	@moving = true
	refresh_viewport
end	

#MOVE: Compute the auto zoom factor
def tooltip_repair_compute(x, y)
	@tooltip_repair = nil
	return unless @something_to_fix
	@ip.pick @view, x, y
	edge = @ip.edge
	vertex = @ip.vertex
	return nil unless vertex || edge
	
	tipinfo = tooltip_repair(vertex)
	tipinfo = tooltip_repair(edge) unless tipinfo
	if !tipinfo && vertex
		vertex.edges.each do |e|
			tipinfo = tooltip_repair(e)
			break if tipinfo
		end
	end	
	@tooltip_repair = tipinfo
end

#MOVE: Mouse leaves the viewport
def onMouseLeave(view)
	@mouseOut = true
	view.invalidate
end

#MOVE: Mouse enters the viewport
def onMouseEnter(view)
	@mouseOut = false
	view.invalidate
end
	
#MOVE: Before Zooming or changing view
def suspend(view)
	@magnifier.suspend(view) if @magnifier && @magnifier_on
end

#MOVE: After changing view
def resume(view)
	@magnifier.resume(view, @ctrl_down) if @magnifier && @magnifier_on
	refresh_viewport
	onMouseMove_zero
end
	
#VIEW: Notification of view changed
def onViewChanged(view)
	return unless @magnifier_on
	@magnifier.onViewChanged(view)
end
	
#---------------------------------------------------------------------------------------------
# SU TOOL: Click Management
#---------------------------------------------------------------------------------------------

#SU TOOL: Button click DOWN
def onLButtonDown(flags, x, y, view)
	#Interrupting geometry construction if running
	return if @suops.interrupt? || @processing_inspection || @processing_fixing
	
	#Palette management
	if super
		refresh_viewport
		return
	end
	
	@button_down = true
	@time_down = Time.now
	@xdown = x
	@ydown = y
	
	#Starting the magnifier
	if !@magnifier_on
		magnifier_start
		onMouseMove_zero
		@time_down_start = Time.now
	else	
		@magnifier.onLButtonDown(flags, x, y, view)
	end	
end

#SU TOOL: Button click UP - Means that we end the selection
def onLButtonUp(flags, x, y, view)
	return if @processing_inspection || @processing_fixing

	#Palette buttons
	if super
		@time_down_start = nil
		onMouseMove_zero
		return
	end
	return unless @button_down
	@button_down = false

	if @magnifier_on
		if @time_down_start && (Time.now - @time_down_start) > 2.0
			magnifier_stop
		else
			@magnifier.onLButtonUp(flags, x, y, view)
		end
		@time_down_start = nil
	end	
end

#SU TOOL: Double Click received
def onLButtonDoubleClick(flags, x, y, view)
	return if super
	if @magnifier_on
		@magnifier.onLButtonDoubleClick(flags, x, y, view)
	end	
end

#---------------------------------------------------------------------------------------------
# KEY: Keyboard handling
#---------------------------------------------------------------------------------------------

#KEY: Return key pressed
def onReturn(view)
	if @magnifier_on
		@magnifier.onReturn(view)
	else
		magnifier_toggle
	end	
	onMouseMove_zero
end

#VCB: Handle VCB input
def onUserText(text, view)
	@magnifier.onUserText(text, view) if @magnifier_on
end

#KEY: Handle Key down
def onKeyDown(key, rpt, flags, view)
	key = Traductor.new_check_key key, flags, false
	
	#Interrupting geometry construction if running
	return if @suops.interrupt?
		
	#Check with Magnifier
	if @magnifier_on && @magnifier.handle_key_down(key)
		refresh_viewport
		return false
	end
	
	@num_shiftdown = 0 unless @num_shiftdown
	@num_shiftdown += 1
	@num_ctrldown = 0 unless @num_ctrldown
	@num_ctrldown += 1
	case key			
	when CONSTRAIN_MODIFIER_KEY
		@time_shift_down = Time.now
		@shift_down = @num_shiftdown
		
	when COPY_MODIFIER_KEY
		@time_ctrl_down = Time.now
		@ctrl_down = @num_ctrldown
			
	when 9
	
	else	
		@shift_down = @ctrl_down = false
		return false
	end	
	refresh_viewport
	false
end

#KEY: Key UP received
def onKeyUp(key, rpt, flags, view)
	key = Traductor.new_check_key key, flags, true
	
	#Check with Magnifier
	if @magnifier_on && @magnifier.handle_key_up(key)
		refresh_viewport
		return false
	end
	
	#Dispatching the event
	case key							
	when CONSTRAIN_MODIFIER_KEY
		if @shift_down && @shift_down == @num_shiftdown && @time_shift_down && (Time.now - @time_shift_down) < 0.4

		end	
		@shift_down = false
		
	when COPY_MODIFIER_KEY
		if @ctrl_down && @ctrl_down == @num_ctrldown && @time_ctrl_down && (Time.now - @time_ctrl_down) < 0.4

		end	
		@ctrl_down = false
		
	when VK_DELETE
		handle_delete
		
	else 
		if @time_ctrl_up && (Time.now - @time_ctrl_up) < 0.1	
			onMouseMove_zero 0
		end	
		@shift_down = @ctrl_down = false
		return false
	end	
	refresh_viewport
	false
end

#DELETE: Manage the operation count after selection entities where deleted by the DEL key
def handle_delete
	if @selection_number > 0
		@geometry_counting += 1
		@geometry_last_ops.push T7[:OPS_DELKey]
	end	
end
		
#---------------------------------------------------------------------------------------
# UNDO: Undo and Rollback Management
#---------------------------------------------------------------------------------------

#UNDO: Cancel and undo methods
def onCancel(flag, view)
	case flag
	when 1, 2	#Undo or reselect the tool
		handle_undo
	when 0		#user pressed Escape	
		handle_escape
	end
end

#UNDO: Handle an undo
def handle_undo
	if rollback_is_valid?
		UI.start_timer(0.1) { rollback }
	else
		@suops.start_operation
	end	
end

#UNDO: Handle the escape key depending on current mode
def handle_escape
	if @magnifier_on && !@magnifier.handle_escape
		magnifier_stop
	end	
end

#UNDO: Execute a rollback
def rollback(perform_undo=false)
	return unless @geometry_counting > 0
	Sketchup.undo if perform_undo
	reparations_recompute_after_undo
	@inspection_allowed = true
	@geometry_counting -= 1
	last_op = @geometry_last_ops.pop
	resolver_rollback

	if @magnifier_on
		@magnifier.handle_undo
	end	
end

#UNDO: Rollback dynamic tooltip
def rollback_tip
	(@geometry_counting <= 0) ? T7[:TIP_RollBack] : @geometry_last_ops.last
end

#UNDO: Signal whether Undo is possible
def rollback_is_valid?
	if @geometry_counting <= 0
		UI.messagebox T6[:OPS_UndoBeyond]
		return false
	end
	true	
end

#UNDO: Check if rollback is possible
def rollback_allowed?
	@geometry_counting > 0
end
		
#---------------------------------------------------------------------------------------------
# DRAW: Drawing Methods
#---------------------------------------------------------------------------------------------

#DRAW: Draw top method
def draw(view)		
	#Updating the wireframe if view has changed
	wireframe_update(view) if @view_tracker.changed?
	
	#Drawing the repairs
	if @something_to_fix
		dessin_wf_draw view
	end	
	
	#Drawing the magnifier
	if @magnifier_on
		@magnifier.draw view
	end
	
	@moving = false
	
	#Drawing the palette
	super
	
	draw_tooltip_repair view
end
	
#DRAW: Draw the tooltip for repair if applicable
def draw_tooltip_repair(view)
	return if @mouse_in_palette
	return unless @tooltip_repair && @xmove
	tip, bkcolor, frcolor = @tooltip_repair
	G6.draw_rectangle_text(view, @xmove, @ymove, tip, bkcolor, frcolor)
end
	
#DRAW: Draw the wireframe to signal the repairs	
def dessin_wf_draw(view)
	#Overlaps
	if @wf_overlap_lines
		view.line_width = 1
		view.line_stipple = ''
		view.drawing_color = @color_wf_overlap
		view.draw GL_LINES, @wf_overlap_lines
	end
	
	#Needle_eye
	if @wf_needle_eye_lines
		view.line_width = 2
		view.line_stipple = ''
		view.drawing_color = @color_wf_needle_eye
		view.draw GL_LINES, @wf_needle_eye_lines
	end
	
	#Coplanar edges
	if @wf_coplanar_edge_lines && @wf_coplanar_edge_lines.length > 0
		view.line_width = 6
		view.line_stipple = '_'
		view.drawing_color = @color_wf_coplanar_edge
		view.draw GL_LINES, @wf_coplanar_edge_lines
	end
	
	#Zero Edges
	if @wf_zero_edge_lines && @wf_zero_edge_lines.length > 0
		view.line_width = 2
		view.line_stipple = ''
		view.drawing_color = @color_wf_zero_edge
		view.draw2d GL_LINES, @wf_zero_edge_lines
	end
	if @wf_zero_edge_contour && @wf_zero_edge_contour.length > 0
		view.line_width = 2
		view.line_stipple = ''
		view.drawing_color = @color_wf_zero_edge_line
		view.draw GL_LINES, @wf_zero_edge_contour
	end

	#Tiny gaps
	if @wf_tiny_gap_lines && @wf_tiny_gap_lines.length > 0
		view.line_width = 2
		view.line_stipple = ''
		view.drawing_color = @color_wf_tiny_gap
		view.draw2d GL_LINES, @wf_tiny_gap_lines
	end
	if @wf_tiny_gap_junctions && @wf_tiny_gap_junctions.length > 0
		view.line_width = 2
		view.line_stipple = '_'
		view.drawing_color = @color_wf_tiny_gap	
		view.draw GL_LINES, @wf_tiny_gap_junctions
	end

	#Split Edges
	if @wf_split_edge_lines && @wf_split_edge_lines.length > 0
		view.line_width = 2
		view.line_stipple = ''
		view.drawing_color = @color_wf_split_edge
		view.draw2d GL_LINES, @wf_split_edge_lines
	end

	#Lost Junctions
	if @wf_lost_junction_lines && @wf_lost_junction_lines.length > 0
		view.line_width = 2
		view.line_stipple = ''
		view.drawing_color = @color_wf_lost_junction
		view.draw2d GL_LINES, @wf_lost_junction_lines
	end
	if @wf_lost_junction_lines_j && @wf_lost_junction_lines_j.length > 0
		view.line_width = 2
		view.line_stipple = '-'
		view.drawing_color = @color_wf_lost_junction
		@wf_lost_junction_lines_j.each do |line|
			view.draw GL_LINE_STRIP, line
		end	
	end
	if @wf_lost_junction_lines_s && @wf_lost_junction_lines_s.length > 0
		view.line_width = 2
		view.line_stipple = ''
		view.drawing_color = @color_wf_lost_junction_s
		@wf_lost_junction_lines_s.each do |line|
			view.draw GL_LINE_STRIP, line
		end	
	end
		
end

#--------------------------------------------------
# WIREFRAME: Wireframe management
#--------------------------------------------------

#WIREFRAME: Update the wireframe asynchronously
def wireframe_update(view)
	if @dessin_timer
		UI.stop_timer @dessin_timer
		@dessin_timer = nil
	end
	@dessin_timer = UI.start_timer(0.0) { wireframe_prepare }
end
	
#WIREFRAME: Reset the wireframe	
def wireframe_reset
	@wf_overlap_lines = nil
	@wf_coplanar_edge_lines = nil
	@wf_zero_edge_lines = nil
	@wf_zero_edge_contour = nil
	@wf_tiny_gap_lines = nil
	@wf_tiny_gap_junctions = nil
	@wf_split_edge_lines = nil
	@wf_needle_eye_lines = nil
	@wf_lost_junction_lines = nil
	@wf_lost_junction_lines_j = nil	
end
	
#WIREFRAME: Prepare the wireframe for drawing	
def wireframe_prepare
	#Initialization
	wireframe_reset
	lines_overlap = []
	lines_needle_eye = []
	lines_coplanar_edge = []
	centers_zero_edge = []
	lines_zero_edges = []
	centers_tiny_gap = []
	junctions_tiny_gap = []
	centers_split_edge = []
	centers_lost_junction = []
	lines_lost_junction = []
	seg_lost_junction = []
	
	#nothing to fix
	return unless @something_to_fix
	
	#Loop on Reparations
	@hsh_all_reparations.each do |rep_id, reparation|
		lent = reparation.lent
		inval = (lent.compact.length != lent.length || lent.find { |e| !e.valid? })
		next if inval
		tr = reparation.tr
		
		case reparation.type
		when :overlap
			e1, e2 = lent
			lines_overlap.push tr * e1.start.position, tr * e1.end.position, tr * e2.start.position, tr * e2.end.position

		when :needle_eye
			lent.each do |e|
				lines_needle_eye.push tr * e.start.position, tr * e.end.position
			end	

		when :lost_junction
			vx1, vx2 = lent
			jpts = reparation.jpts.collect { |pt| tr * pt }
			if jpts.length == 2
				center = Geom.linear_combination 0.5, jpts[0], 0.5, jpts[1]
				seg_lost_junction.push jpts
			else
				center = jpts[1]
				lines_lost_junction.push jpts
			end	
			centers_lost_junction.push center
			lines_lost_junction.push jpts

		when :coplanar_edge
			e = lent.first
			lines_coplanar_edge.push tr * e.start.position, tr * e.end.position
			
		when :zero_edge
			lent.each { |e| lines_zero_edges.push tr * e.start.position, tr * e.end.position }
			e = lent.first
			centers_zero_edge.push Geom.linear_combination(0.5, tr * e.start.position, 0.5, tr * e.end.position)

		when :split_edge
			vx = lent.first
			centers_split_edge.push tr * vx.position

		when :tiny_gap
			vx1, vx2 = lent
			centers_tiny_gap.push Geom.linear_combination(0.5, tr * vx1.position, 0.5, tr * vx2.position)
			junctions_tiny_gap.push tr * vx1.position, tr * vx2.position
			
		end	
	end
		
	#Preparing the wireframe 
	@wf_overlap_lines = lines_overlap.collect { |pt| G6.small_offset @view, pt } unless lines_overlap.empty?
	@wf_needle_eye_lines = lines_needle_eye.collect { |pt| G6.small_offset @view, pt } unless lines_needle_eye.empty?
	@wf_coplanar_edge_lines = lines_coplanar_edge.collect { |pt| G6.small_offset @view, pt } unless lines_coplanar_edge.empty?
	@wf_zero_edge_lines = wireframe_small_circle centers_zero_edge unless centers_zero_edge.empty?
	@wf_zero_edge_contour = lines_zero_edges.collect { |pt| G6.small_offset @view, pt } unless lines_zero_edges.empty?
	@wf_tiny_gap_lines = wireframe_small_circle centers_tiny_gap unless centers_tiny_gap.empty?
	@wf_tiny_gap_junctions = junctions_tiny_gap.collect { |pt| G6.small_offset @view, pt } unless junctions_tiny_gap.empty?
	@wf_split_edge_lines = wireframe_small_circle(centers_split_edge, true) unless centers_split_edge.empty?
	@wf_lost_junction_lines = wireframe_small_circle(centers_lost_junction, true) unless centers_lost_junction.empty?
	@wf_lost_junction_lines_j = lines_lost_junction
	@wf_lost_junction_lines_s = seg_lost_junction
				
	#Refreshing the view
	@view.invalidate
end
	
#DRAW: Compute the lines for small circles or losanges
def wireframe_small_circle(centers, flg_rect=false)
	segs = (flg_rect) ? @rect_segs : @circle_segs
	wf_lines = nil
	if centers && centers.length > 0
		vpx = @view.vpwidth
		vpy = @view.vpheight
		wf_lines = []
		centers.each do |center|
			center2d = @view.screen_coords center
			x, y = center2d.x, center2d.y
			next if x < 0 || x > vpx || y < 0 || y > vpy
			t = Geom::Transformation.translation ORIGIN.vector_to(center2d)
			wf_lines += segs.collect { |pt| t * pt }
		end	
	end
	wf_lines
end
		
#--------------------------------------------------
# MENU: Contextual menu
#--------------------------------------------------

#MENU: Contextual menu
def getMenu(menu)
	cxmenu = Traductor::ContextMenu.new
		
	#Magnifier
	if @magnifier_on
		@magnifier.contextual_menu_contribution(cxmenu)
		cxmenu.add_sepa
		cxmenu.add_item(T7[:MNU_ExitGlassMode]) { magnifier_stop }
		cxmenu.show menu
		return true
	end	
	
	#Exit and Abort
	cxmenu.add_sepa
	cxmenu.add_item(@mnu_abort) { exit_tool }
	cxmenu.add_item(@mnu_exit) { abort_tool }
	
	#Showing the menu
	cxmenu.show menu
	true
end

#============================================================
# OPERATION: handling SU operation
#============================================================
	
#OPERATION: Start an operation based on symbol	
def start_operation(op_symb)
	text = (op_symb.class == Symbol) ? @hsh_ops_titles[op_symb] : op_symb
	@suops.set_title "#{@menutitle} [#{text}]"
	@suops.start_operation
	@geometry_last_ops.push text	
end

#OPERATION: Commit and save the fixing
def commit_operation
	@suops.commit_operation
	@geometry_counting += 1
end

#============================================================
# Top Processing methods
#============================================================

def groups_make_unique(lentities)
	lgroups = lentities.grep Sketchup::Group
	lgroups.each do |e|
		cdef = e.entities.parent
		next unless cdef && cdef.count_instances != 1 && cdef.instances.include?(e)
		e.make_unique unless G6.grouponent_unique?(e)
		groups_make_unique e.entities.to_a
		@ng_unique += 1
	end	

	lcomps = lentities.grep Sketchup::ComponentInstance
	lcomps.each do |e|
		cdef_id = e.definition.entityID
		next if @hsh_compdef[cdef_id]
		@hsh_compdef[cdef_id] = true
		groups_make_unique e.definition.entities.to_a
	end	
end

#ALGO: Inspect the selection and create the resolvers
def handle_initial_selection(selection=nil)
	#Choosing the selection
	selection = @model.selection.to_a unless selection
	selection = @model.active_entities.to_a if selection.empty?

	#Initialization
	@nb_comp = 0
	@nb_group = 0
	@lst_resolvers = []
	
	#Making all groups unique in selection
	@suops.set_title T6[:T_OPS_MakeGroupUnique] 
	@suops.start_operation
	@ng_unique = 0
	@hsh_compdef = {}
	groups_make_unique selection
	if @ng_unique > 0
		@suops.commit_operation
	else
		@suops.abort_operation
	end	
	
	#Screening the selection recursively
	@draw_hidden = @rendering_options["DrawHidden"]
	@hsh_compdef = {}
	process_selection selection, Sketchup.active_model, @tr_id
	
	#Text for title of the floating palette
	title = T7[:PlugName]
	if @model.selection.to_a.empty?
		text = T6[:T_TXT_WholeModel]
	else	
		ltext = []
		nb_edges = @model.selection.grep(Sketchup::Edge).length
		ltext.push "#{T6[:T_TXT_Edges]}: #{nb_edges}" if nb_edges > 0
		ltext.push "#{T6[:T_TXT_Comp]}: #{@nb_comp}" if @nb_comp > 0
		ltext.push "#{T6[:T_TXT_Group]}: #{@nb_group}" if @nb_group > 0
		text = ltext.join ", "
	end	
	title = T7[:PlugName] 
	title += " [#{text}]" if text
	@palette.floating_set_title @pal_float, title
		
	#Getting the message if any
	@message_statistics = nil
	nb_too_small = 0
	factor_scale = nil
	@lst_resolvers.each do |resolver|
		fac, nb = resolver.statistics_info
		next unless fac
		factor_scale = fac if !factor_scale || fac > factor_scale
		nb_too_small += nb
	end	
	if factor_scale
		len = 0.1
		text = T7[:MSG_TooSmall_Edge1, nb_too_small, G6.format_length_general(len)]
		text += "\n" + T7[:MSG_TooSmall_Edge2, factor_scale]
		@message_statistics = text
	end	
	
	#Setting environment after very first analysis
	compute_numbers
	@something_to_fix = false
	@geometry_fixed = false
	@selection.clear
	@selection_number = 0
end

#ALGO: Compute the initial number of edges and faces
def compute_numbers
	nedges = nfaces = 0
	@lst_resolvers.each do |resolver|
		nedges += resolver.get_initial_nb_edges
		nfaces += resolver.get_initial_nb_faces
	end
	if @geometry_fixed
		@tx_edges_after_fix = "#{nedges} "
		@tx_faces_after_fix = "#{nfaces} "	
	else
		@tx_edges_initial = "#{nedges} "
		@tx_faces_initial = "#{nfaces} "
		@tx_edges_original = @tx_edges_initial unless @tx_edges_original
		@tx_faces_original = @tx_faces_initial unless @tx_faces_original
		@tx_edges_after_fix = "-   "
		@tx_faces_after_fix = "-   "
	end	
end

#ALGO: Go through the selection and Create the resolvers		
def process_selection(lentities, comp, t)
	#Getting all edges
	ledges = lfaces = nil
	if comp.instance_of?(Sketchup::ComponentInstance)
		@nb_comp += 1	
	elsif comp.instance_of?(Sketchup::Group)
		@nb_group += 1
	end
	ledges = lentities.grep Sketchup::Edge
	lfaces = lentities.grep Sketchup::Face
	
	#Creating the resolvers
	if (ledges && ledges.length > 0)
		@lst_resolvers.push Resolver.new(@lst_params, ledges, lfaces, comp, t, @hsh_parameters)
	end
	
	#Processing groups
	lgroup = lentities.grep Sketchup::Group
	lgroup.each do |e|
		next if !@draw_hidden && e.hidden?
		e.make_unique unless G6.grouponent_unique?(e)
		process_selection e.entities.to_a, e, t * e.transformation
	end	

	#Processing component instances
	lcomp = lentities.grep Sketchup::ComponentInstance
	lcomp.each do |e|
		next if !@draw_hidden && e.hidden?
		cdef_id = e.definition.entityID
		next if @hsh_compdef[cdef_id]
		@hsh_compdef[cdef_id] = true
		process_selection e.definition.entities.to_a, e, t * e.transformation
	end	
end

#ALGO: Start the inspection of edges based on current settings
def resolver_rollback
	#Resetting the selection at resolver level
	@lst_resolvers.each do |resolver|
		resolver.execute_rollback
	end
	
	#Updating the status
	#@something_to_fix = false
	@geometry_fixed = false
	compute_numbers	
	repair_checking
	parameter_checking
	wireframe_prepare
	refresh_viewport
	
	#Reinpecting if required
	if !@magnifier_on && @hsh_parameters[:reinspect]
		#resolver_inspection
		#refresh_viewport
	end	
end

#Manage the verification after inspection
def repair_checking
	@fake_elements = []
	return unless @hsh_all_reparations
	
	#Counting the number of repairs, typical distance and highlight colors
	@hsh_typical_distance = {}
	@hsh_entities_colors = {}
	@hsh_nb_repairs = Hash.new 0
	@lst_params.each { |symb| @hsh_nb_repairs[symb] = 0 if @hsh_parameters["check_#{symb}".intern] }

	@hsh_all_reparations.each do |rep_id, reparation|
		lent = reparation.lent
		next if lent.compact.length != lent.length || lent.find { |e| !e.valid? }
		type = reparation.type
		@hsh_nb_repairs[type] += 1
		reparation.lent.each do |e|
			next unless e && e.valid?
			eid = e.entityID	
			d = reparation.d
			if d
				@hsh_typical_distance[eid] = d
				if e.instance_of?(Sketchup::Edge)
					@hsh_typical_distance[e.start.entityID] = @hsh_typical_distance[e.end.entityID] = d
				end
			end	
			@hsh_entities_colors[eid] = type			
		end	
		reparation_fake_edges_register reparation
	end	
	return unless @hsh_nb_repairs
	
	#Text and colors for the repairs
	@something_to_fix = false
	@hsh_nb_repairs_text = {}
	@hsh_nb_repairs_color = {}
	@hsh_nb_repairs.each do |symb, nb| 
		nb = @hsh_nb_repairs[symb]
		nb = 0 unless nb
		@hsh_nb_repairs_text[symb] = "#{nb}  "
		if nb > 0
			@something_to_fix = true
		end	
		@hsh_nb_repairs_color[symb] = (nb > 0 && @hsh_parameters["check_#{symb}".intern]) ? 'lightpink' : nil
	end		
end

#REPARATION: Register a fake elements in the list
def reparation_fake_edges_register(reparation)
	jpts = reparation.jpts
	return unless jpts
	type = reparation.type
	tr = reparation.tr
	parent = reparation.parent
	lent_id = reparation.lent_id
	reparation.lst_fake_ids = [] unless reparation.lst_fake_ids
	for i in 0..jpts.length - 2
		id = "edge_#{@fake_elements.length}"
		@fake_elements.push [id, [jpts[i], jpts[i+1]], tr, parent]
		reparation.lst_fake_ids.push id
		@hsh_entities_colors[id] = type
		@hsh_all_rep_ref[id] = reparation.rep_id
	end	
end

#ALGO: Reset all reparations
def reparations_reset
	@hsh_all_reparations = {}
	@hsh_all_rep_ref = {}
	@hsh_typical_distance = {}
	#@fake_elements = []
end

#ALGO: Recompute the reparations after undo
def reparations_recompute_after_undo
	return unless @hsh_all_reparations
	hmap_id = {}
	@lst_resolvers.each { |resolver| resolver.map_id hmap_id }
	
	@hsh_all_reparations.each do |rep_id, rep|
		rep.lent = rep.lent_id.collect { |id| hmap_id[id] }
	end	
end

#ALGO: Start the inspection of edges based on current settings
def resolver_inspection
	@model.selection.clear
	
	#Resetting environment
	reparations_reset
	
	#Creating the visual bar
	hsh = { :delay => 0, :mini_delay => 0.5, :interrupt => true }
	@vbar = Traductor::VisualProgressBar.new T7[:MSG_Inspecting], hsh
	@vbar_nsteps = @lst_resolvers.length * 1.0 + 1

	#Starting the visual bar
	@vbar.start
	UI.set_cursor @id_cursor_hourglass_red
	
	#Calling the resolver for inspection
	@processing_inspection = true
	wireframe_reset
	@suops.define_end_proc { |time| robot_inspection_terminate time }
	@suops.start_execution { robot_inspection }
end

#ALGO: Start the fixing of edges based on current settings
def resolver_fixing
	return unless @something_to_fix

	#Creating the visual bar
	hsh = { :style_color => :bluegreen, :delay => 0, :mini_delay => 0.5, :interrupt => true }
	@vbar = Traductor::VisualProgressBar.new T7[:MSG_Repairing], hsh
	@vbar_nsteps = @lst_resolvers.length * 1.0

	#Starting the visual bar
	@vbar.start
	UI.set_cursor @id_cursor_hourglass_red
	
	#Calling the resolver for inspection
	start_operation :global_fixing
	@processing_fixing = true
	@suops.define_end_proc { |time| robot_fixing_terminate time }
	@suops.start_execution { robot_fixing }
	return
end

#ALGO: Manage a single fixing on a repair	
def execute_single_fixing(lrepair_info)
	hsh_by_type = {}
	resolver = nil
	lrepair_info.each do |repair_info|
		lent, rep_id = repair_info
		rep = @hsh_all_reparations[rep_id]
		next unless rep
		type = rep.type
		resolver = rep.resolver unless resolver
		lst = hsh_by_type[type]
		lst = hsh_by_type[type] = [] unless lst
		lst.push lent
	end
	
	#Performing the single fixing
	start_operation :single_fixing
	resolver.execute_single_fixing hsh_by_type
	commit_operation
	repair_checking
	wireframe_prepare
end
	
#ALGO: Number of repairs
def get_nb_repairs_text(symb) ; @hsh_nb_repairs_text[symb] ; end
def get_nb_repairs_color(symb) ; @hsh_nb_repairs_color[symb] ; end
def get_nb_repairs(symb)
	nb = @hsh_nb_repairs[symb] 
	(nb) ? nb : 0
end

#---------------------------------------------------------------------------------------------
# ROBOT_INSPECTION: Robot for Inspection
#---------------------------------------------------------------------------------------------

#ROBOT_INSPECTION: Main State-robot function to progress along the inspection
def robot_inspection
	@nb_resolvers = @lst_resolvers.length
	begin
		robot_inspection_exec
	rescue Exception => e
		@suops.abort_operation
		Traductor::RubyErrorDialog.invoke e, @menutitle, T7[:ERR_DuringInspection]
		abort_tool
	end
end

#ROBOT_INSPECTION: Main State-robot function to progress along the inspection
def robot_inspection_exec
	while(action, *param = @suops.current_step) != nil	
		case action
		when :_init
			next_step = [:inspection, 0, 0]
		when :inspection	
			return if @suops.yield?
			iresolver, istep = param
			next_step = robot_inspection_execute_single(iresolver, istep)
		end	
		break if @suops.next_step(*next_step)
	end
end

#ROBOT_INSPECTION: Invoke a single resolver for Inspection
def robot_inspection_execute_single(iresolver, istep)
	resolver = @lst_resolvers[iresolver]
	return nil unless resolver
	resolver.execute_inspection @hsh_all_reparations, @hsh_all_rep_ref, @vbar, 1.0 / @vbar_nsteps, iresolver / @vbar_nsteps
	[:inspection, iresolver+1, 0]
end

#ROBOT_INSPECTION: Terminate the inspection
def robot_inspection_terminate(time)
	@processing_inspection = false
	@suops.abort_operation

	@vbar.update_title T7[:TXT_PreparingDisplay]
	pc_beg = (@vbar_nsteps - 1) / @vbar_nsteps
	@vbar.progression pc_beg, T7[:TXT_VisualElements]
	
	#Aborting the inspection
	if time < 0
		reparations_reset
		@inspection_allowed = true
	
	#Inspection went well
	else
		@inspection_allowed = false
	end	
	
	#Updating the numbers and preparing the wireframe
	compute_numbers
	repair_checking
	wireframe_prepare
		
	#Finishing
	UI.set_cursor 0
	@vbar.stop
	
	onMouseMove_zero
end

#---------------------------------------------------------------------------------------------
# ROBOT_FIXING: Robot for Inspection
#---------------------------------------------------------------------------------------------

#ROBOT_FIXING: Main State-robot function to progress along the inspection
def robot_fixing
	@nb_resolvers = @lst_resolvers.length
	begin
		robot_fixing_exec
	rescue Exception => e
		@suops.abort_operation
		Traductor::RubyErrorDialog.invoke e, @menutitle, T7[:ERR_DuringInspection]
		abort_tool
	end
end

#ROBOT_FIXING: Main State-robot function to progress along the inspection
def robot_fixing_exec
	while(action, *param = @suops.current_step) != nil	
		case action
		when :_init
			next_step = [:fixing, 0, 0]
		when :fixing	
			return if @suops.yield?
			iresolver, istep = param
			next_step = robot_fixing_execute_single(iresolver, istep)
		end	
		break if @suops.next_step(*next_step)
	end
end

#ROBOT_FIXING: Invoke a single resolver for Inspection
def robot_fixing_execute_single(iresolver, istep)
	resolver = @lst_resolvers[iresolver]
	return nil unless resolver
	resolver.execute_fixing @vbar, 1.0 / @vbar_nsteps, iresolver / @vbar_nsteps
	[:fixing, iresolver+1, 0]
end

#ROBOT_FIXING: Terminate the inspection
def robot_fixing_terminate(time)
	@processing_fixing = false

	#Aborting the Fixing
	if time < 0
		@suops.abort_operation
		@vbar.stop
		@inspection_allowed = true
		onMouseMove_zero
		return
	end
	
	#Fixing went OK
	UI.set_cursor 0
	@vbar.stop
	
	#Setting the new status
	@hsh_typical_distance = {}
	@hsh_entities_colors = {}
	@something_to_fix = false
	@geometry_fixed = true
	
	#Committing
	commit_operation
	
	#Update the number of edges and faces
	compute_numbers
	
	#Enabling inspection
	@inspection_allowed = true

	#Re-inspection if allowed
	if @hsh_parameters[:reinspect]
		refresh_viewport
		resolver_inspection
	end	
	
	onMouseMove_zero
end

#--------------------------------------------------------------
#--------------------------------------------------------------
# PALETTE: Palette Management
#--------------------------------------------------------------
#--------------------------------------------------------------

#PALETTE: Separator for the Main and floating palette
def pal_separator ;	@palette.declare_separator ; end

#PALETTE: Initialize the main palette and top level method
def init_palette
	hshpal = { :no_message => true, :no_main => true, :key_registry => :edgeinspector }
	@palette = Traductor::Palette.new hshpal
	@draw_local = self.method "draw_button_opengl"
	@blason_proc = self.method "draw_button_blason"
	@ogl = Traductor::OpenGL_6.new
	@pal_float = :pal_floating_param
				
	#Palette for parameters
	palette_floating
	
	#Palette for the magnifier
	@magnifier.palette_contribution(@palette)
	
	#Associating the palette
	set_palette @palette
end

#PALETTE: Show or hide the palette
def palette_show(flag_show)
	@palette.show_hide_floating @pal_float, flag_show
end

#PALETTE: Floating Palette
def palette_floating
	#Parameters for the layout
	bk_color = 'lightblue'
	hi_color = 'lightgreen'
	val_color = 'gold'
	tit_color = 'thistle'
		
	#Declare the floating palette
	flpal = :pal_floating_param
	hsh = { :title => T7[:TIT_FixingResults] }
	@palette.declare_floating flpal, hsh
		
	#Building the texts and tips
	htexts = {}
	wtext = 0
	@lst_params.each do |symb|
		htexts[symb] = text = T7["BOX_Repair_#{symb}".intern]
		w, = G6.simple_text_size text
		wtext = w if w > wtext
	end
	wtext = [wtext + 20, 130].max
	
	wsepa = 8
	wcheck = 16
	wsepa2 = wsepa / 2
	height = 20
	wdessin = 32
	wval = 110
	wresult = 80
	wpcflag = 32
	wsign = height
	wtot = wdessin + wtext + wval + wpcflag + wresult + wsign
	row = -1

	#Table for number of faces and edges
	row += 1
	palette_table_edges_faces(flpal, row, wtot)
	
	#Vertical separator	
	row += 1
	hsh = { :floating => flpal, :passive => true, :height => 2, :width => wtot, :row => row, :bk_color => 'gold' }
	@palette.declare_button :pal_sepf_v0, hsh
	
	#Title row for repairs
	row += 1
	hshb = { :floating => flpal, :row => row, :height => 16 }
	hsh = { :width => wcheck, :bk_color => tit_color, :draw_proc => @draw_local, :tooltip => T6[:T_BUTTON_SelectAll] }
	@palette.declare_button(:pal_rep_select_all, hshb, hsh) { parameter_select_all true }

	hshb = { :floating => flpal, :row => row, :height => 16 }
	hsh = { :width => wcheck, :bk_color => tit_color, :draw_proc => @draw_local, :tooltip => T6[:T_BUTTON_UnSelectAll] }
	@palette.declare_button(:pal_rep_select_none, hshb, hsh) { parameter_select_all false }

	hsh = { :passive => true, :width => wsepa, :bk_color => tit_color, :draw_proc => :separator_V }
	@palette.declare_button(:pal_rep_sepa1, hshb, hsh)

	hsh = { :passive => true, :width => wtext + wval + wpcflag - wsepa - wsepa2, :bk_color => tit_color, 
	        :text => T7[:BOX_TypeRepair], :justif => 'MH', :tooltip => T7[:TIP_TypeRepair] }
	@palette.declare_button(:pal_rep_type, hshb, hsh)

	hsh = { :passive => true, :width => wsepa, :bk_color => tit_color, :draw_proc => :separator_V }
	@palette.declare_button(:pal_rep_sepa2, hshb, hsh)

	hsh = { :passive => true, :width => wsign + wresult - wsepa2, :bk_color => tit_color, 
	        :text => T7[:BOX_Results], :justif => 'MH', :tooltip => T7[:TIP_Results] }
	@palette.declare_button(:pal_rep_found, hshb, hsh)
	
	#Creating the rows for parameters buttons
	@lst_params.each do |symb|
		row += 1
		height2 = height / 2

		text = htexts[symb]
		tip = T7["TIP_Repair_#{symb}".intern]
		
		#Button for flag/value or just flag
		hshb_dessin = { :floating => flpal, :passive => true, :row => row, :width => wdessin, :height => height, 
		                :draw_proc => @draw_local, :tooltip => tip }
		if @lst_params_tol.include?(symb)
			widb = wtext
			hshb_val = { :floating => flpal, :height => height, :width => wval, :justif => 'MH'}
			type = :tol
		elsif @lst_params_tol_big.include?(symb)
			widb = wtext
			hshb_val = { :floating => flpal, :height => height, :width => wval, :justif => 'MH'}
			type = :tol_big
		elsif @lst_params_flag.include?(symb)
			widb = wtext
			hshb_val = { :floating => flpal, :tooltip => 'tip flag', :height => height, :width => wval + wpcflag, :justif => 'MH'}
			type = :flag
		else
			widb = wtext + wval + wpcflag
			hshb_val = nil	
			type = nil
		end
		hshb_text = { :floating => flpal, :hi_color => hi_color, :text => text, :tooltip => tip,
					  :width => widb, :height => height, :justif => 'LH'}
		palette_parameters_button symb, hi_color, hshb_dessin, hshb_text, wpcflag, type, hshb_val
		
		#Button for sign and result
		hshb_sign = { :floating => flpal, :passive => true, :draw_proc => @draw_local, :tooltip => tip, :width => wsign, :height => height }
		hshb_result = { :floating => flpal, :passive => true, :tooltip => tip, :width => wresult, :height => height, :justif => 'RH'}
		palette_parameters_button_result(symb, hshb_sign, hshb_result)		
	end
	
	#Vertical separator	
	row += 1
	hsh = { :floating => flpal, :passive => true, :height => 2, :width => wtot, :row => row, :bk_color => 'gold' }
	@palette.declare_button :pal_sepf_v1, hsh
	
	#Buttons for Inspecting
	row += 1
	hrow = 32
	wsepa = 8
	wbut = (wtot - 3 * hrow - 2 * wsepa) / 2
	hshfloat = { :floating => flpal, :width => wbut, :height => hrow }
	
	pal_symb = "#{flpal}_inspect".intern
	grayed_proc = proc { !@inspection_allowed }
	hsh = { :row => row, :bk_color => 'yellow', :text => T7[:BOX_Inspect], :tooltip => T7[:TIP_Inspect], 
	        :justif => 'MH', :grayed_proc => grayed_proc, :draw_proc => :magnifier }
	@palette.declare_button(pal_symb, hshfloat, hsh) { resolver_inspection }

	#Separator
	hsh = { :floating => flpal, :sepa => true, :passive => true, :height => hrow, :width => wsepa, :draw_proc => :separator_V }
	@palette.declare_button :pal_sepf_0, hsh

	#Buttons for Fixing
	pal_symb = "#{flpal}_fix".intern
	grayed_proc = proc { !@something_to_fix }
	hsh = { :row => row, :bk_color => 'lightgreen', :text => T7[:BOX_Fix], :tooltip => T7[:TIP_Fix], 
	        :justif => 'MH', :grayed_proc => grayed_proc, :draw_proc => :valid }
	@palette.declare_button(pal_symb, hshfloat, hsh) { resolver_fixing }
	
	#Separator
	hsh = { :floating => flpal, :sepa => true, :passive => true, :height => hrow, :width => wsepa, :draw_proc => :separator_V }
	@palette.declare_button :pal_sepf_1, hsh
	
	#Abort and Exit
	hshfloat = { :floating => flpal, :width => hrow, :height => hrow }
	hsh = { :draw_proc => :std_abortexit, :main_color => 'red', :frame_color => 'green', 
			:tooltip => T6[:T_STR_AbortTool] }
	@palette.declare_button(:pal_abort_f, hshfloat, hsh) { abort_tool }

	#Rollback
	grayed_proc = proc { !rollback_allowed? }
	tip_proc = proc { rollback_tip }
	hsh = { :tip_proc => tip_proc, :draw_proc => :rollback, :grayed_proc => grayed_proc }
	@palette.declare_button(:pal_back_f, hshfloat, hsh) { rollback true }

	#Exit tool
	hsh = { :draw_proc => :std_exit, :tooltip => T6[:T_STR_ExitTool] }
	@palette.declare_button(:pal_exit_f, hshfloat, hsh) { exit_tool }
	
	#Message
	palette_statistics_message(flpal, wtot, row+1)
end

#PALETTE: Message zone for warning
def palette_statistics_message(flpal, wtot, row)
	hidden_proc = proc { !@message_statistics }
	hshc = { :floating => flpal, :passive => true, :width => wtot, :hidden_proc => hidden_proc }
	
	hsh = { :row => row, :height => 2, :bk_color => 'gold' }
	@palette.declare_button :pal_sepf_v4, hshc, hsh
	
	text_proc = proc { @message_statistics }
	hsh = { :row => row+1, :height => 44, :bk_color => 'pink', :text_proc => text_proc, :justif => 'LH' }
	@palette.declare_button(:pal_statistics, hshc, hsh)
end

#Button for flag
def palette_parameters_button(symb, hi_color, hshb_dessin, hshb_text, wpcflag, type=nil, hshb_val=nil)
	check_symb = "check_#{symb}".intern
	bk_proc = proc { (@hsh_parameters[check_symb]) ? hi_color : nil }
	
	pal_symb = "pal_param_dessin_#{symb}".intern
	hsh = { :bk_color => bk_proc }
	@palette.declare_button(pal_symb, hshb_dessin, hsh)
	
	pal_symb = "pal_param_tit_#{symb}".intern
	double_click_proc = proc { parameter_set_only check_symb }
	hsh = { :bk_color => bk_proc, :double_click_proc => double_click_proc }
	@palette.declare_button(pal_symb, hshb_text, hsh) { parameter_toggle check_symb }
	
	if type == :tol || type == :tol_big
		palette_parameters_tolerance(symb, check_symb, hshb_val, bk_proc, wpcflag, type)
		
	elsif type == :flag
		flag_symb = "flag_#{symb}".intern
		value_proc = proc { @hsh_parameters[flag_symb] }
		tip = T7["TIP_Repair_Flag_#{symb}".intern]
		text = T7["BOX_Repair_Flag_#{symb}".intern]
		pal_symb = "pal_param_flag_#{symb}".intern
		hsh = { :value_proc => value_proc, :bk_color => bk_proc, :hi_color => bk_proc, :text => text, :tooltip => tip }
		@palette.declare_button(pal_symb, hshb_val, hsh) { 	parameter_set(check_symb, true) ; parameter_toggle(flag_symb) }
	
	end
end

#PALETTE: Button for tolerance value as length
def palette_parameters_tolerance(symb, check_symb, hshb_val, bk_proc, wpcflag, type)
	#hard coded parameter
	if type == :tol_big
		tolmax = nil
		tolincr = nil
		percent_max = @percent_max_big
		percent_min = @percent_min_big
		percent_incr = @percent_incr_big
	else
		tolmax = @tolerance_max
		tolincr = @tolerance_incr
		percent_max = @percent_max
		percent_min = @percent_min
		percent_incr = @percent_incr
	end
	
	#Tolerance as length
	tol_symb = "tolerance_#{symb}".intern
	percent_flag_symb = "percent_flag_#{symb}".intern
	validate_proc = proc { |val| unit_convert val }
	vprompt_proc = proc { unit_prompt symb }
	get_proc = proc { @hsh_parameters[tol_symb].to_f }
	set_proc = proc { |val| parameter_set(check_symb, true) ; parameter_set(tol_symb, val) }
	show_proc = proc { |val| G6.format_length_general(val) }
	hidden_proc = proc { @hsh_parameters[percent_flag_symb] }
	hshi = { :vtype => :float, :vprompt_proc => vprompt_proc, :vmin => @tolerance_min, :vmax => tolmax, :vincr => tolincr, 
			 :vsprintf => show_proc, :vbutsprintf => show_proc, :get_proc => get_proc, :set_proc => set_proc,
			 :validate_proc => validate_proc }
	input = Traductor::InputField.new hshi
	
	pal_symb = "pal_param_val_#{symb}".intern
	hsh = { :input => input, :bk_color => bk_proc, :hidden_proc => hidden_proc, :tip_proc => vprompt_proc }
	@palette.declare_button pal_symb, hshb_val, hsh 	
	
	#Tolerance as percent
	per_symb = "percent_value_#{symb}".intern
	percent_flag_symb = "percent_flag_#{symb}".intern
	vprompt_proc = proc { percent_prompt symb }
	get_proc = proc { @hsh_parameters[per_symb].to_f }
	set_proc = proc { |val| parameter_set(check_symb, true) ; parameter_set(per_symb, val) }
	hshi = { :vtype => :percent, :vprompt_proc => vprompt_proc, :vmin => percent_min, :vmax => percent_max, :vincr => percent_incr, 
			 :get_proc => get_proc, :set_proc => set_proc, :vsprintf => "%0.1f%" }
	input = Traductor::InputField.new hshi
	
	hidden_proc = proc { !@hsh_parameters[percent_flag_symb] }
	pal_symb = "pal_param_percent_#{symb}".intern
	hsh = { :input => input, :bk_color => bk_proc, :hidden_proc => hidden_proc, :tip_proc => vprompt_proc }
	@palette.declare_button pal_symb, hshb_val, hsh 	
	
	#Button to toggle length / percent
	value_proc = proc { !@hsh_parameters[percent_flag_symb] }
	pal_symb = "pal_param_percent_flag_#{symb}".intern
	hsh = { :value_proc => value_proc, :bk_color => bk_proc, :width => wpcflag, :bk_color => bk_proc, 
	        :hi_color => bk_proc, :draw_proc => @draw_local, :tooltip => T7[:TIP_TogglePercentLength] }
	@palette.declare_button(pal_symb, hshb_val, hsh) { parameter_toggle percent_flag_symb }	
end

#Manage the unit format metric or architectural
def unit_format(len)
	archi = G6.model_units_architectural?
	format = "%0.2f"
	if archi
		len = len.to_mm
		text = "#{sprintf format, len}\""
	else
		text = "#{sprintf format, len} mm"
	end	
	text
end

#Convert to the unit format metric or architectural
def unit_convert(sval)
	val = Traductor.string_to_float_formula sval
	return nil unless val
	unless G6.model_units_architectural?
		val = val.mm.to_inch
	end
	val
end

def unit_prompt(symb)
	prompt = T7["PROMPT_Repair_#{symb}".intern]
	if G6.model_units_architectural?
		prompt +=  " (inch)"
	else	
		prompt +=  " (mm)"
	end
	prompt
end

def percent_prompt(symb)
	prompt = T7["PROMPT_Repair_#{symb}".intern]
	prompt +=  " (#{T7[:TIP_PercentTolerance]})"
	prompt
end

#PALETTE: Button for results
def palette_parameters_button_result(symb, hshb_sign, hshb_result)
	bk_proc = proc { get_nb_repairs_color(symb) }
	
	pal_symb = "pal_param_sign_#{symb}".intern
	hsh = { :bk_color => bk_proc }
	@palette.declare_button(pal_symb, hshb_sign, hsh)
	
	pal_symb = "pal_param_result_#{symb}".intern
	text_proc = proc { get_nb_repairs_text(symb) }
	hsh = { :text_proc => text_proc, :bk_color => bk_proc }
	@palette.declare_button(pal_symb, hshb_result, hsh)
end

#PALETTE: Draw the boxes for the numbers of edges and faces
def palette_table_edges_faces(flpal, row, wtot)
	#pal_separator
	
	height = 22
	wbla = 32
	wbar = 2
	hsepa = 8
	wsepa = 8
	htot = wbla * 2 + hsepa
	txe = T6[:T_TXT_Edges]
	txf = T6[:T_TXT_Faces]
	we, = G6.simple_text_size txe
	wf, = G6.simple_text_size txf
	w = [we, wf].max + 15
	color_ini = 'powderblue'
	color_cur = 'lightskyblue'
	color_fix = 'deepskyblue'
	color_bar = 'blue'
	space = '  '
	wnum = (wtot - w - wbla - 2 * wbar - wsepa) / 3
	
	hsh_float = { :floating => flpal, :row => row, :height => height }
	hsh_bar = { :floating => flpal, :passive => true, :row => row, :height => wbar, :bk_color => color_bar }

	#Blason and reinspect flag
	hshc = { :floating => flpal, :row => row, :height => wbla, :width => wbla }
	
	tip = "EdgeInspector" + ' ' + T6[:T_STR_DefaultParamDialog]
	hsh = { :draw_proc => @blason_proc, :tooltip => tip, :rank => 2 }
	@palette.declare_button(:blason, hshc, hsh) { MYPLUGIN.invoke_default_parameter_dialog }

	hsh = { :passive => true, :draw_proc => :separator_H, :height => hsepa, :rank => 1 }
	@palette.declare_button(:pal_sepa_reinspect, hshc, hsh)

	tip = T7[:TIP_Reinspect]
	value_proc = proc { @hsh_parameters[:reinspect] }
	hsh = { :value_proc => value_proc, :draw_proc => @draw_local, :tooltip => tip, :hi_color => 'yellow' }
	@palette.declare_button(:pal_reinspect, hshc, hsh) { parameter_toggle :reinspect }

	#Horizontal separator
	hsh = { :floating => flpal, :passive => true, :row => row, :height => htot, :width => wsepa, :draw_proc => :separator_V }
	@palette.declare_button(:pal_horiz_bar1, hsh)
	
	#Vertical separator
	hsh = { :floating => flpal, :passive => true, :row => row, :height => htot, :width => wbar, :bk_color => color_bar }
	@palette.declare_button(:pal_vert_bar1, hsh)
	
	#Titles
	hshc = { :passive => true, :width => w, :bk_color => color_ini, :justif => 'LH' }
	
	hsh = { :rank => 5, :width => w  }
	@palette.declare_button(:pal_tit_bar, hsh_bar, hsh)

	hsh = { :rank => 4, :tooltip => T7[:TIP_NbEdgesTitle] }
	@palette.declare_button(:pal_tit_void, hsh_float, hshc, hsh)

	hsh = { :rank => 3, :width => w  }
	@palette.declare_button(:pal_tit_mid, hsh_bar, hsh)

	hsh = { :text => txe, :rank => 2, :tooltip => T7[:TIP_NbEdgesTitle] }
	@palette.declare_button(:pal_tit_edges, hsh_float, hshc, hsh)
	
	hsh = { :text => txf, :rank => 1, :tooltip => T7[:TIP_NbFacesTitle] }
	@palette.declare_button(:pal_tit_faces, hsh_float, hshc, hsh)
	
	hsh = { :width => w  }
	@palette.declare_button(:pal_tit_bot, hsh_bar, hsh)

	#Original Numbers
	hshn = { :passive => true, :width => wnum, :bk_color => color_ini, :justif => 'RH' }
	
	hsh = { :rank => 5, :width => wnum  }
	@palette.declare_button(:pal_ini_bar, hsh_bar, hsh)

	text = T7[:BOX_Original] + space
	hsh = { :text => text, :rank => 4, :tooltip => T7[:TIP_NbEdgesOriginal] }
	@palette.declare_button(:pal_nb_ini_tit, hsh_float, hshn, hsh)

	hsh = { :rank => 3, :width => wnum  }
	@palette.declare_button(:pal_ini_mid, hsh_bar, hsh)

	text_proc = proc { @tx_edges_original }
	hsh = { :text_proc => text_proc, :rank => 2, :tooltip => T7[:TIP_NbEdgesOriginal] }
	@palette.declare_button(:pal_nb_ini_edges, hsh_float, hshn, hsh)
	
	text_proc = proc { @tx_faces_original }
	hsh = { :text_proc => text_proc, :rank => 1, :tooltip => T7[:TIP_NbFacesOriginal] }
	@palette.declare_button(:pal_nb_ini_faces, hsh_float, hshn, hsh)
	
	hsh = { :width => wnum  }
	@palette.declare_button(:pal_ini_bot, hsh_bar, hsh)

	#Current Numbers
	hshn = { :passive => true, :width => wnum, :bk_color => color_cur, :justif => 'RH' }

	hsh = { :rank => 5, :width => wnum  }
	@palette.declare_button(:pal_cur_bar, hsh_bar, hsh)

	text = T7[:BOX_BeforeFix] + space
	hsh = { :text => text, :rank => 4, :tooltip => T7[:TIP_NbEdgesCurrent] }
	@palette.declare_button(:pal_nb_cur_tit, hsh_float, hshn, hsh)

	hsh = { :rank => 3, :width => wnum  }
	@palette.declare_button(:pal_cur_mid, hsh_bar, hsh)

	text_proc = proc { @tx_edges_initial }
	hsh = { :text_proc => text_proc, :rank => 2, :tooltip => T7[:TIP_NbEdgesCurrent] }
	@palette.declare_button(:pal_nb_cur_edges, hsh_float, hshn, hsh)
	
	text_proc = proc { @tx_faces_initial }
	hsh = { :text_proc => text_proc, :rank => 1, :tooltip => T7[:TIP_NbFacesCurrent] }
	@palette.declare_button(:pal_nb_cur_faces, hsh_float, hshn, hsh)	

	hsh = { :width => wnum  }
	@palette.declare_button(:pal_cur_bot, hsh_bar, hsh)

	#After Fixing Numbers
	hshn = { :passive => true, :width => wnum, :bk_color => color_fix, :justif => 'RH' }

	hsh = { :rank => 5, :width => wnum  }
	@palette.declare_button(:pal_fix_bar, hsh_bar, hsh)

	text = T7[:BOX_AfterFix] + space
	hsh = { :text => text, :rank => 4, :tooltip => T7[:TIP_NbEdgesFix] }
	@palette.declare_button(:pal_nb_fix_tit, hsh_float, hshn, hsh)

	hsh = { :rank => 3, :width => wnum  }
	@palette.declare_button(:pal_fix_mid, hsh_bar, hsh)

	text_proc = proc { @tx_edges_after_fix }
	hsh = { :text_proc => text_proc, :rank => 2, :tooltip => T7[:TIP_NbEdgesFix] }
	@palette.declare_button(:pal_nb_fix_edges, hsh_float, hshn, hsh)
	
	text_proc = proc { @tx_faces_after_fix }
	hsh = { :text_proc => text_proc, :rank => 1, :tooltip => T7[:TIP_NbFacesFix] }
	@palette.declare_button(:pal_nb_fix_faces, hsh_float, hshn, hsh)	
	
	hsh = { :width => wnum  }
	@palette.declare_button(:pal_fix_bot, hsh_bar, hsh)

	#Vertical separator
	hsh = { :floating => flpal, :passive => true, :row => row, :height => htot, :width => wbar, :bk_color => color_bar }
	@palette.declare_button(:pal_vert_bar2, hsh)	
end

#PALETTE: Drawing the Blason
def draw_button_blason(symb, dx, dy)
	lst_gl = []
	dx2 = dx / 2
	dy2 = dy / 2
	
	lst_gl += @ogl.draw_proc :std_blason, dx, dy, 'khaki', 'gold'
	
	lpti = [[3, 3], [dx2-3, dy2+3]]
	pts = lpti.collect { |a| Geom::Point3d.new *a }
	lst_gl.push [GL_LINE_STRIP, pts, 'blue', 2, '']	

	lpti = [[dx2-2, dy2+5], [dx-3, dy-4]]
	pts = lpti.collect { |a| Geom::Point3d.new *a }
	lst_gl.push [GL_LINE_STRIP, pts, 'red', 2, '']	

	lst_gl += @ogl.draw_proc :magnifier, dx, dy
	
	lst_gl
end

#PALETTE: Custom drawing of buttons
def draw_button_opengl(symb, dx, dy, main_color, frame_color, selected, grayed)
	code = symb.to_s
	lst_gl = []
	dx2 = dx / 2
	dy2 = dy / 2
	color = (grayed) ? 'gray' : frame_color
	
	case code
	
	when /reinspect/i
		lst_gl += @ogl.draw_proc :std_redo, dx, dy, nil, 'purple'
		lst_gl += @ogl.draw_proc :magnifier, dx, dy, nil, nil, 0.6
	
	when /select_all/
		pts = [Geom::Point3d.new(dx2-4, dy2-4), Geom::Point3d.new(dx2+4, dy2+4),
		       Geom::Point3d.new(dx2-4, dy2+4), Geom::Point3d.new(dx2+4, dy2-4)]
		lst_gl.push [GL_LINES, pts, 'black', 1, '']
		pts = G6.pts_square dx2, dy2, 5
		lst_gl.push [GL_LINE_LOOP, pts, 'black', 1, '']
		
	when /select_none/
		pts = G6.pts_square dx2, dy2, 5
		lst_gl.push [GL_LINE_LOOP, pts, 'black', 1, '']
		
	when /_bandeau/i
		pts = []
		pts.push Geom::Point3d.new(0, 0)
		pts.push Geom::Point3d.new(dx, 0)
		pts.push Geom::Point3d.new(dx, dy)
		pts.push Geom::Point3d.new(0, dy)
		lst_gl.push [GL_POLYGON, pts, 'green']
		
	when /param_percent_flag_/i
		dec = 4
		len = dx - dec * 2
		n = 3
		dy2 = dy2 - 3
		dx4 = len/n
		lpti = [[dec, dy2], [dx-dec, dy2]]
		lpti2 = []
		for i in 0..n
			x = dec + i * dx4
			lpti2.push [x, dy2], [x, dy2+4]
		end
		color = 'blue'
		pts = lpti.collect { |a| Geom::Point3d.new *a }
		lst_gl.push [GL_LINES, pts, color, 2, '']
		pts = lpti2.collect { |a| Geom::Point3d.new *a }
		lst_gl.push [GL_LINES, pts, color, 1, '']
		
	when /param_dessin_overlap/i
		y = dy2 + 4
		pts = [Geom::Point3d.new(dx2-7, y+1), Geom::Point3d.new(dx2+7, y+1)]
		lst_gl.push [GL_LINE_STRIP, pts, 'red', 2, '']
		pts = [Geom::Point3d.new(2, y-1), Geom::Point3d.new(dx-2, y-1)]
		lst_gl.push [GL_LINE_STRIP, pts, 'blue', 2, '']

		y = dy2 - 4
		pts = [Geom::Point3d.new(2, y+1), Geom::Point3d.new(dx2+3, y+1)]
		lst_gl.push [GL_LINE_STRIP, pts, 'red', 2, '']
		pts = [Geom::Point3d.new(dx2-3, y-1), Geom::Point3d.new(dx-2, y-1)]
		lst_gl.push [GL_LINE_STRIP, pts, 'blue', 2, '']
		
	when /param_dessin_needle_eye/i
		dx3 = (dx-2) / 3
		dy3 = (dy-2) / 3
		lpti = [[2,2], [dx3, dy2-2], [2* dx3, dy2+dy3], [dx-2, 3]]
		pts = lpti.collect { |a| Geom::Point3d.new *a }
		lst_gl.push [GL_LINE_STRIP, pts, 'red', 2, '']
		lpti = [[2,2], [dx3+1, dy2+2], [2* dx3-2, dy2+dy3-2], [dx-2, 3]]
		pts = lpti.collect { |a| Geom::Point3d.new *a }
		lst_gl.push [GL_LINE_STRIP, pts, 'blue', 2, '']

	when /param_dessin_lost_junction/i
		dx4 = (dx-2) / 4
		dy4 = (dy-2) / 4
		lpti = [[2,2], [2+dx4, dy2]]
		pts = lpti.collect { |a| Geom::Point3d.new *a }
		lst_gl.push [GL_LINE_STRIP, pts, 'red', 2, '']
		lpti = [[2+dx4, dy2], [dx2, dy-2]]
		pts = lpti.collect { |a| Geom::Point3d.new *a }
		lst_gl.push [GL_LINE_STRIP, pts, 'red', 3, '.']
		lpti = [[dx2, dy-2], [dx2+2+dx4, dy2]]
		pts = lpti.collect { |a| Geom::Point3d.new *a }
		lst_gl.push [GL_LINE_STRIP, pts, 'blue', 3, '.']
		lpti = [[dx2+2+dx4, dy2], [dx-2, 2]]
		pts = lpti.collect { |a| Geom::Point3d.new *a }
		lst_gl.push [GL_LINE_STRIP, pts, 'blue', 2, '']
		
	when /param_dessin_coplanar_edge/i
		dec = 5
		pts1 = G6.pts_square dx2, dy2, dx2-dec
		pts2 = [Geom::Point3d.new(dx2, 0), Geom::Point3d.new(dx2, dy)]
		lst_gl.push [GL_POLYGON, pts1, 'white']
		lst_gl.push [GL_LINE_STRIP, pts2, 'red', 2, '-']
		lst_gl.push [GL_LINE_LOOP, pts1, 'blue', 2, '']

	when /param_dessin_zero_edge/i
		y1 = dy2+2
		y2 = dy2-2
		pts = [Geom::Point3d.new(2, y1), Geom::Point3d.new(dx2-3, y1), Geom::Point3d.new(dx2+3, y2), Geom::Point3d.new(dx-2, y2)]
		lst_gl.push [GL_LINES, pts, 'blue', 2, '']
		pts = [Geom::Point3d.new(dx2-3, y1), Geom::Point3d.new(dx2+3, y2)]
		lst_gl.push [GL_LINE_STRIP, pts, 'red', 2, '']

	when /param_dessin_tiny_gap/i
		y = dy2
		lpti = [[2, dy2+4], [dx2-2, dy2], [2, dy2-4]]
		pts = lpti.collect { |a| Geom::Point3d.new *a }
		lst_gl.push [GL_LINE_STRIP, pts, 'blue', 2, '']
		lpti = [[dx-2, dy2+4], [dx2+2, dy2], [dx-2, dy2-4]]
		pts = lpti.collect { |a| Geom::Point3d.new *a }
		lst_gl.push [GL_LINE_STRIP, pts, 'red', 2, '']

	when /param_dessin_split_edge/i
		y = dy2
		pts = [Geom::Point3d.new(2, y), Geom::Point3d.new(dx-2, y)]
		lst_gl.push [GL_LINES, pts, 'blue', 2, '']
		pts = G6.pts_square dx2, dy2, 3
		lst_gl.push [GL_POLYGON, pts, 'red']

	when /param_sign_overlap/
		pts = [Geom::Point3d.new(2, 2), Geom::Point3d.new(dx-2, dy-2)]
		lst_gl.push [GL_LINE_STRIP, pts, @color_wf_overlap, 2, '']

	when /param_sign_needle_eye/
		pts = [Geom::Point3d.new(2, 2), Geom::Point3d.new(dx-2, dy-2)]
		lst_gl.push [GL_LINE_STRIP, pts, @color_wf_needle_eye, 2, '']

	when /param_sign_coplanar_edge/
		pts = [Geom::Point3d.new(2, 2), Geom::Point3d.new(dx-2, dy-2)]
		lst_gl.push [GL_LINE_STRIP, pts, @color_wf_coplanar_edge, 4, '-']

	when /param_sign_zero_edge/
		pts = G6.pts_circle dx2, dy2, 7, 6
		lst_gl.push [GL_LINE_STRIP, pts, @color_wf_zero_edge, 2, '']
		
	when /param_sign_tiny_gap/
		pts = G6.pts_circle dx2, dy2, 7, 6
		lst_gl.push [GL_LINE_STRIP, pts, @color_wf_tiny_gap, 2, '']

	when /param_sign_split_edge/
		pts = G6.pts_circle dx2, dy2, 7, 4
		lst_gl.push [GL_LINE_STRIP, pts, @color_wf_split_edge, 2, '']		
		
	when /param_sign_lost_junction/
		pts = G6.pts_circle dx2, dy2, 7, 4
		lst_gl.push [GL_LINE_STRIP, pts, @color_wf_lost_junction, 2, '']		
		
	end	#case code
	
	lst_gl
end

end	#class EdgeInspectorTool

#============================================================
#============================================================
# Class RESOLVER: Core algorithms to inspect and fix Edges
#============================================================
#============================================================
	
class Resolver	

#-------------------------------------------------------------
# INIT: Instance Class initialization and information methods
#-------------------------------------------------------------

#Initialization
def initialize(lst_repairs_symb, ledges, lfaces, comp, tr, hsh_parameters)
	@model = Sketchup.active_model
	@selection = @model.selection
	@comp = comp
	@name_comp = G6.grouponent_name(comp)
	@tr = tr
	@tr_id = Geom::Transformation.new
	@hsh_parameters = hsh_parameters
	@lst_repairs_symb = lst_repairs_symb
	@len_su_minimum = 0.1
	
	#Initialization
	if comp.instance_of?(Sketchup::Group)
		@is_container = :group
	elsif comp.instance_of?(Sketchup::ComponentInstance)
		@is_container = :component
	else
		@is_container = nil
	end

	#Use native or real distance for inspection
	if @use_native_distance
		@tr_dist = @tr_id
		@tr_dist_inv = @tr.inverse
	else	
		@tr_dist = @tr
		@tr_dist_inv = @tr_id
	end	
	
	#Sampling parameters
	@nbox_div = 21
	@nvec_div = 20
	
	#Registering the edges and faces
	@entities = G6.grouponent_entities @comp
	selection_register ledges, lfaces unless @is_container
	
	#Checking the statistics
	statistics_analyze_selection
	
	#Arrays for methods
	@txt_inspecting = T7[:MSG_Inspecting]
	@txt_repairing = T7[:MSG_Repairing]
	@hsh_method_inspection = {}
	@hsh_method_resolve = {}
	@hsh_repairs_text = {}
	@lst_repairs_symb.each do |symb|
		next unless self.respond_to?("#{symb}_inspection".intern)
		@hsh_method_inspection[symb] = self.method "#{symb}_inspection"
		@hsh_method_resolve[symb] = self.method "#{symb}_resolve"
		@hsh_repairs_text[symb] = T7["BOX_Repair_#{symb}".intern]
	end
	
	#Initializing the selection
	selection_restablish
end

#-------------------------------------------------------------
# STATISTICS: Statistics analysis
#-------------------------------------------------------------

#STATISTICS: register the statistical analysis for a resolver
def statistics_info
	@stat_info
end

#STATISTICS: Compute the average length of the selection
def statistics_analyze_selection(ledall=nil)
	@factor_scale = @stat_info = nil
	ledall = @entities.grep Sketchup::Edge unless ledall
	@len_average = 0
	@len_max = 0
	@len_min = nil
	led_too_small = []
	return if ledall.empty?
	lentot = 0
	ledall.each do |edge|
		len = (@tr_dist * edge.start.position).distance (@tr_dist * edge.end.position)
		lentot += len
		@len_min = len if !@len_min || len < @len_min
		@len_max = len if len > @len_max
		led_too_small.push edge if len < @len_su_minimum
	end	
	ntot = ledall.length
	@len_average = lentot / ledall.length
	nb_too_small = led_too_small.length
	
	#Selection is too small - Should be scale
	if @len_average < @len_su_minimum || nb_too_small.to_f / ntot > 0.1
		factor_scale = ((@len_su_minimum / @len_average).round + 1) * 2
		@stat_info = [factor_scale, nb_too_small]
	end
end

#-------------------------------------------------------------
# SELECTION: Selection Management
#-------------------------------------------------------------
	
#SELECTION: Register the edges which are initially in selection and those which are out	
def selection_register(ledges, lfaces)
	@entities = G6.grouponent_entities @comp

	#Registering Edges
	@hsh_edges_in = {}
	ledges.each { |e| @hsh_edges_in[e.entityID] = e }
	@hsh_edges_out = {}
	ledall = @entities.grep Sketchup::Edge
	ledall.each do |e|
		eid = e.entityID
		@hsh_edges_out[eid] = e unless @hsh_edges_in[eid]
	end	
		
	#Registering Faces
	@hsh_faces_in = {}
	lfaces.each { |e| @hsh_faces_in[e.entityID] = e }
	@hsh_faces_out = {}
	ledall = @entities.grep Sketchup::Face
	ledall.each do |e|
		eid = e.entityID
		@hsh_faces_out[eid] = e unless @hsh_faces_in[eid]
	end	
end

#SELECTION: Recalculate the edges in selection
def selection_restablish
	#For Groups and Component Instances
	if @is_container
		@comp.make_unique if @is_container == :group && !G6.grouponent_unique?(@comp)
		@entities = G6.grouponent_entities @comp
		@initial_edges = @entities.grep Sketchup::Edge
		@initial_faces = @entities.grep Sketchup::Face

	#For top active model
	else
		@entities = @model.active_entities
		@initial_edges = @entities.grep(Sketchup::Edge).find_all { |e| e.valid? && !@hsh_edges_out[e.entityID] }
		@initial_faces = @entities.grep(Sketchup::Face).find_all { |e| e.valid? && !@hsh_faces_out[e.entityID] }
	end	
	@nb_edges_initial = @initial_edges.length
	@nb_faces_initial = @initial_faces.length
end

#SELECTION: Racalculate the edges in selection
def recompute_edges_in
	ledall = @entities.grep Sketchup::Edge
	return ledall if @is_container
	ledall.find_all { |e| e.valid? && !@hsh_edges_out[e.entityID] }
end

#SELECTION: Recalculate the faces in selection	
def recompute_faces_in
	ledall = @entities.grep Sketchup::Face
	return ledall if @is_container 
	ledall.find_all { |e| e.valid? && !@hsh_faces_out[e.entityID] }
end
	
#SELECTION: Information methods
def get_initial_nb_edges ; @nb_edges_initial ; end
def get_initial_nb_faces ; @nb_faces_initial ; end

#SELECTION: Recompute the map of entityID
def map_id(hmap_id)
	@entities.each do |e|
		if e.instance_of?(Sketchup::Edge)
			hmap_id[e.entityID] = e 
			hmap_id[e.start.entityID] = e.start 
			hmap_id[e.end.entityID] = e.end 
		end			
	end	
end

#-------------------------------------------------------------
# EXECUTE: Top Calculation methods
#-------------------------------------------------------------

#EXECUTE: Proceed with a rollback
def execute_rollback
	selection_restablish
end

#EXECUTE: Calculation of effective tolerance
def effective_tolerance(symb)
	flag_symb = "percent_flag_#{symb}".intern
	if @hsh_parameters[flag_symb]
		percent_symb = "percent_value_#{symb}".intern
		percent = @hsh_parameters.fetch percent_symb, 0.01
		tol = @len_average * percent
	else
		len_symb = "tolerance_#{symb}".intern
		tol = @hsh_parameters.fetch len_symb, 0.01
	end
	#[tol, 0.05].max	
	tol
end

#EXECUTE: Getting the values of parameters
def execute_parameters(fixing=false)
	@hsh_flags = {}
	@lst_repairs_symb.each do |symb|
		@hsh_flags[symb] = @hsh_parameters.fetch "check_#{symb}".intern, true
	end
	
	unless fixing
		@use_native_distance = MYDEFPARAM[:DEFAULT_EdgeInspector_NativeDistance]
		@tolerance_overlap = effective_tolerance :overlap
		@tolerance_zero_edge = effective_tolerance :zero_edge
		@tolerance_tiny_gap = effective_tolerance :tiny_gap
	end
	
	@lst_repairs_symb_eff = @lst_repairs_symb.find_all { |symb| @hsh_method_inspection[symb] && @hsh_flags[symb] }
end

#EXECUTE: Analyze all edges et detect required defects	
def execute_inspection(hsh_all_reparations, hsh_all_rep_ref, vbar, vbar_step, elapsed_pc)
	return if @initial_edges.empty?
	@hsh_all_reparations = hsh_all_reparations
	@hsh_all_rep_ref = hsh_all_rep_ref
	@vbar = vbar
	
	#Getting parameters
	execute_parameters
	
	#Visual bar Initialization
	nsteps = @lst_repairs_symb_eff.length * 1.0
	text_ef = " [#{@nb_edges_initial} " + ((@nb_edges_initial < 2) ? T6[:T_TXT_Edge] : T6[:T_TXT_Edges])
	text_ef += ", #{@nb_faces_initial} " + ((@nb_faces_initial < 2) ? T6[:T_TXT_Face] : T6[:T_TXT_Faces]) + ']'
	@vbar.update_title @txt_inspecting + ' - ' + @name_comp + text_ef
	@vbar_pc_span = vbar_step / nsteps
	
	#Global initialization
	@hsh_pvx = nil
	
	#Proceeding with inspection
	@lst_repairs_symb_eff.each_with_index do |symb, i|
		@vbar_pc_beg = elapsed_pc + i * @vbar_pc_span
		@vbar.progression @vbar_pc_beg, @hsh_repairs_text[symb]
		@hsh_method_inspection[symb].call
	end
end
	
#EXECUTE: helper for logging mini_progression	
def vbar_mini_progression(i, out_of, modulo_val)
	return unless i.round.modulo(modulo_val) == 0
	@vbar.mini_progression 1.0 * i / out_of, @vbar_pc_span
end

#EXECUTE: Execute the fixing	
def execute_fixing(vbar, vbar_step, elapsed_pc)
	return if @initial_edges.empty?

	#Getting parameters
	@vbar = vbar
	execute_parameters true
	
	#Visual bar Initialization
	nsteps = @lst_repairs_symb_eff.length * 1.0
	text_ef = " [#{@nb_edges_initial} " + ((@nb_edges_initial < 2) ? T6[:T_TXT_Edge] : T6[:T_TXT_Edges])
	text_ef += ", #{@nb_faces_initial} " + ((@nb_faces_initial < 2) ? T6[:T_TXT_Face] : T6[:T_TXT_Faces]) + ']'
	@vbar.update_title @txt_repairing + ' - ' + @name_comp + text_ef
	@vbar_pc_span = vbar_step / nsteps
	
	#Resolution of defects in order
	@hsh_curves = {}
	@lst_repairs_symb_eff.each_with_index do |symb, i|
		@vbar_pc_beg = elapsed_pc + i * @vbar_pc_span
		@vbar.progression @vbar_pc_beg, @hsh_repairs_text[symb]
		@hsh_method_resolve[symb].call
	end

	#Reconstructive the curves exploded, if applicable
	curve_reconstruct_all
	
	#Re-establishing the selection
	selection_restablish
end
	
def execute_single_fixing(hsh_by_type)
	hsh_by_type.each do |symb, lent|
		case symb
		when :overlap
			overlap_single_fixing(lent)		
		when :needle_eye
			needle_eye_single_fixing(lent)		
		when :coplanar_edge
			coplanar_edge_single_fixing(lent)		
		when :zero_edge
			zero_edge_single_fixing(lent)		
		when :tiny_gap
			tiny_gap_single_fixing(lent)
		when :split_edge
			split_edge_single_fixing lent
		when :lost_junction
			lost_junction_single_fixing lent
		end	
	end
end
	
#-------------------------------------------------------------
# SORT: Classification and Sorting of edges
#-------------------------------------------------------------
	
#SORT: Create pseudo vertices for all edges	
def create_pseudo_vertices(ledges)
	@hsh_pvx = {}
	@hsh_pvx_on_edges = {}
	
	ledges.each do |edge|
		@hsh_pvx_on_edges[edge.entityID] = []
		[edge.start, edge.end].each do |v|
			@hsh_pvx[v.entityID] = [v, v.position]
		end	
	end
	ledges
end
		
#SORT: Classify the edge by vector direction
def edge_classify_by_vector(ledges)
	n = @nvec_div + 1
	n2 = n * n
	@hsh_edge_info = {}
	@hsh_split_alone = []
	@nb_total_edges = 0

	ledges.each do |edge|
		#Edge vector
		ptbeg = edge.start.position
		ptend = edge.end.position
		vec = ptbeg.vector_to(ptend).normalize
		
		#Computing the classification key for direction
		px = (vec.x.abs * @nvec_div).round
		py = (vec.y.abs * @nvec_div).round
		pz = (vec.z.abs * @nvec_div).round
		key = px + py * n + pz * n2
		led_start = edge.start.edges
		led_end = edge.end.edges
		len_start = led_start.length
		len_end = led_end.length
			
		#Storing the information
		hsh = @hsh_split_alone
		hsh[key] = [] unless hsh[key]
		hsh[key].push edge
		@nb_total_edges += 1
		@hsh_edge_info[edge.entityID] = [edge, key, ptbeg, ptend]
	end	
end

#SORT: Classify edge in 3D box space
def classify_in_box(ledges)
	@hsh_boxes = {}
	@hsh_box_info = {}
	
	#Size of the sampling
	n = @nbox_div + 1
	n2 = n * n
	
	#Computing the bounding box
	bb = Geom::BoundingBox.new
	ledges.each { |e| bb.add [e.start.position, e.end.position] }
	pbmin = bb.min
	pbmax = bb.max
	xmin = pbmin.x
	xmax = pbmax.x
	ymin = pbmin.y
	ymax = pbmax.y
	zmin = pbmin.z
	zmax = pbmax.z
	@dx = (xmax - xmin) / @nbox_div
	@dy = (ymax - ymin) / @nbox_div
	@dz = (zmax - zmin) / @nbox_div

	#Classifying the edges
	ledges.each do |edge|
		edge_id = edge.entityID
		ptbeg = edge.start.position
		ptend = edge.end.position

		kx1 = kx2 = ky1 = ky2 = kz1 = kz2 = 0
		if @dx != 0
			lk = [((ptbeg.x - xmin) / @dx).round, ((ptend.x - xmin) / @dx).round]
			lk = lk.reverse if lk.first > lk.last
			kx1, kx2 = lk
		end	
		if @dy != 0
			lk = [((ptbeg.y - ymin) / @dy).round, ((ptend.y - ymin) / @dy).round]
			lk = lk.reverse if lk.first > lk.last
			ky1, ky2 = lk
		end	
		if @dz != 0
			lk = [((ptbeg.z - zmin) / @dz).round, ((ptend.z - zmin) / @dz).round]
			lk = lk.reverse if lk.first > lk.last
			kz1, kz2 = lk
		end	
		
		ls_keybox = []
		for ix in kx1..kx2
			for iy in ky1..ky2
				for iz in kz1..kz2
					key = ix + n * iy + n2 * iz
					ls_keybox.push key
					@hsh_boxes[key] = [] unless @hsh_boxes[key]
					@hsh_boxes[key].push edge
				end
			end
		end	
		@hsh_box_info[edge_id] = ls_keybox
	end	
end

#-------------------------------------------------------------
# REPARATION: Reparation structure Handling
#-------------------------------------------------------------

#REPARATION: Create a reparation
def reparation_create(type, lent, d, jpts=nil)
	reparation = EdgeInspector::Reparation.new
	reparation.type = type
	lent = [lent] unless lent.class == Array
	
	#Entity already referenced in a reparation
	lent.each do |e|
		return if @hsh_all_rep_ref[e.entityID]
	end	
	
	#Storing the reparation details
	reparation.lent = lent
	reparation.lent_id = lent_id = lent.collect { |e| e.entityID }
	reparation.rep_id = rep_id = lent_id.sort.join('-')
	reparation.d = d
	reparation.tr = @tr
	reparation.parent = @comp
	reparation.resolver = self
	reparation.jpts = jpts
				
	#Storing the reparation and the reference from entities
	@hsh_all_reparations[rep_id] = reparation
	lent_id.each { |id| @hsh_all_rep_ref[id] = rep_id }
	reparation
end

#REPARATION: Check if an entity reparation should be ignored
def reparation_ignored?(e)
	return false unless e.valid? && @hsh_all_rep_ref
	rep_id = @hsh_all_rep_ref[e.entityID]
	return false unless rep_id
	rep = @hsh_all_reparations[rep_id]
	return false unless rep
	rep.ignored
end

#-------------------------------------------------------------
# CURVE: Curve Handling
#-------------------------------------------------------------

def curve_explode(edge)
	curve = edge.curve
	return unless curve
	@hsh_curves = {} unless @hsh_curves
	@hsh_curves[curve.entityID] = curve.vertices.collect { |vx| vx.position }
	edge.explode_curve
end

def curve_reconstruct_all
	#puts "Reconstruct curves = #{@hsh_curves.length}"
end

#=============================================================
#-------------------------------------------------------------
# OVERLAP: Core algorithms to remove overlapping edges
#-------------------------------------------------------------
#=============================================================
	
#OVERLAP: Inspect all edges for potential overlaps	
def overlap_inspection	
	#Creating pseudo vertices
	create_pseudo_vertices @initial_edges
	
	#Classifying the edges by vector along the 3 axes
	edge_classify_by_vector @initial_edges
	
	#Comparing edges
	@hsh_redraw_edges = {}
	@hsh_pair_edges = {}
	inb_edge = 0
	@nnt = 0
	@nncpt = 0
	@hsh_overlap_compared = {}
	@hsh_split_alone.each do |ledges|
		next unless ledges
		classify_in_box ledges
		@nn = 0
		@nncomp = 0
		ledges.each do |edge|
			vbar_mini_progression inb_edge, @nb_total_edges, 400
			inb_edge += 1
			edge_id = edge.entityID
			e, key, ptbeg, ptend = @hsh_edge_info[edge.entityID]
			boxes = @hsh_box_info[edge_id]
			boxes.each do |kbox|
				overlap_check_edge edge, ptbeg, ptend, @hsh_boxes[kbox]
			end	
		end
		@nnt += @nn
		@nncpt += @nncomp
	end
	
	#Contributing to the repair and wireframe info
	@hsh_redraw_edges.each do |eid, e|
		v1, pt1 = @hsh_pvx[e.start.entityID]
		v2, pt2 = @hsh_pvx[e.end.entityID]	
		if pt1 != pt2
			e2 = @hsh_pair_edges[eid]
			d = (@tr * e.start.position).distance_to_line [@tr * e2.start.position, @tr * e2.end.position]
			reparation_create(:overlap, [e, e2], d)
		end	
	end
end
		
#OVERLAP: Check potential overlapping of an edge with others	
def overlap_check_edge(edge, ptbeg, ptend, led_others)
	return unless led_others
	edge_id = edge.entityID
	led_others.each do |e|
		eid = e.entityID
		e, key, p1, p2 = @hsh_edge_info[eid]
		next if e == edge
		kk = "#{edge_id}-#{eid}"
		next if @hsh_overlap_compared[kk]
		@hsh_overlap_compared[kk] = true
		@nncomp += 1
		if overlap_compare_edge(edge, e)
			@hsh_overlap_compared[kk] = true
			@nn += 1	
		end
	end	
end

#OVERLAP: Compare two edges for overlapping and calculate the translation needed
def overlap_compare_edge(edge, e)
	#PseudoVertices of the main edge
	vbeg, ptbeg = @hsh_pvx[edge.start.entityID]
	vend, ptend = @hsh_pvx[edge.end.entityID]
	line = [ptbeg, ptend]
	
	#PseudoVertices of the secondary edge
	v1, pt1 = @hsh_pvx[e.start.entityID]
	v2, pt2 = @hsh_pvx[e.end.entityID]
	
	#Edges are not close enough
	p1proj = pt1.project_to_line(line)
	return false if (@tr_dist * pt1).distance(@tr_dist * p1proj) > @tolerance_overlap
	p2proj = pt2.project_to_line(line)
	return false if (@tr_dist * pt2).distance(@tr_dist * p2proj) > @tolerance_overlap
	
	#Edges vertices are close
	if (@tr_dist * pt1).distance(@tr_dist * ptbeg) <= @tolerance_overlap
		p1proj = ptbeg
	elsif (@tr_dist * pt1).distance(@tr_dist * ptend) <= @tolerance_overlap
		p1proj = ptend
	end
	if (@tr_dist * pt2).distance(@tr_dist * ptbeg) <= @tolerance_overlap
		p2proj = ptbeg
	elsif (@tr_dist * pt2).distance(@tr_dist * ptend) <= @tolerance_overlap
		p2proj = ptend
	end
	
	#Second edge is not overlapping with first edge
	g01 = (p1proj == ptbeg || p1proj == ptend)
	g02 = (p2proj == ptbeg || p2proj == ptend)
	g1 = p1proj != ptbeg && p1proj != ptend && G6.point_within_segment?(p1proj, ptbeg, ptend)
	g2 = p2proj != ptbeg && p2proj != ptend && G6.point_within_segment?(p2proj, ptbeg, ptend)
	return false unless (g01 && g02) || g1 || g2
		
	#Storing the overlap informations	
	@hsh_redraw_edges[e.entityID] = e
	@hsh_pair_edges[e.entityID] = edge
	
	@hsh_pvx[e.start.entityID] = [v1, p1proj]
	@hsh_pvx[e.end.entityID] = [v2, p2proj]	
	
	lvx = @hsh_pvx_on_edges[edge.entityID]
	@hsh_pvx_on_edges[e.entityID].each do |vid|
		v, pt = @hsh_pvx[vid]
		next if v == v1 || v == v2
		ptproj = pt.project_to_line(line)
		if (@tr_dist * ptproj).distance(@tr_dist * ptbeg) < @tolerance_overlap
			ptproj = ptbeg
		elsif (@tr_dist * ptproj).distance(@tr_dist * ptend) < @tolerance_overlap
			ptproj = ptend
		end	
		@hsh_pvx[vid] = [v, ptproj]
		lvx.push vid unless lvx.include?(vid)
	end

	true
end
		
#OVERLAP: Resolution and fixing of problems	
def overlap_resolve	
	#Computing the transformations
	lst_vertices = []
	lst_vectors = []
	nsteps = 1 + @hsh_redraw_edges.length
	
	@hsh_pvx.each do |vid, pvx|
		v, pt = pvx
		next unless v.valid?
		next if v.edges.find { |e| reparation_ignored?(e) }
		pt0 = v.position
		next if pt == pt0
		vec = pt0.vector_to pt
		lst_vertices.push v
		lst_vectors.push vec
	end	
	
	#Edges to be recreated
	lst_redraw_lines = []
	@hsh_redraw_edges.each do |eid, e|
		next unless e.valid?
		next if reparation_ignored?(e)
		curve_explode e
		v1, pt1 = @hsh_pvx[e.start.entityID]
		v2, pt2 = @hsh_pvx[e.end.entityID]	
		lst_redraw_lines.push [pt1, pt2]
	end

	#Applying Transformation for vertices
	vbar_mini_progression 1, nsteps, 1
	@entities.transform_by_vectors lst_vertices, lst_vectors
		
	#Redrawing edges which have been moved	
	ledges = []	
	g = @entities.add_group
	lst_redraw_lines.each_with_index do |a, i|
		pt1, pt2 = a
		vbar_mini_progression i+1, nsteps, 200
		g.entities.add_line(pt1, pt2) unless pt1 == pt2
	end	
	g.explode
end
	
#OVERLAP: Single fixing
def overlap_single_fixing(lledges)
	lvert = []
	lvec = []
	lines = []
	lledges.each do |edge1, edge2|
		line = [edge2.start.position, edge2.end.position]
		vx1 = edge1.start
		vx2 = edge1.end
		pt1 = vx1.position
		pt2 = vx2.position
		ptproj1 = pt1.project_to_line line
		ptproj2 = pt2.project_to_line line
		lvert.push vx1, vx2
		lvec.push pt1.vector_to(ptproj1), pt2.vector_to(ptproj2)
		lines.push [ptproj1, ptproj2]
	end
	
	#Moving the vertices
	@entities.transform_by_vectors lvert, lvec
	
	#Redrawing the edge
	g = @entities.add_group
	lines.each do |ptproj1, ptproj2|
		g.entities.add_line(ptproj1, ptproj2)
	end	
	g.explode
end
	
#=============================================================
#-------------------------------------------------------------
# ZERO_EDGE: Core algorithms to remove tiny edges
#-------------------------------------------------------------
#=============================================================

#ZERO_EDGE: Inspection	
def zero_edge_inspection
	tol = @tolerance_zero_edge
	
	#Sequencing the tiny edges
	lst_zero_sequences = zero_edge_sequencing
	
	#Tiny edge sequences
	lst_zero_sequences.each do |dtot, led, reason|	
		reparation_create(:zero_edge, led, dtot)
	end
end

#ZERO_EDGE: Sequencing and classification of sequences
def zero_edge_sequencing(ledges=nil)
	tol = @tolerance_zero_edge
	hsh_zero_edges = {}
	ledges = @initial_edges unless ledges
	
	ledges.each do |edge|
		pt1 = edge.start.position unless pt1
		pt2 = edge.end.position unless pt2
		d = (@tr * pt1).distance(@tr * pt2)
		dcomp = (@tr_dist * pt1).distance(@tr_dist * pt2)
		hsh_zero_edges[edge.entityID] = [edge, d] if dcomp > 0 && dcomp < tol	
	end
	
	#Finding the sequences and reason
	hseq = {}
	hsh_used = {}
	lseq = []
	hsh_zero_edges.each do |eid, a|
		edge, d = a
		next if hsh_used[eid]
		hsh_used[eid] = true
		led_right, reason_right = zero_edge_pursue(edge, edge.end, hsh_zero_edges, hsh_used)
		led_right.each { |e| hsh_used[e.entityID] = true }
		led_left, reason_left = zero_edge_pursue(edge, edge.start, hsh_zero_edges, hsh_used)
		led_left.each { |e| hsh_used[e.entityID] = true }
		led = led_left.reverse + [edge] + led_right
		if reason_right == :hook || reason_left == :hook
			reason = :hook
		elsif reason_right == :loop || reason_left == :loop
			reason = :loop
		else
			reason = :between
		end	
		dtot = 0
		led.each { |e| dtot += e.length }
		lseq.push [dtot, led, reason]
	end
	
	#Sorting the sequences
	lseq.sort { |a, b| a.first <=> b.first }
end

#ZERO_EDGE: Pursue a tiny edge in a direction
def zero_edge_pursue(edge0, vx0, hsh_zero_edges, hsh_used)
	edge_ini = edge0
	led = []
	while true
		vxedges = vx0.edges.find_all { |e| e != edge0 }
		case vxedges.length
		when 0
			return [led, :hook, nil]
		when 1
			edge0 = vxedges.first
			return [led, :loop] if hsh_used[edge0.entityID]
			return [led, :long, edge0] unless hsh_zero_edges[edge0.entityID]
			return [led, :loop] if edge0 == edge_ini
			vx0 = edge0.other_vertex(vx0)
			led.push edge0
		else
			return [led, :crossing]
		end
	end
	[led, :hook]
end

#ZERO_EDGE: Merge edge1 with edge2 - edge2 disappears
def zero_edge_merge(e1, e2)
	curve_explode e1
	curve_explode e2
	
	#Middle vertex
	lvx = e1.vertices & e2.vertices
	vx0 = lvx.first
	return nil unless vx0
	
	#Moving the middle vertex of to the end of edge2
	t0 = Time.now
	vx2 = e2.other_vertex(vx0)
	pt0 = vx0.position
	pt2 = vx2.position
	vec = pt0.vector_to(pt2)
	line = [pt2, pt2.offset(vec.axes[1], 30)]
	@entities.transform_by_vectors [vx0], [vec]

	#Transforming fake lines into edges with zero length to remove edge2
	t0 = Time.now
	g = @entities.add_group
	e = g.entities.add_line *line
	g.entities.transform_by_vectors [e.end], [e.end.position.vector_to(e.start.position)]
	g.explode
	
	#Checking the result
	(e1.valid?) ? e1 : nil
end

#ZERO_EDGE: absorb the last edge after reduction
def zero_edge_absorb_final(edge)
	vxbeg = edge.start
	vxend = edge.end
	led_beg = vxbeg.edges
	led_end = vxend.edges
	ptbeg = vxbeg.position
	ptend = vxend.position
	ptmid = Geom.linear_combination 0.5, ptbeg, 0.5, ptend
	
	#Edge has a lonely termination
	if led_beg.length == 1 || led_end.length == 1
		edge.erase!
		return
	end
	
	#Edge has a crossing on at least one side
	if led_beg.length > 2 || led_end.length > 2
		zero_edge_disolve_edge(edge, ptmid)
		return
	end
	
	#Edge is on a contour
	[1].each do |a|
		edge1 = led_beg.find { |e| e != edge }
		edge2 = led_end.find { |e| e != edge }
		vx1 = edge1.other_vertex(vxbeg)
		vx2 = edge2.other_vertex(vxend)
		vec1 = vx1.position.vector_to ptbeg
		vec2 = vx2.position.vector_to ptend
		line1 = [ptbeg, vec1]
		line2 = [ptend, vec2]
		lpt = Geom.closest_points line1, line2
		ptmid1, ptmid2 = lpt
		break unless ptmid1
		ptinter = Geom.linear_combination 0.5, ptmid1, 0.5, ptmid2
		vec_inter1 = ptbeg.vector_to ptinter
		vec_inter2 = ptend.vector_to ptinter
		break if vec1 % vec_inter1 < 0 || vec2 % vec_inter2 < 0
		vec = ptmid.vector_to ptinter
		d = edge.length
		ptinter = ptmid.offset(vec, d) if ptmid.distance(ptinter) > d
		zero_edge_disolve_edge(edge, ptinter)
		@entities.add_cpoint ptinter
		return
	end
	
	#Default if problem
	ptmid = Geom.linear_combination 0.5, ptbeg, 0.5, ptend
	zero_edge_disolve_edge(edge, ptmid)
end

#ZERO_EDGE: Shrink the edge to the <ptmid> point
def zero_edge_disolve_edge(edge, ptmid)
	t0 = Time.now
	vxbeg = edge.start
	vxend = edge.end
	vec_beg = vxbeg.position.vector_to ptmid
	vec_end = vxend.position.vector_to ptmid
	line = [ptmid, ptmid.offset(vec_beg.axes[1], 30)]
	#@entities.transform_by_vectors [vxbeg], [vec_beg]
	#@entities.transform_by_vectors [vxend], [vec_end]
	@entities.transform_by_vectors [vxbeg, vxend], [vec_beg, vec_end]
	
	t0 = Time.now
	g = @entities.add_group
	e = g.entities.add_line *line
	g.entities.transform_by_vectors [e.end], [e.end.position.vector_to(e.start.position)]
	g.explode
end

#ZERO_EDGE: Reduce a sequence to a single edge or none
def zero_edge_reduce_sequence(seq)
	return seq.first if seq.length == 1
	
	#Loop for reducing the sequence left and right
	while true
		break if seq.length <= 1
		if seq.length == 2
			e = zero_edge_merge *seq
			seq = [e]
			break
		end
		dev1 = zero_edge_deviation seq[0], seq[1]
		dev2 = zero_edge_deviation seq[-2], seq[-1]
		if dev1 < dev2
			e1 = seq.shift
			e2 = seq.shift
			e = zero_edge_merge e1, e2
			seq = [e] + seq if e && zero_edge_tiny?(e)
		else
			e2 = seq.pop
			e1 = seq.pop
			e = zero_edge_merge e1, e2
			seq = seq + [e] if e && zero_edge_tiny?(e)
		end
		seq = seq.find_all { |e| e.valid? }
	end
	
	#Checking the edge left if any
	e = seq.first
	(e && zero_edge_tiny?(e)) ? e : nil
end

#ZERO_EDGE: check if an edge is tiny (with a factor)
def zero_edge_tiny?(edge)
	pt1 = edge.start.position
	pt2 = edge.end.position
	dcomp = (@tr_dist * pt1).distance(@tr_dist * pt2)
	dcomp < @tolerance_zero_edge * 2
end

#ZERO_EDGE: Compute the deviation of two edges
def zero_edge_deviation(e1, e2)
	lvx = e1.vertices & e2.vertices
	vx0 = lvx.first
	return Math::PI unless vx0
	vx1 = e1.other_vertex(vx0)
	vx2 = e2.other_vertex(vx0)
	pt0 = vx0.position
	vec1 = vx1.position.vector_to pt0
	vec2 = pt0.vector_to vx2.position
	return 0 unless vec1.valid? && vec2.valid?
	vec1.angle_between(vec2)
end

#ZERO_EDGE: Resolution for a single sequence
def zero_edge_single_fixing(lledges)
	ledges = lledges.flatten
	zero_edge_resolve ledges
end
	
#ZERO_EDGE: Resolution for all sequences
def zero_edge_resolve(ledges=nil)
	tol = @tolerance_zero_edge
	ledges = recompute_edges_in unless ledges
	
	lst_zero_sequences = zero_edge_sequencing ledges
	
	nsteps = lst_zero_sequences.length
	lst_zero_sequences.each_with_index do |a, i|
		dtot, seq, reason = a
		vbar_mini_progression i, nsteps, 200
		next if seq.find { |e| reparation_ignored?(e) }
		
		#Reducing the sequence to one edge
		edge_alone = zero_edge_reduce_sequence(seq)
		next unless edge_alone && edge_alone.valid?
		
		#Resolving the last edge
		if reason == :between
			curve_explode edge_alone
			zero_edge_absorb_final edge_alone
		else
			edge_alone.erase!
		end
	end	
end

#=============================================================
#-------------------------------------------------------------
# TINY_GAP: Core algorithms to detect and fix tiny gaps
#-------------------------------------------------------------
#=============================================================
	
#TINY_GAP: Inspection	
def tiny_gap_inspection
	@hsh_pvx_tg = @hsh_pvx
	tiny_gap_inspection_resolution @initial_edges
end
	
#TINY_GAP: Resolution	
def tiny_gap_resolve
	@hsh_pvx_tg = nil
	ledges = recompute_edges_in
	tiny_gap_inspection_resolution ledges, true
end

#TINY_GAP: Check the small gaps and record or fix them	
def tiny_gap_inspection_resolution(ledges, flg_resolve=false)
	tol = @tolerance_tiny_gap
	tol *= 1.01 if flg_resolve
	
	#Getting the vertices
	hsh_vx_used = {}
	ledges.each do |edge|
		next unless edge.valid?
		[edge.start, edge.end].each { |vx| hsh_vx_used[vx.entityID] = vx }
	end
	
	#Computing the global bounding box
	if @is_container
		bbtot = @comp.bounds
	else
		bbtot = Geom::BoundingBox.new
		ledges.each { |edge| bbtot.add edge.start.position, edge.end.position }
	end
		
	#Preparing the information
	linfo = []
	hsh_vx_used.each do |vx_id, vx|
		pt = (@hsh_pvx_tg) ? @hsh_pvx_tg[vx_id][1] : vx.position
		linfo.push [@tr * pt, vx]
	end
	
	#Classifying the vertices in smaller boxes
	nsteps = hsh_vx_used.length * 2
	space_grid = G6::SpaceGrid.new(bbtot, 20, tol)
	hsh_bb = space_grid.classify_points_vertices(linfo) { |ipt| vbar_mini_progression ipt, nsteps, 50 }
	
	#Comparing the vertices
	ll_prox = tiny_gap_compare_vertices(hsh_bb, tol, nsteps) 
	
	#Resolution
	if flg_resolve
		tiny_gap_repair_all ll_prox
		
	#Inspection: Storing the results
	else
		ll_prox.each do |a1, a2|
			vx1, pt1 = a1
			vx2, pt2 = a2
			d = pt1.distance(pt2)
			reparation_create(:tiny_gap, [vx1, vx2], d, [pt1, pt2])
		end
	end
	
	#Method cleanup
	hsh_vx_used = space_grid = hsh_bb = nil
end
	
#TINY_GAP: Compare vertices based on proximity in the space grid
def tiny_gap_compare_vertices(hsh_bb, tol, nsteps)
	#Counting the steps
	nmax = 0
	hsh_bb.each do |key, linfo|
		n = linfo.length - 2
		next unless linfo && linfo.length > 1
		for i in 0..n
			for j in i+1..n+1
				nmax += 1
			end	
		end
	end
	nsteps2 = 1.0 * nsteps / 2
	rsteps = nsteps2 / nmax
	
	#Comparing the vertices
	istep = 0
	hvx_info = {}
	hsh_key_used = {}
	hsh_bb.each do |key, linfo|
		next unless linfo && linfo.length > 1
		
		n = linfo.length - 2
		for i in 0..n
			pti, vxi = linfo[i]	
			vxi_id = vxi.entityID
			for j in i+1..n+1
				vbar_mini_progression nsteps2 + istep * rsteps, nsteps, 400
				istep += 1
				vxj_id = vxi.entityID
				ptj, vxj = linfo[j]
				vkey = tiny_gap_key(vxi, vxj)
				next if hsh_key_used[vkey]
				hsh_key_used[vkey] = true
				next if vxi == vxj
				d = (@tr_dist_inv * pti).distance(@tr_dist_inv * ptj)
				if d < tol && (d > 0 || vxi.position == vxj.position)				
					next if vxi.common_edge(vxj) || tiny_gap_connected?(vxi, vxj, tol * 1.5)
					ll = hvx_info[vxi_id]
					ll = hvx_info[vxi_id] = [] unless ll
					ll.push [d, [vxi, pti], [vxj, ptj]]
					ll = hvx_info[vxj_id]
					ll = hvx_info[vxj_id] = [] unless ll
					ll.push [d, [vxi, pti], [vxj, ptj]]
				end	
			end
		end	
	end	
	
	#Sorting the information on junctions by distance
	lj = []
	hvx_info.each do |vxid, ll|
		ll = ll.sort { |a, b| a.first <=> b.first }
		lj.push ll.first
	end
	lj = lj.sort { |a, b| a.first <=> b.first }
			
	#Keeping only the best junctions, one per vertex
	hsh_used = {}
	llprox = []
	lj.each do |info|
		d, a1, a2 = info
		vx1, = a1
		vx2, = a2
		vx1_id = vx1.entityID
		vx2_id = vx2.entityID
		next if hsh_used[vx1_id] || hsh_used[vx2_id]
		hsh_used[vx1_id] = hsh_used[vx2_id] = true
		llprox.push [a1, a2]
	end

	llprox
end

#TINY GAP: Compute a unique key for a pair of vertices
def tiny_gap_key(vx1, vx2)
	vx1_id = vx1.entityID
	vx2_id = vx2.entityID
	(vx1_id < vx2_id) ? "#{vx1_id} - #{vx2_id}" : "#{vx2_id} - #{vx1_id}"
end
	
#TINY_GAP: Check if the 2 vertices are not joined by edges, faces directly or indirectly	
def tiny_gap_connected?(vx1, vx2, tol)
	ledges1 = vx1.edges.find_all { |e| !e.smooth? && !e.soft? && !e.hidden? }
	ledges2 = vx2.edges.find_all { |e| !e.smooth? && !e.soft? && !e.hidden? }
	
	#Both vertices are alone
	return true if ledges1.length + ledges2.length == 2
	
	#Sharing the same faces
	return true unless (vx1.faces & vx2.faces).empty?

	#Sharing the same curve
	hcurve1 = {}
	vx1.edges.each { |e| curve = e.curve ; hcurve1[curve.entityID] = curve.entityID if curve }
	hcurve2 = {}
	vx2.edges.each { |e| curve = e.curve ; hcurve2[curve.entityID] = curve.entityID if curve }
	return true unless (hcurve1.values & hcurve2.values).empty?
	
	#Flat junction
	return true if tiny_gap_flat_at_vertex(vx1) || tiny_gap_flat_at_vertex(vx2)
	
	#Pointing in the right direction
	psmin = 0.95
	vecdir = vx1.position.vector_to(vx2.position).normalize
	lvec1 = ledges1.collect { |e| ovx = e.other_vertex(vx1) ; vx1.position.vector_to(ovx.position).normalize }
	lvec2 = ledges2.collect { |e| ovx = e.other_vertex(vx2) ; vx2.position.vector_to(ovx.position).normalize }
	vec1 = G6.vector_exact_average(lvec1)
	vec2 = G6.vector_exact_average(lvec2)
	return false if vec1 % vec2 < -psmin && vecdir % vec1 < 0 && vecdir % vec2 > 0
	
	true	
end
	
#TINY_GAP: Check if the shape is roughly flat at the vertex	
def tiny_gap_flat_at_vertex(vx)
	ledges = vx.edges
	return false if ledges.length == 1
	ledges = ledges.find_all { |e| !e.smooth? && !e.soft? && !e.hidden? }
	return true if ledges.length < 2
	e1 = ledges[0]
	vec1 = e1.start.position.vector_to(e1.end.position).normalize
	
	for i in 1..ledges.length-1
		e = ledges[i]
		vec = e.start.position.vector_to(e.end.position).normalize
		return false if (vec % vec1).abs < 0.8
	end
	true
end
	
#TINY_GAP: Repair the tiny gaps	
def tiny_gap_repair_all(ll_prox)
	return if ll_prox.empty?
	hsh_vx_pos = {}
	ll_prox.each do |a1, a2|
		vx1, = a1
		vx2, = a2
		tiny_gap_repair(vx1, vx2)
	end
end
	
#TINY_GAP: Repair a single tiny gap	
def tiny_gap_repair(vx1, vx2)
	vx1_id = vx1.entityID
	vx2_id = vx2.entityID
	return if reparation_ignored?(vx1) || reparation_ignored?(vx2)
	pt1 = vx1.position
	pt2 = vx2.position
	return if pt1 == pt2
	vec = pt1.vector_to(pt2)
	
	#Getting the elements connected
	ed1 = vx1.edges[0]
	ed2 = vx2.edges[0]
	allcon1 = ed1.all_connected.grep(Sketchup::Edge)
	allcon2 = ed2.all_connected.grep(Sketchup::Edge)
	
	#Elements connected: Move the vertices
	if allcon2.include?(ed1)
		lvec = []
		lvert = []
		if vx1.edges.length < vx2.edges.length
			lvert.push vx1
			lvec.push vec
		elsif vx1.edges.length > vx2.edges.length
			lvert.push vx2
			lvec.push vec.reverse
		else
			ptmid = Geom.linear_combination 0.5, pt1, 0.5, pt2
			lvert.push vx1, vx2
			lvec.push pt1.vector_to(ptmid), pt2.vector_to(ptmid)
		end
		@entities.transform_by_vectors lvert, lvec
	
	#Elements disjoint: move one of the group
	else
		if allcon1.length == allcon2.length
			ptmid = Geom.linear_combination 0.5, pt1, 0.5, pt2
			t = Geom::Transformation.translation pt1.vector_to(ptmid)
			@entities.transform_entities t, allcon1
			t = Geom::Transformation.translation pt2.vector_to(ptmid)
			@entities.transform_entities t, allcon2
		elsif allcon1.length > allcon2.length
			t = Geom::Transformation.translation vec.reverse
			@entities.transform_entities t, allcon2
		else
			t = Geom::Transformation.translation vec
			@entities.transform_entities t, allcon1
		end	
	end	
	
	#Cleanup at junction to remove the congruent vertices
	vx = [vx1, vx2].find { |v| v.valid? }
	return unless vx
	g = @entities.add_group
	pt1 = vx.position
	pt2 = pt1.offset vec.axes[1], 20
	e = g.entities.add_line(pt1, pt2)
	g.entities.transform_by_vectors [e.end], [e.end.position.vector_to(e.start.position)]
	g.explode
end

#TINY_GAP: Single fixing 
def tiny_gap_single_fixing(llvx)
	llprox = []
	llvx.each do |vx1, vx2|
		llprox.push [[vx1, vx1.position], [vx2, vx2.position]]
	end
	tiny_gap_repair_all(llprox)	
end

#==================================================================
#------------------------------------------------------------------
# SPLIT_EDGE: Core algorithms to detect and fix lonely vertices
#------------------------------------------------------------------
#==================================================================
	
#SPLIT_EDGE: Inspection	
def split_edge_inspection
	split_edge_inspection_resolution @initial_edges
end
	
#SPLIT_EDGE: Resolution	
def split_edge_resolve
	ledges = recompute_edges_in
	split_edge_inspection_resolution ledges, true
end

#SPLIT_EDGE: Check the small gaps and record or fix them	
def split_edge_single_fixing(llvx)
	llvx.flatten.each do |vx|
		split_edge_resolve_at_vertex vx if vx.valid?
	end	
end

#SPLIT_EDGE: Check the small gaps and record or fix them	
def split_edge_inspection_resolution(ledges, flg_resolve=false)
	#Getting the vertices
	hsh_vx_used = {}
	ledges.each do |edge|
		[edge.start, edge.end].each { |vx| hsh_vx_used[vx.entityID] = vx }
	end
	
	#Checking the vertices
	lvx_alone = []
	hsh_vx_used.each do |vx_id, vx|
		edges = vx.edges
		next if edges.length != 2
		e1, e2 = edges
		vx1 = e1.other_vertex vx
		vx2 = e2.other_vertex vx
		vec1 = vx1.position.vector_to vx.position
		vec2 = vx.position.vector_to vx2.position
		next unless vec1.valid? && vec2.valid?
		lvx_alone.push vx if vec1.parallel?(vec2)
	end
		
	#Resolution
	if flg_resolve
		lvx_alone.each do |vx|
			split_edge_resolve_at_vertex vx unless reparation_ignored?(vx)
		end	
		
	#Inspection: Storing the results
	else
		lvx_alone.each do |vx|
			reparation_create(:split_edge, vx, nil)
		end
	end
	
	#Method cleanup
	hsh_vx_used = nil
end
	
#SPLIT_EDGE: Compare vertices based on proximity in the space grid
def split_edge_resolve_at_vertex(vx)
	return unless vx.valid?
	e1, e2 = vx.edges
	curve_explode(e1)
	curve_explode(e2)
	vx1 = e1.other_vertex vx
	ptx = vx.position
	vec = ptx.vector_to vx1.position
	ptend = ptx.offset vec.axes[0], 0.01
	edge = @entities.add_line ptx, ptend  
	edge.erase!
end
	
#====================================================================
#--------------------------------------------------------------------
# COPLANAR_EDGE: Core algorithms to detect and fix coplanar edges
#--------------------------------------------------------------------
#====================================================================
	
#COPLANAR_EDGE: Inspection	
def coplanar_edge_inspection
	coplanar_edge_inspection_resolution @initial_edges
end
	
#COPLANAR_EDGE: Resolution	
def coplanar_edge_resolve
	ledges = recompute_edges_in
	coplanar_edge_inspection_resolution ledges, true
end

#COPLANAR_EDGE: Resolution of individual coplanar edges	
def coplanar_edge_single_fixing(ledges)
	coplanar_edge_inspection_resolution ledges.flatten, true
end

#COPLANAR_EDGE: Check the small gaps and record or fix them	
def coplanar_edge_inspection_resolution(ledges, flg_resolve=false)
	dotmax = 0.9999999991	#based on thomthom advice
	same_mat = @hsh_parameters[:flag_coplanar_edge]

	#Checking the edges
	ledges_coplanar = []
	ledges.each do |edge|
		next unless edge.valid?
		next if reparation_ignored?(edge)
		faces = edge.faces
		next if faces.length != 2
		f1, f2 = faces
		next if same_mat && (f1.material != f2.material || f1.back_material != f2.back_material)
		dot = f1.normal % f2.normal
		ledges_coplanar.push edge if dot > dotmax
	end
		
	#Resolution
	if flg_resolve
		ls_erase = []
		ledges_coplanar.each { |edge| ls_erase.push edge if edge.valid? }
		@entities.erase_entities ls_erase	
		
	#Inspection: Storing the results
	else
		ledges_coplanar.each do |edge|
			reparation_create(:coplanar_edge, edge, nil)
		end
	end
end
		
#====================================================================
#--------------------------------------------------------------------
# NEEDLE_EYE: Core algorithms to detect and fix needle eyes and ears
#--------------------------------------------------------------------
#====================================================================
	
#NEEDLE_EYE: Inspection	
def needle_eye_inspection
	needle_eye_inspection_resolution @initial_edges
end
	
#NEEDLE_EYE: Resolution	
def needle_eye_resolve
	ledges = recompute_edges_in
	needle_eye_inspection_resolution ledges, true
end

#NEEDLE_EYE: Resolution of individual coplanar edges	
def needle_eye_single_fixing(ledges)
	return unless @llbad_edges
	ledges = ledges.flatten
	lerase = []
	ledges.each do |e|
		eid = e.entityID
		@llbad_edges_id.each do |lse|
			if lse.include?(eid)
				lerase.concat lse
				break
			end
		end	
	end
	return if lerase.empty?
	ledges = recompute_edges_in.find_all { |e| e.faces.length == 0 }
	hsh_edges = {}
	ledges.each { |e| hsh_edges[e.entityID] = e }
	lerase = lerase.collect { |eid| hsh_edges[eid] }
	lerase = lerase.find_all { |e| e && e.valid? }
	@entities.erase_entities lerase
end

#NEEDLE_EYE: Check the small gaps and record or fix them	
def needle_eye_inspection_resolution(ledges, flg_resolve=false)
	#Edge not connected with faces
	ledges = ledges.find_all { |e| e.faces.length == 0 }
	return if ledges.empty?
	
	#Edge detector
	detector = G6::NeedleEyeDetector.new

	#Checking the edges
	good_edges = detector.detect_error_in_edges(ledges, @comp)
	
	#Getting the bad edges if any
	hinfo = detector.get_info
	return if hinfo.length == 0
	bad_edges = []
	hinfo.each { |pid, lst| bad_edges.concat lst[1] }
		
	#Splitting the bad edges into contiguous sequences
	hsh_bad_edges = {}
	bad_edges.each { |e| hsh_bad_edges[e.entityID] = true }
	
	@llbad_edges = []
	bad_edges0 = bad_edges.clone
	while true
		edge0 = bad_edges0.shift
		break unless edge0
		lprev = needle_eye_split_edges(edge0, edge0.start, hsh_bad_edges, bad_edges0)	
		lnext = needle_eye_split_edges(edge0, edge0.end, hsh_bad_edges, bad_edges0)
		ledges = lprev + [edge0] + lnext
		@llbad_edges.push ledges
	end	
	@llbad_edges_id = []
	@llbad_edges.each do |ls|
		@llbad_edges_id.push ls.collect { |e| e.entityID }
	end	
	
	#Resolution
	if flg_resolve
		lerase = []
		@llbad_edges.each do |lbad|
			next if lbad.find { |e| reparation_ignored?(e) }
			lerase.concat lbad
		end	
		@entities.erase_entities lerase
		
	#Inspection: Storing the results
	else
		@llbad_edges.each do |lbad|
			reparation_create(:needle_eye, lbad, nil)
		end	
	end
end
	
#NEEDLE_EYE: Split the bad edges into sequence of correlated edges	
def needle_eye_split_edges(edge0, vx0, hsh_bad_edges, bad_edges)
	ledges = []
	hsh_edges_used = {}
	hsh_edges_used[edge0.entityID] = true
	edge_cur = edge0
	while edge_cur
		edge_cur = nil
		vx0.edges.each do |e|
			eid = e.entityID
			next if hsh_edges_used[eid]
			hsh_edges_used[eid] = true
			next unless hsh_bad_edges[eid]
			ledges.push e
			bad_edges.delete e
			edge_cur = e
			vx0 = e.other_vertex vx0
		end
	end	
	ledges
end
	
#====================================================================
#--------------------------------------------------------------------
# LOST_JUNCTION: Core algorithms to detect and fix lost junctions
#--------------------------------------------------------------------
#====================================================================
	
#LOST_JUNCTION: Inspection	
def lost_junction_effective_tolerance
	symb = :lost_junction
	@percent_lost_junction = @tol_lost_junction = nil
	flag_symb = "percent_flag_#{symb}".intern
	if @hsh_parameters[flag_symb]
		percent_symb = "percent_value_#{symb}".intern
		@percent_lost_junction = @hsh_parameters.fetch percent_symb, 0.05
	else
		len_symb = "tolerance_#{symb}".intern
		@tol_lost_junction = @hsh_parameters.fetch len_symb, 0.01
	end
end
	
#LOST_JUNCTION: Inspection	
def lost_junction_inspection
	lost_junction_effective_tolerance
	lost_junction_inspection_resolution @initial_edges
end
	
#LOST_JUNCTION: Resolution	
def lost_junction_resolve
	ledges = recompute_edges_in
	lost_junction_inspection_resolution ledges, true
end

#LOST_JUNCTION: Resolution of individual coplanar edges	
def lost_junction_single_fixing(lvx)
	vx1, vx2 = lvx[0]
	vx1_info = lost_junction_vertex_info(vx1)
	vx2_info = lost_junction_vertex_info(vx2)
	junction = lost_junction_match_vertices(vx1_info, vx2_info)
	lost_junction_repair_all([junction]) if junction
end

#LOST_JUNCTION: Check the small gaps and record or fix them	
def lost_junction_inspection_resolution(ledges, flg_resolve=false)
	#Edge not connected with faces
	ledges = ledges.find_all { |e| e.faces.length == 0 }
	return if ledges.empty?
	
	#Termination vertices
	hvertices_info = {}
	ledges.each do |edge|
		[edge.start, edge.end].each do |vx| 
			next if vx.edges.length != 1
			info = lost_junction_vertex_info(vx)
			vxid = vx.entityID
			hvertices_info[vxid] = lost_junction_vertex_info(vx) unless hvertices_info[vxid]
		end	
	end
	return if hvertices_info.length == 0
	
	#Matching the vertices
	hvxinfo = {}
	lvertices_info = hvertices_info.values
	ljunctions = []
	n = lvertices_info.length - 1
	for i in 0..n-1
		vxi_info = lvertices_info[i]
		vxi = vxi_info[0]
		for j in i+1..n
			vxj_info = lvertices_info[j]
			vxj = vxj_info[0]
			junction = lost_junction_match_vertices(vxi_info, vxj_info)
			lost_junction_store_info(hvxinfo, junction) if junction
		end
	end	
	
	#Sorting the information on junctions by distance
	lj = []
	hvxinfo.each do |vxid, ls|
		ls = ls.sort { |a, b| a.first <=> b.first }
		lj.push ls.first
	end	
	lj = lj.sort { |a, b| a.first <=> b.first }
	
	#Keeping only the best junctions, one per vertex
	hsh_used = {}
	ljunctions = []
	lj.each do |junction|
		d, vx1, vx2 = junction
		vx1_id = vx1.entityID
		vx2_id = vx2.entityID
		next if hsh_used[vx1_id] || hsh_used[vx2_id]
		hsh_used[vx1_id] = hsh_used[vx2_id] = true
		ljunctions.push junction
	end
	
	#Resolution
	if flg_resolve
		lost_junction_repair_all(ljunctions)
		
	#Inspection: Storing the results
	else
		ljunctions.each do |jinfo|
			d, vx1, vx2, jpts = jinfo
			reparation_create :lost_junction, [vx1, vx2], d, jpts
		end	
	end
end

#LOST_JUNCTION: Store the information in a Hash array for each vertex of the junction
def lost_junction_store_info(hvxinfo, junction)
	d, vx1, vx2 = junction
	vx1_id = vx1.entityID
	vx2_id = vx2.entityID
	ls = hvxinfo[vx1_id]
	ls = hvxinfo[vx1_id] = [] unless ls
	ls.push junction
	ls = hvxinfo[vx2_id]
	ls = hvxinfo[vx2_id] = [] unless ls
	ls.push junction
end

#LOST_JUNCTION: Proceed with the repairing of a lost junction
def lost_junction_repair_all(ljunctions)
	lvert = []
	lvec = []
	lines = []
	ljunctions.each do |jinfo|
		d, vx1, vx2, jpts = jinfo
		
		#Adding a junction segment
		if jpts.length == 2
			@entities.add_line *jpts
			next
		end
		
		#Joining the segments by moving the vertices
		next if reparation_ignored?(vx1) || reparation_ignored?(vx2)
		pt1, ptmid, pt2 = jpts
		lvert.push vx1, vx2
		pt1 = vx1.position
		pt2 = vx2.position
		next if pt1 == pt2
		vec1 = pt1.vector_to(ptmid)
		lvec.push vec1, pt2.vector_to(ptmid)
		lines.push [ptmid, ptmid.offset(vec1.axes[1], 20)]
	end
	
	#Moving the vertices
	@entities.transform_by_vectors lvert, lvec
	
	#Cleanup at junction to remove the congruent vertices
	g = @entities.add_group
	lines.each do |pt1, pt2|
		e = g.entities.add_line(pt1, pt2)
		g.entities.transform_by_vectors [e.end], [e.end.position.vector_to(e.start.position)]
	end	
	g.explode
	
	#Method cleanup
	lvert = lvec = lines = hsh_vx_pos = nil	
end

#LOST_JUNCTION: Compute the information at vertex
def lost_junction_vertex_info(vx)
	edge0 = vx.edges.first
	vx0 = vx
	
	#Finding the path from the vertex
	ledges = []
	d = 0
	while edge0
		ledges.push edge0
		vx_next = edge0.other_vertex vx
		d += edge0.length
		angle_max = 140.degrees
		next_edge = nil
		vx_next.edges.each do |e|
			next if e == edge0 || e.faces.length > 0
			if vx_next.edges.length == 2
				next_edge = e
				break
			end	
			angle = G6.edges_angle_at_vertex(edge0, e, vx_next)
			if angle > angle_max
				angle_max = angle
				next_edge = e
			end	
		end
		edge0 = next_edge
		vx = vx_next
	end
	
	#information on vertex
	vx1 = ledges[0].other_vertex vx0
	vec = vx1.position.vector_to vx0.position
	[vx0, ledges, d, vec, vx0.position]
end

#LOST_JUNCTION: Match two vertices and return the path between each other
def lost_junction_match_vertices(vx1_info, vx2_info)
	vx1, ledges1, d1, vec1, pt1 = vx1_info
	vx2, ledges2, d2, vec2, pt2 = vx2_info
	return nil if pt1 == pt2
	d = pt1.distance pt2
	vec = pt1.vector_to pt2
	
	#Check on direction
	return nil if vec % vec1 < 0 || vec % vec2 > 0
	
	#Check on distance
	return nil if @percent_lost_junction && d /(d1+d2) > @percent_lost_junction
	return nil if @tol_lost_junction && d > @tol_lost_junction
	
	#Segments are perfectly aligned
	line1 = [pt1, vec1]
	line2 = [pt2, vec2]
	
	if pt1.on_line?(line2)
		ptmid = Geom.linear_combination 0.5, pt1, 0.5, pt2
		return((vec % vec2 < 0) ? [d, vx1, vx2, [pt1, ptmid, pt2]] : nil)
	end

	#Check on intersection and distance
	ptinter = Geom.intersect_line_line line1, line2
	return nil unless ptinter
	
	di1 = pt1.distance(ptinter)
	di2 = pt2.distance(ptinter)
	di = di1 + di2
	return nil if @percent_lost_junction && di /(d1+d2) > @percent_lost_junction
	return nil if @tol_lost_junction && di > @tol_lost_junction

	dm = d * 2
	return nil if di1 > dm || di2 > dm
	
	#Deciding about adding a segment or moving the vertices
	d1 = (ledges1.collect { |e| e.length }).min
	d2 = (ledges1.collect { |e| e.length }).min
	r = d / [d1, d2].min
	if r < 1.5 && r > 0.75
		return [d, vx1, vx2, [pt1, pt2]]
	end	
	
	if pt1.vector_to(ptinter) % vec1 < 0 || pt2.vector_to(ptinter) % vec2 < 0
		ptinter = Geom.linear_combination 0.5, pt1, 0.5, pt2
	end	
	[d, vx1, vx2, [pt1, ptinter, pt2]]
end

end	#class Resolver

end	#End Module EdgeInspector

end	#End Module F6_FredoTools
