=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2011 Fredo6 - Designed and written Jul 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_FredoTools__MarkVertices.rb
# Original Date	:   15 Jul 2011 (based on work published in Nov 2009)
# Description	:   Mark all vertices of the selection with a guide point
#                   This is useful to visualize the vertices
# IMPORTANT		:	DO NOT TRANSLATE STRINGS in the source code
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

module MarkVertices

#====================================================================================================
#----------------------------------------------------------------------------------------------------
# Plugin Implementation
#----------------------------------------------------------------------------------------------------
#====================================================================================================

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

def self._execution(symb)
	execute
end

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------


#Initialize and execute the marking of vertices
def self.execute(selection=nil)
	@hsh_vertices = {}
	@hsh_entities = {}
	@model = Sketchup.active_model
	@menutitle = T7[:PlugName]
	selection = @model.selection unless selection
	
	#Create the list of vertices
	process_vertices selection
		
	#No vertices
	if !@hsh_vertices.values.find { |hsh| hsh.length > 0 }
		UI.messagebox T6[:MSG_NoVertices]
		return
	end
	
	#processing the components
	Sketchup.set_status_text T6[:MSG_Processing], SB_VCB_LABEL
	@model.start_operation @menutitle

	n = 0
	@hsh_vertices.each do |key, hsh|
		entities = @hsh_entities[key]
		hsh.each { |k, pt| entities.add_cpoint pt }
		n += hsh.length
	end
	
	@model.commit_operation
	Sketchup.set_status_text T6[:MSG_Finished], SB_VCB_LABEL
	Sketchup.set_status_text "#{n} #{T6[:T_TXT_Vertices]}", SB_VCB_VALUE
end

#Process vertices on the selection recursively
def self.process_vertices(lentities, comp=nil)
	hsh = get_hash(comp)
	return unless hsh
	lentities.each do |e|
		if e.instance_of?(Sketchup::Edge)
			create_vertices hsh, [e.start, e.end]
		elsif e.instance_of?(Sketchup::Face)
			create_vertices hsh, e.vertices.to_a
		elsif e.instance_of?(Sketchup::Group)
			process_vertices e.entities, e
		elsif e.instance_of?(Sketchup::ComponentInstance)
			process_vertices e.definition.entities, e
		end
	end	
end

#Get the hash table to store vertices	
def self.get_hash(comp)
	if comp.instance_of?(Sketchup::ComponentInstance)
		key = comp.definition.entityID
		hsh = @hsh_vertices[key]
		return nil if hsh
		entities = comp.definition.entities
	elsif comp.instance_of?(Sketchup::Group)
		key = comp.entityID
		entities = comp.entities
	else	
		key = '__the__model'
		entities = @model.active_entities
	end	
	hsh = @hsh_vertices[key]
	return hsh if hsh
	
	hsh = @hsh_vertices[key] = {} unless hsh
	@hsh_entities[key] = entities
	hsh
end

#Create vertices	
def self.create_vertices(hsh, lvx)
	lvx.each { |vx| hsh[vx.entityID] = vx.position }
end
	
end	#End Module MarkVertices

end	#End Module F6_FredoTools
