=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2014 Fredo6 - Designed and written Mar 2014 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_FredoTools__MoveAlong.rb
# Original Date	:   11 Mar 2014 
# Description	:   Analog to SU Move tool, but can force the direction along a defined plane
# IMPORTANT		:	DO NOT TRANSLATE STRINGS in the source code
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools
				   
module MoveAlong

#Texts for the module
T7[:TIP_ToggleCopyMode] = "Toggle Copy versus Move mode"
T7[:TIP_ToggleDeferMode] = "Toggle interactive move of selection"
T7[:TIP_ToggleAlignMode] = "Toggle alignment of object to target direction by pivoting"
T7[:TIP_ToggleExtendedMode] = "Toggle Extended selection mode for Edges"

T7[:TIP_RotationRedo] = "Redo Rotation"
T7[:TIP_MovementRedo] = "Redo Move"
T7[:TIP_NativeMove] = "Call the native Sketchup Move tool"
T7[:TIP_AutoSmooth] = "Automatic Smooth of new created edges (quad conventions)"

T7[:MSG_CopyMode] = "Ctrl for Copy mode"
T7[:MSG_DeferMode] = "TAB for Wireframe mode (deferred)"
T7[:MSG_ExtendedMode] = "Shift for Extended selection"

T7[:MSG_Rotate] = "Click and drag to Rotate"

T7[:MSG_DeformationByVertex] = 	"Deformation from a single vertex"
T7[:MSG_Preselection] = "Pre-selection"
T7[:MSG_DynamicSelection] = "Interactive selection"

T7[:TIP_PivotedTip] = "Flip orientation (or + and Enter in VCB)"

#====================================================================================================
#----------------------------------------------------------------------------------------------------
# Plugin Implementation
#----------------------------------------------------------------------------------------------------
#====================================================================================================

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

def self._execution(symb)
	Sketchup.active_model.select_tool MoveAlongTool.new
end

#=============================================================================================
#=============================================================================================
# Class MoveAlongTool: main class for the Interactive tool
#=============================================================================================
#=============================================================================================

class MoveAlongTool < Traductor::PaletteSuperTool

#----------------------------------------------------------------------------------------------------
# INIT: Initialization
#----------------------------------------------------------------------------------------------------

#INIT: Class instance initialization
def initialize(*args)
	#Basic Initialization
	@tr_id = Geom::Transformation.new
	@model = Sketchup.active_model
	@selection = @model.selection
	@entities = @model.active_entities
	@view = @model.active_view
	@dico_name = "Fredo6_FredoTools_MoveAlong"
	@dico_attr = "Parameters"
	init_flags
	MYDEFPARAM.add_notify_proc(self.method("notify_from_defparam"), true)	
	@duration_long_click = 0.8
	@wireframe_max_elts = 800
	
	#Creating the protractor shape
	hsh = { :default_color => 'red' }
	@protractor = G6::ShapeProtractor.new hsh
	
	#Parsing the arguments
	args.each { |arg| arg.each { |key, value|  parse_args(key, value) } if arg.class == Hash }

	#Loading parameters
	parameter_load
	
	#Static initialization
	init_cursors
	init_colors
	init_messages

	#Initializing the Key manager
	@keyman = Traductor::KeyManager.new() { |event, key, view| key_manage(event, key, view) }

	#Initializing the Zoom Manager
	@zoom_void = G6::ZoomVoid.new
	
	#Create the Origin-Target Picker
	hsh = { :notify_proc => self.method("notify_from_otpicker") }
	@otpicker = Traductor::OriginTargetPicker.new @hsh_parameters, hsh
	@precision = @otpicker.get_precision
	notify_from_defparam
	
	#Creating the palette manager and texts
	init_palette
end

def init_flags
	@flag_anchor_text = false
end
	
#INIT: Parse the arguments of the initialize method
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /from/i
		@invoked_from = value
	end	
end

#INIT: Initialize texts and messages
def init_messages
	@menutitle = T7[:PlugName]
	
	@mnu_exit = T6[:T_BUTTON_Exit]	
	@mnu_abort = T6[:T_STR_AbortTool]	
	
	@msg_origin = [T6[:T_MSG_OriginPick], T7[:MSG_CopyMode], T7[:MSG_DeferMode], T7[:MSG_ExtendedMode], T6[:T_MSG_PickFromModel], 
	               T6[:T_MSG_ArrowModeAxis], T6[:T_MSG_DoubleClickRepeat]].join " - "
	@msg_target = [T6[:T_MSG_TargetPick], T7[:MSG_CopyMode], T7[:MSG_DeferMode], T6[:T_MSG_ShiftLockMode], T6[:T_MSG_ArrowModeAxis]].join " - "
	@msg_rotation = T7[:MSG_Rotate]
	
	@tip_copy_mode = T7[:TIP_ToggleCopyMode] + " [#{T6[:T_KEY_CTRL]}]"
	@tip_defer_mode = T7[:TIP_ToggleDeferMode] + " [#{T6[:T_KEY_TAB]}]"
	@tip_align_mode = T7[:TIP_ToggleAlignMode] + " [#{T6[:T_KEY_Shift]}-#{T6[:T_KEY_TAB]}]"
	@tip_extended_mode = T7[:TIP_ToggleExtendedMode] + " [#{T6[:T_KEY_ESC]}]"
	
	@tip_rotation_redo = T7[:TIP_RotationRedo] + " [#{T6[:T_KEY_DoubleClick]}]"
	@tip_movement_redo = T7[:TIP_MovementRedo] + " [#{T6[:T_KEY_DoubleClick]}]"
			
	@txt_offset = T6[:T_TXT_Offset]
	@txt_offset_remote = T6[:T_TXT_Distance]
	@txt_angle = T6[:T_TXT_Angle]
	
	@msg_preselection = T7[:MSG_Preselection]
	@msg_dynselection = T7[:MSG_DynamicSelection]
	@msg_vxselection = T7[:MSG_DeformationByVertex]
	@msg_doubleclick_exit = T6[:T_INFO_DoubleClickExit]
end

#INIT: Initialize texts and messages
def init_colors
	@color_box_distance = ['lightgrey', 'black']
	@color_box_angle = ['pink', 'red']
	
	@color_but_title = 'lightblue'
	@color_but_hi = 'lightgreen'	
	
	@color_wireframe_comp = 'purple'
	@color_wireframe_elt = 'teal'
end

#INIT: Initialize the cursors
def init_cursors
	@id_cursor_plane = MYPLUGIN.create_cursor "MoveAlong_cursor_arrow_plane", 0, 0	
	@id_cursor_plane_extended = MYPLUGIN.create_cursor "MoveAlong_cursor_arrow_plane_extended", 0, 0	
	@id_cursor_plane_plus = MYPLUGIN.create_cursor "MoveAlong_cursor_arrow_plane_plus", 0, 0	
	@id_cursor_plane_plus_extended = MYPLUGIN.create_cursor "MoveAlong_cursor_arrow_plane_plus_extended", 0, 0	

	@id_cursor_vector = MYPLUGIN.create_cursor "MoveAlong_cursor_arrow_vector", 0, 0	
	@id_cursor_vector_extended = MYPLUGIN.create_cursor "MoveAlong_cursor_arrow_vector_extended", 0, 0	
	@id_cursor_vector_plus = MYPLUGIN.create_cursor "MoveAlong_cursor_arrow_vector_plus", 0, 0	
	@id_cursor_vector_plus_extended = MYPLUGIN.create_cursor "MoveAlong_cursor_arrow_vector_plus_extended", 0, 0
	
	@id_cursor_rotation = 643

	@id_cursor = 0
end

#--------------------------------------------------------------
# PARAMETER: Tolerance and other settings and parameters
#--------------------------------------------------------------

#INIT: Notification from default parameters
def notify_from_defparam(key=nil, val=nil)
	
end

#Persistent parameters and defaults
@@hsh_parameters_persist = {} unless defined?(@@hsh_parameters_persist) && @@hsh_parameters_persist.length > 0

#PARAMETER: Load the parameters from the model attribute and the defaults
def parameter_load
	@hsh_parameters = {}
	
	#Defaults
	@hsh_param_default = {
		:defer_mode => false,
		:option_autosmooth => true,
		:extended_mode => false,
		:align_mode => false
	}
		
	#Calculating the appropriate parameters
	@hsh_parameters = {}
	@hsh_parameters.update @hsh_param_default
	sparam = Traductor::Default.read @dico_name, @dico_attr
	hsh = eval(sparam) rescue {}
	hsh = {} unless hsh
	@@hsh_parameters_persist.update hsh
	@hsh_parameters.update @@hsh_parameters_persist 
	
	@defer_mode = @hsh_parameters[:defer_mode]
	@extended_mode = @hsh_parameters[:extended_mode]
	@option_autosmooth = @hsh_parameters[:option_autosmooth]
end

#PARAMETER: Save the parameters as a model attribute
def parameter_save
	@otpicker.option_update_hash(@hsh_parameters)
	@hsh_parameters[:defer_mode] = @defer_mode 
	@hsh_parameters[:extended_mode] = @extended_mode 
	@hsh_parameters[:option_autosmooth] = @option_autosmooth 

	@@hsh_parameters_persist.update @hsh_parameters
	sparam = @@hsh_parameters_persist.inspect.gsub('"', "'")
	Traductor::Default.store @dico_name, @dico_attr, sparam
end

#----------------------------------------------------------------------------------------------------
# AUTOSMOOTH: Manage AutoSmooth option
#----------------------------------------------------------------------------------------------------

def autosmooth_toggle
	@option_autosmooth = !@option_autosmooth
end	

#----------------------------------------------------------------------------------------------------
# EXTENDED: Manage Extended selection mode
#----------------------------------------------------------------------------------------------------

#EXTENDED: Toggle the Extended mode flag
def extended_toggle
	@extended_mode = !@extended_mode
	extended_manage_after
end

#EXTENDED: Tasks after changing the Extended mode flag
def extended_manage_after
	onMouseMove_zero
end

#----------------------------------------------------------------------------------------------------
# COPY: Manage Copy mode
#----------------------------------------------------------------------------------------------------

#COPY: Toggle the Copy Mode
def copy_toggle
	@copy_mode = !@copy_mode
	copy_manage_after
end

#COPY: Set the Copy Mode
def copy_set(mode)
	@copy_mode = mode
	copy_manage_after
end

#COPY: Tasks after changing the Copy mode flag
def copy_manage_after
	if @dragging
		movement_set_back
		movement_prepare
	end	
	onMouseMove_zero
end	

#----------------------------------------------------------------------------------------------------
# DEFER : Manage Defer mode
#----------------------------------------------------------------------------------------------------

#DEFER: Toggle the defer mode
def defer_toggle
	@defer_mode = !@defer_mode
	defer_manage_after
end

#DEFER: Manage the change of Defer mode
def defer_manage_after
	if @dragging
		movement_set_back
		movement_prepare
	end	
	onMouseMove_zero
end	

#----------------------------------------------------------------------------------------------------
# ALIGN : Manage Alignment mode
#----------------------------------------------------------------------------------------------------

#ALIGN: Toggle the Align mode
def align_toggle
	@align_mode = !@align_mode
	align_manage_after
end

#ALIGN: Manage the change of Align mode
def align_manage_after
	if @dragging
		movement_set_back
		movement_prepare
	elsif @geometry_generated
		movement_repeat true
	end	
	@otpicker.show_implicit_plane @align_mode
	onMouseMove_zero
end	

#----------------------------------------------------------------------------------------------------
# NOTIFY: Call back from Otpicker
#----------------------------------------------------------------------------------------------------

def notify_from_otpicker(event)
	case event
	when :shift_down
		@keyman.shift_down?
	when :ctrl_down	
		@keyman.ctrl_down?
	when :onMouseMove_zero
		onMouseMove_zero
	else
		:no_event
	end	
end

#----------------------------------------------------------------------------------------------------
# ACTIVATION: Plugin Activation / Deactivation
#----------------------------------------------------------------------------------------------------

#ACTIVATION: Tool activation
def activate
	LibFredo6.register_ruby "FredoTools::MoveAlong"
	Traductor::Hilitor.stop_all_active
			
	#Initializing the Executor of Operation
	hsh = { :title => @menutitle, :palette => @palette, :no_commit => true }
	@suops = Traductor::SUOperation.new hsh

	#Initializing the environment
	reset_env
	
	#Handling the initial selection
	@selection = @model.selection
	preselection = (@selection.length == 0) ? nil : @selection.to_a
	@active_selection = @otpicker.preselection_set(preselection)
	compute_message_selection
		
	#Refreshing the view
	refresh_viewport
end

#ACTIVATION: Deactivation of the tool
def deactivate(view)
	parameter_save
	
	if @aborting && @geometry_generated
		text = T6[:T_MSG_ConfirmAbortText] + "\n\n" + T6[:T_MSG_ConfirmAbortQuestion]
		if UI.messagebox(text, MB_YESNO) != 6
			@suops.abort_operation
		end	
	elsif @geometry_generated
		commit_operation
	else
		@suops.abort_operation
	end	
	preselection = @otpicker.preselection_get
	@selection.clear unless preselection
	view.invalidate
	@view.remove_observer self
end

#ACTIVATION: Reset the picking environment
def reset_env
	@mode = @otpicker.set_mode(:origin)
	@pt_origin = @pt_target = nil
	@otpicker.reset
	@dragging = false
	@rotation_mode = false
	@bb_extents = Geom::BoundingBox.new.add @model.bounds
	@red_cross_info = nil
end	

#ACTIVATION: Exit the tool
def exit_tool
	@aborting = false
	@model.select_tool nil
end

#ACTIVATION: Abort the operation and exit
def abort_tool
	@aborting = true
	@model.select_tool nil
end

#ACTIVATION: Activate the native SU Move tool
def native_tool
	Sketchup.send_action "selectMoveTool:"
end

#--------------------------------------------------------------
# VIEWPORT: View Management
#--------------------------------------------------------------

#VIEWPORT: Computing Current cursor
def onSetCursor
	#Palette cursors
	ic = super
	return (ic != 0) if ic
		
	if @rotation_mode
		@id_cursor = @id_cursor_rotation
		
	elsif @otpicker.option_vector?
		if @extended_mode
			@id_cursor = (@copy_mode) ? @id_cursor_vector_plus_extended : @id_cursor_vector_extended
		else	
			@id_cursor = (@copy_mode) ? @id_cursor_vector_plus : @id_cursor_vector
		end	
	elsif @otpicker.option_planar?
		if @extended_mode
			@id_cursor = (@copy_mode) ? @id_cursor_plane_plus_extended : @id_cursor_plane_extended
		else	
			@id_cursor = (@copy_mode) ? @id_cursor_plane_plus : @id_cursor_plane
		end	
	end
	
	#Other cursors depending on state
	UI.set_cursor @id_cursor
end

#VIEWPORT: Refresh the viewport
def refresh_viewport
	onSetCursor
	show_message
	@view.invalidate
end
	
#VIEWPORT: Show message in the palette
def show_message
	#Mouse if either in the palette or out of the viewport
	if @mouseOut
		@palette.set_message nil
		return
	elsif @mouse_in_palette
		@palette.set_message @palette.get_main_tooltip, 'aliceblue'
		return
	end	
	
	if @rotation_mode || @action_ongoing == :rotation
		show_message_rotation
	else
		show_message_movement
	end	
end

#VIEWPORT: Show message for Rotation
def show_message_rotation
	Sketchup.set_status_text @txt_angle, SB_VCB_LABEL
	txangle = (@rot_angle) ? "#{sprintf('%0.1f', @rot_angle.radians)} deg." : ""
	Sketchup.set_status_text txangle, SB_VCB_VALUE
	Sketchup.set_status_text @msg_rotation
end

#VIEWPORT: Show message for Move
def show_message_movement
	#Displaying the Offset or Distance
	tx_lab, tx_len = @otpicker.vcb_label_length
	Sketchup.set_status_text tx_lab, SB_VCB_LABEL
	Sketchup.set_status_text tx_len, SB_VCB_VALUE

	#Main status bar - Help text
	text = (@mode == :origin) ? @msg_origin : @msg_target
	Sketchup.set_status_text text
	
	#Displaying tooltip in the palette
	tip_view = @otpicker.get_view_tooltip
	tip_dir = @otpicker.direction_tooltip
	tip = tip_view
	if tip_dir && tip_dir != tip_view
		if tip
			tip += " - " + tip_dir
		else
			tip = tip_dir
		end
	end
	code = 'lightgrey'
	if !tip_view	
		tip = @msg_doubleclick_exit if @mode == :origin
		code = 'palegreen'
	elsif tip_dir
		code = 'lightblue'
	else
		code = 'lightyellow'
	end	
	@palette.set_message tip, code
	
	#Displaying information on selection
	@palette.set_tooltip @message_selection, @message_selection_code
end

#---------------------------------------------------------------------------------------------
# MOVE: Mouse Move Methods
#---------------------------------------------------------------------------------------------

#MOVE: Mouse Movements
def onMouseMove_zero(flag=nil) ; onMouseMove(@flags, @xmove, @ymove, @view) if @xmove ; end
def onMouseMove(flags, x, y, view)
	#Event for the palette
	@xmove = x
	@ymove = y
	@flags = flags
	
	#Mouse in palette
	if super
		@mouse_in_palette = true
		refresh_viewport
		return	
	end
	@mouse_in_palette = false
			
	return if @moving
	
	#Tracking the origin with possible auto-selection
	if @mode == :origin
		#Pick origin with inference
		@origin_info = @otpicker.origin_pick view, x, y, @extended_mode, @keyman.shift_down?, @copy_mode
		@pt_origin, @type_origin = @origin_info
		@bb_extents.add @pt_origin if @pt_origin
		@active_selection = @otpicker.origin_active_selection
		compute_message_selection
		rotation_compute_crosses view
		rotation_cross_picked(view, x, y)
		@vec_parallel = @otpicker.remote_parallel_vector
		
	elsif @rotation_mode && @dragging
		rotation_execute(view, x, y)
		
	#Tracking the target	
	elsif @mode == :target
		@pt_target, = @otpicker.target_pick(view, x, y, !@defer_mode)

		#Updating the Bounding box for drawing
		@bb_extents.add @pt_target if @pt_target
		
		#Executive the move
		movement_on_going
		if @defer_mode
			sel = selection_non_defer
			movement_perform_move(sel) unless sel.empty?
		else
			movement_execute
		end	
	end
	
	#Refreshing the view
	@moving = true
	refresh_viewport
end	

#MOVE: Mouse leaves the viewport
def onMouseLeave(view)
	@mouseOut = true
	view.invalidate
end

#MOVE: Mouse enters the viewport
def onMouseEnter(view)
	@mouseOut = false
	view.invalidate
end
	
#MOVE: Before Zooming or changing view
def suspend(view)
	@keyman.reset
	@zoom_void.suspend(view, @xmove, @ymove)
end

#MOVE: After changing view
def resume(view)
	@keyman.reset
	@zoom_void.resume(view)
	refresh_viewport
end
	
#VIEW: Notification of view changed
def onViewChanged(view)

end
	
#VIEW: Compute the message selection	
def compute_message_selection
	if @otpicker.preselection_get
		@message_selection = @msg_preselection
		@message_selection_code = 'lightblue'
	elsif @pt_origin && @active_selection && @active_selection.grep(Sketchup::Vertex).length == 1
		@message_selection = @msg_vxselection
		@message_selection_code = 'thistle'
	else
		@message_selection = @msg_dynselection
		@message_selection_code = 'lightgreen'
	end		
end
	
#---------------------------------------------------------------------------------------------
# VCB: VCB Management
#---------------------------------------------------------------------------------------------

#VCB: Enable or disable the VCB
def enableVCB?
	true
end

#VCB: Handle VCB input
def onUserText(text, view)
	remote = @otpicker.vcb_origin_remote?
	@vcb = Traductor::VCB.new
	if @rotation_mode ||@action_ongoing == :rotation
		@vcb.declare_input_format :fangle, "f"				#angle in degrees
		@vcb.declare_input_format :angle, "a"				#angle round for round mode
	elsif @mode == :origin && remote
		@vcb.declare_input_format :distance, "l"			#Distance from remote	
	else
		@vcb.declare_input_format :offset, "l"				#offset length
		@vcb.declare_input_format :multiple, "i__x"			#multiple copy
		@vcb.declare_input_format :multiple2, "star"		#multiple copy
		@vcb.declare_input_format :divide, "slash"			#divide copy
		@vcb.declare_input_format :flip, "_+" if @pivoted	#Flip pivoting
		@vcb.declare_input_format :flip_origin, "_+" if @pivoted_at_origin	#Flip pivoting
	end
	
	nb_errors = @vcb.process_parsing(text)
	
	if nb_errors > 0
		lmsg = []
		@vcb.each_error { |symb, s| lmsg.push "<#{s}>" }
		msg = "#{T6[:T_ERROR_InVCB]} \"#{text}\"  --> #{lmsg.join(' - ')}"
		@palette.set_message msg, 'E'
		view.invalidate
	else
		action_from_VCB
	end	
end

#VCB: Execute actions from the VCB inputs	
def action_from_VCB
	angle = multiple = divide = distance = nil
	loffset = []
	@vcb.each_result do |symb, val, suffix|
		case symb
		when :offset
			loffset.push val
		when :distance
			distance = val
		when :multiple, :multiple2
			multiple = val
		when :divide
			multiple = val
			divide = true
		when :angle
			angle = val
		when :fangle
			angle = val.degrees
		when :flip
			movement_pivot_flip
			return
		when :flip_origin
			movement_pivot_flip
			return
		end
	end
	
	#Changing the angle
	if @rotation_mode || @action_ongoing == :rotation
		rotation_post_execute angle if angle
		return
	end	
	
	#Changing the remote distance
	if distance
		movement_change_from_remote distance
		return
	end
	
	#Changing the offset and multiple
	noffset = loffset.length
	if noffset > 0 || multiple
		if noffset <= 1
			movement_post_execute loffset[0], multiple, divide
		else
			vec = Geom::Vector3d.new(*loffset[0..2])
			if vec.valid?
				offset = vec.length
			else
				offset = nil
			end	
			movement_set_vector vec, multiple, divide	
		end	
	end	
end	

#---------------------------------------------------------------------------------------------
# SU TOOL: Click Management
#---------------------------------------------------------------------------------------------

#SU TOOL: Button click DOWN
def onLButtonDown(flags, x, y, view)
	#Palette management
	if super
		refresh_viewport
		return
	end
	
	#Storing the reference
	@button_down = true
	@xdown = x
	@ydown = y	
	
	#Changing mode
	if @mode == :origin
		onMouseMove(0, x, y, view) unless @pt_origin
		action_start
		@time_down_start = Time.now
	end	
	@time_down = Time.now
end

#SU TOOL: Button click UP - Means that we end the selection
def onLButtonUp(flags, x, y, view)
	#Palette buttons
	if super
		@time_down_start = nil
		onMouseMove_zero
		return
	end
	return unless @button_down
	@button_down = false
	
	#Logic for dragging
	far_from_origin = (@xdown - x).abs > 3 || (@ydown - y).abs > 3
	if @time_down && (Time.now - @time_down) > @duration_long_click && !far_from_origin
		reset_env
		onMouseMove_zero
		@otpicker.direction_pick_from_model unless @rotation_mode
	elsif @dragging
		if !@time_down_start || (Time.now - @time_down_start) > 0.3 
			action_stop
		end
	end	
end

#SU TOOL: Double Click received
def onLButtonDoubleClick(flags, x, y, view)
	return if super
	if @rotation_mode
		rotation_repeat
	
	elsif @type_origin != :void_origin
		if @align_mode && movement_align_at_origin
		
		else
			movement_redo
		end	
	else
		exit_tool
	end	
end

#---------------------------------------------------------------------------------------------
# ACTION: Control of the Movement or Rotation
#---------------------------------------------------------------------------------------------

#ACTION: Start either Move or Rotate
def action_start
	if @rotation_mode
		rotation_start
	else
		movement_start if @pt_origin
	end	
	@time_action_start = Time.now
end

#ACTION: Stop either Move or Rotate
def action_stop
	if @rotation_mode
		rotation_stop
	elsif !@pt_target && movement_align_at_origin
		return
	else	
		movement_store
		movement_stop
	end	
end

#---------------------------------------------------------------------------------------------
# SELECTION: Management of current selection
#---------------------------------------------------------------------------------------------

#SELECTION: Cumulate selection in mode :origin
def selection_cumulate
	@active_selection = @otpicker.preselection_cumulate
end

#SELECTION: Extend the selection from a single vertex
def selection_extend_from_vertex(vx)
	hsh_edges = {}
	vx.edges.each do |edge|
		curve = edge.curve
		if curve
			curve.edges.each { |e| hsh_edges[e.entityID] = e }
		else
			hsh_edges[edge.entityID] = edge
		end	
	end
	hsh_edges.values
end

#SELECTION: Calculate the selection to put into a copy group
def selection_safe(selection)
	vx = selection[0]
	if vx.instance_of?(Sketchup::Vertex)
		selection = selection_extend_from_vertex(vx)
		@hi_curve_pts = nil		
	end
	@otpicker.suselection_add selection, true
	selection
end

#SELECTION: calculate the selection non subject to defer mode
def selection_non_defer
	if @active_selection && @active_selection[0].instance_of?(Sketchup::Vertex)
		return @active_selection
	end	
	sel = @active_selection.grep(Sketchup::ConstructionLine) + @active_selection.grep(Sketchup::ConstructionPoint)
	sel += @active_selection.grep(Sketchup::Text) + @active_selection.grep(Sketchup::Dimension)
	sel += @active_selection.grep(Sketchup::DimensionLinear) + @active_selection.grep(Sketchup::DimensionRadial)
	sel
end

#SELECTION: Register the original selection
def selection_register_original
	@no_setback = true
	@original_selection_ids = []
	@original_selection.each do |e| 
		next unless e.valid?
		@original_selection_ids.push e.entityID
		@no_setback = false if e.instance_of?(Sketchup::Face) || e.instance_of?(Sketchup::Edge) ||
                              e.instance_of?(Sketchup::Vertex)
	end	
end

#SELECTION: Restore the original selection, using a map of entityID
def selection_restore_original
	return unless @original_selection
	
	#Finding the invalid element in the original selection
	ls_inval = []
	@original_selection.each_with_index do |e, i| 
		ls_inval.push i unless e.valid?
	end
	
	return @original_selection if ls_inval.empty?
	
	#Building a map
	hmap_ids = {}
	@entities.to_a.each { |e| hmap_ids[e.entityID] = e }
	
	#Retrieving the invalid elements and restoring them in the selection
	ls_inval.each do |i|
		id = @original_selection_ids[i]
		e = hmap_ids[id]
		@original_selection[i] = (e) ? e : nil		
	end
	@original_selection = @original_selection.find_all { |e| e }
	selection_register_original
	@original_selection
end

#SELECTION: Calculated the extended selection to put into a group for intersections
def selection_extended(sel)
	hsh = {}
	sel.each do |e|
		if e.instance_of?(Sketchup::Vertex)
			e.faces.each { |f| hsh[f.entityID] = f }
			e.edges.each { |ee| hsh[ee.entityID] = ee }
		elsif e.instance_of?(Sketchup::Edge)
			curve = e.curve
			ledges = []
			ledges = (curve) ? curve.edges : [e]
			ledges.each do |ee| 
				hsh[ee.entityID] = ee 
				ee.faces.each { |f| hsh[f.entityID] = f }
			end	
		elsif e.instance_of?(Sketchup::Face)
			hsh[e.entityID] = e
			e.edges.each { |ee| hsh[ee.entityID] = ee }
		end
	end
	hsh.values
end

#---------------------------------------------------------------------------------------------
# MOVEMENT: Control of the Movement
#---------------------------------------------------------------------------------------------

#MOVEMENT: Initialize movement
def movement_prepare
	@hsh_snapshot_edges = nil
	
	#Mapping the id of the dynamic pre-selection if needed
	if @copy_mode || @multiple
		g = @entities.add_group selection_safe(@original_selection)
		@copy_group = g.copy
		g.explode
		@active_selection = [@copy_group]
	elsif movement_move_grouping?
		g = @entities.add_group selection_safe(@original_selection)
		@move_group = g
		@active_selection = [@move_group]		
	else
		@active_solution = @original_selection
		@hsh_snapshot_edges = {}
		@hsh_snapshot_vx = {}
		@entities.grep(Sketchup::Edge).each do |e| 
			@hsh_snapshot_edges[e.entityID] = e 
			vx1_id = e.start.entityID
			vx2_id = e.end.entityID
			@hsh_snapshot_vx["#{vx1_id}-#{vx2_id}"] = true
		end	
	end	
	@otpicker.exclusion_register(@active_selection)
	
	#Entering the dragging mode
	@tr_movement = @tr_id
	@dragging = true
	@mode = @otpicker.set_mode(:target)
end

#MOVEMENT: Check if selection can be encapsulated in a group for moving
def movement_move_grouping?
	return false unless @no_setback
	sel = @original_selection
	n = sel.length
	comps = sel.grep(Sketchup::ComponentInstance) + sel.grep(Sketchup::Group)
	return false if comps.length == n
	
	if n == 1
		elt0 = sel[0]
		
		#text with no anchor must be encapsulated - bug in transform_entities
		return true if elt0.instance_of?(Sketchup::Text) && !elt0.point
	end
	
	false
end

#MOVEMENT: Cancel movement
def movement_cancel
	abort_operation
	start_operation
	reset_env
	@active_selection = @otpicker.preselection_restore
	onMouseMove_zero
end

#MOVEMENT: Restore the initial state
def movement_set_back	
	#Cancelling any previous movement
	abort_operation
	start_operation
	@tr_movement = @tr_id
	
	#Restoring the selection if the copy group was on
	@active_selection = selection_restore_original
	wireframe_compute
	@otpicker.exclusion_register(@active_selection)
	@copy_group = nil	
end

#MOVEMENT: start the move
def movement_start
	return unless @active_selection
	
	#Saving the original selection
	@multiple = nil
	@pivoted = false
	@pivoted_at_origin = nil
	@vec_pivot_flip = false
	@original_selection = @active_selection
	selection_register_original
	wireframe_compute
	
	@pt_origin_prev = @pt_origin
	
	#Commiting the previous operation if any
	commit_operation
	start_operation
	
	#Initializing the move environment
	movement_prepare
	
	refresh_viewport
end

#MOVEMENT: Store the information of current movement
def movement_store
	@otpicker.vcb_freeze
	@vec_move_prev = @pt_origin.vector_to(@pt_target) if @pt_target
end

#MOVEMENT: Finish the move
def movement_stop
	return unless @dragging && @pt_target
	
	#Deferred Move
	if @defer_mode && !@multiple
		movement_execute 
	end	
		
	#Interactive_move
	vec_move = @pt_origin.vector_to @pt_target
	unless vec_move.valid?
		reset_env
		return
	end	
	
	#Exploding the copy group if any
	if @copy_group
	
		#Handling multiple copies
		if @multiple
			ncopy = @multiple.abs
			offset = vec_move.length
			if @divide
				offset0 = -offset
				offset = offset / ncopy
			else
				offset0 = 0
			end	
			if ncopy > 0 && offset > 0
				lsi = (2..ncopy).to_a
				if @multiple < 0
					if @divide
						lsi += (-ncopy+1..0).to_a 
					else
						lsi += (-ncopy..-1).to_a 
					end
				end	
				lsi.each do |i|
					pt = @pt_origin.offset(vec_move, (i-1) * offset + offset0)
					t = Geom::Transformation.translation @pt_origin.vector_to(pt)
					g = @copy_group.copy
					@entities.transform_entities t, g
					G6.grouponent_explode g
				end
			end
		end
		
		#Resetting the selection
		@active_selection = G6.grouponent_explode(@copy_group)
		if @otpicker.preselection_get
			if (!@multiple || @multiple == 1)
				@active_selection = @otpicker.preselection_restore
			else
				@active_selection = @otpicker.preselection_set nil
			end	
		end	
		@copy_group = nil
		
	#Group copy
	elsif @move_group
		@active_selection = G6.grouponent_explode(@move_group)
	
	#Ensuring intersections are properly done
	elsif !@no_setback
		sel = selection_extended(@active_selection)
		g = @entities.add_group sel
		@active_selection = G6.grouponent_explode(g)
	end
	
	#Autosmooth for new created edges
	if @hsh_snapshot_edges
		@entities.grep(Sketchup::Edge).each do |e|
			next if e.faces.empty?
			next if @hsh_snapshot_edges[e.entityID]
			vx1_id = e.start.entityID
			vx2_id = e.end.entityID
			next if @hsh_snapshot_vx["#{vx1_id}-#{vx2_id}"] || @hsh_snapshot_vx["#{vx2_id}-#{vx1_id}"]
			e.smooth = e.soft = true
			e.casts_shadows = false
		end
	end	
	
	@otpicker.suselection_add @active_selection, true
	@otpicker.preselection_set_when_cumulative	
	
	#Resetting the environment
	@lst_wireframe_elt = @lst_wireframe_comp = nil
	@geometry_generated = true
	reset_env
	
	#Refreshing the view
	refresh_viewport
end

#MOVEMENT: Perform the interactive move
def movement_execute
	return unless @dragging && @pt_target && @active_selection

	#Abort operation if natural geometry not in copy mode
	movement_set_back unless @copy_group || @no_setback
	
	#Execute the move of entities
	movement_perform_move
end

#MOVEMENT: Move the entities
def movement_perform_move(sel=nil)
	sel = @active_selection unless sel
	return unless @pt_target
	vec = @pt_origin.vector_to @pt_target
	return unless vec.valid?
	@action_ongoing = :move
	
	#Move anchor point of text
	if @flag_anchor_text && sel.length == 1 && sel[0].instance_of?(Sketchup::Text)
		pt, symb, info_comp = @otpicker.origin_info
		eltx, comp, tr = info_comp
		if symb == :text_anchor && !comp
			eltx.point = @pt_target
			return
		end	
	end
	
	#Otherwise move entities
	tinv = @tr_movement.inverse
	trot = movement_pivot_tr
	t = trot * Geom::Transformation.translation(vec)
	@entities.transform_entities t * tinv, sel
	@tr_movement = t
end

#MOVEMENT: store pivot directions
def movement_on_going
	@vec_pivot_origin, @code_pivot_origin = @otpicker.direction_pivot_origin
	@vec_pivot_target, @code_pivot_target = @otpicker.direction_pivot_target
end	

#MOVEMENT: Compute the pivot transformation for alignment if applicable
def movement_pivot_tr
	@pivoted = false
	return @tr_id unless @align_mode
	
	#Directions at origin and target
	vec_origin = @vec_pivot_origin
	vec_target = @vec_pivot_target
	return @tr_id unless vec_origin && vec_target
	
	#Flipping if required
	if @vec_pivot_flip
		vec_target = vec_target.reverse
	end	
	axes_origin = vec_origin.reverse.axes
	axes_target = vec_target.axes
	tr_ori = Geom::Transformation.axes @pt_target, *axes_origin
	tr_targ = Geom::Transformation.axes @pt_target, *axes_target
	trot = tr_targ * tr_ori.inverse
	
	#Final transformation
	@pivoted = true
	trot
end

#MOVEMENT: Post flipping of selection when alignment was done previously
def movement_pivot_flip
	return unless @align_mode
	@vec_pivot_flip = !@vec_pivot_flip
	if @pivoted
		movement_set_back
		movement_repeat true, nil
	elsif @pivoted_at_origin
		@geometry_generated = false
		movement_pivot_at_origin *@pivoted_at_origin
	end
end

#MOVEMENT: Perform a post execute operation
def movement_post_execute(offset, multiple=nil, divide=nil)
	#Resetting the previous selection
	@active_selection = @original_selection
	@multiple = multiple
	@divide = divide
	
	#Store current information
	movement_store if @dragging
		
	#Perform the move	
	movement_repeat true, offset
	
	#Resetting the SU selection
	unless @multiple
		@otpicker.suselection_add @active_selection, true
	end	
end

#MOVEMENT: Perform a post execute operation
def movement_set_vector(new_vec, multiple=nil, divide=nil)
	#Resetting the previous selection
	@multiple = multiple
	@divide = divide
	
	#Store current information and perform the move
	if @pt_target
		@active_selection = @original_selection
		movement_store if @dragging
		movement_repeat true, nil, new_vec
	else	
		movement_start
		movement_repeat false, nil, new_vec
	end
	
	#Resetting the SU selection
	unless @multiple
		@otpicker.suselection_add @active_selection, true
	end	
end

#MOVEMENT: Repeat the movement based on the previous vector and offset if defined
def movement_repeat(cancel=nil, offset=nil, new_vec=nil, new_origin=nil)
	return unless @active_selection

	#Getting the applicable origin and target
	@pt_origin, @pt_target = @otpicker.vcb_change_specifications(offset, new_origin, new_vec)
	
	#Not valid post-execution - Resetting the previous environment
	if !@pt_origin || !@pt_target || @pt_origin == @pt_target
		reset_env
		return
	end
	
	#Cancelling and restarting operation if vector has not changed
	if new_vec
		@vec_move_prev = new_vec
	else	
		new_vec = @vec_move_prev
	end	
	if cancel && @vec_move_prev.valid? && new_vec.valid? && new_vec.parallel?(@vec_move_prev)
		movement_set_back
	else
		commit_operation
		start_operation
	end
	
	#Performing the transformation
	movement_prepare
	movement_perform_move
	movement_stop
	
	#Storing the multiple and divide flags
	@multiple_prev = @multiple
	@divide_prev = @divide		
end

#MOVEMENT: Redo the movement on selection
def movement_redo
	return unless @active_selection && @pt_origin_prev
	@multiple = @multiple_prev
	@divide = @divide_prev
	movement_repeat
end

#MOVEMENT: Change the origin from remote point
def movement_change_from_remote(distance)
	new_ori, targ = @otpicker.vcb_change_specifications(distance)
	return if distance == 0
	
	@origin_info = @otpicker.origin_set_info [new_ori, :remote, nil]
	@pt_origin, = new_ori
	movement_start
	onMouseMove_zero
end

#MOVEMENT: Perform nothing or pivoting in Origin mode
def movement_align_at_origin
	return false unless @align_mode && @pt_origin
	vec_ori, = @otpicker.direction_pivot_origin
	if @vec_parallel && vec_ori && @original_selection
		movement_pivot_at_origin @pt_origin, @vec_parallel, vec_ori
	end	
	reset_env
	@otpicker.remote_reset
	onMouseMove_zero
	true
end

#MOVEMENT: Perform a pivoting at origin
def movement_pivot_at_origin(pt_origin, vec_ref, vec_ori)
	vec_ref0 = vec_ref
	vec_ori0 = vec_ori
	angle = vec_ref.angle_between vec_ori
	vec_ref = vec_ref.reverse if angle > 0.5 * Math::PI
	vec_ref = vec_ref.reverse if @vec_pivot_flip
	
	axes_ref = vec_ref.axes
	axes_ori = vec_ori.axes
	tr_ref = Geom::Transformation.axes pt_origin, *axes_ref
	tr_ori = Geom::Transformation.axes pt_origin, *axes_ori
	trot = tr_ref * tr_ori.inverse
	
	if @geometry_generated
		commit_operation
		start_operation
	else	
		movement_set_back
	end	
	
	#Performing the rotation
	@entities.transform_entities trot, @original_selection
	@geometry_generated = true
	@pivoted_at_origin = [pt_origin, vec_ref0, vec_ori0]
end

#---------------------------------------------------------------------------------------------
# ROTATION: Handle rotation of groups and components
#---------------------------------------------------------------------------------------------

#ROTATION: Compute the crosses if applicable
def rotation_compute_crosses(view)
	ptinter, top_parent, ipanel, pts = @otpicker.origin_info_top_parent

	#No top parent under mouse
	unless top_parent
		@red_cross_info = nil
		return
	end	
	
	#Already calculated
	top_parent_c, ipanel_c = @red_cross_info
	return if top_parent_c == top_parent && ipanel == ipanel_c && rotation_crosses_showable?(view)
	
	#Calculating the crosses
	dim = 6
	ratio = 0.7
	lines = []
	pt1, pt2, pt3, pt4 = pts
	vecx = pt1.vector_to pt2
	vecy = pt1.vector_to pt4
	vecz = vecx * vecy
	lenx2 = vecx.length * 0.5 * ratio
	leny2 = vecy.length * 0.5 * ratio
	ptmid = Geom.linear_combination 0.5, pt1, 0.5, pt3
	pts_cross = []
	[[vecx, lenx2], [vecx, -lenx2], [vecy, leny2], [vecy, -leny2]].each do |vec, len|
		pts_cross.push ptmid.offset(vec, len)
	end
	
	#Storing the information
	@rot_tr = @tr_id
	@red_cross_info = [top_parent, ipanel, pts, ptmid, pts_cross, [vecx, vecy, vecz]]
	
	#Checking if the crosses bounds is not too small
	rotation_crosses_showable?(view)
end

#ROTATION: Check if the rotation crosses should be shown
def rotation_crosses_showable?(view)
	return false unless @red_cross_info
	top_parent, ipanel, pts, ptmid, pts_cross = @red_cross_info
	pix = 20
	ptmid2d = view.screen_coords ptmid
	ptmid2d.z = 0
	pts_cross.each do |pt|
		pt2d = view.screen_coords pt
		pt2d.z = 0
		if ptmid2d.distance(pt2d) < pix
			@red_cross_info = nil
			return false
		end	
	end	
	true
end

#ROTATION: Check if a red cross is picked
def rotation_cross_picked(view, x, y)
	@rotation_mode = false
	return nil unless @red_cross_info
	
	#Priority to bounding box and vertex
	pt, type = @origin_info
	case type.to_s
	when /bbox/, /center/, /vertex/, /edge/, /cpoint/, /end_cline/, /midpoint/, /inter/
		return
	end
	
	top_parent, ipanel, pts, ptmid, pts_cross, lvec = @red_cross_info
	
	dmin = @precision + 1 
	ptxy = Geom::Point3d.new x, y, 0
	icross_found = nil
	pts_cross.each_with_index do |ptc, icross|
		ptc2d = view.screen_coords ptc
		ptc2d.z = 0
		d = ptc2d.distance ptxy
		if d < dmin
			dmin = d
			icross_found = icross
		end	
	end
	
	return nil unless icross_found
	
	#Setting the protractor
	ptcross = pts_cross[icross_found]
	pt1, pt2, pt3, pt4 = pts_cross
	@rot_center = Geom.linear_combination 0.5, pt1, 0.5, pt2
	@rot_normal = pt1.vector_to(pt2) * pt1.vector_to(pt4)
	@rot_basedir = @rot_basedir_ori = @rot_center.vector_to(ptcross)
	face = nil
	origin_info = @otpicker.origin_info
	face, = origin_info[2] if origin_info
	face = nil unless face.instance_of?(Sketchup::Face)
	@protractor.set_placement @rot_center, @rot_normal, @rot_basedir, face
	
	@rotation_mode = true
	@rot_comp = top_parent
	@refresh_viewport
end

#Rotation Repeat
def rotation_repeat
	rotation_post_execute @rot_angle_prev if @rot_angle_prev
end

#ROTATION: Start the rotation
def rotation_start
	#Entering the dragging mode
	@rot_tr = @tr_id
	@dragging = true
	@mode = @otpicker.set_mode(:target)
	
	#Starting the operation
	commit_operation
	start_operation	
end

#ROTATION: Stopping the rotation
def rotation_stop
	@geometry_generated = true
	@rot_angle_prev = @rot_angle
	reset_env
end

#ROTATION: Executing the rotation
def rotation_execute(view, x, y)
	@action_ongoing = :rotation

	#Calculating the angle
	normal = (view.camera.direction % @rot_normal > 0) ? @rot_normal.reverse : @rot_normal
	plane = [@rot_center, normal]
	pt = Geom.intersect_line_plane view.pickray(x, y), plane
	vec = @rot_center.vector_to pt
	angle = @rot_basedir_ori.angle_between vec
	angle = -angle if angle != 0 && (@rot_basedir_ori * vec) % normal < 0
	
	#Rounding up the angle
	center2d = view.screen_coords @rot_center
	center2d.z = 0
	ptxy = Geom::Point3d.new x, y, 0
	d = center2d.distance ptxy
	radius2d = 1.2 * @protractor.get_radius2d
	factor = d / radius2d
	adjusted_angle = @protractor.adjust_angle(angle.abs, factor)
	
	#Performing the transformation
	@rot_angle = (angle < 0) ? -adjusted_angle : adjusted_angle
	rotation_perform
end
	
#ROTATION: Perform the rotation of the component	
def rotation_perform
	normal = (@view.camera.direction % @rot_normal > 0) ? @rot_normal.reverse : @rot_normal
	trot = Geom::Transformation.new @rot_center, normal, @rot_angle
	t = @rot_tr.inverse * trot
	@rot_basedir = trot * @rot_basedir_ori
	#@entities.transform_entities t, @rot_comp
	@entities.transform_entities t, @active_selection
	@rot_tr = trot	
end
	
#ROTATION: Executing the rotation
def rotation_post_execute(angle)
	@rot_tr = @tr_id unless @action_ongoing == :rotation
	@rot_angle = angle
	
	#Committing the previous operation if any
	commit_operation
	start_operation

	#Performing the rotation
	rotation_perform	
	rotation_stop
end
	
#ROTATION: Draw the protractor and base direction lines
def rotation_draw(view)
	return unless @rotation_mode
	
	size = view.pixels_to_model 100, @rot_center
	
	#Draw the guideline
	lvec = [@rot_basedir_ori]
	lvec.push @rot_basedir if @rot_basedir != @rot_basedir_ori
	
	#Draw the current and original base dir
	center2d = view.screen_coords @rot_center
	center2d.z = 0
	lvec.each_with_index do |vec, ivec|
		pt = @rot_center.offset vec, size
		pt2d = view.screen_coords pt
		pt2d.z = 0
		vec2d = center2d.vector_to pt2d
		wid = 2 * view.vpwidth
		pt1 = center2d.offset vec2d, wid
		if (ivec == 1)
			pt2 = center2d
			view.line_width = 1
		else	
			pt2 = center2d.offset(vec2d, -wid)
			view.line_width = 2
		end	
		view.line_stipple = '_'
		view.drawing_color = 'black'
		view.draw2d GL_LINE_STRIP, [pt1, pt2]
	end
	
	#Drawing the protractor
	@protractor.draw view, 1.5	
	
	#Tip for angle
	if @rot_angle
		txangle = "#{sprintf('%0.1f', @rot_angle.radians)} deg."
		bkcolor, frcolor = @color_box_angle
		G6.draw_rectangle_text(view, center2d.x, center2d.y, txangle, bkcolor, frcolor)
	end	
end

#ROTATION: Draw the red crosses
def rotation_draw_marks(view)
	return unless @red_cross_info
	
	top_parent, ipanel, pts, ptmid, pts_cross, lvec = @red_cross_info
	vecx, vecy, vecz = lvec
	
	#Calculating the crosses
	dim = 6
	lines = []
	pts_cross.each do |ptc|
		size = view.pixels_to_model dim, ptc
		lines.push ptc.offset(vecx, -size), ptc.offset(vecx, size), ptc.offset(vecy, -size), ptc.offset(vecy, size)
	end
	
	#Drawing the crosses
	view.line_stipple = ''
	view.line_width = 2
	view.drawing_color = 'red'
	view.draw GL_LINES, lines.collect { |pt| @rot_tr * pt }
end

#---------------------------------------------------------------------------------------------
# WIREFRAME: Wireframe for Defer mode
#---------------------------------------------------------------------------------------------

#WIREFRAME: Compute the wireframe for the active selection
def wireframe_compute
	selection = @original_selection
	return unless selection && selection.length > 0
	
	@lst_wireframe_elt = []
	@lst_wireframe_comp = []
	@wireframe_level = 0
	
	#Groups and components
	wireframe_grouponent_contribution(selection, @tr_id)
	
	#Faces, edges and images
	wireframe_element_contribution selection
end

#Check if enough for the wireframe
def wireframe_stop?
	@lst_wireframe_comp.length + @lst_wireframe_elt.length > @wireframe_max_elts
end

#WIREFRAME: Build the wireframe for faces, edges and images
def wireframe_element_contribution(ls_entities, t=nil)
	return if wireframe_stop?
	
	if t == nil
		t = @tr_id
		lst_wireframe = @lst_wireframe_comp
	else
		lst_wireframe = @lst_wireframe_elt
	end
	
	#Faces
	hedges = {}
	faces = ls_entities.grep(Sketchup::Face)
	faces.each do |face|
		next unless face.valid?
		face.edges.each do |edge|
			next unless edge.valid? && G6.edge_plain?(edge)
			eid = edge.entityID
			next if hedges[eid]
			hedges[eid] = true
			lst_wireframe.push t * edge.start.position, t * edge.end.position
			return if wireframe_stop?
		end	
	end

	#Edges
	ledges = ls_entities.grep(Sketchup::Edge)
	ledges.each do |edge|
		next unless edge.valid? && (G6.edge_plain?(edge) || edge.faces.length == 0)
		next if hedges[edge.entityID]
		lst_wireframe.push t * edge.start.position, t * edge.end.position
		return if wireframe_stop?
	end

	#Images
	limages = ls_entities.grep(Sketchup::Image)
	limages.each do |image|
		bb = image.bounds
		lst_wireframe.concat G6.bbox_quads_lines(bb)[1].collect { pt| t * pt }
		return if wireframe_stop?
	end
end

#WIREFRAME: Compute the wireframe recursively for groups and components
def wireframe_grouponent_contribution(ls_entities, t=nil)
	ls_next = []

	lscomp = ls_entities.grep(Sketchup::ComponentInstance)
	lscomp.each do |comp|
		tt = t * comp.transformation
		@lst_wireframe_comp.concat G6.grouponent_box_lines(@view, comp, tt)
		return if wireframe_stop?
		ee = comp.definition.entities
		wireframe_element_contribution(ee, tt)
		ls_next.push [ee, tt]
	end	

	lsgroup = ls_entities.grep(Sketchup::Group)
	lsgroup.each do |group|
		tt = t * group.transformation
		@lst_wireframe_comp.concat G6.grouponent_box_lines(@view, group, tt)
		return if wireframe_stop?
		ee = group.entities
		wireframe_element_contribution(ee, tt)
		ls_next.push [ee, tt]
	end	
	
	return if ls_next.empty?
	ls_next.each { |a| wireframe_grouponent_contribution *a }
end

#WIREFRAME: Draw the wireframe
def wireframe_draw(view, lst_wireframe, color)
	return unless @defer_mode && @pt_target
	
	#Wireframe for groups, components and edges
	if lst_wireframe && lst_wireframe.length > 0
		vec = @pt_origin.vector_to @pt_target
		trot = movement_pivot_tr
		t = trot * Geom::Transformation.translation(vec)
		
		if @copy_mode
			view.line_width = 3
			view.line_stipple = ''
			view.drawing_color = color
		else	
			view.line_width = 2
			view.line_stipple = '-'
			view.drawing_color = color
		end	
		view.draw2d GL_LINES, lst_wireframe.collect { |pt| view.screen_coords(t * pt) }
	end
end

#---------------------------------------------------------------------------------------------
# KEY: Keyboard handling
#---------------------------------------------------------------------------------------------

#KEY: Return key pressed
def onReturn(view)
	@otpicker.direction_pick_from_model
	refresh_viewport
end

#KEY: Manage key events
def key_manage(event, key, view)
	case event
	
	#SHIFT key
	when :shift_down
		@otpicker.direction_changed
	when :shift_toggle
		if @mode == :origin
			extended_toggle
		else	
			@otpicker.direction_toggle_lock
		end	
	when :shift_up
		@otpicker.direction_changed
		selection_cumulate unless @shift_tab_done
		@shift_tab_done = false
	
	#CTRL key	
	when :ctrl_down	
		@copy_mode_at_down = @copy_mode
		copy_toggle
	when :ctrl_up
		copy_set @copy_mode_at_down
	when :ctrl_other_up
		copy_set @copy_mode_at_down
	
	#ALT key
	when :alt_down
		@otpicker.no_inference_toggle
	when :alt_up	
		@otpicker.no_inference_set false
		
	#Arrows	
	when :arrow_down	
		@otpicker.option_from_arrow key, @keyman.shift_down?, @keyman.ctrl_down?
	
	#Backspace	
	when :backspace
		@otpicker.inference_reset

	#TAB	
	when :tab
		defer_toggle
	when :shift_tab
		@shift_tab_done = true
		align_toggle
	end
end

#KEY: Handle Key down
def onKeyDown(key, rpt, flags, view)
	ret_val = @keyman.onKeyDown(key, rpt, flags, view)
	onSetCursor
	@view.invalidate
	ret_val
end

#KEY: Handle Key up
def onKeyUp(key, rpt, flags, view)
	ret_val = @keyman.onKeyUp(key, rpt, flags, view)
	onSetCursor
	@view.invalidate
	ret_val
end
	
#---------------------------------------------------------------------------------------
# UNDO: Undo and Rollback Management
#---------------------------------------------------------------------------------------

#UNDO: Cancel and undo methods
def onCancel(flag, view)
	case flag
	when 1, 2	#Undo or reselect the tool
		handle_undo
	when 0		#user pressed Escape	
		handle_escape
	end
end

#UNDO: Handle an undo
def handle_undo
	reset_env
	UI.start_timer(0.1) { rollback }
end

#UNDO: Handle the escape key depending on current mode
def handle_escape
	if @otpicker.remote_active_dir
		@otpicker.remote_reset
	elsif @dragging
		movement_cancel
	elsif @otpicker.preselection_get
		@active_selection = @otpicker.preselection_set(nil)
		onMouseMove_zero
	elsif @mode == :origin
		@otpicker.direction_cancel_all
	end	
end

#UNDO: Execute a rollback
def rollback(perform_undo=false)
	if perform_undo
		movement_set_back
		abort_operation
		Sketchup.undo unless @geometry_generated
	end	
	@geometry_generated = false
	preselection = @otpicker.preselection_get
	if preselection
		preselection = preselection.find_all { |e| e.valid? }
		preselection = nil if preselection.empty?
		@active_selection = @otpicker.preselection_set preselection
	else
		@active_selection = nil
		@otpicker.suselection_clear
	end	
	reset_env
	@pivoted = false
	@pivoted_at_origin = nil
	onMouseMove_zero
end
		
#---------------------------------------------------------------------------------------------
# DRAW: Drawing Methods
#---------------------------------------------------------------------------------------------

#DRAW: Get the extents for drawing in the viewport
def getExtents
	@bb_extents
end

#DRAW: Draw top method
def draw(view)		
	#Rotation Mode
	rotation_draw_marks view
	if @rotation_mode
		rotation_draw view
	
	#Movement mode
	else
		if @mode == :target
			wireframe_draw view, @lst_wireframe_comp, @color_wireframe_comp
			wireframe_draw view, @lst_wireframe_elt, @color_wireframe_elt
		end	
		@otpicker.draw_all view
	end
	
	#Flag for fluidity
	@moving = false
	
	#Drawing the palette
	super
end

#--------------------------------------------------
# MENU: Contextual menu
#--------------------------------------------------

#MENU: Contextual menu
def getMenu(menu)
	cxmenu = Traductor::ContextMenu.new

	#Privileged directions
	cxmenu.add_sepa
	@otpicker.menu_contribution cxmenu
	
	#Copy and Defer modes
	cxmenu.add_sepa
	cxmenu.add_item(@tip_copy_mode) { copy_toggle }
	cxmenu.add_item(@tip_defer_mode) { defer_toggle }
	cxmenu.add_item(@tip_align_mode) { align_toggle }
	cxmenu.add_item(@tip_extended_mode) { extended_toggle }
	
	#Redo movement
	if @active_selection
		cxmenu.add_sepa
		if @rotation_mode
			cxmenu.add_item(@tip_rotation_redo) { rotation_repeat }	if @rot_angle_prev
		elsif @otpicker.vcb_redo_possible?
			cxmenu.add_item(@tip_movement_redo) { movement_start ; movement_redo }	
		end		
	end
	
	#Exit and Abort
	cxmenu.add_sepa
	cxmenu.add_item(@mnu_abort) { exit_tool }
	cxmenu.add_item(@mnu_exit) { abort_tool }
	
	#Showing the menu
	cxmenu.show menu
	true
end

#============================================================
# OPERATION: handling SU operation
#============================================================
	
#OPERATION: Start an operation based on symbol	
def start_operation
	@suops.start_operation nil, false
end

#OPERATION: Commit and save the fixing
def commit_operation
	return unless @geometry_generated
	@copy_group = nil
	@suops.commit_operation
	@geometry_generated = false
end

#OPERATION: Abort the operation
def abort_operation
	@suops.abort_operation
end

#-------------------------------------------------------------- 
#--------------------------------------------------------------
# Palette Management
#--------------------------------------------------------------
#--------------------------------------------------------------

#PALETTE: Separator for the Main and floating palette
def pal_separator ;	@palette.declare_separator ; end

#PALETTE: Initialize the main palette and top level method
def init_palette
	hshpal = { :width_message => 250, :width_message_min => 150, :key_registry => :toposhaper }
	@palette = Traductor::Palette.new hshpal
	@draw_local = self.method "draw_button_opengl"
	@blason_proc = self.method "draw_button_blason"
	@symb_blason = :blason
	@bk_color_tool = "lemonchiffon"
		
	#Blason Button
	tip = T7[:PlugName] + ' ' + T6[:T_STR_DefaultParamDialog]
	hsh = { :blason => true, :draw_proc => @blason_proc, :tooltip => tip }
	@palette.declare_button(@symb_blason, hsh) { MYPLUGIN.invoke_default_parameter_dialog }
	
	#Plane and Vector settings
	pal_separator
	@otpicker.palette_contribution_planar @palette
	pal_separator
	@otpicker.palette_contribution_vector @palette
	
	#Copy, Defer and Extened modes
	palette_copy
	palette_defer
	palette_extended
	palette_align
	palette_options
	palette_actions
	palette_native

	#Floating palette for flipping
	palette_floating_flip
	
	#Abort and Exit
	pal_separator
	hsh = { :draw_proc => :std_abortexit, :main_color => 'red', :frame_color => 'green', 
	        :tooltip => T6[:T_STR_AbortTool] }
	@palette.declare_button(:pal_abort, hsh) { abort_tool }

	#Rollback
	tip = T6[:T_STR_Undo] + " [#{T6[:T_KEY_CTRL]}-Z]"
	hsh = { :tooltip => tip, :draw_proc => :rollback }
	@palette.declare_button(:pal_back, hsh) { rollback true }
	
	#Exit tool
	hsh = { :draw_proc => :std_exit, :tooltip => T6[:T_STR_ExitTool] }
	@palette.declare_button(:pal_exit, hsh) { exit_tool }
		
	#Associating the palette
	set_palette @palette
end

#PALETTE: Button for Copy mode
def palette_copy
	pal_separator
	hgt = 32
	wid = hgt

	val_proc = proc { @copy_mode }
	hsh = { :tooltip => @tip_copy_mode, :height => hgt, :width => wid, :draw_proc => @draw_local,
			:hi_color => @color_but_hi, :value_proc => val_proc }
	@palette.declare_button(:pal_copy_mode, hsh) { copy_toggle }
end

#PALETTE: Button for Defer mode
def palette_defer
	pal_separator
	hgt = 32
	wid = hgt

	val_proc = proc { @defer_mode }
	hsh = { :tooltip => @tip_defer_mode, :height => hgt, :width => wid, :draw_proc => @draw_local,
			:hi_color => @color_but_hi, :value_proc => val_proc }
	@palette.declare_button(:pal_defer_mode, hsh) { defer_toggle }
end

#PALETTE: Button for Defer mode
def palette_align
	pal_separator
	hgt = 32
	wid = hgt

	val_proc = proc { @align_mode }
	hsh = { :tooltip => @tip_align_mode, :height => hgt, :width => wid, :draw_proc => @draw_local,
			:hi_color => @color_but_hi, :value_proc => val_proc }
	@palette.declare_button(:pal_align_mode, hsh) { align_toggle }
end

#PALETTE: Button for Defer mode
def palette_extended
	pal_separator
	hgt = 32
	wid = hgt

	val_proc = proc { @extended_mode }
	hsh = { :tooltip => @tip_extended_mode, :height => hgt, :width => wid, :draw_proc => :arrow_RL,
			:hi_color => @color_but_hi, :value_proc => val_proc }
	@palette.declare_button(:pal_extended_mode, hsh) { extended_toggle }
end

def palette_native
	pal_separator
	hsh = { :tooltip => T7[:TIP_NativeMove], :draw_proc => :std_edition_move, :main_color => 'black' }
	@palette.declare_button(:pal_native, hsh) { native_tool }
end	

#PALETTE: Button for Actions
def palette_options
	pal_separator
	
	#Title button for Options
	symb_master = :pal_options
	hsh = { :type => 'multi_free', :passive => true, :text => T6[:T_BOX_Options], :tooltip => T6[:T_TIP_Options],
	        :bk_color => @color_but_title }
	@palette.declare_button(symb_master, hsh)
	
	hshp = { :parent => symb_master, :width => 28, :hi_color => @color_but_hi }

	#Auto-smooth of edges
	symb = "#{symb_master}_AutoSmooth".intern
	value_proc = proc { @option_autosmooth }
	hsh = { :value_proc => value_proc, :draw_proc => @draw_local, :tooltip => T7[:TIP_AutoSmooth] }
	@palette.declare_button(symb, hshp, hsh) { autosmooth_toggle }
	
	#Remote Inferences
	@otpicker.palette_contribute_button(:inference_remote, @palette, symb_master, hshp)
	
	#Inferences on Components
	@otpicker.palette_contribute_button(:inference_on_comp, @palette, symb_master, hshp)
	
	#No Inferences
	@otpicker.palette_contribute_button(:no_inference, @palette, symb_master, hshp)	
end
	
#PALETTE: Button for Actions
def palette_actions
	pal_separator

	#Title button for Actions
	symb_master = :pal_actions
	hsh = { :type => 'multi_free', :passive => true, :text => T6[:T_BOX_Actions], :tooltip => T6[:T_TIP_Actions],
	        :bk_color => @color_but_title }
	@palette.declare_button(symb_master, hsh)
	
	hshp = { :parent => symb_master, :width => 28, :hi_color => @color_but_hi }

	#Toggle Hidden geometry
	@otpicker.palette_contribute_button(:toggle_show_hidden, @palette, symb_master, hshp)
	
	#Erase all inference marks
	@otpicker.palette_contribute_button(:reset_all_marks, @palette, symb_master, hshp)
end

#PALETTE: Floating palette for Calculation operations (algo)
def palette_floating_flip
	wh = 20
	
	#Declare the floating palette
	flpal = :pal_flip_floating
	hshfloat = { :floating => flpal, :row => 0 }
	hidden_proc = proc { (!@pivoted && !@pivoted_at_origin)|| @dragging }
	hsh = { :auto_move => [:up], :hidden_proc => hidden_proc }
	@palette.declare_floating flpal, hsh
	
	#Button for flipping
	symb = "#{flpal}_flip".intern
	hsh = { :tooltip => T7[:TIP_PivotedTip], :width => wh, :height => wh, :draw_proc => @draw_local }
	@palette.declare_button(symb, hshfloat, hsh) { movement_pivot_flip }
end

#PALETTE: Drawing the Blason
def draw_button_blason(symb, dx, dy)
	lst_gl = []
	dx2 = dx/2
	dy2 = dy/2
	
	lpti = [[1, dy2], [dx2, dy-1], [dx-1, dy2], [dx2, 1]]
	pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
	lst_gl.push [GL_POLYGON, pts, 'lightyellow']

	dec = 2
	lpti = [[1, dy2+dec], [dx2-dec, dy-1], [dx2+dec, dy-1], [dx-1, dy2+dec], [dx-1, dy2-dec], [dx2+dec, 1], [dx2-dec, 1], [1, dy2-dec]]
	pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
	lst_gl.push [GL_LINES, pts, 'lightblue', 1, '']	
	
	dec = 4
	y = dy2-dec-7
	lpti = [[dx2, y], [dx2-dec, y+dec], [dx2+dec, y+dec]]
	pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
	pts2 = G6.pts_square dx2, y+dec+2, 2
	lst_gl.push [GL_POLYGON, pts, 'red']	
	lst_gl.push [GL_POLYGON, pts2, 'red']	
	lst_gl.push [GL_LINE_LOOP, pts, 'black', 1, '']	
	lst_gl.push [GL_LINE_LOOP, pts2, 'black', 1, '']	
	t = Geom::Transformation.rotation Geom::Point3d.new(dx2, dy2, 0), Z_AXIS, 90.degrees
	for i in 1..3
		pts = pts.collect { |pt| t * pt }
		pts2 = pts2.collect { |pt| t * pt }
		lst_gl.push [GL_POLYGON, pts, 'red']	
		lst_gl.push [GL_POLYGON, pts2, 'red']	
		lst_gl.push [GL_LINE_LOOP, pts, 'black', 1, '']	
		lst_gl.push [GL_LINE_LOOP, pts2, 'black', 1, '']	
	end

	lst_gl
end

#PALETTE: Custom drawing of buttons
def draw_button_opengl(symb, dx, dy, main_color, frame_color, selected)
	code = symb.to_s
	lst_gl = []
	dx2 = dx / 2
	dy2 = dy / 2
	grayed = @palette.button_is_grayed?(symb)
	color = (grayed) ? 'gray' : frame_color
	
	case code
	when /pal_copy_mode/
		dx3 = dx/3
		pts = G6.pts_rectangle 1, dx2-2, dx3, dx2
		lst_gl.push [GL_POLYGON, pts, 'blue']		
		lst_gl.push [GL_LINE_LOOP, pts, 'black', 1, '']		
		pts = G6.pts_rectangle dx-dx3-1, dx2-2, dx3, dx2
		lst_gl.push [GL_POLYGON, pts, 'green']		
		lst_gl.push [GL_LINE_LOOP, pts, 'black', 1, '']	
		dec = 5
		lpti = [[dx2-dec, dec+1], [dx2+dec, dec+1], [dx2, 1], [dx2, 2*dec+1]]
		pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
		lst_gl.push [GL_LINES, pts, 'black', 2, '']	

	when /pal_defer_mode/
		dx3 = dx/3
		pts = G6.pts_rectangle 1, 2, dx3, dy-4
		lst_gl.push [GL_POLYGON, pts, 'blue']		
		lst_gl.push [GL_LINE_LOOP, pts, 'black', 1, '']		
		pts = G6.pts_rectangle dx-dx3-1, 2, dx3, dy-4
		lst_gl.push [GL_LINE_LOOP, pts, 'purple', 2, '-']	
		
	when /autosmooth/i
		pts = G6.pts_rectangle 2, 2, dx2, dy-2
		lst_gl.push [GL_LINE_LOOP, pts, 'gray', 1, '']		
		lpti = [[2, 2], [dx2, dy-2]]
		pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
		lst_gl.push [GL_LINE_STRIP, pts, 'blue', 2, '.']	
		lpti = [[dx-9, 2], [dx-4, 12], [dx-4, 10], [dx, 2], [dx-6, 5], [dx-2, 5]]
		pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
		lst_gl.push [GL_LINES, pts, 'red', 2, '']	

	when /pal_align_mode/
		dx3 = dx/3
		pts1 = G6.pts_rectangle 4, 0, dx3, dy-4
		trot = Geom::Transformation.rotation Geom::Point3d.new(dx2, dy2, 0), Z_AXIS, -30.degrees
		pts1 = pts1.collect { |pt| trot * pt }
		lpti = [[dx-1, 0], [dx-1, dy-2]]
		pts2 = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
		pts3 = G6.pts_circle dx2+3, dy2, dy2-2
		pts3 = pts3[5..7]
		pt1 = pts3[0]
		x1, y1 = pt1.to_a
		dec = 4
		pts4 = [Geom::Point3d.new(x1-dec, y1+2), pt1, Geom::Point3d.new(x1-2, y1-dec)]
		lst_gl.push [GL_POLYGON, pts1, 'red']		
		lst_gl.push [GL_LINE_LOOP, pts1, 'black', 1, '']	
		lst_gl.push [GL_LINE_STRIP, pts2, 'navy', 2, '_']	
		color = 'darkgray'
		lst_gl.push [GL_LINE_STRIP, pts3, color, 2, '']	
		lst_gl.push [GL_LINE_STRIP, pts4, color, 2, '']	
		
	when /floating_flip/
		dec = 0
		lpti = [[0, 0], [dx2-1, 0], [dx2-1, dy]]
		pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
		lst_gl.push [GL_POLYGON, pts, 'yellow']		
		lst_gl.push [GL_LINE_LOOP, pts, 'gray', 1, '']	

		lpti = [[dx2+1, 0], [dx, dec], [dx2+1, dy]]
		pts = lpti.collect { |x, y| Geom::Point3d.new(x, y, 0) }
		lst_gl.push [GL_POLYGON, pts, 'red']		
		lst_gl.push [GL_LINE_LOOP, pts, 'gray', 1, '']	
		
		
	end	#case code
	
	lst_gl
end

end	#class MoveAlongTool

end	#End Module MoveAlong

end	#End Module F6_FredoTools
