=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2011 Fredo6 - Designed and written Aug 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_FredoTools__ReportLabelArea.rb
# Original Date	:   26 Aug 2011 (based on work published in June 2011)
# Description	:   Built reports on Areas for selection or whole model
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

module ReportLabelArea

#Texts for the module
T7[:LABEL_NbFace] = "#Faces"
T7[:LABEL_Elements] = "Elements"
T7[:MSG_GenerationExport] = "Export File generated: %1"
T7[:MSG_GenerationExport_Error] = "Error in generation of CSV file: %1"
T7[:MSG_TotalArea] = "TOTAL AREA"
T7[:MSG_AverageArea] = "AVERAGE AREA"
T7[:MSG_DefaultMaterial] = "Default Material"
T7[:MSG_PW_Calculating] = "Calculating Area (%1)"
T7[:MSG_PW_OpeningReport] = "Opening Report"
T7[:MSG_PW_PleaseWait] = "..................... Please wait....."
T7[:MSG_CalculatingArea] = "Calculating the Area"	
T7[:TXT_ToplevelGeometry] = "Top-level Geometry"
T7[:TXT_EmbeddedCompGroup] = "Embedded Components and Groups"

T7[:REP_By_Material] = "by Material"
T7[:REP_By_Container] = "by Container"
T7[:REP_By_Layer] = "by Layer"

T7[:TIP_AddSelection] = "Click to Add to the selection"	
T7[:TIP_RemoveSelection] = "Click to Remove from the selection"	
T7[:TIP_Exit] = "Double Click to Exit Tool"	
T7[:TIP_Reset] = "Long Click to Clear Selection"	
T7[:TIP_Drag] = "Click and Drag to create Label"	
T7[:TIP_ReplaceLabel] = "Click to update text label with units"	
T7[:TIP_EditLabel] = "Double Click to edit the text of label"	
T7[:TIP_MoveLabel] = "Click and Drag to move Label"	
T7[:TIP_MoveAnchor] = "Click and Drag to move Anchor"	
T7[:TIP_LabelAll] = "Substitute current units in all labels of the selection"
T7[:TIP_Layer] = "Layer where the labels are created"

T7[:TIP_Recurse] = "Recurse: Calculate areas recursively down through components and groups"
T7[:TIP_SingleFace] = "Toggle selection of Single face or surfaces"
T7[:TIP_MultiArrow] = "Toggle multiple arrow for labels when multiple selection"
T7[:BUTTON_Recurse] = "Recurse (Shift)"
T7[:BUTTON_SingleFace] = "Single Face"
T7[:BUTTON_MultiArrow] = "Multi Arrow Labels"
T7[:DLG_TextLabel] = "New Text for the label"

T7[:MNU_EditLabel] = "Edit Text Label"	
T7[:MNU_UpdateLabel] = "Update Text Label"	
T7[:MNU_EraseLabel] = "Erase Label"	
T7[:MNU_AddSelection] = "Add to Selection"	
T7[:MNU_RemoveSelection] = "Remove from Selection"	
T7[:MNU_ClearSelection] = "Clear Selection (Esc)"	

T7[:OPS_LabelAll] = "Label All"
T7[:OPS_LabelErase] = "Label Erase"
T7[:OPS_LabelUpdate] = "Label Update"
T7[:OPS_LabelMove] = "Label Move"
T7[:OPS_LabelCreate] = "Label Create"

T7[:TIP_AutoUnitArea] = "Infer the most appropriate format"

#====================================================================================================
#----------------------------------------------------------------------------------------------------
# Plugin Implementation
#----------------------------------------------------------------------------------------------------
#====================================================================================================

HTML = Traductor::HTML

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

#Top level execution for menu commands
def self._execution(symb)
	case symb
	when :ReportLabelArea_report
		report_dialog 
	when :ReportLabelArea_label
		label_dialog
	end	
end

#----------------------------------------------------------------------------------------
# Persistence
#----------------------------------------------------------------------------------------

#Save the units across call to dialog boxes
def self.save_units(symb_unit, decimals)
	@symb_unit = symb_unit
	@decimals = decimals
	@model_unit = Sketchup.active_model
end

#Get the persistent units
def self.get_units
	return [nil, nil] if @model_unit != Sketchup.active_model
	[@symb_unit, @decimals]
end

#Save the current report
def self.save_current_report(symb)
	@current_report = symb
end

#Get the persisted cuirrent report
def self.get_current_report
	@current_report = :by_material unless @current_report	
	@current_report
end

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

#Initialize and execute the ReportLabelArea functionality
def self.report_dialog(selection=nil)
	ReportAreaTopDialog.invoke
end

#Initialize and execute the ReportLabelArea functionality
def self.label_dialog(selection=nil)
	Sketchup.active_model.select_tool LabelTool.new
end

#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
# Top Dialog box for Report Area
#--------------------------------------------------------------------------------------------------------------			 
#--------------------------------------------------------------------------------------------------------------			 

class ReportAreaTopDialog

PseudoData = Struct.new :nb_faces, :other_nb_faces, :tot_nb_faces, :face_area, :other_area, :tot_area
PseudoMat = Struct.new :su_mat, :nb_faces, :areas, :face, :name, :png_file
PseudoLayer = Struct.new :su_layer, :nb_faces, :area, :name

VisualReport = Struct.new :symb, :name, :ltable, :lst_areas

@@top_dialog = nil
@@unique_key = "ReportArea_Report_DLG"

#Invoke the Report Dialog
def ReportAreaTopDialog.invoke(selection=nil)
	@@top_dialog = self.new(selection) unless Traductor::Wdlg.check_instance_displayed(@@unique_key)
end	

#initialization of the dialog box 
def initialize(selection=nil)
	t0 = Time.now
	pls = Traductor::PleaseWaitTool.new "ReportLabelArea::Report"
	pls.message T7[:MSG_PW_Calculating, Traductor.text_model_or_selection] + T7[:MSG_PW_PleaseWait], 'yellow'
	return pls.exit unless calculate_area selection
	pls.message T7[:MSG_PW_OpeningReport] + T7[:MSG_PW_PleaseWait], 'lightgreen'
	@wdlg = create_dialog_top
	pls.exit
	update_status_bar
	true
end

#--------------------------------------------------------------------------------------------------------------
# Calculation of surface
#--------------------------------------------------------------------------------------------------------------			 

#Calculate the area of a selection
def calculate_area(selection=nil)
	@model = Sketchup.active_model
	@layer0 = @model.layers["Layer0"]
	selection = @model.selection unless selection
	entities = (selection == nil || selection.empty?) ? @model.active_entities : selection
	Sketchup.set_status_text T7[:MSG_Processing], SB_VCB_LABEL
	Sketchup.set_status_text T7[:MSG_CalculatingArea]

	#Initialization	
	@total_area = 0
	@total_faces = 0
	@hsh_materials = {}
	@hsh_layers = {}
	@uac = UnitAreaController.new *ReportLabelArea.get_units

	#Text for categories
	@hsh_texts = {}
	@hsh_texts[:top] = "Top Level"
	@hsh_texts[:group] = T6[:T_TXT_GROUP]
	@hsh_texts[:comp_inst] = T6[:T_TXT_COMP_INST]
	@hsh_texts[:comp_def] = T6[:T_TXT_COMPONENT]
	@hsh_texts[:mat] = T6[:T_TXT_MATERIAL]
	@hsh_texts[:back_mat] = T6[:T_TXT_BACK_MATERIAL]	
	@hsh_texts[:layer] = T6[:T_TXT_LAYER]	
	
	#Processing the selection	
	hoptions = { :entity_proc => method("face_process_proc"), :pelt_init_proc => method("pelt_init_proc") }
	@processor = Traductor::SelectionProcessor.new selection, hoptions 
	@processor.apply method("pelt_proc")
	@top_pelt = @processor.top_pelt
	
	#NO face in the selection
	if @total_area == 0
		UI.messagebox(T7[:MSG_NoFaceInSelection]) 
		return false
	end	
	
	#Finding the best unit
	@average_area = @total_area / @total_faces
	@uac.find_best_unit @average_area unless ReportLabelArea.get_units[0]
	
	#Finish the calculation
	update_status_bar
	true
end

#Procedure for Face processing
def face_process_proc(after, pelt, face)
	return unless after && face.instance_of?(Sketchup::Face)
	
	#Storing info for Elements
	data = pelt.data
	data.nb_faces += 1
	area = G6.face_area face, pelt.tr
	data.face_area += area
	@total_area += area
	@total_faces += 1
	
	#Storing info for materials
	mat = face.material
	store_for_mat face.material, face, area, 0
	store_for_mat face.back_material, face, area, 1
	
	#Storing info for layers
	store_for_layers(face.layer, area) unless face.layer == @layer0 || pelt.layers.include?(face.layer)
	if pelt.layers.length == 1 && pelt.layers[0] == @layer0
		store_for_layers @layer0, area
	else	
		layers = pelt.layers.delete_if { |a| a == @layer0 }
		layers.each { |layer| store_for_layers(layer, area) }
	end	
end

#Store information for materials
def store_for_mat(mat, face, area, ipos)
	id = (mat) ? mat.entityID : 0
	pmat = @hsh_materials[id] 
	unless pmat
		pmat = @hsh_materials[id] = PseudoMat.new
		pmat.areas = [0, 0]
		pmat.nb_faces = [0, 0]
		pmat.name = (mat) ? mat.display_name.strip : "--#{T7[:MSG_DefaultMaterial]}--"
		pmat.su_mat = mat
		pmat.face = face
	end
	pmat.areas[ipos] += area
	pmat.nb_faces[ipos] += 1
end

#Store information for materials
def store_for_layers(layer, area)
	id = (layer) ? layer.entityID : 0
	player = @hsh_layers[id] 
	unless player
		player = @hsh_layers[id] = PseudoLayer.new
		player.area = 0
		player.nb_faces = 0
		player.name = layer.name
		player.su_layer = layer
	end
	player.area += area
	player.nb_faces += 1
end

#Process a pseudo element	
def pelt_proc(after, pelt)
	return unless after
	data = pelt.data
	pelt.childrens.each do |p| 
		pdata = p.data
		data.other_area += pdata.face_area + pdata.other_area
		data.other_nb_faces += pdata.nb_faces + pdata.other_nb_faces
	end	
	data.tot_nb_faces = data.nb_faces + data.other_nb_faces
	data.tot_area = data.face_area + data.other_area
end	

#Initialization of data at creation of a pseudo element
def pelt_init_proc(pelt)
	data = pelt.data = PseudoData.new
	data.nb_faces = data.other_nb_faces = 0
	data.face_area = data.other_area = 0
	Sketchup.set_status_text "#{@uac.format_area @total_area}", SB_VCB_VALUE
end	

#Displaying the message in the status bar
def update_status_bar
	txarea = @uac.format_area @total_area
	Sketchup.set_status_text T7[:MSG_Finished], SB_VCB_LABEL
	Sketchup.set_status_text "#{txarea}", SB_VCB_VALUE
	text = "#{@total_faces} #{T7[:T_TXT_Faces]} --> #{T7[:MSG_TotalArea]} = #{txarea}"
	text += " - #{T7[:MSG_AverageArea]} = #{@uac.format_area(@average_area)}"
	Sketchup.set_status_text text
end

#--------------------------------------------------------------------------------------------------------------
# Dialog box configuration
#--------------------------------------------------------------------------------------------------------------			 

#Create the dialog box
def create_dialog_top
	init_dialog_top
	wdlg_key = @@unique_key
	@wdlg = Traductor::Wdlg.new @title, wdlg_key, false
	@wdlg.set_unique_key @@unique_key
	@wdlg.set_size @wid_total, @hgt_total
	@wdlg.set_background_color 'whitesmoke'
	@wdlg.set_callback self.method('topdialog_callback') 
	@wdlg.set_on_close { on_close_top() }
	refresh_dialog_top
	@wdlg.show
	@wdlg
end

#Initialize parameters of the dialog box
def init_dialog_top
	#Column width and Heights
	@hgt_table = 300
	@wid_extra = (RUN_ON_MAC) ? 40 : 80
	@wid_col_name = 300
	@wid_col_area = 200
	@wid_col_nbface = 60
	@wid_total = @wid_col_name + @wid_col_area + @wid_col_nbface + @wid_extra + 10
	@hgt_total = @hgt_table + 230	
	@title = T7[:MNU_Report]
	
	@txt_toplevel_geometry = T7[:TXT_ToplevelGeometry]
	@txt_embedded = T7[:TXT_EmbeddedCompGroup]
	@tip_group = T7[:T_TXT_GROUP]
	@tip_inst = T7[:T_TXT_COMP_INST]
	
	@lst_reports = [[:by_material, T7[:REP_By_Material]],
	                [:by_layer, T7[:REP_By_Layer]],
	                [:by_container, T7[:REP_By_Container]]]
	@lst_reports.each { |a| create_visual_report *a }

	@space4 = "&nbsp;" * 4
	@space_img = "&nbsp;" * 7	
end

#Initialize a report structure
def create_visual_report(symb, name)
	vreport = VisualReport.new
	vreport.symb = symb
	vreport.name = name
	@hsh_reports = {} unless @hsh_reports
	@hsh_reports[symb] = vreport
	vreport
end

#Refresh the dialog box
def refresh_dialog_top
	html = format_html_top @wdlg
	@wdlg.set_html html
end

#Notification of window closure
def on_close_top
	@@top_dialog = nil
	purge_temp_files
end

#Close the dialog box
def close_dialog_top
	@wdlg.close
end

#Call back for Dialog
def topdialog_callback(event, type, id, svalue)
	case event
		
	#Command buttons
	when /onclick/i
		if @uac.callback_check(id, @total_area / @total_faces)
			update_areas
			return svalue
		end
		case id
		when /ButtonExport/i
			export_as_csv
		when 'ButtonDone'
			@wdlg.close
		when 'ButtonPrint'
			@wdlg.print
		end	

	when /onchange/i
		case id
		when /ComboReport/i
			change_report svalue.intern
		end	
		
	when /onKeyUp/i	#Escape
		@wdlg.close if svalue =~ /\A27\*/
	when /onKeyDown/i	#Return key
		@wdlg.close if svalue =~ /\A13\*/
		
	end
	svalue
end

#Change the report
def change_report(symb)
	@wdlg.execute_script "please_wait() ;"
	ReportLabelArea.save_current_report symb
	UI.start_timer(0, false) { refresh_dialog_top }
end

#Update the areas after a change of unit or decimal
def update_areas
	@uac.update_tables @wdlg
	symb, unit, decimal, factor = @uac.get_current_param
	@wdlg.execute_script "update_areas('#{unit}', #{decimal}, #{factor}) ;"
	ReportLabelArea.save_units symb, decimal
	update_status_bar
end

#Build the HTML for Dialog
def format_html_top(wdlg)
	#Creating the HTML stream	
	html = Traductor::HTML.new
		
	#Initialization	
	@tw = Sketchup.create_texture_writer
	@tmpdir = LibFredo6.tmpdir
	space2 = "&nbsp;&nbsp;"
	@units = @uac.current_name
	
	#Creating the main table
	@xtable = Traductor::HTML_Xtable.new "XT0", html, wdlg
	ltable = prepare_report_table
	hoptions = option_xtable
	txt_table = @xtable.format_table ltable, hoptions
	
	#Styles used in the dialog box
	html.create_style 'Title', nil, 'B', 'K: navy', 'F-SZ: 16', 'text-align: center'
	html.create_style 'Header', nil, 'B', 'F-SZ: 11', 'BG: khaki'
	html.create_style 'Footer', nil, 'B', 'F-SZ: 11', 'BG: thistle'
	html.create_style 'HElement', 'Header', 'K: navy', 'text-align: left'
	html.create_style 'HArea', 'Header', 'K: navy', 'text-align: right'
	html.create_style 'HNbFace', 'Header', 'K: navy', 'text-align: right'
	
	html.create_style 'Element', nil, 'B', 'K: black', 'text-align: left'
	html.create_style 'ComboReport', nil, 'B', 'K: black', 'text-align: left', 'F-SZ: 11', 'K: blue', 'BG: powderblue', 'width: 250px'
	html.create_style 'Area', nil, 'B', 'K: navy', 'text-align: right'
	html.create_style 'NbFace', nil, 'I', 'K: green', 'text-align: right'
	html.create_style 'Level1', nil, 'F-SZ: 11'
	html.create_style 'Level2', nil, 'F-SZ: 10'
	
	@uac.html_class_style html
	
	html.create_style 'Button', nil, 'F-SZ: 10'
	html.create_style 'ButtonExport', nil, 'F-SZ: 9', 'BG: lightyellow'
	
	#Inserting the main table and unit choosers
	table_d = @uac.format_table_decimal
	table_u = @uac.format_table_units
	html.body_add "<div><table width='100%' cellspacing='0' cellpadding='0'><tr>"
	html.body_add "<td width='50px' align='left' valign='bottom'>#{@xtable.html_expand_buttons}</td>"
	html.body_add "<td align='right' width='70%' align='right'>#{table_u}</td>"
	html.body_add "<td align='right' valign='center'>#{table_d}</td>"
	html.body_add "</tr></table></div>"
	html.body_add "<div>", txt_table, "</div>"
	
	#Creating the dialog box button	
	butdone = HTML.format_button T7[:T_BUTTON_Done], "ButtonDone", 'Button', nil
	butexport = HTML.format_button T7[:T_BUTTON_ExportCSV], "ButtonExport", 'ButtonExport', nil, T7[:TIP_GenerationExportTxt]
	butprint = HTML.format_button T7[:T_BUTTON_Print], "ButtonPrint", 'Button', nil
	html.body_add "<table class='T_NOPRINT_Style' width='99%' cellpadding='6px'><tr>"
	html.body_add "<td width='33%' align='left'>", butprint, "</td>"
	html.body_add "<td width='33%' align='center'>", butexport, "</td>"
	html.body_add "<td align='right'>", butdone, "</td>"
	html.body_add "</tr></table>"
		
	#Creating a fixed div for showing images
	@size_mag = 100
	top = (RUN_ON_MAC) ? 120 : 100
	perc = (RUN_ON_MAC) ? 96 : 100
	style = "position:absolute; left: #{@wid_col_name}px; top: #{top}px ; border: 1px solid gray ; background-color : gray"
	style += "padding: 0px ; width: #{@size_mag}px ; height: #{@size_mag}px ; display: none"
	text = "<div id='Div_Magnifier' style='#{style}'>"
	text += "<img id='Magnifier' width='#{perc}%' height='#{perc}%' />"
	text += "</div>"
	text += "<div id='Div_Magnifier_solid' style='#{style}'></div>"
	html.body_add text
	
	#Please Wait message
	w = 200
	h = 100
	msg = "Please Wait...."
	style = "position:absolute; left: #{(@wid_total - w) * 0.5}px; top: #{(@hgt_total - h) * 0.5}px ; border: 1px solid gray ;"
	style += "width: #{w}px ; height: #{h}px ; display: none ; background-color: green ; color: white ; font-weight: bold"
	text = "<div id='PleaseWaitDiv' style='#{style}' >"
	text += "<table height='100%' width='100%'><tr><td align='center' valign='center'>#{msg}</td></tr></table></div>"
	html.body_add text
	
	#Special_scripts
	html.script_add @uac.special_scripts
	html.script_add special_scripts
	
	#Returning the HTML object
	html	
end

#Specific scripts for Downloading information
def special_scripts()
	#Storing the areas for each field
	dot, comma = Traductor.dot_comma true
	text = "var dot = '#{dot}' ;"
	text += "var comma = '#{comma}' ;"
	text += "var hsh_ids = [] ;"
	@hreport.lst_areas.each do |a|
		next unless a
		ipos, id, area = a
		text += "\nhsh_ids['#{id}'] = #{area} ;"
	end
	text += "\nhsh_ids['FOOTER_AREA'] = #{@total_area} ;"

	#Function to update the areas
	text += %Q~

function update_areas(unit, decimal, factor) {
	var multiple = 1.0 ;
	for (var i = 0 ; i < decimal ; i++) multiple = multiple * 10.0 ;
	for (var key in hsh_ids) {
		var obj = document.getElementById (key) ;
		if (obj == null) continue ;
		newval = hsh_ids[key] * factor ;
		sval = format_float(newval, decimal, multiple) ;
		obj.innerHTML = sval + ' ' + unit ;
		obj.title = newval.toString() + ' ' + unit ;
	}
}

function format_float(val, decimal, multiple) {
	var srest = null ;
	var smain = null ;
	if (decimal == 0) 
		smain = Math.round(val).toString() ;
	else {
		smain = (Math.round(val * multiple) / multiple).toString() ;
		var ls = smain.split ('.') ;
		smain = ls[0] ;
		srest = ls[1] ;
		if (!srest) srest = '' ;
		if (srest.length < decimal) srest = (srest + '00000000000000').substr(0, decimal) ;
	}	
	var newval = '' ;
	leng = smain.length ;
	smain = reverse_string (smain) ;
	for (var i = 0 ; i < leng ; i++) {
		if ((i > 0) && (i % 3 == 0)) newval = newval + comma ;
		newval = newval + smain.substr(i,1) ;
	}	
	newval = reverse_string (newval) ;
	if (srest) newval = newval + dot + srest ;
	return newval ;
}

function reverse_string(s) {
	return s.split("").reverse().join('') ;
}

function show_image(obj) {
	var src = obj.src ;
	var img = document.getElementById ('Magnifier') ;
	img.src = src ;
	var dv = document.getElementById ('Div_Magnifier') ;
	dv.style.display = "" ;
}

function show_image_solid(obj) {
	var color = obj.style.backgroundColor ;
	var dv = document.getElementById ('Div_Magnifier_solid') ;
	dv.style.backgroundColor = color ;
	dv.style.display = "" ;
}

function hide_image() {
	document.getElementById ('Div_Magnifier').style.display = 'none' ;
}

function hide_image_solid() {
	document.getElementById ('Div_Magnifier_solid').style.display = 'none' ;
}

function please_wait() {
	document.getElementById ('PleaseWaitDiv').style.display = '' ;
}

~
	
	text
end

#--------------------------------------------------------------------------------------------------------------
# Managing Xtable for reports
#--------------------------------------------------------------------------------------------------------------			 

#Def purge temporary files
def purge_temp_files
	@hsh_materials.each do |key, pmat| 
		f = pmat.png_file
		File.unlink f if f && FileTest.exist?(f)
	end	
end

#Prepare the table for the current report
def prepare_report_table
	#Retrrieving the current ltable - If it does not exist, build it once
	@current_report = ReportLabelArea.get_current_report
	@hreport = @hsh_reports[@current_report]
	prepare_xtable__by unless @hreport.ltable
	
	#Update the table with current units
	ltable = @hreport.ltable
	@hreport.lst_areas.each_with_index do |a, i|
		next unless a
		ipos, id, area, color = a
		ltable[i][ipos] = build_text_area id, area, color
	end
	
	#return the ltable
	@hreport.ltable
end

#Options for the Xtable
def option_xtable
	#Specification for columns and headers
	combo = HTML.format_combobox(@current_report, @lst_reports, "ComboReport", "ComboReport")
	h1 = []
	h1.push({ :content => combo, :style => "HElement" })
	h1.push({ :content => T7[:LABEL_NbFace], :style => "HNbFace" })
	h1.push({ :content => T7[:T_TXT_Area], :style => "HArea" })

	tit = "#{T7[:MSG_TotalArea]} (#{@top_pelt.name})"
	foot_area = build_text_area "FOOTER_AREA", @total_area, 'green'
	f1 = []
	f1.push({ :content => tit, :style => "HElement" })
	f1.push({ :content => "#{Traductor.format_number_by_3 @total_faces}", :style => "HNbFace" })
	f1.push({ :content => foot_area[0], :tip => foot_area[1], :style => "HArea" })
	
	c = []
	c.push({ :style => "Element", :width => @wid_col_name })
	c.push({ :style => "NbFace", :width => @wid_col_nbface })
	c.push({ :style => "Area", :width => @wid_col_area })

	lv0 = {}
	lv1 = { :style => "Level1", :css_style => "border-top: 2px solid steelblue" }
	lv2 = { :style => "Level2", :css_style => "border-top: 1px solid gray" }

	#Returning the Options
	hoptions = { :columns => c, :headers => h1, :footers => f1, :levels => [lv0, lv1, lv2], :body_height => "#{@hgt_table}px" }	
end

#--------------------------------------------------------------------------------------------------------------
# Specific Reports
#--------------------------------------------------------------------------------------------------------------			 

#top switch for preparing report
def prepare_xtable__by
	case @current_report
	when :by_container
		prepare_xtable__by_container
	when :by_layer
		prepare_xtable__by_layer
	else
		prepare_xtable__by_material	
	end
end

#Prepare the Xtable for the report By Material
def prepare_xtable__by_material
	@hreport.lst_areas = lst_areas = []
	@hreport.ltable = ltable = []
	
	[:mat, :back_mat].each do |key|
		front = (key == :mat)
		ipos = (front) ? 0 : 1
		text_key = @hsh_texts[key]
		
		#Total Area
		ltable.push [1, [text_key], "#{Traductor.format_number_by_3 @total_faces}"]
		lst_areas.push [3, "#{key}_T", @total_area, 'navy']
		
		#Areas by materials
		list_materials[ipos].each_with_index do |pmat, i|
			lmat = build_text_material pmat, front
			ltable.push [2, lmat, "#{Traductor.format_number_by_3 pmat.nb_faces[ipos]}"]
			lst_areas.push [3, "#{key}_#{i}", pmat.areas[ipos], 'gray']
		end	
	end
	
	ltable
end

#Prepare the Xtable for the report By Container
def prepare_xtable__by_container
	@hreport.lst_areas = lst_areas = []
	@hreport.ltable = ltable = []
	
	lst_c = @processor.get_component_instances.sort { |a, b| a.name <=> b.name }
	lst_g = @processor.get_groups.sort { |a, b| a.name <=> b.name }
	
	fill_pelt ltable, lst_areas, 1, @top_pelt unless @top_pelt == lst_c[0] || @top_pelt == lst_g[0]
	(lst_c + lst_g).each { |pelt| fill_pelt ltable, lst_areas, 1, pelt }
	
	ltable
end

#Prepare the Xtable for the report By Layer
def prepare_xtable__by_layer
	@hreport.lst_areas = lst_areas = []
	@hreport.ltable = ltable = []

	@lst_layers = @hsh_layers.values.sort { |a, b| a.name <=> b.name }
	a0 = @lst_layers.find { |a| a.name == "Layer0" }
	if a0
		@lst_layers.delete a0
		@lst_layers = [a0] + @lst_layers
	end	
	
	@lst_layers.each_with_index do |player, i|
		ltable.push [1, player.name, "#{Traductor.format_number_by_3 player.nb_faces}"]
		lst_areas.push [3, "Layer_#{i}", player.area, 'blue']
	end	
	
	ltable
end

#--------------------------------------------------------------------------------------------------------------
# Utilities for formatting the reports
#--------------------------------------------------------------------------------------------------------------			 

#Compute the list of materials and back_materials
def list_materials
	return @lst_pmat if @lst_pmat
	@lst_pmat = []
	[0, 1].each do |ipos|
		lst_pmat = @hsh_materials.values.find_all { |pmat| pmat.nb_faces[ipos] != 0 }	
		lst_pmat.sort! { |a, b| a.name <=> b.name }
		@lst_pmat.push lst_pmat
	end
	@lst_pmat
end

#Create the ltable entries for the element
def fill_pelt(ltable, lst_areas, level, pelt)
	data = pelt.data
	area = data.face_area
	other_area = data.other_area
	tot_area = area + other_area
	nb_faces = data.nb_faces
	other_nb_faces = data.other_nb_faces
	tot_nb_faces = nb_faces + other_nb_faces
	return if tot_nb_faces == 0
	
	if pelt.su_obj.instance_of?(Sketchup::Group)
		color = 'olive'
		tip = @tip_group
	elsif pelt.su_obj.instance_of?(Sketchup::ComponentInstance)
		color = 'purple'
		tip = @tip_inst
	else
		color = 'crimson'
		tip = nil
	end
	
	hstyle = { :style => "color: #{color}" }
	ltable.push [-level, [pelt.name, tip, hstyle], ["#{Traductor.format_number_by_3 tot_nb_faces}", nil, hstyle]]
	lst_areas.push [3, "#{pelt.key}_T", tot_area, color]
	
	if nb_faces > 0
		ltable.push [level+1, [@txt_toplevel_geometry], "#{Traductor.format_number_by_3 nb_faces}"]
		lst_areas.push [3, "#{pelt.key}_G", area, 'seagreen']
	end	
	if other_nb_faces > 0
		ltable.push [level+1, [@txt_embedded], "#{Traductor.format_number_by_3 other_nb_faces}"]
		lst_areas.push [3, "#{pelt.key}_O", other_area, 'gray']
	end	
end

#Build the HTML text for an area
def build_text_area(id, area, color=nil)
	color = 'navy' unless color
	tiparea = "#{@uac.value_area(area)} #{@units}"
	harea = "<span id='#{id}' style='color: #{color}'>#{@uac.format_area(area)}</span>"
	[harea, tiparea]
end

#Build the HTML text for a material or back_material
def build_text_material(pmat, front=true)
	mat = pmat.su_mat
	face = pmat.face
	
	if mat
		color = mat.color
		name = mat.name.strip
		name = $1 + $2 if name =~ /\[(.+)\](.*)/
		fname = name
		if name =~ /<(.+)>(.*)/
			name = "&lt;#{$1}&gt;#{$2}" if name =~ /<(.+)>(.*)/
			fname = $1 + $2
		end	
		mpath = File.join @tmpdir, "RPA6_Mat#{fname}.png"
		unless pmat.png_file
			if mat.texture && face
				@tw.load face, front
				@tw.write face, front, mpath
			else
				begin
					mat.write_thumbnail mpath, 256		#Only for SU8 M1+
				rescue
					mpath = nil
				end
			end	
			pmat.png_file = mpath
		end	
		case mat.materialType
		when 1
			tip = 'textured'
		when 2
			tip = 'colorized textured'
1		else
			tip = 'solid'
		end
		tip += " - alpha = #{mat.alpha}" unless mat.alpha == 1.0

	else
		name = pmat.name
		tip = name
		mpath = Traductor::MYPLUGIN.picture_get "Thumbnail_Default_Mat.png"
	end
	hcolor = HTML.color color
	
	#Creating the HTML text for image or pure color
	if mpath
		hw = "width='20' height='14' style='border: 1px solid gray'"
		action = "onmouseover='show_image(this)' onmouseout='hide_image()'"
		img = "<img src='#{HTML.image_file mpath}' #{hw} align='middle' #{action}/>"
	else
		action = "onmouseover='show_image_solid(this)' onmouseout='hide_image_solid()'"
		img = "<span style='background-color: #{hcolor} ; cursor: default' #{action}>#{@space_img}</span>"
	end	

	#Building the text
	["#{img}#{@space4}#{name}", tip]
end

#--------------------------------------------------------------------------------------------------------------
# Export section
#--------------------------------------------------------------------------------------------------------------			 

#Export the file as txt
def export_as_csv
	#Default file name
	path = Sketchup.active_model.path
	rootname = (path) ? File.basename(path, "skp") : "Untitled"
	rootname += "_Areas_#{Time.now.strftime "%d-%b-%Y %Hh%Mm%Ss"}.csv"
	@current_unit = @uac.current_name
	
	#Asking for the export file
	dir = Sketchup.active_model.path
	if dir.empty?
		dir = nil
	else	
		dir = File.dirname dir
	end	
	fpath = UI.savepanel T7[:T_BUTTON_ExportCSV], dir, rootname
	return unless fpath
	
	#Formatting the export file
	fpath = fpath.gsub(/\\/, '/')
	begin
		File.open(fpath, "w") do |f|
			mpath = @model.path
			mpath = "Untitled" unless mpath && mpath.length > 0
			f.puts "#{T7[:T_TXT_Model]},#{mpath}"
			f.puts "#{T7[:T_TXT_Date]},#{Time.now.strftime "%d-%b-%Y %Hh%M:%S"}"
			f.puts " "
			export__by f
		end
		status = UI.messagebox T7[:MSG_GenerationExport, fpath] + "\n\n" + T7[:T_STR_DoYouWantOpenFile], MB_YESNO
		Traductor.openURL fpath if status == 6
	rescue
		UI.messagebox T7[:MSG_GenerationExport_Error, fpath]
	end
end

#Top switch for preparing report
def export__by(f)
	case @current_report
	when :by_container
		export__by_container f
	when :by_layer
		export__by_layer f
	else
		export__by_material f	
	end
end

#Contribution to export by Materials
def export__by_material(f)
	[0, 1].each do |ipos|
		f.puts ""
		text_key = @hsh_texts[[:mat, :back_mat][ipos]]
		u = @current_unit.unpack("U*").pack("C*")
		f.puts "#{text_key},,#{@uac.value_area @total_area},#{u}"
		@lst_pmat[ipos].each_with_index do |pmat, i|
			name = pmat.name.gsub ',', ' '
			f.puts ",#{name},#{@uac.value_area pmat.areas[ipos]},#{u}"
		end	
	end	
end

#Contribution to export by Container
def export__by_container(f)
	lst_c = @processor.get_component_instances.sort { |a, b| a.name <=> b.name }
	lst_g = @processor.get_groups.sort { |a, b| a.name <=> b.name }
	u = @current_unit.unpack("U*").pack("C*")
	
	lbig = [[:comp_inst, lst_c], [:group, lst_g]]
	lbig = [[:top, [@top_pelt]]] + lbig unless @top_pelt == lst_c[0] || @top_pelt == lst_g[0]
	
	lbig.each do |a|
		key, lst = a
		if lst.length > 0
			f.puts @hsh_texts[key]
			lst.each do |pelt| 
				name = pelt.name.gsub ',', ' '
				data = pelt.data
				f.puts ",#{name},,#{data.tot_nb_faces},#{@uac.value_area data.tot_area},#{u}"
				f.puts ",,#{@txt_toplevel_geometry},#{data.nb_faces},#{@uac.value_area data.face_area},#{u}" if data.nb_faces > 0
				f.puts ",,#{@txt_embedded},#{data.other_nb_faces},#{@uac.value_area data.other_area},#{u}" if data.other_nb_faces > 0
			end	
		end
	end	
end

#Contribution to export by Materials
def export__by_layer(f)
	f.puts ""
	text_key = @hsh_texts[:layer]
	u = @current_unit.unpack("U*").pack("C*")
	f.puts "#{text_key},,#{@total_faces},#{@uac.value_area @total_area},#{u}"
	@lst_layers.each do |player|
		name = player.name.gsub ',', ' '
		f.puts ",#{name},#{player.nb_faces},#{@uac.value_area player.area},#{u}"
	end	
end

end	#Class ReportAreaTopDialog

#========================================================================================
#========================================================================================
# Class LabelTool: Interactive tool to show area and put labels
#========================================================================================
#========================================================================================

class LabelTool

DragInfo = Struct.new :text, :origin, :area, :entities, :tr, :layer
MoveInfo = Struct.new :state, :ptvec, :vec2d, :area

@@dico_label = "FredoTools_ReportLabelArea"
@@attr_label = "Area"
@@attr_label_id = "Companion"
@@cur_layer = "Layer0"
@@hgt_y_mac = (RUN_ON_MAC) ? 4 : 0

@@single_face = nil
@@recurse = nil
@@multi_arrow = nil

def initialize
	#Unit Controller
	@uac = UnitAreaController.new *ReportLabelArea.get_units
	prepare_layers
	@square = @uac.square_character
	
	#Cursors
	@id_cursor_exit = Traductor.create_cursor "Cursor_Arrow_Exit"
	@id_cursor_face = Traductor.create_cursor "Cursor_Arrow_Face"	
	@id_cursor_face_single = Traductor.create_cursor "Cursor_Arrow_Face_Single"	
	@id_cursor_container = Traductor.create_cursor "Cursor_Arrow_Container"	
	@id_cursor_label = Traductor.create_cursor "Cursor_Arrow_Label"	
	@id_cursor_move = Traductor.create_cursor "Cursor_Move_24"	

	#Colors
	@capa_colored_face = (Sketchup.version >= "8.0")
	@color_dragging = 'yellow'
	@color_draggingR = 'orange'
	@color_anchor = 'gold'
	@precision_anchor = 8
	@time_long_click = 0.6
	@show_boxlines = MYDEFPARAM[:DEFAULT_ReportArea_ShowBoxlines]
	
	@color_selection_face = 'dodgerblue'
	@color_selection_frame = 'lightblue'
	@color_selection_box = 'blue'
	
	@color_picked_face_add = 'seagreen'
	@color_picked_frame_add = 'lightgreen'
	@color_picked_box_add = 'green'
	
	@color_picked_face_remove = 'hotpink'
	@color_picked_frame_remove = 'pink'
	@color_picked_box_remove = 'red'

	#Tooltips
	@tip_add_selection = T7[:TIP_AddSelection]
	@tip_remove_selection = T7[:TIP_RemoveSelection]
	@tip_face = T7[:T_TXT_Face]
	@tip_faces = T7[:T_TXT_Faces]	
	@tip_exit = T7[:TIP_Exit]
	@tip_reset = T7[:TIP_Reset]
	@tip_drag = T7[:TIP_Drag]
	@tip_replace_label = T7[:TIP_ReplaceLabel]
	@tip_edit_label = T7[:TIP_EditLabel]
	@tip_move_label = T7[:TIP_MoveLabel]
	@tip_move_anchor = T7[:TIP_MoveAnchor]
	@tip_layer = T6[:T_TXT_LAYER]
	
	#text for operations
	@ops_label = "LabelArea"
	@ops_label_all = T7[:OPS_LabelAll]
	@ops_label_erase = T7[:OPS_LabelErase]
	@ops_label_update = T7[:OPS_LabelUpdate]
	@ops_label_create = T7[:OPS_LabelCreate]
	@ops_label_move = T7[:OPS_LabelMove]
	
	#Loading persistent parameters
	persistence_load
	
	#Dynamic environment
	@tr_id = Geom::Transformation.new
	@hsh_elt_info = {}
	reset_env
end

#Reset the environment
def reset_env
	@tr = nil
	@parent = nil
	@elt_label = nil
	@picked_faces = nil
	@total_selection = []
	@hsh_selected = {}
	@already = false
	@key_picked = nil
	@key_previous = nil
	@single_previous = !@single_face
	@recurse_previous = !@recurse
	@current_area = 0
	@total_area = 0
	@dragging = nil
	@label_to_move = 0
	@anchor_to_move = 0
	@id_cursor = @id_cursor_exit
end

#Clear the current selection
def clear_selection
	@hsh_selected.each { |key, info| info.selected = false }
	reset_env
	after_pick
end

#Evaluate the initial selection
def initial_selection
	return
	@selection.each do |e|
		if e.instance_of?(Sketchup::Face)
			@current_selection.push [[e], @model, @tr_id] 
		elsif e.instance_of?(Sketchup::Group) || e.instance_of?(Sketchup::ComponentInstance)
			@current_selection.push [e, e, e.transformation] 
		end	
	end
	@total_area = calculate_area_container @selection, @model, @tr_id
	@txt_current_area = @uac.format_area @current_area
end

#Exit the tool
def exit_tool
	@model.select_tool nil
end

#List for Combo for layers
def prepare_layers
	@layer_same = "Same Layer"
	@layer_area = "Layer for Area Labels"
	layers = Sketchup.active_model.layers.to_a.collect { |layer| layer.name }
	layers.delete "Layer0"
	layers.delete @layer_area
	@layers = [@layer_same, @layer_area, "Layer0"] + layers
	@@cur_layer = nil unless layers.include?(@@cur_layer) 
	@@cur_layer = @layer_same unless @@cur_layer
end

def layers() ; @layers ; end
def get_cur_layer() ; @@cur_layer ; end
def set_cur_layer(layer_name) ; @@cur_layer = layer_name ; end

#---------------------------------------------------------------------------------------------
# Recurse and Single Face parameters
#---------------------------------------------------------------------------------------------

#Load parameters from persistence
def persistence_load
	@recurse = (@@recurse == nil) ? MYDEFPARAM[:DEFAULT_ReportArea_Recurse] : @@recurse
	@single_face = (@@single_face == nil) ? MYDEFPARAM[:DEFAULT_ReportArea_SingleFace] : @@single_face
	@multi_arrow = (@@multi_arrow == nil) ? MYDEFPARAM[:DEFAULT_ReportArea_MultiArrow] : @@multi_arrow
end

#Save parameters to persistence
def persistence_save
	@@recurse = @recurse
	@@single_face = @single_face
	@@multi_arrow = @multi_arrow
end

#---------------------------------------------------------------------------------------------
# Recurse and Single Face parameters
#---------------------------------------------------------------------------------------------

def recurse? ; @recurse ; end
def toggle_recurse ; @recurse = !@recurse ; @dialog.update_recurse(@recurse) ; end

def single_face? ; @single_face ; end
def toggle_single_face ; @single_face = !@single_face unless @button_down ; @dialog.update_single_face(@single_face) ; end

def multi_arrow? ; @multi_arrow ; end
def toggle_multi_arrow ; @multi_arrow = !@multi_arrow ; @dialog.update_multi_arrow(@multi_arrow) ; end

#---------------------------------------------------------------------------------------------
# Specific methods
#---------------------------------------------------------------------------------------------

#Manage the click down
def handle_click_down
	if @elt_label
		if @label_anchor
			@anchor_to_move = 1
		elsif !@elt_empty	
			@label_to_move = 1
		end	
	elsif @picked_faces || @picked_container
		@drag_origin = compute_dragging_origin
	else
		#clear_selection
	end	
end

#Compute the plane of a face after transformation
def face_transformed_plane(face, tr)
	pt = face.vertices[0].position
	pt2 = pt.offset face.normal, 1
	pt_t = tr * pt
	pt2_t = tr * pt2
	[pt_t, pt_t.vector_to(pt2_t)]
end

#Manage the click up: put pick elements in selection
def handle_click_up(flags)
	#Finish move for label
	if @elt_label
		if @anchor_to_move == 2
			finish_anchor_move
		elsif @label_to_move == 2
			finish_label_move
		else	
			update_label_text
		end	
		@label_to_move = 0
		@anchor_to_move = 0
		return	
	end	
	
	#Finish dragging for creating label
	if @dragging
		finish_dragging
		return
	end
	
	#Test for Long click
	if Time.now.to_f - @time_click_down > @time_long_click
		clear_selection unless @picked_faces || @picked_container
		return
	end	
	
	#Put faces or container in the selection and test for long click in empty space
	if @picked_faces || @picked_container
		merge_with_selection
		after_pick
	end	
end
	
#Button click - On a face, label the face, otherwise end the selection
def hanleDoubleClick(flags, x, y, view)
	if @elt_label
		edition_label
	elsif @picked_faces
		merge_with_selection
		@view.invalidate
		create_label_no_arrow	
	else	
		exit_tool
	end	
end
	
#Update context and display after a pick	
def after_pick
	compute_current_name
	compute_tooltip_cursor
	update_dialog
	@view.invalidate	
end
	
#Compute the tooltip for the view	
def compute_tooltip_cursor
	txsel = (@already) ? @tip_remove_selection : @tip_add_selection
	tip_extra = ""
	if @elt_label
		if @label_anchor
			tip = @tip_move_anchor
		else
			tip_extra = "#{@tip_layer}: #{@elt_label.layer.name}\n"
			tip = @tip_replace_label + "\n" + @tip_edit_label + "\n" + @tip_move_label
		end	
		@id_cursor = @id_cursor_label
	elsif @picked_faces || @picked_container
		tip = "#{@current_name} ........ #{@txt_current_area}" + "\n" + txsel + "\n" + @tip_drag
		@id_cursor = (@picked_faces) ? ((@single_face) ? @id_cursor_face_single : @id_cursor_face) : @id_cursor_container
	else
		tip = @tip_reset + "\n" + @tip_exit
		@id_cursor = @id_cursor_exit
	end
	@view.tooltip = tip_extra + tip
	Sketchup.set_status_text tip.sub("\n", "   --  ")
end
	
#Compute the  name of the current picked element	
def compute_current_name
	if @picked_faces
		case (n = @picked_faces.length)
		when 1
			@current_name = "1 #{@tip_face}"
		else
			@current_name = "#{n} #{@tip_faces}"
		end
	elsif @picked_container
		@current_name = G6.grouponent_name(@picked_container)
	else
		@current_name = ""
	end	
end

#Manage actions on a mouse move
def handle_move(flags, x, y, view)
	if @button_down
		if (@xdown - x).abs > 4 || (@ydown - y).abs > 4
			if @anchor_to_move > 0
				start_anchor_move if @anchor_to_move == 1
			elsif @label_to_move > 0
				start_label_move if @label_to_move == 1
			elsif !@dragging	
				start_dragging unless @elt_label
			end	
		end	
		process_anchor_move(view, x, y)
		process_label_move(view, x, y)
		@view.invalidate
	else
		pick_current_element(flags, x, y, view)
		compute_tooltip_cursor
		update_dialog
		@view.invalidate	
	end
end

#Interactive pick of label, faces or container
def pick_current_element(flags, x, y, view)
	return if @button_down
	what_is_under_mouse(flags, x, y, view)
	elt = [@elt_label, @picked_faces, @picked_container].find { |e| e }
	@key_picked = compute_key elt, @tr
	@picked_new = (elt) ? [elt, @parent, @tr] : nil
	return if @key_previous == @key_picked && (@picked_new == nil || @elt_label || (@picked_faces && @single_previous == @single_face) || (@picked_container && @recurse_previous == @recurse))
		
	@selection.clear
	if @elt_label
		@selection.add @elt_label 
		find_label_id_companions
		@elt_label_new_text, @elt_label_new_area = calculate_new_label_text(@elt_main_label)
		
	elsif @picked_new
		evaluate_area
		@txt_current_area = @uac.format_area @current_area
		pick_part_of_selection
		@frozen = false
	else
		@current_area = 0
		@txt_current_area = ""
	end	
	
	@key_previous = @key_picked
	@single_previous = @single_face
	@recurse_previous = @recurse
	after_pick
end
	
#Identify which objects are under the mouse
def what_is_under_mouse(flags, x, y, view)
	ph = view.pick_helper
	ph.do_pick x, y, 0
	
	@elt_label = nil
	@label_anchor = nil
	@picked_faces = nil
	@picked_container = nil
	elt = nil
	edge = ph.picked_edge
	elt = ph.picked_element
	
	#Label
	if elt.class == Sketchup::Text
		@elt_label = elt = ph.picked_element
		@elt_label_id = get_label_id @elt_label

	#Edge	
	elsif edge && G6.edge_plain?(edge) && edge.parent != @model
		@picked_elt = elt = edge

	#Face	
	elsif ph.picked_face
		@picked_elt = elt = ph.picked_face
		@picked_faces = (@single_face) ? [elt] : G6.face_neighbours(elt) 
		@picked_faces.sort! { |a, b| a.entityID <=> b.entityID }
	end
	
	@tr = @tr_id
	@parent = nil
	return unless elt
	
	#Finding the parent and transformation
	for i in 0..ph.count
		ls = ph.path_at(i)
		if ls && ls.include?(elt)
			@parent = ls[-2]
			@tr = ph.transformation_at(i)
			break
		end
	end	
	
	#Finding if there is an anchor point selected
	if @elt_label && @elt_label.vector && @elt_label.vector.valid?
		ptxy = @view.screen_coords @tr * @elt_label.point
		@label_anchor = ptxy if (ptxy.x - x).abs < @precision_anchor &&  (ptxy.y - y).abs < @precision_anchor
	end
	
	@picked_container = @parent unless @elt_label || @picked_faces
end

#---------------------------------------------------------------------------------------
# Manage Selection
#---------------------------------------------------------------------------------------

#Check the status of the current picked elements
def pick_part_of_selection
	infoc = @hsh_elt_info[@key_picked]
	return unless infoc
	lst = (@recurse) ? infoc.lst_contentR : infoc.lst_content
	@instructions_picked = instructions_compute lst
	if lst.find { |e| !e.selected }
		@already = nil
	else
		@already = infoc
	end	
end
	
#Build the initial selection	
def build_initial_selection
	#Handling the faces at top level of active model
	faces = @initial_selection.find_all { |e| e.instance_of?(Sketchup::Face) }	
	llfaces = split_by_contiguous_faces(faces)
	llfaces.each do |lfaces|
		@tr = @tr_id
		@key_picked = compute_key lfaces, @tr
		@picked_new = [lfaces, nil, @tr]
		@picked_faces = lfaces
		evaluate_area
		merge_with_selection
	end
	@picked_faces = @picked_new = nil
	
	#Adding groups and components at top level
	lcontainers = @initial_selection.find_all { |e| e.instance_of?(Sketchup::Group) || e.instance_of?(Sketchup::ComponentInstance) }
	lcontainers.each do |container|
		@tr = container.transformation
		@key_picked = compute_key container, @tr
		@picked_new = [container, nil, @tr]
		@picked_container = container
		evaluate_area
		merge_with_selection
	end
	@picked_container = @picked_new = nil
end
	
#Split a list of faces in contiguous groups	
def split_by_contiguous_faces(faces)
	return faces if faces.length < 2
	
	hsh_faces = {}
	faces.each { |f| hsh_faces[f.entityID] = f }
	
	groups = []
	lfaces = faces.clone
	while lfaces.length > 0
		face = lfaces.shift
		ls = face.all_connected.find_all { |f| f.instance_of?(Sketchup::Face) && hsh_faces[f.entityID] }
		groups.push([face] + ls)
		lfaces -= ls
	end	
	groups
end
	
#Integrate the current picked element to the total selection	
def merge_with_selection
	return unless @picked_new
	pick_part_of_selection
	if @already
		remove_from_selection
	else
		add_to_selection
	end	
	recompute_total_selection
	@frozen = true
end
		
#Add the current picked element to the selection	
def add_to_selection
	infoc = @hsh_elt_info[@key_picked]
	return unless infoc
	lst = (@recurse) ? infoc.lst_contentR : infoc.lst_content
	infoc.recursed = @recurse
	lst.each do |info|
		info.absorbed = true
		next if info.selected
		info.selected = true
		@hsh_selected[info.key] = info
	end
	unless infoc.elt.class == Array
		infoc.selected = true
		infoc.absorbed = true
		@hsh_selected[infoc.key] = infoc
	end	
	
	#Adjusting the total selection to remove redundant elements
	@total_selection.clone.each do |info|
		lst = info.lst_content
		n = 0
		lst.each do |a|
			n += 1 if a.absorbed
			a.absorbed = false
		end	
		@total_selection.delete info if n == lst.length
	end
	@hsh_selected.values.each { |a| a.absorbed = false }
	
	@total_selection.push infoc
end

#Remove the picked element from the selection
def remove_from_selection
	#Removing the elements and its content
	infoc = @hsh_elt_info[@key_picked]
	return unless infoc
	lst = (@recurse) ? infoc.lst_contentR : infoc.lst_content
	lst.each do |info|
		next unless info.selected
		info.selected = false
		@hsh_selected.delete info.key
	end
	unless @picked_new.class == Array
		infoc.selected = false
		@hsh_selected.delete infoc.key
	end
	
	#Removing the empty containers
	lscontainer = @hsh_selected.values.find_all { |e| e.is_container }
	lscontainer.each do |info|
		unless info.lst_content.find { |e| e.selected }
			info.selected = false
			@hsh_selected.delete info.key
		end	
	end
	
	@total_selection.delete_if { |a| a == infoc }
end

#Compute the total selection - Calculate the drawing elements
def recompute_total_selection
	triangles = []
	frames = []
	boxlines = []
	@total_area = 0
	@hsh_selected.delete_if { |key, info| !info.selected }
	@hsh_selected.each do |key, info|
		if info.is_face
			triangles.concat info.triangles
			frames.concat info.frames
			@total_area += info.area
		else	
			boxlines.concat info.boxlines if info.boxlines
		end	
	end
	@instructions_selection = [triangles, frames, boxlines]
end

#---------------------------------------------------------------------------------------
# Substitution for All labels
#---------------------------------------------------------------------------------------

#Sunstitute units for all labels
def substitute_all_labels
	entities = (@initial_selection.empty?) ? @model.entities : @initial_selection
	hcomp = {}
	ls_labels = substitute_all_labels_explore entities, {}, []
	return if ls_labels.empty?
	
	#Replacing the labels
	@model.start_operation @ops_label_all
	ls_labels.each do |e|
		new_text, area = calculate_new_label_text e
		stamp_label_text e, area if area
		e.text = new_text if e.text != new_text
	end	
	@model.commit_operation	
end

#Compute the list of all labels
def substitute_all_labels_explore(entities, hcomp, ls_labels)	
	entities.each do |e|
		if e.instance_of?(Sketchup::Text)
			ls_labels.push e if e.text && e.text.strip.length > 0
		elsif e.instance_of?(Sketchup::Group)
			substitute_all_labels_explore e.entities, hcomp, ls_labels
		elsif e.instance_of?(Sketchup::ComponentInstance)
			cdef = e.definition
			unless hcomp[cdef.entityID]
				hcomp[cdef.entityID] = true
				substitute_all_labels_explore cdef.entities, hcomp, ls_labels
			end	
		end
	end	
	ls_labels
end

#Erase the current label
def erase_label
	ls_label = []
	ls_label.push @elt_label if @elt_label.valid?
	
	#Finding Companion labels
	ls_label += @lst_id_companions if @elt_label == @elt_main_label
	
	#Erasing all labels
	return if ls_label.empty?
	@model.start_operation @ops_label_erase
	ls_label.each { |sulabel| sulabel.erase! if sulabel.valid?}
	@model.commit_operation
end

#---------------------------------------------------------------------------------------
# Moving an existing Label
#---------------------------------------------------------------------------------------

#begin the move of a label
def start_label_move
	return unless @elt_label && @elt_label.point
	@label_to_move = 2
	pt_arrow = @elt_label.point.offset @elt_label.vector
	pt_arrow2d = @view.screen_coords pt_arrow
	ptdown = Geom::Point3d.new @xdown, @ydown, 0
	@vec_label_move = ptdown.vector_to pt_arrow2d
	ray = @view.pickray @xdown, @ydown
	@plane_label_move = [pt_arrow, ray[1]]
	@model.start_operation @ops_label_move
end

#Move the label interactively with the mouse
def process_label_move(view, x, y)
	return unless @label_to_move == 2
	
	#Labels with leader
	if @elt_label.vector && @elt_label.vector.valid?
		ptxy = Geom::Point3d.new x, y, 0
		pt_arrow2d = (@vec_label_move.valid?) ? ptxy.offset(@vec_label_move) : ptxy
		ray = @view.pickray pt_arrow2d.x, pt_arrow2d.y
		pt = Geom.intersect_line_plane ray, @plane_label_move
		@elt_label.vector = @elt_label.point.vector_to pt
		
		@lst_id_companions.each do |sulabel|
			sulabel.vector = sulabel.point.vector_to pt
		end
	
	#Label with NO leader
	else
		ptxy = Geom::Point3d.new x, y, 0
		pt_arrow2d = (@vec_label_move.valid?) ? ptxy.offset(@vec_label_move) : ptxy
		pt = compute_dragging_origin pt_arrow2d.x, pt_arrow2d.y
		@elt_label.point = G6.small_offset view, pt, 10
	end
end

#Finish the move of a label and commit
def finish_label_move
	@model.commit_operation
end

#Compute the moving companions on the label being moved
def find_label_id_companions
	if @elt_label_id
		entities = G6.entities_from_parent @parent
		@lst_id_companions = entities.find_all { |e| e.instance_of?(Sketchup::Text) && get_label_id(e) == @elt_label_id }
		main_label = @lst_id_companions.find { |e| e.text && e.text.strip != "" }
		@elt_main_label = (main_label) ? main_label : @elt_label
		@lst_id_companions.delete @elt_label
	else
		@lst_id_companions = []
		@elt_main_label = @elt_label
	end	
end

#---------------------------------------------------------------------------------------
# Moving an existing Anchor
#---------------------------------------------------------------------------------------

#begin the move of a label
def start_anchor_move
	@anchor_to_move = 2
	@model.start_operation @ops_label_move
end

#Move the label interactively with the mouse
def process_anchor_move(view, x, y)
	return unless @anchor_to_move == 2
	pivot = @elt_label.point.offset @elt_label.vector
	@ipdown.pick view, x, y
	pt = @ipdown.position
	@elt_label.point = @tr.inverse * pt
	@elt_label.vector = @elt_label.point.vector_to(pivot)
	@label_anchor = view.screen_coords pt
end

#Finish the move of a label and commit
def finish_anchor_move
	@model.commit_operation
end

#---------------------------------------------------------------------------------------------
# Contextual Menu
#---------------------------------------------------------------------------------------------

#Entries for the contextual menu
def getMenu(menu)
	sepa = false
	if @elt_label
		menu.add_item(T7[:MNU_EditLabel]) { edition_label }
		menu.add_item(T7[:MNU_UpdateLabel]) { update_label_text }
		menu.add_item(T7[:MNU_EraseLabel]) { erase_label }
		sepa = true
	elsif @picked_new
		pick_part_of_selection
		if @already
			menu.add_item(T7[:MNU_RemoveSelection]) { merge_with_selection }
			sepa = true
		else	
			menu.add_item(T7[:MNU_AddSelection]) { merge_with_selection }
			sepa = true
		end	
	end
	
	#Clear Selection
	if @total_selection.length > 0
		menu.add_separator if sepa
		menu.add_item(T7[:MNU_ClearSelection]) { clear_selection }
		sepa = true
	end
	
	#Options
	menu.add_separator if sepa
	menu.add_item(menu_onoff_text(T7[:BUTTON_Recurse], @recurse)) { toggle_recurse }
	menu.add_item(menu_onoff_text(T7[:BUTTON_SingleFace], @single_face)) { toggle_single_face }
	menu.add_item(menu_onoff_text(T7[:BUTTON_MultiArrow], @multi_arrow)) { toggle_multi_arrow }
	
	#Exit Tool
	menu.add_separator
	menu.add_item(T6[:T_STR_ExitTool]) { exit_tool }
end

def menu_onoff_text(text, flag)
	onoff = (flag) ? T6[:T_MNU_On] : T6[:T_MNU_Off]
	text + " --> #{onoff}"
end
	
#---------------------------------------------------------------------------------------------
# Dragging during the creation of a label
#---------------------------------------------------------------------------------------------

#Start the dragging process to create a label
def start_dragging
	pick_part_of_selection
	if @already 
		area = @total_area
		return if area == 0
		@lst_companions = compute_dragging_companions
	else
		area = @current_area 
		@lst_companions = []
	end	
	return if area == 0
	text = @uac.format_area area
	@dragging = DragInfo.new
	@dragging.origin = @drag_origin
	@dragging.tr = @tr.inverse
	@dragging.entities = G6.grouponent_entities @parent
	if @picked_container
		@dragging.text = "#{@current_name}  #{text}"
		@dragging.layer = @picked_container.layer
	else
		@dragging.text = text
		@dragging.layer = @picked_faces[0].layer
	end
	@dragging.area = area
end

#Get the content of an element
def info_get_content(info)
	(info.recursed) ? info.lst_contentR : info.lst_content
end

#check if the current selected element is part of another
def part_of_selected_element?(info1, infot)
	lst1 = info_get_content(info1)
	lstt = info_get_content(infot)
	lst_inter = lst1 & lstt
	(lst_inter.length > 0)
end

#Compute the dragging Companions
def compute_dragging_companions
	#finding the companions
	lst_companion = []
	@total_selection.each do |a|
		lst_companion.push a unless a == @already || part_of_selected_element?(@already, a)
	end
	
	#Calculating the anchor points for all faces
	lsbary = []
	lst_companion.each do |info|
		ls = []
		info_get_content(info).each do |a| 
			bary = compute_bary(a)
			ls.push bary if bary
		end	
		lsbary.push ls unless ls.empty?
	end
	
	lsbary
end

#Compute the barycenter of a face based on its outer loop
def compute_bary(info)
	return nil unless info.is_face
	return info.bary if info.bary
	face = info.elt
	tr = info.tr
	lpts = face.outer_loop.vertices.collect { |v| tr * v.position }
	info.bary = G6.curve_barycenter(lpts + [lpts[0]])
end

#Finish the dragging process and create the label	
def finish_dragging
	within = @ctrl_down || @shift_down
	create_new_label within
	@dragging = nil
	@view.invalidate
end

#Compute the origin for potential dragging
def compute_dragging_end(origin)
	eye = @view.camera.eye
	vec_orig = eye.vector_to origin
	len_orig = vec_orig.length
	a = len_orig
	
	ray = @view.pickray(@x, @y)
	vec_end = ray[1].clone
	angle_eye = vec_orig.angle_between vec_end
	
	origin_2d = @view.screen_coords origin
	ptxy = Geom::Point3d.new @x, @y
	pixels = ptxy.distance origin_2d
	ecart = @view.pixels_to_model pixels, origin
	pttarg = eye.offset vec_end, len_orig
	ecart2 = @view.pixels_to_model pixels, pttarg
	ecart = 0.5 * (ecart + ecart2)
	
	vec_end.length = len_orig - 0.5 * ecart * Math.cos(angle_eye)
	vec_end - vec_orig
end

#Compute the origin for potential dragging
def compute_dragging_origin(x=nil, y=nil)
	x = @xdown unless x
	y = @ydown unless y
	
	ray = @view.pickray(x, y)
	origin = nil
	
	#Try with view ray intersection
	if @picked_elt.class == Sketchup::Face
		origin = Geom.intersect_line_plane ray, face_transformed_plane(@picked_elt, @tr)
	elsif @picked_elt.class == Sketchup::Edge
		ls = Geom.intersect_line_line ray, [@picked_elt.start, @picked_elt.end].collect { |v| @tr * v.position }
		origin = ls[1] if ls
	end
	
	#Try with model raytes otherwise
	unless origin
		ll = @model.raytest ray
		@drag_origin = ll[0] if ll
	end	
	
	#Finally, rely on input point
	unless origin
		@ipdown.pick @view, x, y
		origin = @ipdown.position
	end	
	origin
end
	
#---------------------------------------------------------------------------------------------
# Creation and Update of Labels
#---------------------------------------------------------------------------------------------
	
#Create a Label object. Area is store as an attribute
def create_new_label(within=false)
	@ipdown.pick @view, @x, @y
	vector = compute_dragging_end @drag_origin
	
	@model.start_operation @ops_label_create
	
	#Level of creation
	if within
		ee = @dragging.entities
		tr = @dragging.tr
	else		
		ee = @model.active_entities
		tr = @tr_id
	end	
	
	#Creating the main label
	origin = tr * @dragging.origin
	vector = tr * vector
	sulabel = ee.add_text @dragging.text, origin, vector
	stamp_label_text sulabel, @dragging.area
	set_label_layer sulabel, @dragging.layer
	id = stamp_label_id sulabel
	
	#Creating the companion label if any
	if @multi_arrow
		pt_end = origin.offset vector
		@lst_closest_bary.each do |pt|
			orig = tr * pt
			vec = orig.vector_to pt_end
			sulabel = ee.add_text "", orig, vec	
			set_label_layer sulabel, @dragging.layer
			stamp_label_id sulabel, id
		end
	end
	
	@model.commit_operation
end

#Create a Label object. Area is store as an attribute
def create_label_no_arrow
	within = @ctrl_down || @shift_down

	@drag_origin = compute_dragging_origin
	
	@model.start_operation @ops_label_create
	if within
		ee = G6.grouponent_entities @parent
		tr = @tr.inverse
	else		
		ee = @model.active_entities
		tr = @tr_id
	end	
	pt = G6.small_offset(@view, @drag_origin, 10)
	text = @uac.format_area @current_area
	sulabel = ee.add_text text, tr * pt
	stamp_label_text sulabel, @current_area
	set_label_layer sulabel, @picked_faces[0].layer
	@model.commit_operation
end

#Update the text of the SU Label
def update_label_text
	return unless @elt_label_new_text
	sulabel = @elt_main_label
	@model.start_operation @ops_label_update
	sulabel.text = @elt_label_new_text
	stamp_label_text sulabel, @elt_label_new_area
	@model.commit_operation
end

#Get or set the Area attributes for a Text Label SU obj 
def stamp_label_text(sulabel, area=nil)
	return nil unless sulabel
	if area
		sulabel.set_attribute @@dico_label, @@attr_label, "#{area}"
	else	
		sarea = sulabel.get_attribute @@dico_label, @@attr_label
		area = (sarea) ? sarea.to_f : nil
	end
	area
end

#Mark the label with a unique id (across SU sessions)
def stamp_label_id(sulabel, id=nil)
	id = Time.now.to_f unless id
	sulabel.set_attribute @@dico_label, @@attr_label_id, "#{id}"
	id
end

#Mark the label with a unique id (across SU sessions)
def get_label_id(sulabel)
	sulabel.get_attribute @@dico_label, @@attr_label_id
end

#Set the layer for the label
def set_label_layer(sulabel, sulayerdef=nil)
	layers = @model.layers
	if @@cur_layer == @layer_same
		layer = sulayerdef
	elsif @@cur_layer == @layer_area
		layer = layers[@layer_area]
		layer = layers.add @layer_area unless layer
	else	
		layer = layers[@@cur_layer]
	end
	layer = layers["Layer0"] unless layer
	sulabel.layer = layer
end

#Update the fields in the dialog box
def update_dialog
	if @elt_label
		@dialog.update_label @elt_main_label.text, @elt_label_new_text
	else
		txcur = @txt_current_area
		txtot = @uac.format_area(@total_area)
		@dialog.update_selection @current_name, txcur, txtot
	end
end

#---------------------------------------------------------------------------------------------
# Evaluation of area
#---------------------------------------------------------------------------------------------

EltInfo = Struct.new :key, :elt, :is_face, :is_container, :tr, :bounds, :area, :areaR, :lst_content, :lst_contentR, :selected, :recursed,
                     :absorbed, :triangles, :frames, :boxlines, :bary

#Calculate a unique key for the current picked entity
def compute_key(p, t)
	(p.class == Array) ? "#{p[0]}-#{p.length}#{t.to_a}" : "#{p}-#{t.to_a}"
end

#Return the info structure for an element. Create it if it does not exist
def elt_info(elt, tr)
	key = compute_key(elt, tr)
	info = @hsh_elt_info[key]
	return info if info
	info = EltInfo.new
	info.elt = elt
	info.is_face = (elt.instance_of?(Sketchup::Face))
	info.is_container = (elt.instance_of?(Sketchup::Group) || elt.instance_of?(Sketchup::ComponentInstance))
	info.tr = tr
	info.key = key
	info.selected = false
	@hsh_elt_info[key] = info
	info
end

#---------------------------------------------------------------------------------------------
# Evaluation of area
#---------------------------------------------------------------------------------------------

#Calculate the area of the picked elements
def evaluate_area
	if @picked_faces
		@current_area = calculate_area_faces @picked_faces, @tr
	elsif @picked_container
		@current_area = calculate_area_container @picked_container, @tr
	else
		@current_area = 0
	end
	@current_area
end

#Calculate the area of a list of faces
def calculate_area_faces(faces, tr)
	info = elt_info(faces, tr)
	return info.area if info.area
	
	area = 0
	lst = []
	faces.each do |face| 
		infof = elt_info face, tr
		a = infof.area
		a = infof.area = G6.face_area(face, tr) unless a
		area += a
		lst.push infof
	end	
	info.area = area
	info.lst_content = info.lst_contentR = lst
	area
end

def calculate_area_single_face(face, tr)
	info = elt_info face, tr
	a = info.area
	a = info.area = G6.face_area(face, tr) unless a
	a
end

#Calculate the area of a list of faces
def calculate_area_container(container, tr)
	info = recurse_container G6.grouponent_entities(container), container, tr
	(@recurse) ? info.areaR : info.area
end

#Create conatiners info recursively
def recurse_container(entities, comp, t)
	#Checking if already computed
	key = compute_key comp, t
	infoc = @hsh_elt_info[key]
	if infoc 
		areak = (@recurse) ? infoc.areaR : infoc.area
		return infoc if areak
	end	
	
	#Creating the component element
	infoc = elt_info comp, t
	area = areaR = 0
	lst_faces = []
	lst_content = []
	infoc.boxlines = container_box_lines comp, t * comp.transformation.inverse if @show_boxlines && defined?(comp.transformation)
	
	entities.each do |e|
		if e.instance_of?(Sketchup::Face)
			area = area + calculate_area_single_face(e, t)
			lst_faces.push elt_info(e,t)
		elsif e.instance_of?(Sketchup::Group) || e.instance_of?(Sketchup::ComponentInstance)
			next unless @recurse
			ents = (e.instance_of?(Sketchup::ComponentInstance)) ? e.definition.entities : e.entities
			info = recurse_container(ents, e, t * e.transformation)
			lst_content = lst_content + [info] + info.lst_contentR
			areaR += info.areaR
		end			
	end
	if @recurse
		infoc.areaR = area + areaR
		infoc.lst_contentR = lst_faces + lst_content
	end	
	infoc.area = area
	infoc.lst_content = lst_faces
	infoc
end

#---------------------------------------------------------------------------------------------
# Replacement of Label
#---------------------------------------------------------------------------------------------

#Calculate the new text to substitute to current text with current units
def calculate_new_label_text(sulabel)
	curtx = sulabel.text
	area = stamp_label_text sulabel
	tx, area = parse_label_text curtx, area
	return nil unless tx
	newtx = tx
	newtx = tx.gsub("%1", @uac.format_area(area)) if area
	[newtx, area]
end

#Parse the text of the label to locate the unit and area value
def parse_label_text(text, stamp_area=nil)
	return nil unless text && text.strip != ""
	ls = text.split @square
	s = ls[0].reverse
	txarea = "%1"
	area = stamp_area
	return [text, stamp_area] unless s =~ /\d/

	unit = $`.strip.reverse
	symb = @uac.symb_unit_from_text unit
	return [text, stamp_area] unless symb
	s = $& + $'
	if s =~ /\s*[^0-9 .,]/
		sarea = $`.strip.reverse
		rest = ($& + $').reverse
	elsif s =~ /(\s*)[0-9 .,]*/
		sarea = s.strip.reverse
		rest = $1.reverse
	else
		return [text, stamp_area]
	end
	
	unless stamp_area
		sarea = sarea.gsub(' ', '')
		area = Traductor.string_to_float(sarea)
		if area
			fact = @uac.factor_convertion symb
			area = area / fact
			txarea = @uac.format_area(area)
		end	
	end	
	text = "#{rest}#{txarea}" + ls[1..-1].join("")
	[text, area]
end

#Edit a text label
def edition_label
	sulabel = @elt_main_label
	prompts = [T7[:DLG_TextLabel] + "  ------------------>"]
	text = sulabel.text
	text = "" unless text && text.strip != ''
	results = [text]
	results = UI.inputbox prompts, results, T7[:MNU_Label]
	return unless results

	@elt_label_new_text = results[0]
	update_label_text
end

#---------------------------------------------------------------------------------------------
# Drawing methods
#---------------------------------------------------------------------------------------------

def draw_picked(view)
	#Drawing component if any
	spec_pick = (@already) ? [@color_picked_box_remove, 3, ''] : [@color_picked_box_add, 3, '']
	if @parent
		col_container = (@dragging && (@ctrl_down || @shift_down) && @parent) ? [@color_draggingR, 1, ''] : ['gray', 1, '']
		color, width, stipple = (@picked_faces || @elt_label) ? col_container : spec_pick
		view.drawing_color = color
		view.line_width = width
		view.line_stipple = stipple
		G6.draw_component view, @parent, @tr
	end
	
	#Drawing faces
	if @picked_faces || @picked_container
		if @already
			colors = [@color_picked_face_remove, @color_picked_frame_remove, @color_picked_box_remove]
		else	
			colors = [@color_picked_face_add, @color_picked_frame_add, @color_picked_box_add]
		end	
		instructions_draw view, @instructions_picked, *colors
	end	
end

#Draw the total current selection
def draw_total_selection(view)
	return if @total_selection.empty?
	colors = [@color_selection_face, @color_selection_frame, @color_selection_box]
	instructions_draw view, @instructions_selection, *colors	
end

#Compute the triangles and frames for a list of entities
def instructions_compute(lst_info, instructions=nil)
	instructions = [[], [], []] unless instructions
	triangles, frames, boxlines = instructions
	plain_only = (lst_info.length > 1)
	lst_info.each do |info|
		elt = info.elt
		tr = info.tr
		if info.is_face
			info.triangles = single_face_triangles(elt, tr) unless info.triangles
			triangles.concat info.triangles 
			if plain_only
				info.frames = single_face_frame(elt, tr, plain_only) unless info.frames
				frames.concat info.frames
			else
				info.frames = single_face_frame(elt, tr) unless info.frames
				frames.concat single_face_frame(elt, tr)
			end
		else
			boxlines.concat info.boxlines if info.boxlines
		end	
	end	
	instructions
end

#Draw faces and frame lines
def instructions_draw(view, instructions, color_face, color_frame, color_box)
	view.line_stipple = ''
	triangles, frames, boxlines = instructions	
	
	if color_face && triangles.length > 0
		view.drawing_color = color_face
		view.draw GL_TRIANGLES, triangles 
	end
	
	if color_frame && frames.length > 0
		view.drawing_color = color_frame
		view.line_width = (@capa_colored_face) ? 1 : 3
		view.draw GL_LINES, frames.collect { |pt| G6.small_offset view, pt}
	end
	
	if color_box && boxlines.length > 0
		view.drawing_color = color_box
		view.line_width = 3
		view.draw GL_LINES, boxlines.collect { |pt| G6.small_offset view, pt, 2}
	end
end

#Compute the triangles of a single face
def single_face_triangles(face, tr)
	mesh = face.mesh
	pts = mesh.points
	triangles = []
	mesh.polygons.each do |p|
		triangles += p.collect { |i| tr * pts[i.abs-1] }
	end	
	triangles
end

#Compute the line loops for framing a face
def single_face_frame(face, tr, plain_only=false)
	frames = []
	face.edges.each { |edge| frames.push(tr * edge.start.position, tr * edge.end.position) }
	frames
end

#Compute the lines for the box of a container
def container_box_lines(comp, tr)
	bb = comp.bounds
	ptsbox = [0, 1, 3, 2, 4, 5, 7, 6].collect { |i| tr * bb.corner(i) }
	boxlines = [0, 1, 1, 2, 2, 3, 3, 0, 0, 4, 4, 5, 5, 6, 6, 7, 7, 4, 1, 5, 2, 6, 3, 7].collect { |i| ptsbox[i] } 
end

#draw a mark at label anchor
def draw_label_anchor(view)
	return unless @label_anchor
	pt = view.screen_coords(@tr * @elt_label.point)
	pts = G6.pts_circle pt.x, pt.y, 5
	view.drawing_color = @color_anchor
	view.draw2d GL_POLYGON, pts
end

#Draw a text label when dragging
def draw_dragging(view)
	origin3d = @dragging.origin
	return unless origin3d
	origin = view.screen_coords origin3d
	pts = G6.pts_square origin.x, origin.y, 3
	view.drawing_color = ((@ctrl_down || @shift_down) && @parent) ? @color_draggingR : @color_dragging
	view.draw2d GL_POLYGON, pts
	view.line_stipple = ''
	dec = 13
	ptxy = Geom::Point3d.new @x, @y, 0
	vec = origin.vector_to ptxy
	widtext = @dragging.text.length * 7
	if vec % X_AXIS > 0
		ptend = ptxy.offset X_AXIS, dec 
		pttext = ptend.offset X_AXIS, 4
	else
		ptend = ptxy.offset X_AXIS, -dec 
		pttext = ptend.offset X_AXIS, -(widtext + 4)
	end
	vecend = compute_dragging_end origin3d
	drag_end = origin3d.offset vecend
	
	#Drawing the companion leaders if any
	if @multi_arrow
		closest_bary_companions drag_end
		view.line_width = 2
		@lst_closest_bary.each do |pt|
			view.draw GL_LINE_STRIP, [pt, drag_end]
			pt2d = view.screen_coords(pt)
			pts = G6.pts_square pt2d.x, pt2d.y, 3
			view.draw2d GL_POLYGON, pts
		end
	end
	
	#drawing the main labe leader and text box
	view.line_width = 2
	view.draw GL_LINE_STRIP, [origin3d, drag_end]
	view.draw2d GL_LINE_STRIP, [ptxy, ptend]
	pttext.y -= 8
	pts = G6.pts_rectangle pttext.x, pttext.y, widtext, 20
	view.draw2d GL_POLYGON, pts
	pttext.x += 4
	view.draw_text pttext, @dragging.text
end

#Calculate the closest bary point to the end of the label
def closest_bary_companions(ptend)
	@lst_closest_bary = []
	return if @lst_companions.empty?
	@lst_companions.each do |ls|
		lss = ls.collect { |pt| [ptend.distance(pt), pt] }
		lss.sort! { |a, b| a[0] <=> b[0] }
		@lst_closest_bary.push lss[0][1]
	end
end

#---------------------------------------------------------------------------------------------
# Standard Tool methods
#---------------------------------------------------------------------------------------------

#Activation of the tool
def activate
	LibFredo6.register_ruby "ReportLabelArea::Label"
	@dialog = LabelDialog.new @uac, self
	@model = Sketchup.active_model
	@view = @model.active_view
	@selection = @model.selection
	@initial_selection = @selection.to_a.clone
	@ph = @view.pick_helper
	@ip = Sketchup::InputPoint.new
	@ipdown = Sketchup::InputPoint.new
	build_initial_selection
	UI.start_timer(0) { after_pick }
	UI.start_timer(1) { after_pick }
end

#Deactivation
def deactivate(view)
	@dialog.close_dialog_top
	reset_env
	persistence_save
	view.invalidate
end

#Current cursor
def onSetCursor
	UI::set_cursor @id_cursor
end

#Mouse Movements
def onMouseMove_zero(flags=0) ; onMouseMove(flags, @x, @y, @view) ; end
def onMouseMove(flags, x, y, view)
	return unless x
	@x = x
	@y = y
	handle_move flags, x, y, view
end	

def onLButtonDoubleClick(flags, x, y, view)
	hanleDoubleClick(flags, x, y, view)
end

#Button click - Means that we end the selection
def onLButtonDown(flags, x, y, view)
	@button_down = true
	@xdown = x
	@ydown = y
	@time_click_down = Time.now.to_f
	handle_click_down
	view.invalidate
end

#Button click - Means that we end the selection
def onLButtonUp(flags, x, y, view)
	@button_down = false
	@xup = x
	@yup = y
	handle_click_up flags
end

def draw(view)
	draw_picked view unless @frozen
	draw_total_selection view
	unless @button_down
		G6.draw_rectangle_text(view, @x, @y, @elt_label_new_text, 'powderblue', 'blue') if @elt_label && @elt_label_new_text
		G6.draw_rectangle_text(view, @x, @y, @txt_current_area, 'lightyellow', 'orange') if @picked_faces
		G6.draw_rectangle_text(view, @x, @y, "#{@current_name} --> #{@txt_current_area}", 'pink', 'red') if @picked_container
	end	
	draw_label_anchor view
	draw_dragging view if @dragging
end

#Transfer key from dialog box
def transfer_key(event, key)
	if RUN_ON_MAC
		case key
		when 16
			key = CONSTRAIN_MODIFIER_KEY
		when 18
			key = COPY_MODIFIER_KEY
		end
	end	
	if event =~ /down/i
		onKeyDown key, 0, 0, @view
	else	
		onKeyUp key, 0, 0, @view
	end	
end

#Trap Modifier keys for extended and Keep selection
def onKeyDown(key, rpt, flags, view)
	key = Traductor.check_key key, flags, false
	case key			
	when CONSTRAIN_MODIFIER_KEY
		@shift_down = true
		onMouseMove_zero 1
	when COPY_MODIFIER_KEY
		@ctrl_down = true
		onMouseMove_zero 1
	else
		@shift_down = @ctrl_down = false
	end	
end

def onKeyUp(key, rpt, flags, view)
	key = Traductor.check_key key, flags, true
	case key	
	when VK_DELETE, 8
		erase_label if @elt_label
		
	when CONSTRAIN_MODIFIER_KEY
		toggle_recurse if @shift_down
		@time_shift_up = Time.now
		onMouseMove_zero 0
	when COPY_MODIFIER_KEY
		toggle_single_face if @ctrl_down
		@time_ctrl_up = Time.now
		onMouseMove_zero 0
	else 
		if @time_ctrl_up && (Time.now - @time_ctrl_up) < 0.1	
			toggle_single_face
			onMouseMove_zero 0
		end	
		if @time_shift_up && (Time.now - @time_shift_up) < 0.1	
			toggle_recurse
			onMouseMove_zero 0
		end	
	end	
	@shift_down = @ctrl_down = false
end

#Cancellation and undo
def onCancel(flag, view)
	if flag == 0
		clear_selection
	else	
		
	end	
end

end	#class LabelTool

#========================================================================================
#========================================================================================
# Class LabelTool: Interactive tool to show area and put labels
#========================================================================================
#========================================================================================

class LabelDialog

@@top_dialog = nil
@@unique_key = "ReportLabelArea_Label_DLG"

#Invoke the Label Dialog
def LabelDialog.invoke(uac, tool)
	@@top_dialog = self.new(uac) #unless Traductor::Wdlg.check_instance_displayed(@@unique_key)
	@@top_dialog
end	

#initialization of the dialog box 
def initialize(uac, tool)
	@uac = uac
	@tool = tool
	@layers = @tool.layers
	@wdlg = create_dialog_top
end
#--------------------------------------------------------------------------------------------------------------
# Dialog box configuration
#--------------------------------------------------------------------------------------------------------------			 

#Create the dialog box
def create_dialog_top
	init_dialog_top
	wdlg_key = @@unique_key
	@wdlg = Traductor::Wdlg.new @title, wdlg_key, false
	@wdlg.set_unique_key @@unique_key
	@wdlg.no_auto_resize
	@wdlg.set_size @wid_total, @hgt_total
	@wdlg.set_background_color 'lavender'
	@wdlg.set_callback self.method('topdialog_callback') 
	@wdlg.set_on_close { on_close_top() }
	@wdlg.set_position 0, 0
	refresh_dialog_top
	@wdlg.show
	@wdlg
end

#Initialize parameters of the dialog box
def init_dialog_top
	#Column width and Heights
	@wid_extra = (RUN_ON_MAC) ? 40 : 80
	@hgt_extra = (RUN_ON_MAC) ? 30 : 0
	@wid_total = 750 + @wid_extra + 10
	@hgt_total = 130 + @hgt_extra
	@title = T7[:MNU_Label]
	
	@space4 = "&nbsp;" * 4
	@space_img = "&nbsp;" * 7
end

#Refresh the dialog box
def refresh_dialog_top
	html = format_html_top @wdlg
	@wdlg.set_html html
end

#Notification of window closure
def on_close_top
	@@top_dialog = nil
	Sketchup.active_model.select_tool nil
end

#Close the dialog box
def close_dialog_top
	@wdlg.close
end

#Call back for Report Dialog
def topdialog_callback(event, type, id, svalue)
	case event
		
	#Command buttons
	when /onclick/i
		if @uac.callback_check(id)
			update_units
			return svalue
		end
		case id
		when /ButtonLabelAll/i
			@tool.substitute_all_labels
		when /ButtonDone/i
			@wdlg.close
		when 'ID_CLEAR_SELECTION'
			@tool.clear_selection
		end	

	#Combo layer
	when /onchange/i
		case id
		when /ButtonRecurse/
			@tool.toggle_recurse
		when /ButtonSingleFace/
			@tool.toggle_single_face
		when /ButtonMultiArrow/
			@tool.toggle_multi_arrow
		when /ComboLayer/i
			@tool.set_cur_layer svalue
		end	
		
	when /onKeyUp/i, /onKeyDown/i	#Escape, return, Space and Ctrl, Shift
		case svalue
		when /\A27\*/, /\A13\*/, /\A32\*/
			@wdlg.close
		when /\A(16)\*/, /\A(17)\*/, /\A(18)\*/
			@tool.transfer_key(event, $1.to_i)
		end	
		
	end
	svalue
end

#update the units
def update_units
	@uac.update_tables @wdlg
	symb, unit, decimal, factor = @uac.get_current_param
	ReportLabelArea.save_units symb, decimal
	@tool.update_dialog
end

#Build the HTML for Report Dialog
def format_html_top(wdlg)
	#Creating the HTML stream	
	html = Traductor::HTML.new
		
	#Initialization	
	space2 = "&nbsp;&nbsp;"
	@units = @uac.current_name
		
	#Styles used in the dialog box
	html.create_style 'Title', nil, 'B', 'K: navy', 'F-SZ: 16', 'text-align: center'
	
	html.create_style 'SmallText', nil, 'B', 'F-SZ: 10', 'K: black', 'text-align: left'
	html.create_style 'SmallTextB', nil, 'B', 'F-SZ: 10', 'K: blue', 'text-align: left'
	html.create_style 'SelectionCurrent', nil, 'B', 'F-SZ: 11', 'K: blue', 'BG: lightgreen', 'text-align: left'
	html.create_style 'SelectionTotal', nil, 'B', 'F-SZ: 11', 'K: black', 'BG: lightblue', 'text-align: left'
	html.create_style 'Image', nil, 'B', 'F-SZ: 11', 'K: black', 'BG: lightblue'
	html.create_style 'Layer', nil, 'B', 'F-SZ: 10', 'K: black', 'BG: lightblue'
	html.create_style 'LabelCurrent', nil, 'B', 'F-SZ: 11', 'K: blue', 'BG: khaki', 'text-align: left'
	html.create_style 'LabelNew', nil, 'B', 'F-SZ: 11', 'K: black', 'BG: powderblue', 'text-align: left'
	html.create_style 'Element', nil, 'B', 'K: black', 'text-align: left'
	html.create_style 'SpanGreen', nil, 'BG: palegreen'
	html.create_style 'SpanRed', nil, 'BG: pink'
	html.create_style 'Area', nil, 'B', 'K: navy', 'text-align: right'
	html.create_style 'NbFace', nil, 'I', 'K: green', 'text-align: right'
	
	@uac.html_class_style html
	
	html.create_style 'Button', nil, 'F-SZ: 10'
	html.create_style 'ButtonY', nil, 'F-SZ: 10', 'BG: yellow'
		
	#Inserting the main table and unit choosers
	txkey = (RUN_ON_MAC) ? " (alt)" : " (Ctrl)"
	butsingleface = HTML.format_checkbox(@tool.single_face?, T7[:BUTTON_SingleFace] + txkey, "ButtonSingleFace", 'SmallText', nil, T7[:TIP_SingleFace])
	butrecurse = HTML.format_checkbox(@tool.recurse?, T7[:BUTTON_Recurse], "ButtonRecurse", 'SmallText', nil, T7[:TIP_Recurse])
	butmultiarrow = HTML.format_checkbox(@tool.multi_arrow?, T7[:BUTTON_MultiArrow], "ButtonMultiArrow", 'SmallText', nil, T7[:TIP_MultiArrow])
	cl = (@tool.single_face?) ? 'SpanGreen' : 'SpanRed'
	spansingleface = HTML.format_span butsingleface, "SPAN_SingleFace", cl, nil, T7[:TIP_SingleFace]
	cl = (@tool.recurse?) ? 'SpanGreen' : 'SpanRed'
	spanrecurse = HTML.format_span butrecurse, "SPAN_Recurse", cl, nil, T7[:TIP_Recurse]
	spanlayer = HTML.format_span T6[:T_TXT_Layer], nil, "SmallText", nil, T7[:TIP_Layer]
	
	butlabelall = HTML.format_button "Label All", "ButtonLabelAll", 'ButtonY', nil, T7[:TIP_LabelAll]
	combolayer = HTML.format_combobox @tool.get_cur_layer, @layers, "ComboLayer", "Layer", nil, T7[:TIP_Layer]
	butdone1 = HTML.format_button T7[:T_BUTTON_Done], "ButtonDone1", 'Button', nil
	butdone2 = HTML.format_button T7[:T_BUTTON_Done], "ButtonDone2", 'Button', nil
	
	table_d = @uac.format_table_decimal
	table_u = @uac.format_table_units
	html.body_add "<div style='position: absolute ; top: 1px ; height: 100px ; width: 98%'>"
	html.body_add "<table width='100%' cellspacing='2px' cellpadding='0'><tr>"
	html.body_add "<td align='right' rowspan='2' width='50%' align='right'>#{table_u}</td>"
	html.body_add "<td align='left' rowspan='2' width='15%' valign='center'>#{table_d}</td>"
	html.body_add "<td align='right' width='35%' valign='center'>#{spanlayer}#{space2}#{combolayer}</td></tr>"
	html.body_add "</tr><td align='right' colspan='3' valign='center'>#{spanrecurse}#{space2}#{spansingleface}</td>"
	html.body_add "</tr>"
	html.body_add "</table>"
	
	wid_field = "width='35%'"
	img_clear = HTML.image_file Traductor::MYPLUGIN.picture_get("Button_Clear")
	himg_clear = HTML.format_imagelink img_clear, 16, 16, "ID_CLEAR_SELECTION", nil, nil, T7[:MNU_ClearSelection]
	nodisp = ""
	style = "style='border: 1px solid blue ; margin-right: 4px'"
	tx_field_total = "Total Area in Selection"
	field_name = HTML.format_span "", "ID_NAME"
	field_current = HTML.format_span "", "ID_CURRENT"
	field_total = HTML.format_span "", "ID_TOTAL"
	html.body_add "<table width='100%' cellspacing='2px' cellpadding='0'>"
	html.body_add "<tr #{nodisp} id='TR_Selection_1'>"
	html.body_add "<td id='SelectionName' colspan='2' class='SmallTextB' #{wid_field}>#{field_name}</td>"
	html.body_add "<td class='SmallText' #{wid_field}>#{tx_field_total}</td>"
	html.body_add "<td align='right'>&nbsp;</td>"
	html.body_add "<td align='right'>#{butmultiarrow}</td>"
	html.body_add "</tr>"
	html.body_add "<tr #{nodisp} id='TR_Selection_2'>"
	html.body_add "<td #{style} class='SelectionCurrent' #{wid_field}>#{space2}#{field_current}</td>"
	html.body_add "<td width='4%' align='center'>>></td>"
	html.body_add "<td #{style} class='SelectionTotal' #{wid_field}>#{space2}#{field_total}</td>"
	html.body_add "<td #{style} class='Image' align='center' width='4%'>#{himg_clear}</td>"
	html.body_add "<td align='right'>#{butlabelall}#{space2}#{butdone1}</td>"
	html.body_add "</tr>"
	
	nodisp = "style='display: none'"
	tx_label_current = "Label - Current Text"
	tx_label_new = "Label - New Text - Click to update"
	field_label_current = HTML.format_span "", "ID_LABEL_CURRENT"
	field_label_new = HTML.format_span "", "ID_LABEL_NEW"
	html.body_add "<tr #{nodisp} id='TR_Label_1'>"
	html.body_add "<td class='SmallTextB' #{wid_field}>#{tx_label_current}</td>"
	html.body_add "<td width='4%' align='center'>&nbsp;</td>"
	html.body_add "<td class='SmallText' #{wid_field}>#{tx_label_new}</td>"
	html.body_add "<td align='right'>&nbsp;</td>"
	html.body_add "</tr>"
	html.body_add "<tr #{nodisp} id='TR_Label_2'>"
	html.body_add "<td #{style} class='LabelCurrent' #{wid_field}>#{space2}#{field_label_current}</td>"
	html.body_add "<td width='5%' align='center'>>></td>"
	html.body_add "<td #{style} class='LabelNew' #{wid_field}>#{space2}#{field_label_new}</td>"
	html.body_add "<td align='right'>#{butlabelall}#{space2}#{butdone2}</td>"
	html.body_add "</tr>"
	
	html.body_add "</table>"
				
	#Special_scripts
	html.script_add special_scripts
	html.script_add @uac.special_scripts
	
	#Returning the HTML object
	html	
end

#Special script to go with the Unit Area controller
def special_scripts()
	text = %Q~

function update_for_selection(name, txcurrent, txtotal) {
	document.getElementById ('ID_NAME').innerHTML = name ;
	document.getElementById ('ID_CURRENT').innerHTML = txcurrent ;
	document.getElementById ('ID_TOTAL').innerHTML = txtotal ;
	field_hide_fields ('', 'none') ;
}

function update_for_label(txcurrent, txnew) {
	document.getElementById ('ID_LABEL_CURRENT').innerHTML = txcurrent ;
	document.getElementById ('ID_LABEL_NEW').innerHTML = txnew ;
	field_hide_fields ('none', '') ;
}

function field_hide_fields(sel, lab) {
	document.getElementById ('TR_Selection_1').style.display = sel ;
	document.getElementById ('TR_Selection_2').style.display = sel ;
	document.getElementById ('TR_Label_1').style.display = lab ;
	document.getElementById ('TR_Label_2').style.display = lab ;
}

function update_cb_recurse(val) {
	var obj = document.getElementById ('ButtonRecurse') ;
	obj.checked = val ;
	obj = document.getElementById ('SPAN_Recurse') ;
	if (val) obj.style.backgroundColor = 'palegreen' ;
	else obj.style.backgroundColor = 'pink' ;
}

function update_cb_single_face(val) {
	var obj = document.getElementById ('ButtonSingleFace') ;
	obj.checked = val ;
	obj = document.getElementById ('SPAN_SingleFace') ;
	if (val) obj.style.backgroundColor = 'palegreen' ;
	else obj.style.backgroundColor = 'pink' ;
}

function update_cb_multi_arrow(val) {
	var obj = document.getElementById ('ButtonMultiArrow') ;
	obj.checked = val ;
}

~
	
	text
end

#update the Selection fields
def update_selection(name, txcurrent, txtotal)
	@wdlg.execute_script "update_for_selection('#{name}', '#{txcurrent}', '#{txtotal}') ;"
end

#Update the Label fields
def update_label(txcurrent, txnew)
	txcurrent = HTML.safe_text txcurrent
	txnew = HTML.safe_text txnew
	@wdlg.execute_script "update_for_label('#{txcurrent}', '#{txnew}') ;"
end

def update_recurse(val)
	sval = (val) ? "true" : "false"
	@wdlg.execute_script "update_cb_recurse(#{sval}) ;"
end

def update_single_face(val)
	sval = (val) ? "true" : "false"
	@wdlg.execute_script "update_cb_single_face(#{sval}) ;"
end

def update_multi_arrow(val)
	sval = (val) ? "true" : "false"
	@wdlg.execute_script "update_cb_multi_arrow(#{sval}) ;"
end

end	#class LabelDialog

#========================================================================================
#========================================================================================
# Class UnitAreaController
#========================================================================================
#========================================================================================

class UnitAreaController

UnitInfo = Struct.new :symb, :categ, :name, :long_name, :short_template, :long_template, :factor

#Instance initialization
def initialize(symb_unit=nil, decimals=nil)
	@nb_decimals = 5
	@square = [178].pack "U"
	@lst_units = []
	@hsh_units = {}
	
	register_unit :mm, :decimal, "mm*", "millimeter*"
	register_unit :cm, :decimal, "cm*", "centimeter*"
	register_unit :dm, :decimal, "dm*", "decimeter*"
	register_unit :m, :decimal, "m*", "meter*"
	register_unit :a, :decimal, "a", "are"
	register_unit :ha, :decimal, "ha", "hectare"
	register_unit :km, :decimal, "km*", "kilometer*"
	
	register_unit :inch, :archi, "inch*"
	register_unit :feet, :archi, "feet*"
	register_unit :yard, :archi, "yard*"
	register_unit :acre, :archi, "acre"
	register_unit :mile, :archi, "mile*"
	
	@lst_units_decimal = @lst_units.find_all { |info| @hsh_units[info].categ != :archi }
	@lst_units_archi = @lst_units.find_all { |info| @hsh_units[info].categ == :archi }
	@nb_units = @lst_units.length
	
	#Setting the initial units
	symb_unit = @lst_units[0] unless symb_unit
	decimals = 0 unless decimals
	set_current_unit symb_unit
	set_decimals decimals
	
	#Colors
	@color_on = 'red'
	@bgcolor_on = 'yellow'
	@color_off = 'gray'
	@bgcolor_off = 'gainsboro'
	@dot, @comma = Traductor.dot_comma true
end

#Registration of a unit definition
def register_unit(symb, categ, short_template, long_template=nil)
	info = UnitInfo.new
	info.symb = symb
	info.categ = categ
	info.short_template = short_template
	info.long_template = long_template
	info.name = short_template.sub('*', @square)
	info.long_name = (long_template) ? long_template.sub('*', @square) : info.name
	info.factor = factor_convertion symb
	@lst_units.push symb
	@hsh_units[symb] = info
end

#Get Informations
def name(symb) ; info = @hsh_units[symb] ; (info) ? info.name : "" ; end
def long_name(symb) ; info = @hsh_units[symb] ; (info) ? info.long_name : "" ; end
	
def get_current_param ; [@units, @unit_name, @decimals, @factor_unit] ; end
def current_name ; @unit_name ; end
def square_character ; @square ; end
	
#----------------------------------------------------------------------------------------------------
# Units, reports and formatting
#----------------------------------------------------------------------------------------------------

#Set the current unit (u as a symbol)
def set_current_unit(u)
	return false if u == @units
	@units = u
	@unit_name = @hsh_units[u].name
	@factor_unit = factor_convertion(@units)
	true
end

#Set the current number of decimals
def set_decimals(d)
	return false if d == @decimals
	@decimals = d
	@multiple = 10.0 ** @decimals
	@pformat_unit = "%3.#{@decimals}f"
	true
end

#Nice print format for areas with unit conversion	
def value_area(area) ; area * @factor_unit ; end
def format_area(area)
	a = area * @factor_unit
	"#{Traductor.format_number_by_3 a, @decimals} #{@unit_name}"
end
		
#Figure out the best unit	
def find_best_unit(area, force=false)
	model = Sketchup.active_model
	
	#Keep current units if model not changed
	return if !force && model == @model_units && @units
	
	#Finding out the most appropriate area unit
	@model_units = model
	options = model.options["UnitsOptions"]
	format = options["LengthFormat"]
	
	case format
	when 1, 2, 3	#Architectural, Engineering, Fractional --> inches, foot
		lsunit = @lst_units_archi
	else			#Decimal
		lsunit = @lst_units_decimal
	end
	unit = lsunit[-1]
	decimal = 0
	lsunit.each do |unit|
		factor = factor_convertion unit
		a = area * factor
		if a < 999
			[100, 10, 1, 0.1, 0.01].each_with_index do |v, i|
				if a > v
					decimal = i
					break
				end
			end
			break
		end	
	end
	
	#Setting the default units
	@model_units = model
	set_current_unit(unit) || set_decimals(decimal) 
end

#Compute the unit conversion factor
def factor_convertion(symb_unit)
	case symb_unit
	when :inch, :feet, :km, :mm, :cm, :mile, :yard, :m
		f = eval "1.0.#{symb_unit.to_s}"
	when :dm	
		return factor_convertion(:cm) / 100.0
	when :ha	
		return factor_convertion(:m) / 10000.0
	when :acre	
		return factor_convertion(:feet) / 43560.0
	when :a	
		return factor_convertion(:m) / 100.0
	else
		f = 1.0
	end
	1.0 / f / f
end

def symb_unit_from_text(txunit, default=nil)
	txunit = txunit.strip.downcase
	case txunit
	when /millimeter/, "mm"
		symb = :mm
	when /centimeter/, "cm"
		symb = :cm
	when /kilometer/, "km"
		symb = :km
	when /decimeter/, "dm"
		symb = :dm
	when /meter/, "m"
		symb = :m
	when /hectare/, "ha"
		symb = :ha
	when /are/, "a"
		symb = :a
	when /acre/
		symb = :acre
	when /feet/, /foot/
		symb = :feet
	when /inch/
		symb = :inch
	when /yard/, "yd"
		symb = :yard
	when /mile/
		symb = :mile
	else
		symb = default
	end
	symb	
end

#----------------------------------------------------------------------------------------
# Formatting of HTML tables for units and decimals
#----------------------------------------------------------------------------------------

#Update the tables for units and decimals
def update_tables(wdlg)
	wdlg.execute_script "uac_update_tables('#{@unit_name}', #{@decimals})"
end

def html_class_style(html)
	html.create_style 'UAC_Decimal', nil, 'B', 'K: black', 'F-SZ: 11', 'border: 1px solid lightgrey'
	html.create_style 'UAC_Unit', nil, 'B', 'K: black', 'F-SZ: 11', 'border: 1px solid black'
	html.create_style 'UAC_UnitAuto', 'UAC_Unit', 'BG: lightgreen'
end

def callback_check(id, sample_area=nil)
	return nil unless id =~ /(UAC_.+)_/
	sid = $1
	
	case id
	when /UAC_Decimal_(\d+)/i		
		return nil unless set_decimals($1.to_i)
	when /UAC_Auto_Unit/i
		return nil unless sample_area && find_best_unit(sample_area, true)
	when /UAC_Unit_(\d+)/i
		return nil unless set_current_unit(@lst_units[$1.to_i])
	else
		return nil
	end	
	[@units, @decimals]
end


#Format the clikable table for the units
def format_table_units
	#initialization
	lenmax = [@lst_units_decimal.length, @lst_units_archi.length].max
	action = HTML.format_actions ['onclick']
	wid = (100.0 / lenmax).round
	n = 0
	#Building the table
	text = "<table width='99%' cellpadding='1px'>"
	[@lst_units_decimal, @lst_units_archi].each do |ls|
		text += "<tr>"
		for i in 0..lenmax-1
			u = ls[i]
			color, bgcolor = (u == @units) ? [@color_on, @bgcolor_on] : [@color_off, @bgcolor_off]
			style = "cursor:pointer ; background-color: #{bgcolor} ; color: #{color}"
			if n == lenmax * 2 - 1
				tip = T7[:TIP_AutoUnitArea]
				text += "<td align='center' width='#{wid}%' id='UAC_Auto_Unit' #{action} style ='cursor:pointer'"
				text += " class='UAC_UnitAuto' title='#{tip}'>?</td>"
			elsif u
				uname = @hsh_units[u].name
				tip = @hsh_units[u].long_name
				text += "<td align='center' width='#{wid}%' id='UAC_Unit_#{n}' #{action} style ='#{style}'"
				text += " class='UAC_Unit' title='#{tip}'>#{uname}</td>"
			else
				text += "<td width='#{wid}%'>&nbsp;</td>"
			end	
			n += 1
		end	
		text += "</tr>"
	end	
	text += "</table>"
	text
end

#Format the clikable table for the decimals
def format_table_decimal
	color = 'yellow'
	decimal = @decimals
	text = ""
	action = HTML.format_actions ['onclick']
	for i in 0..@nb_decimals
		s = '0'
		if i == 0
			s += @dot
			tip = T7[:T_TXT_NO_Decimal]
		elsif i == 1
			tip = "1 #{T7[:T_TXT_Decimal]}"
		else 
			tip = "#{i} #{T7[:T_TXT_Decimals]}"
		end	
		color, bgcolor = (i <= decimal) ? [@color_on, @bgcolor_on] : [@color_off, @bgcolor_off]
		style = "cursor:pointer ; background-color: #{bgcolor} ; color: #{color}"
		text += "<span id='UAC_Decimal_#{i}' #{action} style ='#{style}' class='UAC_Decimal' title='#{tip}'>&nbsp;#{s}&nbsp;</span>"
	end
	text
end

#Special script to go with the Unit Area controller
def special_scripts()
	#Transfering constants between Ruby and JS
	text = ""
	text += "var uac_nb_decimals = #{@nb_decimals} ;"
	text += "var uac_nb_units = #{@nb_units} ;"
	text += "var uac_color_on = '#{@color_on}' ;"
	text += "var uac_bgcolor_on = '#{@bgcolor_on}' ;"
	text += "var uac_color_off = '#{@color_off}' ;"
	text += "var uac_bgcolor_off = '#{@bgcolor_off}' ;"

	#Function to update the areas
	text += %Q~

function uac_update_tables(unit, decimal) {
	for (var i = 0 ; i <= uac_nb_decimals ; i++) {
		key = 'UAC_Decimal_' + i.toString() ;
		var obj = document.getElementById (key) ;
		if (i <= decimal) {
			obj.style.backgroundColor = uac_bgcolor_on ;
			obj.style.color = uac_color_on ; 
			}
		else {
			obj.style.backgroundColor = uac_bgcolor_off ;
			obj.style.color = uac_color_off ;
			}
	}
	for (var i = 0 ; i < uac_nb_units ; i++) {
		key = 'UAC_Unit_' + i.toString() ;
		var obj = document.getElementById (key) ;
		if (obj.innerHTML == unit) {
			obj.style.backgroundColor = uac_bgcolor_on ;
			obj.style.color = uac_color_on ; 
			}
		else {
			obj.style.backgroundColor = uac_bgcolor_off ;
			obj.style.color = uac_color_off ;
			}
	}
}

~
	
	text
end

end	#class UnitAreaController

end	#End Module ReportLabelArea

end	#End Module F6_FredoTools
