=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2012 Fredo6 - Designed and written Feb 2012 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_FredoTools__ReverseOrientFaces.rb
# Original Date	:   25 Mar 2012
# Description	:   Reverse and orient faces with various options
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

module ReverseOrientFaces

#Texts for the module
T7[:TIP_ReverseMode] = "Reverse Faces"
T7[:TIP_OrientMode] = "Orient Faces"
T7[:INFO_ClickReverse] = "CLICK to Reverse faces"
T7[:INFO_DragReverse] = "DRAG to continue Reversing faces"
T7[:INFO_ClickOrient] = "CLICK to start Orienting faces"
T7[:INFO_DragOrient] = "DRAG to continue Orienting faces"
T7[:MNU_SwitchToReverse] = "Switch to Reverse mode"
T7[:MNU_SwitchToOrient] = "Switch to Orient mode"
T7[:MSG_SelectedFaces] = "Selected Faces: %1"

#====================================================================================================
#----------------------------------------------------------------------------------------------------
# Plugin Implementation
#----------------------------------------------------------------------------------------------------
#====================================================================================================

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

#Launch Method from the ReverseOrientFaces command (icon or menu)
def self._execution(symb)
	Sketchup.active_model.select_tool ReverseOrientFacesTool.new
end

#----------------------------------------------------------------------------------------------------
# ReverseOrientFacesTool: Class managing the SU Tool
#----------------------------------------------------------------------------------------------------

class ReverseOrientFacesTool < Traductor::PaletteSuperTool

DrawFrame = Struct.new :frames, :dashed, :contour, :ll_frames, :ll_dashed, :ll_contour, :scene

#Tool Initialization
def initialize
	init_cursor
	init_colors
	init_messages
	@lst_face_selection = [:single, :surface, :connected, :same_color, :orientation]
end

#Initialize colors
def init_colors
	@color_message = 'khaki'
	@color_face_reverse = 'green'
	@color_face_orient = 'blue'
	
	if G6.su_capa_color_polygon
		@color_picked_face = Sketchup::Color.new 'lightgreen'
		@color_picked_face.alpha = 0.2
	else
		@color_picked_face = nil
	end	
end

def init_cursor
	@id_cursor_reverse = MYPLUGIN.create_cursor "ReverseOrientFaces_cursor_Reverse", 0, 0
	@id_cursor_orient = MYPLUGIN.create_cursor "ReverseOrientFaces_cursor_Orient", 0, 0
	@id_cursor = @id_cursor_reverse
end

#Initialize messages to be displayed in the palette message bar
def init_messages
	@txt_front = T6[:T_TXT_Recto]
	@txt_back = T6[:T_TXT_Verso]
	@hsh_messages = {}
	@hsh_messages[:exit] = T6[:T_INFO_DoubleClickExit]
	@hsh_messages[:click_reverse] = T7[:INFO_ClickReverse]
	@hsh_messages[:drag_reverse] = T7[:INFO_DragReverse]
	@hsh_messages[:click_orient] = T7[:INFO_ClickOrient]
	@hsh_messages[:drag_orient] = T7[:INFO_DragOrient]
end

#Setting the navigation environment
def reset_env
	@tr = @tr_r = @tr_id = Geom::Transformation.new
	@parent_down = nil
	@parent = nil
	@picked_faces = nil
	reset_drawing_env
	@key_picked = nil
	@key_previous = nil
	@initial_face = nil
	@last_face = nil
end	
	
#---------------------------------------------------------------------------------------------
# Plugin Parameters: Persistence, modes and flags
#---------------------------------------------------------------------------------------------

#Load parameters from persistence
def persistence_load
	@reverse_mode = true
	@option_face_selection = :surface
	sparam = Sketchup.read_default "FredoTools_ReverseOrientFaces", "Param"
	return unless sparam
	begin
		hsh = eval sparam
		@reverse_mode = hsh[:reverse_mode] if hsh.has_key?(:reverse_mode)
		@option_face_selection = hsh[:option_face_selection] if hsh.has_key?(:option_face_selection)
	rescue
	end	
end

#Save parameters to persistence
def persistence_save
	hsh = {}
	hsh[:reverse_mode] = @reverse_mode
	hsh[:option_face_selection] = @option_face_selection
	sparam = hsh.inspect.gsub('"', "'")
	Sketchup.write_default "FredoTools_ReverseOrientFaces", "Param", sparam
end

#Toggle the mode between Reverse and Orient
def toggle_reverse_mode
	@reverse_mode = !@reverse_mode
	after_pick
	onSetCursor
end

#Cycle thorugh the option for face selection
def toggle_face_selection(incr)
	ipos = @lst_face_selection.index(@option_face_selection)
	ipos = 1 unless ipos
	ipos = (ipos + incr).modulo(@lst_face_selection.length)
	set_face_selection @lst_face_selection[ipos]
end

#Switch the display mode to monochome
def set_render_mode
	@old_render_mode = @rendering_options["RenderMode"]
	@rendering_options["RenderMode"] = 5
end

#Restore the initial render mode
def restore_render_mode
	@rendering_options["RenderMode"] = @old_render_mode
end

#Toggle the render mode between monchrome and textured
def toggle_render_mode
	@rendering_options["RenderMode"] = (@rendering_options["RenderMode"] == 5) ? 2 : 5
end

def set_face_selection(mode=nil)
	mode = :surface unless mode
	@option_face_selection = mode
	reset_drawing_env
	onMouseMove_zero
end	
	
def toggle_draw_hidden
	@rendering_options["DrawHidden"] = !@rendering_options["DrawHidden"]
end

def start_over
	@key_previous = nil
end
	
#---------------------------------------------------------------------------------------------
# Show Messages
#---------------------------------------------------------------------------------------------

#Show message in the palette
def show_message
	#mouse if either in the palette or out of the viewport
	if @mouseOut
		@palette.set_message ""
		return
	elsif @mouse_in_palette
		@palette.set_message @palette.get_main_tooltip, 'aliceblue'
		return
	end	
		
	#Building the text to be displayed
	lst = []
	if !@picked_faces
		lst.push :exit
	elsif @initial_face
		if @reverse_mode
			lst.push((@button_down) ? :drag_reverse : :click_reverse)
		else
			lst.push((@button_down) ? :drag_orient : :click_orient)
		end	
	end	
		
	ltx = lst.collect { |symb| @hsh_messages[symb] }
	text = ltx.join(" -- ")
	@palette.set_message text, @color_message
end
	
#---------------------------------------------------------------------------------------------
# Undo Management
#---------------------------------------------------------------------------------------------

def handle_undo
	start_over
	onMouseMove_zero
end

#---------------------------------------------------------------------------------------------
# Picking Element on Screen
#---------------------------------------------------------------------------------------------

#Manage actions on a mouse move
def handle_move(flags, x, y, view)	
	pick_current_element(flags, x, y, view)
	after_pick	
end

#Identify which objects (face and edge)  are under the mouse
def picking_under_mouse(flags, x, y, view)
	@ph.do_pick x, y, 0	
	ph_face = @ph.picked_face
				
	#Finding the parent and transformation
	tr = @tr_id
	parent = nil
	return nil unless ph_face
	for i in 0..@ph.count
		ls = @ph.path_at(i)
		if ls && ls.include?(ph_face)
			parent = ls[-2]
			tr = @ph.transformation_at(i)
			break
		end
	end	
	tr = @tr_id unless tr
	[ph_face, tr, parent]
end
	
#Interactive pick of edges, faces or container
def pick_current_element(flags, x, y, view)
	#Getting the faces and edges under the mouse
	@picked_faces = nil
	if @button_down && @painting
		@picked_faces = picking_with_drag(flags, x, y, view)
		return unless @picked_faces
		@key_picked = "#{@picked_faces.last.object_id}-#{@tr.to_a}"
	else
		@initial_face, @tr, @parent = picking_under_mouse(flags, x, y, view)
		@tr_r = @tr.inverse if @tr

		#Analyzing the picked face or edge
		if @initial_face
			@initial_face_front = front_face_is_visible?(@initial_face)
			@text_initial_face = (@initial_face_front) ? @txt_front : @txt_back
		else
			@text_initial_face = nil
		end
		
		#Getting the faces
		if @initial_face
			@picked_faces = @hsh_extend_faces["#{@initial_face.entityID}--#{@tr.to_a}"]
			@key_picked = "#{@picked_faces.object_id}-#{@tr.to_a}"
			unless @picked_faces
				@picked_faces = picking_extend_faces(@initial_face, @option_face_selection) 
				@picked_faces.each { |f| @hsh_extend_faces["#{f.entityID}--#{@tr.to_a}"] = @picked_faces }
				@key_picked = "#{@picked_faces.object_id}-#{@tr.to_a}"
				compute_frames_faces
			end	
		end		
	end	
	
	#Checking if entities picked are new
	return if @key_previous == @key_picked
	@key_previous = @key_picked
	
	#New selection - Highlighting or painting	
	if @button_down
		if @painting && @parent == @parent_down
			continue_painting if @picked_faces
		else
			@picked_faces = nil
		end	
	end	
end

#Special method for dragging - Use the trace of mouse movements to collect faces selected
def picking_with_drag(flags, x, y, view)
	#Reference of picked point in 2D	
	@pt_buf = Geom::Point3d.new(@xdown, @ydown, 0) unless @pt_buf
	ptxy = Geom::Point3d.new x, y, 0
	vec = ptxy.vector_to @pt_buf
	return nil unless vec.valid?
	segxy = [ptxy, @pt_buf]
	
	#Compute start face
	next_face = nil
	facebuf, trbuf, parentbuf = picking_under_mouse(flags, @pt_buf.x, @pt_buf.y, view)
	facebuf = nil if parentbuf != @parent_down
	next_face = facebuf
	@pt_buf = ptxy.clone
	return nil unless next_face 
	
	#Finding progressively contiguous faces
	t0 = Time.now
	picked_faces = []
	hsh_edges = {}
	lst = [next_face]
	while lst.length > 0
		next_face = lst.shift
		picked_faces.push next_face
		next_face.edges.each do |edge|
			edge_id = edge.entityID
			next if hsh_edges[edge_id]
			hsh_edges[edge_id] = true
			seg_edge = [edge.start.position, edge.end.position].collect { |pt| pt2d = @view.screen_coords(@tr * pt) ; pt2d.z = 0 ; pt2d }
			if G6.intersect_segment_segment(segxy, seg_edge)
				newface = face_visible_from_edge next_face, edge
				lst.push newface if newface
			end	
		end
	end
	return nil if picked_faces.empty?
	
	#Extending the picked faces
	ls_faces = []
	picked_faces.each do |face|
		next if ls_faces.include?(face)
		ls_faces += picking_extend_faces(face, @option_face_selection)
	end
	ls_faces
end

#Find the next visible face bordering the edge
def face_visible_from_edge(face0, edge)
	ptmid = @tr * Geom.linear_combination(0.5, edge.start.position, 0.5, edge.end.position)
	ptmid2d = @view.screen_coords ptmid
	edge.faces.each do |f|
		next if f == face0
		vec_in = G6.transform_vector G6.normal_in_to_edge(edge, f), @tr
		size = @view.pixels_to_model 2, ptmid
		ptdec = ptmid.offset vec_in, size
		pt2d = @view.screen_coords(ptdec)
		ray = @view.pickray pt2d.x, pt2d.y
		pt, item = @model.raytest ray
		return f if pt && (@tr.inverse * pt).on_plane?(f.plane)
	end	
	nil
end

#Extending the initial face
def picking_extend_faces(face0, option_face_selection)
	#Extending faces
	return nil unless face0
	case option_face_selection
	when :surface
		lsfaces = [face0] + G6.face_neighbours(face0)
	when :connected
		lsfaces = face0.all_connected.find_all { |e| e.instance_of?(Sketchup::Face) } 
	when :same_color
		lsfaces = adjacent_faces_same([face0])
	when :orientation
		lsfaces = adjacent_faces_orientation([face0])
	else
		lsfaces = [face0]
	end	
	lsfaces	
end
	
#Find the adjacent faces with same color	
def adjacent_faces_same(faces)
	face0 = faces[0]
	front0 = front_face_is_visible?(face0)
	mat0 = (front0) ? face0.material : face0.back_material
	
	hsh_faces = {}
	faces.each do |face|
		next if hsh_faces[face.entityID]
		lface = [face]	
		while lface.length > 0
			f = lface.shift
			next if hsh_faces[f.entityID]
			hsh_faces[f.entityID] = f
			f.edges.each do |e|
				e.faces.each do |ff| 
					lface.push ff unless ff == f || hsh_faces[ff.entityID] || !same_as_initial_face?(front0, mat0, ff)
				end	
			end
		end	
	end	
	hsh_faces.values
end

#Find the adjacent faces with same color	
def adjacent_faces_orientation(faces)
	hsh_faces = {}
	faces.each do |face|
		next if hsh_faces[face.entityID]
		lface = [face]	
		while lface.length > 0
			f = lface.shift
			next if hsh_faces[f.entityID]
			hsh_faces[f.entityID] = f
			f.edges.each do |e|
				e.faces.each do |ff| 
					next if hsh_faces[ff.entityID]
					edge, = ff.edges & f.edges
					next unless edge
					lface.push ff unless ff == f || hsh_faces[ff.entityID] || edge.reversed_in?(f) == edge.reversed_in?(ff)
				end	
			end
		end	
	end	
	hsh_faces.values
end

#Check if the face has the same material as the initial face (and optionally same stamp / uv_mode)
def same_as_initial_face?(front, mat, face)
	(front) ? face.material == mat : face.back_material == mat
end

#Check if edge prop follows the specifications	
def acceptable_edge?(edge)
	return true if @action_uv == :mark_diagonal && @ph.picked_face
	return false unless edge
	return true if edge.soft? && @option_edge_prop =~ /S/
	return true if edge.smooth? && @option_edge_prop =~ /M/
	return true if edge.hidden? && @option_edge_prop =~ /H/
	return true if !edge.soft? && !edge.smooth? && !edge.hidden? && @option_edge_prop =~ /P/
	false
end

#---------------------------------------------------------------------------------------------
# Contextual Menu
#---------------------------------------------------------------------------------------------

#Entries for the contextual menu
def getMenu(menu)
	#Options
	mnutext = (@reverse_mode) ? T7[:MNU_SwitchToOrient] : T7[:MNU_SwitchToReverse]
	menu.add_item(mnutext) { toggle_reverse_mode }
	
	#Exit Tool
	menu.add_separator
	menu.add_item(T6[:T_STR_ExitTool]) { exit_tool }
end
	
#---------------------------------------------------------------------------------------------
# Navigation logic
#---------------------------------------------------------------------------------------------

#Compute the Tooltip for view
def compute_tooltip
	tip = (@picked_faces) ? T7[:MSG_SelectedFaces, @picked_faces.length.to_s] : ""
	@palette.set_tooltip tip
end

#Compute the cursor
def compute_cursor
	@id_cursor = (@reverse_mode) ? @id_cursor_reverse : @id_cursor_orient
end

#Update context and display after a pick	
def after_pick
	compute_cursor
	compute_tooltip
	show_message
	@view.invalidate	
end

#Manage the click down
def handle_click_down(flags, x, y, view)
	face, = picking_under_mouse(flags, x, y, view)
	if face
		@parent_down = @parent
		start_painting
	end	
end

#Manage the click up: put pick elements in selection
def handle_click_up(flags, x, y, view)
	finish_painting if @painting
	onMouseMove_zero
end
	
#Button click - On a face, label the face, otherwise end the selection
def handleDoubleClick(flags, x, y, view)
	exit_tool unless @picked_faces
end

#Return or Enter key pressed
def handle_onReturn

end

#---------------------------------------------------------------------------------------------
# Processing Logic
#---------------------------------------------------------------------------------------------

#Start Processing faces
def start_painting()
	return unless @picked_faces && @picked_faces.length > 0
	@painting = true
	@lst_faces = []
	@hsh_session_faces = {}
	@hsh_painted = {}
	@hsh_edges_used = {}

	#Starting the processing
	@model.start_operation "ReverseOrientFaces: " + ((@reverse_mode) ? T7[:TIP_ReverseMode] : T7[:TIP_OrientMode])
	continue_painting
end

#Subsequent processing
def continue_painting
	continue_session @picked_faces, @initial_face
end

#Finish Processing method
def finish_painting()
	@model.commit_operation	
	@painting = false
	reset_drawing_env
end

#Top method to process faces progressively
def continue_session(faces, last_face=nil)
	return unless faces
	
	#Initializing the faces
	faces.each { |face| @hsh_session_faces[face.entityID] = face }
	
	#Checking if there is at least one new face to paint
	return unless @hsh_session_faces.keys.find { |key| !@hsh_painted[key] }

	#Painting faces progressively
	@last_face = last_face if last_face
	while true
		compute_next_faces @last_face if @last_face
		face, edge = get_next_face
		break unless face
		if @reverse_mode
			G6.reverse_face_with_uv(face)
		else
			auto_flip face
		end	
		@hsh_painted[face.entityID] = face
		@last_face = face
	end
	
	@hsh_family_faces = @hsh_painted
	@hsh_session_faces = @hsh_painted.clone
end

#Update the list of next faces
def compute_next_faces(face)
	return unless face
	face.edges.each do |edge|
		@hsh_edges_used[edge.entityID] = face
		edge.faces.each do |f|
			next if f == face
			id = f.entityID
			next if @hsh_painted[id] || !@hsh_session_faces[id]
			@lst_faces.push [f, edge, face]
		end	
	end
end		

#Get the next face to process, with its starting edge	
def get_next_face
	#Finding it in the current list
	while @lst_faces.length > 0
		face, edge, @last_face = @lst_faces.shift
		return [face, edge] unless @hsh_painted[face.entityID]
	end
	
	#Finding a processed face and edge where to start from otherwise
	@hsh_session_faces.each do |key, face|
		next if @hsh_painted[key]
		face.edges.each do |edge|
			@last_face = @hsh_edges_used[edge.entityID]
			return [face, edge] if @last_face
		end
	end	
	
	#Nor more face
	nil
end

#Auto reverse the face if its side is not in continuity with previous
def auto_flip(face)
	return unless @last_face && face != @last_face
	edge, = face.edges & @last_face.edges
	G6.reverse_face_with_uv(face) if edge && edge.reversed_in?(face) == edge.reversed_in?(@last_face)	
end

#Determine if the front (or back) face is the visible face
def front_face_is_visible?(face)
	pt2d = @view.screen_coords(@tr * face.vertices[0].position)
	ray = @view.pickray pt2d.x, pt2d.y
	(G6.transform_vector(face.normal, @tr) % ray[1] <= 0)	
end

#---------------------------------------------------------------------------------------------
# Drawing methods
#---------------------------------------------------------------------------------------------

#Top method to execute drawing
def handle_drawing(view)
	drawing_frames_faces view unless @button_down
	drawing_parent view
	drawing_hi_face view
	drawing_face_side view
end

#Draw a label to tell the side of picked face
def drawing_face_side(view)
	return if @button_down || !@initial_face || !@x
	if @initial_face_front
		bkcolor = 'yellow'
		frcolor = 'orange'
	else
		bkcolor = 'lightgrey'
		frcolor = 'black'
	end	
	G6.draw_rectangle_text view, @x, @y, @text_initial_face, bkcolor, frcolor
end

#Draw highlights for faces and edges
def drawing_hi_face(view)
	return unless !@button_down && @picked_faces && @picked_faces.length > 1
	pts = []
	@initial_face.edges.each do |e|
		pts.push @tr * e.start.position, @tr * e.end.position
	end	
	view.drawing_color = (@reverse_mode) ? @color_face_reverse: @color_face_orient
	view.line_width = 2
	view.line_stipple = ''
	view.draw GL_LINES, pts.collect { |pt| G6.small_offset(view, pt, 2) }
end

#Draw the parent container box if any
def drawing_parent(view)
	if @parent
		view.drawing_color, view.line_width, view.line_stipple = ['gray', 1, '']
		G6.draw_component view, @parent, @tr
	end
end

#---------------------------------------------------------------------------------------------
# Utility methods for managing faces highlight
#---------------------------------------------------------------------------------------------

#Reset the drawing caches
def reset_drawing_env
	@hsh_extend_faces = {}
	@hsh_frames_face = {}
	@key_picked = nil
end

#Recompute some display elements when view or scene is changed
def on_change_of_view
	@hsh_frames_face.each { |key, drframe| drframe.scene = nil }
end

#Compute the triangles and frames for a list of entities
def compute_frames_faces
	return unless @picked_faces
	drframe = @hsh_frames_face[@key_picked]
	return if drframe
	
	draw_frames = []
	draw_dashed = []
	draw_contour = []
	
	if @option_face_selection == :single
		@picked_faces.each do |face|
			face.edges.each { |edge| draw_contour.push(@tr * edge.start.position, @tr * edge.end.position) }
		end
	else
		hfaces = {}
		@picked_faces.each { |f| hfaces[f.entityID] = true }
		@picked_faces.each do |face|
			face.edges.each do |edge| 
				if edge_is_border(edge, hfaces)
					draw_contour.push(@tr * edge.start.position, @tr * edge.end.position) 
				elsif G6.edge_plain?(edge)
					draw_frames.push(@tr * edge.start.position, @tr * edge.end.position) 
				else
					draw_dashed.push(@tr * edge.start.position, @tr * edge.end.position) 
				end	
			end	
		end
	end	
	
	#Storing in a structure
	drframe = DrawFrame.new	
	drframe.frames = draw_frames
	drframe.dashed = draw_dashed
	drframe.contour = draw_contour
	@hsh_frames_face[@key_picked] = drframe
end

#Draw the frames
def drawing_frames_faces(view)
	return unless @picked_faces
	drframe = @hsh_frames_face[@key_picked]
	return unless drframe
	drawing_frames view, drframe
end
	
#Generic draw of farmes	
def drawing_frames(view, drframe)	
	#Checking for recalculation
	scene = drframe.scene
	unless scene && scene == @model.pages.selected_page
		drframe.ll_dashed = drframe.dashed.collect { |pt| G6.small_offset view, pt}
		drframe.ll_frames = drframe.frames.collect { |pt| G6.small_offset view, pt}
		drframe.ll_contour = drframe.contour.collect { |pt| G6.small_offset view, pt}
		drframe.scene = @model.pages.selected_page
	end
		
	view.drawing_color = (@reverse_mode) ? @color_face_reverse : @color_face_orient
		
	#Drawing the frames	
	unless drframe.dashed.empty?
		view.line_width = 1
		view.line_stipple = '_'
		view.draw GL_LINES, drframe.ll_dashed
	end	
	unless drframe.frames.empty?
		view.line_width = 1
		view.line_stipple = ''
		view.draw GL_LINES, drframe.ll_frames
	end	
	unless drframe.contour.empty?
		view.line_width = 3
		view.line_stipple = ''
		view.draw GL_LINES, drframe.ll_contour
	end	
end

#Check if the edge is a border
def edge_is_border(edge, hfaces)
	n = 0
	edge.faces.each do |f|
		n += 1 if hfaces[f.entityID]
		return false if n > 1
	end
	true
end

#---------------------------------------------------------------------------------------------
# Standard Tool methods
#---------------------------------------------------------------------------------------------

#SU Activation of the tool
def activate
	LibFredo6.register_ruby "ReverseOrientFaces"
	
	#Setting the environment
	@model = Sketchup.active_model	
	@view = @model.active_view
	@selection = @model.selection
	@rendering_options = @model.rendering_options
	@ph = @view.pick_helper
	reset_env
	reset_drawing_env
	
	#Setting the default parameters
	persistence_load

	#Palette
	init_palette
		
	#Getting in Monochrome
	set_render_mode
	
	#Initialization	
	@button_down = false
	@painting = false
	start_over
	after_pick
end

#Deactivation
def deactivate(view)
	restore_render_mode
	persistence_save
	view.invalidate
end

#Current cursor
def onSetCursor
	ic = super
	return (ic != 0) if ic
	UI.set_cursor @id_cursor
end

#Handle Return key
def onReturn(view)
	handle_onReturn
end
	
def resume(view)
	on_change_of_view
	onMouseMove_zero
end
	
#Mouse Movements
def onMouseMove_zero(flags=0) ; onMouseMove(flags, @x, @y, @view) ; end
def onMouseMove(flags, x, y, view)
	@time_move = nil
	#return if @mouseOut
	#Checking validity of the move
	if super
		@mouse_in_palette = true
		show_message
		return
	end	
	@mouse_in_palette = false
	@flags = flags
	return unless x
	@time_move = Time.now

	reconcile_button_state(flags, x, y, view)
	
	return if @drawing_ 
	@drawing_ = true

	@x = x
	@y = y
	handle_move flags, x, y, view
end	

#Check if the state of mouse is compatible with internal flags
def reconcile_button_state(flags, x, y, view)
	if RUN_ON_MAC
		onLButtonUp(flags, x, y, view) if flags == 256 && @button_down
	else
		onLButtonUp(flags, x, y, view) if flags.modulo(2) == 0 && @button_down
	end
end

#Button click - Means that we end the selection
def onLButtonDown(flags, x, y, view)
	return if super	
	@button_down = true
	@xdown = x
	@ydown = y
	@pt_buf = nil
	@time_click_down = Time.now.to_f
	handle_click_down(flags, x, y, view)
	
	view.invalidate
end

#Button click - Means that we end the selection
def onLButtonUp(flags, x, y, view)
	return if super	
	return unless @button_down
	@button_down = false
	@xup = x
	@yup = y
	@pt_buf = nil
	handle_click_up(flags, x, y, view)
end

#Double Click received
def onLButtonDoubleClick(flags, x, y, view)
	return if super
	handleDoubleClick(flags, x, y, view)
end

#Cancellation and undo
def onCancel(flag, view)
	reset_drawing_env
	if flag == 0
		Sketchup.undo
		handle_undo
	else	
		UI.start_timer(0) { handle_undo }
	end
end

#Exit the tool
def exit_tool
	@model.select_tool nil
end

#Draw method for Polyline tool
def draw(view)	
	#Drawing in the view
	handle_drawing(view)
	
	#Drawing the palette
	super
	
	#Synchro with Move
	@drawing_ = false	
end

#Handle Key down
def onKeyDown(key, rpt, flags, view)
	key = Traductor.check_key key, flags, false
	@num_keydown = 0 unless @num_keydown
	@num_keydown += 1
	case key			
	when CONSTRAIN_MODIFIER_KEY
		@shift_down = true
		onMouseMove_zero 1
	when COPY_MODIFIER_KEY
		@ctrl_down = @num_keydown
		onMouseMove_zero 1

	when VK_DOWN, VK_LEFT
		toggle_face_selection -1
	when VK_UP, VK_RIGHT
		toggle_face_selection 1
		
	when 13
		if @shift_down || @ctrl_down
			
		end	
	else	
		@shift_down = @ctrl_down = false
	end	
end

def onKeyUp(key, rpt, flags, view)
	key = Traductor.check_key key, flags, true
	case key	
	when VK_DELETE, 8
		
	when CONSTRAIN_MODIFIER_KEY
		@time_shift_up = Time.now
		onMouseMove_zero 0
		@shift_down = false
		
		
	when COPY_MODIFIER_KEY
		@time_ctrl_up = Time.now
		if @ctrl_down && @ctrl_down == @num_keydown
			toggle_reverse_mode
		end	
		onMouseMove_zero 0
		@ctrl_down = false
		
	else 
		@shift_down = @ctrl_down = false
	end	
end

def onMouseLeave(view)
	@mouseOut = true
	show_message
end

def onMouseEnter(view)
	@mouseOut = false
end

#--------------------------------------------------------------
#--------------------------------------------------------------
# Palette Management
#--------------------------------------------------------------
#--------------------------------------------------------------

#---------------------------------------------------------------------
# Creation of the palette
#---------------------------------------------------------------------

#Initialize the main palette and top level method
def init_palette
	hshpal = { :width_message => 200, :key_registry => :reverse_orient_faces }
	@palette = Traductor::Palette.new hshpal
	@draw_local = self.method "draw_button_opengl"
	@blason_proc = self.method "draw_button_blason"
	@symb_blason = :blason
		
	#Blason Button
	hsh = { :blason => true, :tooltip => @title, :draw_proc => @blason_proc, :tooltip => T7[:PlugName] + ' ' + T6[:T_STR_DefaultParamDialog] }
	@palette.declare_button(@symb_blason, hsh) { MYPLUGIN.invoke_default_parameter_dialog }

	#Reverse and Orient Mode
	pal_separator
	hshb = { :height => 32, :width => 32 }
	proc = proc { @reverse_mode }
	hsh = { :value_proc => proc, :draw_proc => @draw_local, :tooltip => T7[:TIP_ReverseMode] }
	@palette.declare_button(:pal_reverse_mode, hsh, hshb) { toggle_reverse_mode }	
	proc = proc { !@reverse_mode }
	hsh = { :value_proc => proc, :draw_proc => @draw_local, :tooltip => T7[:TIP_OrientMode] }
	@palette.declare_button(:pal_orient_mode, hsh, hshb) { toggle_reverse_mode }	
	
	#Option for faces
	palette_face_selection
	
	#Option for the SU environment
	pal_separator
	proc = proc { @rendering_options["DrawHidden"] }
	hsh = { :value_proc => proc, :draw_proc => @draw_local, :tooltip => T6[:T_TIP_DrawHidden] }
	@palette.declare_button(:pal_draw_hidden, hsh) { toggle_draw_hidden }	
	proc = proc { @rendering_options["RenderMode"] == 5 }
	hsh = { :value_proc => proc, :draw_proc => @draw_local, :tooltip => T6[:T_TIP_Monochrome] }
	@palette.declare_button(:pal_monochrome, hsh) { toggle_render_mode }	

	#Exit tool
	pal_separator
	hsh = { :draw_proc => :std_exit, :tooltip => T6[:T_STR_ExitTool] }
	@palette.declare_button(:pal_exit, hsh) { @model.select_tool nil }
		
	#Associating the palette
	set_palette @palette
end

#Separator for the Main and floating palette
def pal_separator ;	@palette.set_current_family ; @palette.declare_separator ; end

#Palette buttons for Face Selection
def palette_face_selection
	hsh = { :passive => true, :height => 32, :draw_proc => :separator_V, :width => 12 }	
	@palette.declare_button :pal_face_sepa, hsh

	wid = 24
	symb_master = :pal_face_selection
	
	hsh = { :passive => true, :type => 'multi_free', :text => T6[:T_BOX_FaceSelection], :height => 16, :tooltip => T6[:T_BOX_FaceSelection] }
	@palette.declare_button(symb_master, hsh)

	hshp = { :parent => symb_master, :width => wid, :draw_proc => @draw_local }
	
	hsh = { :value_proc => (proc { @option_face_selection == :single }), :tooltip => T6[:T_TIP_FaceSelectionSingle],
            :draw_proc => :std_face_selection_single }
	@palette.declare_button(:pal_face_selection_single, hshp, hsh) { set_face_selection :single }
	hsh = { :value_proc => (proc { @option_face_selection == :surface }), :tooltip => T6[:T_TIP_FaceSelectionSurface],
	        :draw_proc => :std_face_selection_surface }
	@palette.declare_button(:pal_face_selection_surface, hshp, hsh) { set_face_selection :surface }
	hsh = { :value_proc => (proc { @option_face_selection == :connected }), :tooltip => T6[:T_TIP_FaceSelectionConnected],
	        :draw_proc => :std_face_selection_connected }
	@palette.declare_button(:pal_face_selection_connected, hshp, hsh) { set_face_selection :connected }
	hsh = { :value_proc => (proc { @option_face_selection == :same_color }), :tooltip => T6[:T_TIP_FaceSelectionSameColor],
	        :draw_proc => :std_face_selection_same_color }
	@palette.declare_button(:pal_face_selection_same_color, hshp, hsh) { set_face_selection :same_color }
	hsh = { :value_proc => (proc { @option_face_selection == :orientation }), :tooltip => T6[:T_TIP_FaceSelectionOrientation],
	        :draw_proc => :std_face_selection_orientation }
	@palette.declare_button(:pal_face_selection_orientation, hshp, hsh) { set_face_selection :orientation }
end

#--------------------------------------------------------------
# Custom drawing for palette buttons
#--------------------------------------------------------------

#Custom drawing of buttons
def draw_button_opengl(symb, dx, dy, main_color, frame_color, selected)
	code = symb.to_s
	lst_gl = []
	dx2 = dx / 2
	dy2 = dy / 2
	grayed = @palette.button_is_grayed?(symb)
	color = (grayed) ? 'gray' : frame_color
	
	case code
		
	when /pal_draw_hidden/i
		color = (selected) ? 'blue' : 'black'
		pts = []
		pts.push Geom::Point3d.new(0, 0)
		pts.push Geom::Point3d.new(dx, 0)
		pts.push Geom::Point3d.new(dx, dy)
		pts.push Geom::Point3d.new(0, dy)
		lst_gl.push [GL_LINE_LOOP, pts, 'gray', 1, '']	
		lst_gl.push [GL_LINE_STRIP, [pts[0], pts[2]], color, 2, '_']	
		lst_gl.push [GL_LINE_STRIP, [pts[1], pts[3]], color, 2, '_']	

	when /pal_monochrome/i
		color = (selected) ? 'blue' : 'black'
		lcolors = ['black', 'darkgray', 'gray', 'lightgrey', 'white']
		origin = Geom::Point3d.new(dx2, dy / 3)
		trot = Geom::Transformation.rotation origin, Z_AXIS, 90.degrees
		for i in 0..4
			pts = G6.pts_circle origin.x, origin.y, dy2 - 2 * i
			pts = pts[0..6].collect { |pt| trot * pt }
			lst_gl.push [GL_POLYGON, pts, lcolors[i]]
			#lst_gl.push [GL_LINE_LOOP, pts, color, 1, '']
		end	
		
	#Arrows to switch families	
	when /pal_orient_mode/i
		fcolor = @rendering_options["FaceFrontColor"]
		bcolor = @rendering_options["FaceBackColor"]
		dx3 = dx / 3
		dy3 = dy / 3
		frcolor = (selected) ? 'blue' : 'black'
		lscolor = [fcolor, bcolor]
		for i in 0..2
			x0 = dx3 * i + 1
			x1 = x0 + dx3
			for j in 0..2
				pts = []
				k = (i+j).modulo(2)
				y0 = dy3 * j
				y1 = y0 + dy3
				pts.push Geom::Point3d.new(x0, y0)
				pts.push Geom::Point3d.new(x0, y1)
				pts.push Geom::Point3d.new(x1, y1)
				pts.push Geom::Point3d.new(x1, y0)
				lst_gl.push [GL_POLYGON, pts, lscolor[k]]
				lst_gl.push [GL_LINE_LOOP, pts, frcolor, 1, '']
			end	
		end

	when /pal_reverse_mode/i
		frcolor = (selected) ? 'green' : 'black'
		fcolor = @rendering_options["FaceFrontColor"]
		bcolor = @rendering_options["FaceBackColor"]
		pts = []
		pts.push Geom::Point3d.new(1, 1)
		pts.push Geom::Point3d.new(dx2-1, 1)
		pts.push Geom::Point3d.new(dx2-1, dy2)
		pts.push Geom::Point3d.new(1, dy2)
		lst_gl.push [GL_POLYGON, pts, fcolor]	
		lst_gl.push [GL_LINE_LOOP, pts, frcolor, 1, '']	
		pts = []
		pts.push Geom::Point3d.new(dx2+1, dy2)
		pts.push Geom::Point3d.new(dx-1, dy2)
		pts.push Geom::Point3d.new(dx-1, dy-1)
		pts.push Geom::Point3d.new(dx2+1, dy-1)
		lst_gl.push [GL_POLYGON, pts, bcolor]	
		lst_gl.push [GL_LINE_LOOP, pts, frcolor, 1, '']	
		
		origin = Geom::Point3d.new dx2, dy2
		trot = Geom::Transformation.rotation origin, Z_AXIS, 45.degrees
		line = []
		line.push Geom::Point3d.new(5, dy2)
		line.push Geom::Point3d.new(dx-5, dy2)
		tr1 = []
		tr1.push Geom::Point3d.new(-1, dy2)
		tr1.push Geom::Point3d.new(5, dy2 + 5)
		tr1.push Geom::Point3d.new(5, dy2 - 5)
		tr2 = []
		tr2.push Geom::Point3d.new(dx+1, dy2)
		tr2.push Geom::Point3d.new(dx-5, dy2 + 5)
		tr2.push Geom::Point3d.new(dx-5, dy2 - 5)
		
		line = line.collect { |pt| trot * pt }
		tr1 = tr1.collect { |pt| trot * pt }
		tr2 = tr2.collect { |pt| trot * pt }
		lst_gl.push [GL_POLYGON, tr1, frcolor]	
		lst_gl.push [GL_POLYGON, tr2, frcolor]	
		lst_gl.push [GL_LINE_STRIP, line, frcolor, 3, '']	
		
	end	#case code
	
	lst_gl
end

#Custom drawing for the Blason
def draw_button_blason(symb, dx, dy)
	lst_gl = []
	dx13 = dx/3
	dx23 = 2 * dx/3
	dy2 = dy/2	
	frcolor = 'black'
	fcolor = @rendering_options["FaceFrontColor"]
	bcolor = @rendering_options["FaceBackColor"]
	
	pts = []
	pts.push Geom::Point3d.new(2, dy2+2)
	pts.push Geom::Point3d.new(dx23, dy2+2)
	pts.push Geom::Point3d.new(dx23, dy-2)
	pts.push Geom::Point3d.new(2, dy-2)
	lst_gl.push [GL_POLYGON, pts, bcolor]	
	lst_gl.push [GL_LINE_LOOP, pts, frcolor, 1, '']	
	
	pts = []
	pts.push Geom::Point3d.new(dx13-2, 3)
	pts.push Geom::Point3d.new(dx-4, 3)
	pts.push Geom::Point3d.new(dx-4, dy2)
	pts.push Geom::Point3d.new(dx13-2, dy2)
	lst_gl.push [GL_POLYGON, pts, fcolor]	
	lst_gl.push [GL_LINE_LOOP, pts, frcolor, 1, '']	
	
	lst_gl
end

end	#class ReverseOrientFacesTool

end	#End Module ReverseOrientFaces

end	#End Module F6_FredoTools
