=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2011 Fredo6 - Designed and written Aug 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_FredoTools__RevertCurve.rb
# Original Date	:   23 Aug 2011 (based on work published in Feb 2010)
# Description	:   Inverse the orientation of selected curves
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

module RevertCurve

#====================================================================================================
#----------------------------------------------------------------------------------------------------
# Plugin Implementation
#----------------------------------------------------------------------------------------------------
#====================================================================================================

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

def self._execution(symb)
	execute
end

#----------------------------------------------------------------------------------------------------
# Plugin Top Level
#----------------------------------------------------------------------------------------------------

#Initialize and execute the Revert of a curve
def self.execute(selection=nil)
	model = Sketchup.active_model
	entities = model.active_entities
	selection = model.selection unless selection
	
	#Finding the curves
	hsh_curves = {}
	selection.each do |e|
		next unless e.instance_of?(Sketchup::Edge)
		curve = e.curve
		hsh_curves[curve.object_id] = curve if curve
	end
	return if hsh_curves.length == 0
	
	#Reverting the curves
	selection.clear
	model.start_operation T7[:PlugName]
	
	hsh_curves.each do |key, curve|
		lspt = curve.vertices.collect { |v| v.position }
		
		#Curve is standalone with no faces
		if curve.vertices.first.edges.length == 1 && curve.vertices.last.edges.length == 1 && !(curve.edges.find { |e| !e.faces.empty? })
			curve.move_vertices lspt.reverse
			selection.add curve.edges
			next
		end

		#Curves with faces. Storing the faces that we be deleted for restoration
		linfo_faces = []
		curve.edges.each do |e|
			e.faces.each { |f| linfo_faces.push [f.vertices.collect { |v| v.position }, f.material, f.back_material] }
		end	
		
		#Erasing the edges of the current curve and constructing the new one in reverse order
		##curve.edges.each { |e| e.erase! }
		grp = entities.add_group
		grp.entities.add_curve lspt.reverse
		le = grp.explode
		le = le.find_all { |e| e.instance_of?(Sketchup::Edge) }
		
		#Restoring the faces
		linfo_faces.each do |lf|
			next
			lpt, mat, bmat = lf
			face = entities.add_face lpt if lpt
			face.material = mat if mat
			face.back_material = bmat if bmat
		end	
		selection.add le
	end
	
	model.commit_operation
end

end	#End Module RevertCurve

end	#End Module F6_FredoTools
